@@ -1,5 +1,5 @@
-/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
-/* vim: set ts=8 sts=4 et sw=4 tw=99: */
+/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
+/* vim: set ts=8 sts=2 et sw=2 tw=80: */
 /* This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
@@ -16,23 +16,16 @@
 #include "nsGlobalWindow.h"
 #include "nsNetUtil.h"
 #include "nsPIDOMWindow.h"
-#include "nsILoadContext.h"
-#include "nsIDocShell.h"
 #include "nsIMutableArray.h"
 #include "nsIScriptError.h"
-#include "nsISensitiveInfoHiddenURI.h"
-
-static_assert(nsIScriptError::errorFlag == JSREPORT_ERROR &&
-                  nsIScriptError::warningFlag == JSREPORT_WARNING &&
-                  nsIScriptError::exceptionFlag == JSREPORT_EXCEPTION &&
-                  nsIScriptError::strictFlag == JSREPORT_STRICT &&
-                  nsIScriptError::infoFlag == JSREPORT_USER_1,
-              "flags should be consistent");
+#include "mozilla/BasePrincipal.h"
 
 nsScriptErrorBase::nsScriptErrorBase()
     : mMessage(),
       mMessageName(),
       mSourceName(),
+      mCssSelectors(),
+      mSourceId(0),
       mLineNumber(0),
       mSourceLine(),
       mColumnNumber(0),
@@ -40,11 +33,14 @@ nsScriptErrorBase::nsScriptErrorBase()
       mCategory(),
       mOuterWindowID(0),
       mInnerWindowID(0),
-      mTimeStamp(0),
+      mMicroSecondTimeStamp(0),
       mInitializedOnMainThread(false),
-      mIsFromPrivateWindow(false) {}
+      mIsFromPrivateWindow(false),
+      mIsFromChromeContext(false),
+      mIsPromiseRejection(false),
+      mIsForwardedFromContentProcess(false) {}
 
-nsScriptErrorBase::~nsScriptErrorBase() {}
+nsScriptErrorBase::~nsScriptErrorBase() = default;
 
 void nsScriptErrorBase::AddNote(nsIScriptErrorNote* note) {
   mNotes.AppendObject(note);
@@ -60,35 +56,30 @@ void nsScriptErrorBase::InitializeOnMainThread() {
     if (window) {
       nsPIDOMWindowOuter* outer = window->GetOuterWindow();
       if (outer) mOuterWindowID = outer->WindowID();
-
-      nsIDocShell* docShell = window->GetDocShell();
-      nsCOMPtr<nsILoadContext> loadContext = do_QueryInterface(docShell);
-
-      if (loadContext) {
-        // Never mark exceptions from chrome windows as having come from
-        // private windows, since we always want them to be reported.
-        nsIPrincipal* winPrincipal = window->GetPrincipal();
-        mIsFromPrivateWindow = loadContext->UsePrivateBrowsing() &&
-                               !nsContentUtils::IsSystemPrincipal(winPrincipal);
-      }
+      mIsFromChromeContext = ComputeIsFromChromeContext(window);
+      mIsFromPrivateWindow = ComputeIsFromPrivateWindow(window);
     }
   }
 
   mInitializedOnMainThread = true;
 }
 
+NS_IMETHODIMP
+nsScriptErrorBase::InitSourceId(uint32_t value) {
+  mSourceId = value;
+  return NS_OK;
+}
+
 // nsIConsoleMessage methods
 NS_IMETHODIMP
-nsScriptErrorBase::GetMessageMoz(char16_t** result) {
-  nsresult rv;
-
+nsScriptErrorBase::GetMessageMoz(nsAString& aMessage) {
   nsAutoCString message;
-  rv = ToString(message);
-  if (NS_FAILED(rv)) return rv;
-
-  *result = UTF8ToNewUnicode(message);
-  if (!*result) return NS_ERROR_OUT_OF_MEMORY;
+  nsresult rv = ToString(message);
+  if (NS_FAILED(rv)) {
+    return rv;
+  }
 
+  CopyUTF8toUTF16(message, aMessage);
   return NS_OK;
 }
 
@@ -117,6 +108,24 @@ nsScriptErrorBase::GetSourceName(nsAString& aResult) {
   return NS_OK;
 }
 
+NS_IMETHODIMP
+nsScriptErrorBase::GetCssSelectors(nsAString& aResult) {
+  aResult.Assign(mCssSelectors);
+  return NS_OK;
+}
+
+NS_IMETHODIMP
+nsScriptErrorBase::SetCssSelectors(const nsAString& aCssSelectors) {
+  mCssSelectors = aCssSelectors;
+  return NS_OK;
+}
+
+NS_IMETHODIMP
+nsScriptErrorBase::GetSourceId(uint32_t* result) {
+  *result = mSourceId;
+  return NS_OK;
+}
+
 NS_IMETHODIMP
 nsScriptErrorBase::GetSourceLine(nsAString& aResult) {
   aResult.Assign(mSourceLine);
@@ -148,13 +157,36 @@ nsScriptErrorBase::GetCategory(char** result) {
 }
 
 NS_IMETHODIMP
-nsScriptErrorBase::GetStack(JS::MutableHandleValue aStack) {
+nsScriptErrorBase::GetHasException(bool* aHasException) {
+  *aHasException = false;
+  return NS_OK;
+}
+
+NS_IMETHODIMP
+nsScriptErrorBase::GetException(JS::MutableHandle<JS::Value> aException) {
+  aException.setUndefined();
+  return NS_OK;
+}
+
+NS_IMETHODIMP
+nsScriptErrorBase::SetException(JS::Handle<JS::Value> aStack) {
+  return NS_ERROR_NOT_IMPLEMENTED;
+}
+
+NS_IMETHODIMP
+nsScriptErrorBase::GetStack(JS::MutableHandle<JS::Value> aStack) {
   aStack.setUndefined();
   return NS_OK;
 }
 
 NS_IMETHODIMP
-nsScriptErrorBase::SetStack(JS::HandleValue aStack) { return NS_OK; }
+nsScriptErrorBase::SetStack(JS::Handle<JS::Value> aStack) { return NS_OK; }
+
+NS_IMETHODIMP
+nsScriptErrorBase::GetStackGlobal(JS::MutableHandle<JS::Value> aStackGlobal) {
+  aStackGlobal.setUndefined();
+  return NS_OK;
+}
 
 NS_IMETHODIMP
 nsScriptErrorBase::GetErrorMessageName(nsAString& aErrorMessageName) {
@@ -168,16 +200,6 @@ nsScriptErrorBase::SetErrorMessageName(const nsAString& aErrorMessageName) {
   return NS_OK;
 }
 
-NS_IMETHODIMP
-nsScriptErrorBase::Init(const nsAString& message, const nsAString& sourceName,
-                        const nsAString& sourceLine, uint32_t lineNumber,
-                        uint32_t columnNumber, uint32_t flags,
-                        const char* category) {
-  return InitWithWindowID(
-      message, sourceName, sourceLine, lineNumber, columnNumber, flags,
-      category ? nsDependentCString(category) : EmptyCString(), 0);
-}
-
 static void AssignSourceNameHelper(nsString& aSourceNameDest,
                                    const nsAString& aSourceNameSrc) {
   if (aSourceNameSrc.IsEmpty()) return;
@@ -201,18 +223,34 @@ static void AssignSourceNameHelper(nsIURI* aSourceURI,
   }
 }
 
+NS_IMETHODIMP
+nsScriptErrorBase::Init(const nsAString& message, const nsAString& sourceName,
+                        const nsAString& sourceLine, uint32_t lineNumber,
+                        uint32_t columnNumber, uint32_t flags,
+                        const nsACString& category, bool fromPrivateWindow,
+                        bool fromChromeContext) {
+  InitializationHelper(message, sourceLine, lineNumber, columnNumber, flags,
+                       category, 0 /* inner Window ID */, fromChromeContext);
+  AssignSourceNameHelper(mSourceName, sourceName);
+
+  mIsFromPrivateWindow = fromPrivateWindow;
+  mIsFromChromeContext = fromChromeContext;
+  return NS_OK;
+}
+
 void nsScriptErrorBase::InitializationHelper(
     const nsAString& message, const nsAString& sourceLine, uint32_t lineNumber,
     uint32_t columnNumber, uint32_t flags, const nsACString& category,
-    uint64_t aInnerWindowID) {
+    uint64_t aInnerWindowID, bool aFromChromeContext) {
   mMessage.Assign(message);
   mLineNumber = lineNumber;
   mSourceLine.Assign(sourceLine);
   mColumnNumber = columnNumber;
   mFlags = flags;
   mCategory = category;
-  mTimeStamp = JS_Now() / 1000;
+  mMicroSecondTimeStamp = JS_Now();
   mInnerWindowID = aInnerWindowID;
+  mIsFromChromeContext = aFromChromeContext;
 }
 
 NS_IMETHODIMP
@@ -221,9 +259,10 @@ nsScriptErrorBase::InitWithWindowID(const nsAString& message,
                                     const nsAString& sourceLine,
                                     uint32_t lineNumber, uint32_t columnNumber,
                                     uint32_t flags, const nsACString& category,
-                                    uint64_t aInnerWindowID) {
+                                    uint64_t aInnerWindowID,
+                                    bool aFromChromeContext) {
   InitializationHelper(message, sourceLine, lineNumber, columnNumber, flags,
-                       category, aInnerWindowID);
+                       category, aInnerWindowID, aFromChromeContext);
   AssignSourceNameHelper(mSourceName, sourceName);
 
   if (aInnerWindowID && NS_IsMainThread()) InitializeOnMainThread();
@@ -235,9 +274,10 @@ NS_IMETHODIMP
 nsScriptErrorBase::InitWithSanitizedSource(
     const nsAString& message, const nsAString& sourceName,
     const nsAString& sourceLine, uint32_t lineNumber, uint32_t columnNumber,
-    uint32_t flags, const nsACString& category, uint64_t aInnerWindowID) {
+    uint32_t flags, const nsACString& category, uint64_t aInnerWindowID,
+    bool aFromChromeContext) {
   InitializationHelper(message, sourceLine, lineNumber, columnNumber, flags,
-                       category, aInnerWindowID);
+                       category, aInnerWindowID, aFromChromeContext);
   mSourceName = sourceName;
 
   if (aInnerWindowID && NS_IsMainThread()) InitializeOnMainThread();
@@ -251,9 +291,10 @@ nsScriptErrorBase::InitWithSourceURI(const nsAString& message,
                                      const nsAString& sourceLine,
                                      uint32_t lineNumber, uint32_t columnNumber,
                                      uint32_t flags, const nsACString& category,
-                                     uint64_t aInnerWindowID) {
+                                     uint64_t aInnerWindowID,
+                                     bool aFromChromeContext) {
   InitializationHelper(message, sourceLine, lineNumber, columnNumber, flags,
-                       category, aInnerWindowID);
+                       category, aInnerWindowID, aFromChromeContext);
   AssignSourceNameHelper(sourceURI, mSourceName);
 
   if (aInnerWindowID && NS_IsMainThread()) InitializeOnMainThread();
@@ -309,7 +350,8 @@ nsScriptErrorBase::ToString(nsACString& /*UTF8*/ aResult) {
   static const char error[] = "JavaScript Error";
   static const char warning[] = "JavaScript Warning";
 
-  const char* severity = !(mFlags & JSREPORT_WARNING) ? error : warning;
+  const char* severity =
+      !(mFlags & nsIScriptError::warningFlag) ? error : warning;
 
   return ToStringHelper(severity, mMessage, mSourceName, &mSourceLine,
                         mLineNumber, mColumnNumber, aResult);
@@ -337,7 +379,13 @@ nsScriptErrorBase::GetInnerWindowID(uint64_t* aInnerWindowID) {
 
 NS_IMETHODIMP
 nsScriptErrorBase::GetTimeStamp(int64_t* aTimeStamp) {
-  *aTimeStamp = mTimeStamp;
+  *aTimeStamp = mMicroSecondTimeStamp / 1000;
+  return NS_OK;
+}
+
+NS_IMETHODIMP
+nsScriptErrorBase::GetMicroSecondTimeStamp(int64_t* aTimeStamp) {
+  *aTimeStamp = mMicroSecondTimeStamp;
   return NS_OK;
 }
 
@@ -355,6 +403,44 @@ nsScriptErrorBase::GetIsFromPrivateWindow(bool* aIsFromPrivateWindow) {
   return NS_OK;
 }
 
+NS_IMETHODIMP
+nsScriptErrorBase::GetIsFromChromeContext(bool* aIsFromChromeContext) {
+  NS_WARNING_ASSERTION(NS_IsMainThread() || mInitializedOnMainThread,
+                       "This can't be safely determined off the main thread, "
+                       "returning an inaccurate value!");
+  if (!mInitializedOnMainThread && NS_IsMainThread()) {
+    InitializeOnMainThread();
+  }
+  *aIsFromChromeContext = mIsFromChromeContext;
+  return NS_OK;
+}
+
+NS_IMETHODIMP
+nsScriptErrorBase::GetIsPromiseRejection(bool* aIsPromiseRejection) {
+  *aIsPromiseRejection = mIsPromiseRejection;
+  return NS_OK;
+}
+
+NS_IMETHODIMP
+nsScriptErrorBase::InitIsPromiseRejection(bool aIsPromiseRejection) {
+  mIsPromiseRejection = aIsPromiseRejection;
+  return NS_OK;
+}
+
+NS_IMETHODIMP
+nsScriptErrorBase::GetIsForwardedFromContentProcess(
+    bool* aIsForwardedFromContentProcess) {
+  *aIsForwardedFromContentProcess = mIsForwardedFromContentProcess;
+  return NS_OK;
+}
+
+NS_IMETHODIMP
+nsScriptErrorBase::SetIsForwardedFromContentProcess(
+    bool aIsForwardedFromContentProcess) {
+  mIsForwardedFromContentProcess = aIsForwardedFromContentProcess;
+  return NS_OK;
+}
+
 NS_IMETHODIMP
 nsScriptErrorBase::GetNotes(nsIArray** aNotes) {
   nsresult rv = NS_OK;
@@ -368,18 +454,39 @@ nsScriptErrorBase::GetNotes(nsIArray** aNotes) {
   return NS_OK;
 }
 
+/* static */
+bool nsScriptErrorBase::ComputeIsFromPrivateWindow(
+    nsGlobalWindowInner* aWindow) {
+  // Never mark exceptions from chrome windows as having come from private
+  // windows, since we always want them to be reported.
+  nsIPrincipal* winPrincipal = aWindow->GetPrincipal();
+  return aWindow->IsPrivateBrowsing() && !winPrincipal->IsSystemPrincipal();
+}
+
+/* static */
+bool nsScriptErrorBase::ComputeIsFromChromeContext(
+    nsGlobalWindowInner* aWindow) {
+  nsIPrincipal* winPrincipal = aWindow->GetPrincipal();
+  return winPrincipal->IsSystemPrincipal();
+}
+
 NS_IMPL_ISUPPORTS(nsScriptError, nsIConsoleMessage, nsIScriptError)
 
 nsScriptErrorNote::nsScriptErrorNote()
-    : mMessage(), mSourceName(), mLineNumber(0), mColumnNumber(0) {}
+    : mMessage(),
+      mSourceName(),
+      mSourceId(0),
+      mLineNumber(0),
+      mColumnNumber(0) {}
 
-nsScriptErrorNote::~nsScriptErrorNote() {}
+nsScriptErrorNote::~nsScriptErrorNote() = default;
 
 void nsScriptErrorNote::Init(const nsAString& message,
-                             const nsAString& sourceName, uint32_t lineNumber,
-                             uint32_t columnNumber) {
+                             const nsAString& sourceName, uint32_t sourceId,
+                             uint32_t lineNumber, uint32_t columnNumber) {
   mMessage.Assign(message);
   AssignSourceNameHelper(mSourceName, sourceName);
+  mSourceId = sourceId;
   mLineNumber = lineNumber;
   mColumnNumber = columnNumber;
 }
@@ -397,6 +504,12 @@ nsScriptErrorNote::GetSourceName(nsAString& aResult) {
   return NS_OK;
 }
 
+NS_IMETHODIMP
+nsScriptErrorNote::GetSourceId(uint32_t* result) {
+  *result = mSourceId;
+  return NS_OK;
+}
+
 NS_IMETHODIMP
 nsScriptErrorNote::GetLineNumber(uint32_t* result) {
   *result = mLineNumber;
