@@ -6,28 +6,36 @@
 
 #include "mozilla/dom/cache/Context.h"
 
+#include "CacheCommon.h"
+
 #include "mozilla/AutoRestore.h"
+#include "mozilla/dom/SafeRefPtr.h"
 #include "mozilla/dom/cache/Action.h"
 #include "mozilla/dom/cache/FileUtils.h"
 #include "mozilla/dom/cache/Manager.h"
 #include "mozilla/dom/cache/ManagerId.h"
+#include "mozilla/dom/quota/Assertions.h"
+#include "mozilla/dom/quota/DirectoryLock.h"
 #include "mozilla/dom/quota/QuotaManager.h"
+#include "mozilla/dom/quota/ResultExtensions.h"
+#include "mozilla/ipc/PBackgroundSharedTypes.h"
 #include "mozIStorageConnection.h"
-#include "nsIFile.h"
 #include "nsIPrincipal.h"
 #include "nsIRunnable.h"
+#include "nsIThread.h"
 #include "nsThreadUtils.h"
 
 namespace {
 
 using mozilla::dom::cache::Action;
-using mozilla::dom::cache::QuotaInfo;
+using mozilla::dom::cache::CacheDirectoryMetadata;
 
 class NullAction final : public Action {
  public:
-  NullAction() {}
+  NullAction() = default;
 
-  virtual void RunOnTarget(Resolver* aResolver, const QuotaInfo&,
+  virtual void RunOnTarget(mozilla::SafeRefPtr<Resolver> aResolver,
+                           const mozilla::Maybe<CacheDirectoryMetadata>&,
                            Data*) override {
     // Resolve success immediately.  This Action does no actual work.
     MOZ_DIAGNOSTIC_ASSERT(aResolver);
@@ -37,11 +45,10 @@ class NullAction final : public Action {
 
 }  // namespace
 
-namespace mozilla {
-namespace dom {
-namespace cache {
+namespace mozilla::dom::cache {
 
 using mozilla::dom::quota::AssertIsOnIOThread;
+using mozilla::dom::quota::DirectoryLock;
 using mozilla::dom::quota::OpenDirectoryListener;
 using mozilla::dom::quota::PERSISTENCE_TYPE_DEFAULT;
 using mozilla::dom::quota::PersistenceType;
@@ -88,15 +95,16 @@ class Context::Data final : public Action::Data {
 class Context::QuotaInitRunnable final : public nsIRunnable,
                                          public OpenDirectoryListener {
  public:
-  QuotaInitRunnable(Context* aContext, Manager* aManager, Data* aData,
-                    nsISerialEventTarget* aTarget, Action* aInitAction)
-      : mContext(aContext),
-        mThreadsafeHandle(aContext->CreateThreadsafeHandle()),
-        mManager(aManager),
+  QuotaInitRunnable(SafeRefPtr<Context> aContext, SafeRefPtr<Manager> aManager,
+                    Data* aData, nsISerialEventTarget* aTarget,
+                    SafeRefPtr<Action> aInitAction)
+      : mContext(std::move(aContext)),
+        mThreadsafeHandle(mContext->CreateThreadsafeHandle()),
+        mManager(std::move(aManager)),
         mData(aData),
         mTarget(aTarget),
-        mInitAction(aInitAction),
-        mInitiatingEventTarget(GetCurrentThreadEventTarget()),
+        mInitAction(std::move(aInitAction)),
+        mInitiatingEventTarget(GetCurrentSerialEventTarget()),
         mResult(NS_OK),
         mState(STATE_INIT),
         mCanceled(false) {
@@ -108,6 +116,12 @@ class Context::QuotaInitRunnable final : public nsIRunnable,
     MOZ_DIAGNOSTIC_ASSERT(mInitAction);
   }
 
+  Maybe<DirectoryLock&> MaybeDirectoryLockRef() const {
+    NS_ASSERT_OWNINGTHREAD(QuotaInitRunnable);
+
+    return ToMaybeRef(mDirectoryLock.get());
+  }
+
   nsresult Dispatch() {
     NS_ASSERT_OWNINGTHREAD(QuotaInitRunnable);
     MOZ_DIAGNOSTIC_ASSERT(mState == STATE_INIT);
@@ -128,8 +142,6 @@ class Context::QuotaInitRunnable final : public nsIRunnable,
     mInitAction->CancelOnInitiatingThread();
   }
 
-  void OpenDirectory();
-
   // OpenDirectoryListener methods
   virtual void DirectoryLockAcquired(DirectoryLock* aLock) override;
 
@@ -150,7 +162,7 @@ class Context::QuotaInitRunnable final : public nsIRunnable,
     nsresult Result() const { return mResult; }
 
    private:
-    ~SyncResolver() {}
+    ~SyncResolver() = default;
 
     bool mResolved;
     nsresult mResult;
@@ -169,7 +181,6 @@ class Context::QuotaInitRunnable final : public nsIRunnable,
     STATE_INIT,
     STATE_GET_INFO,
     STATE_CREATE_QUOTA_MANAGER,
-    STATE_OPEN_DIRECTORY,
     STATE_WAIT_FOR_DIRECTORY_LOCK,
     STATE_ENSURE_ORIGIN_INITIALIZED,
     STATE_RUN_ON_TARGET,
@@ -197,15 +208,16 @@ class Context::QuotaInitRunnable final : public nsIRunnable,
     mInitAction = nullptr;
   }
 
-  RefPtr<Context> mContext;
-  RefPtr<ThreadsafeHandle> mThreadsafeHandle;
-  RefPtr<Manager> mManager;
+  SafeRefPtr<Context> mContext;
+  SafeRefPtr<ThreadsafeHandle> mThreadsafeHandle;
+  SafeRefPtr<Manager> mManager;
   RefPtr<Data> mData;
   nsCOMPtr<nsISerialEventTarget> mTarget;
-  RefPtr<Action> mInitAction;
+  SafeRefPtr<Action> mInitAction;
   nsCOMPtr<nsIEventTarget> mInitiatingEventTarget;
   nsresult mResult;
-  QuotaInfo mQuotaInfo;
+  Maybe<mozilla::ipc::PrincipalInfo> mPrincipalInfo;
+  Maybe<CacheDirectoryMetadata> mDirectoryMetadata;
   RefPtr<DirectoryLock> mDirectoryLock;
   State mState;
   Atomic<bool> mCanceled;
@@ -215,29 +227,17 @@ class Context::QuotaInitRunnable final : public nsIRunnable,
   NS_DECL_NSIRUNNABLE
 };
 
-void Context::QuotaInitRunnable::OpenDirectory() {
-  NS_ASSERT_OWNINGTHREAD(QuotaInitRunnable);
-  MOZ_DIAGNOSTIC_ASSERT(mState == STATE_CREATE_QUOTA_MANAGER ||
-                        mState == STATE_OPEN_DIRECTORY);
-  MOZ_DIAGNOSTIC_ASSERT(QuotaManager::Get());
-
-  // QuotaManager::OpenDirectory() will hold a reference to us as
-  // a listener.  We will then get DirectoryLockAcquired() on the owning
-  // thread when it is safe to access our storage directory.
-  mState = STATE_WAIT_FOR_DIRECTORY_LOCK;
-  QuotaManager::Get()->OpenDirectory(PERSISTENCE_TYPE_DEFAULT,
-                                     mQuotaInfo.mGroup, mQuotaInfo.mOrigin,
-                                     quota::Client::DOMCACHE,
-                                     /* aExclusive */ false, this);
-}
-
 void Context::QuotaInitRunnable::DirectoryLockAcquired(DirectoryLock* aLock) {
   NS_ASSERT_OWNINGTHREAD(QuotaInitRunnable);
+  MOZ_DIAGNOSTIC_ASSERT(aLock);
   MOZ_DIAGNOSTIC_ASSERT(mState == STATE_WAIT_FOR_DIRECTORY_LOCK);
   MOZ_DIAGNOSTIC_ASSERT(!mDirectoryLock);
 
   mDirectoryLock = aLock;
 
+  MOZ_DIAGNOSTIC_ASSERT(mDirectoryLock->Id() >= 0);
+  mDirectoryMetadata->mDirectoryLockId = mDirectoryLock->Id();
+
   if (mCanceled) {
     Complete(NS_ERROR_ABORT);
     return;
@@ -284,11 +284,6 @@ NS_IMPL_ISUPPORTS(mozilla::dom::cache::Context::QuotaInitRunnable, nsIRunnable);
 // +----------+-----------+                 |
 //            |                             |
 // +----------v-----------+                 |
-// |    OpenDirectory     |  Resolve(error) |
-// |    (Orig Thread)     +-----------------+
-// +----------+-----------+                 |
-//            |                             |
-// +----------v-----------+                 |
 // | WaitForDirectoryLock |  Resolve(error) |
 // |    (Orig Thread)     +-----------------+
 // +----------+-----------+                 |
@@ -319,31 +314,38 @@ Context::QuotaInitRunnable::Run() {
   // May run on different threads depending on the state.  See individual
   // state cases for thread assertions.
 
-  RefPtr<SyncResolver> resolver = new SyncResolver();
+  SafeRefPtr<SyncResolver> resolver = MakeSafeRefPtr<SyncResolver>();
 
   switch (mState) {
     // -----------------------------------
     case STATE_GET_INFO: {
       MOZ_ASSERT(NS_IsMainThread());
 
-      if (mCanceled) {
-        resolver->Resolve(NS_ERROR_ABORT);
-        break;
+      auto res = [this]() -> Result<Ok, nsresult> {
+        if (mCanceled) {
+          return Err(NS_ERROR_ABORT);
+        }
+
+        nsCOMPtr<nsIPrincipal> principal = mManager->GetManagerId().Principal();
+
+        mozilla::ipc::PrincipalInfo principalInfo;
+        QM_TRY(
+            MOZ_TO_RESULT(PrincipalToPrincipalInfo(principal, &principalInfo)));
+
+        mPrincipalInfo.emplace(std::move(principalInfo));
+
+        mState = STATE_CREATE_QUOTA_MANAGER;
+
+        MOZ_ALWAYS_SUCCEEDS(
+            mInitiatingEventTarget->Dispatch(this, nsIThread::DISPATCH_NORMAL));
+
+        return Ok{};
+      }();
+
+      if (res.isErr()) {
+        resolver->Resolve(res.inspectErr());
       }
 
-      RefPtr<ManagerId> managerId = mManager->GetManagerId();
-      nsCOMPtr<nsIPrincipal> principal = managerId->Principal();
-      nsresult rv = QuotaManager::GetInfoFromPrincipal(
-          principal, &mQuotaInfo.mSuffix, &mQuotaInfo.mGroup,
-          &mQuotaInfo.mOrigin);
-      if (NS_WARN_IF(NS_FAILED(rv))) {
-        resolver->Resolve(rv);
-        break;
-      }
-
-      mState = STATE_CREATE_QUOTA_MANAGER;
-      MOZ_ALWAYS_SUCCEEDS(
-          mInitiatingEventTarget->Dispatch(this, nsIThread::DISPATCH_NORMAL));
       break;
     }
     // ----------------------------------
@@ -355,49 +357,67 @@ Context::QuotaInitRunnable::Run() {
         break;
       }
 
-      if (QuotaManager::Get()) {
-        OpenDirectory();
-        return NS_OK;
-      }
+      QM_TRY(QuotaManager::EnsureCreated(), QM_PROPAGATE,
+             [&resolver](const auto rv) { resolver->Resolve(rv); });
 
-      mState = STATE_OPEN_DIRECTORY;
-      QuotaManager::GetOrCreate(this);
-      break;
-    }
-    // ----------------------------------
-    case STATE_OPEN_DIRECTORY: {
-      NS_ASSERT_OWNINGTHREAD(QuotaInitRunnable);
+      auto* const quotaManager = QuotaManager::Get();
+      MOZ_DIAGNOSTIC_ASSERT(quotaManager);
 
-      if (NS_WARN_IF(!QuotaManager::Get())) {
-        resolver->Resolve(NS_ERROR_FAILURE);
-        break;
-      }
+      QM_TRY_UNWRAP(
+          auto principalMetadata,
+          quotaManager->GetInfoFromValidatedPrincipalInfo(*mPrincipalInfo));
+
+      mDirectoryMetadata.emplace(std::move(principalMetadata));
+
+      // Open directory
+      RefPtr<DirectoryLock> directoryLock = quotaManager->CreateDirectoryLock(
+          PERSISTENCE_TYPE_DEFAULT, *mDirectoryMetadata,
+          quota::Client::DOMCACHE,
+          /* aExclusive */ false);
+
+      // DirectoryLock::Acquire() will hold a reference to us as a listener. We
+      // will then get DirectoryLockAcquired() on the owning thread when it is
+      // safe to access our storage directory.
+      mState = STATE_WAIT_FOR_DIRECTORY_LOCK;
+      directoryLock->Acquire(this);
 
-      OpenDirectory();
       break;
     }
     // ----------------------------------
     case STATE_ENSURE_ORIGIN_INITIALIZED: {
       AssertIsOnIOThread();
 
-      if (mCanceled) {
-        resolver->Resolve(NS_ERROR_ABORT);
-        break;
+      auto res = [this]() -> Result<Ok, nsresult> {
+        if (mCanceled) {
+          return Err(NS_ERROR_ABORT);
+        }
+
+        QuotaManager* quotaManager = QuotaManager::Get();
+        MOZ_DIAGNOSTIC_ASSERT(quotaManager);
+
+        QM_TRY(MOZ_TO_RESULT(quotaManager->EnsureStorageIsInitialized()));
+
+        QM_TRY(
+            MOZ_TO_RESULT(quotaManager->EnsureTemporaryStorageIsInitialized()));
+
+        QM_TRY_UNWRAP(mDirectoryMetadata->mDir,
+                      quotaManager
+                          ->EnsureTemporaryOriginIsInitialized(
+                              PERSISTENCE_TYPE_DEFAULT, *mDirectoryMetadata)
+                          .map([](const auto& res) { return res.first; }));
+
+        mState = STATE_RUN_ON_TARGET;
+
+        MOZ_ALWAYS_SUCCEEDS(
+            mTarget->Dispatch(this, nsIThread::DISPATCH_NORMAL));
+
+        return Ok{};
+      }();
+
+      if (res.isErr()) {
+        resolver->Resolve(res.inspectErr());
       }
 
-      QuotaManager* qm = QuotaManager::Get();
-      MOZ_DIAGNOSTIC_ASSERT(qm);
-      nsresult rv = qm->EnsureOriginIsInitialized(
-          PERSISTENCE_TYPE_DEFAULT, mQuotaInfo.mSuffix, mQuotaInfo.mGroup,
-          mQuotaInfo.mOrigin, getter_AddRefs(mQuotaInfo.mDir));
-      if (NS_FAILED(rv)) {
-        resolver->Resolve(rv);
-        break;
-      }
-
-      mState = STATE_RUN_ON_TARGET;
-
-      MOZ_ALWAYS_SUCCEEDS(mTarget->Dispatch(this, nsIThread::DISPATCH_NORMAL));
       break;
     }
     // -------------------
@@ -408,7 +428,7 @@ Context::QuotaInitRunnable::Run() {
 
       // Execute the provided initialization Action.  The Action must Resolve()
       // before returning.
-      mInitAction->RunOnTarget(resolver, mQuotaInfo, mData);
+      mInitAction->RunOnTarget(resolver.clonePtr(), mDirectoryMetadata, mData);
       MOZ_DIAGNOSTIC_ASSERT(resolver->Resolved());
 
       mData = nullptr;
@@ -417,7 +437,7 @@ Context::QuotaInitRunnable::Run() {
       // the marker file.  If it wasn't opened successfully, then no need to
       // create a marker file anyway.
       if (NS_SUCCEEDED(resolver->Result())) {
-        MOZ_ALWAYS_SUCCEEDS(CreateMarkerFile(mQuotaInfo));
+        MOZ_ALWAYS_SUCCEEDS(CreateMarkerFile(*mDirectoryMetadata));
       }
 
       break;
@@ -426,7 +446,8 @@ Context::QuotaInitRunnable::Run() {
     case STATE_COMPLETING: {
       NS_ASSERT_OWNINGTHREAD(QuotaInitRunnable);
       mInitAction->CompleteOnInitiatingThread(mResult);
-      mContext->OnQuotaInit(mResult, mQuotaInfo, mDirectoryLock.forget());
+      mContext->OnQuotaInit(mResult, mDirectoryMetadata,
+                            mDirectoryLock.forget());
       mState = STATE_COMPLETE;
 
       // Explicitly cleanup here as the destructor could fire on any of
@@ -436,7 +457,9 @@ Context::QuotaInitRunnable::Run() {
     }
     // -----
     case STATE_WAIT_FOR_DIRECTORY_LOCK:
-    default: { MOZ_CRASH("unexpected state in QuotaInitRunnable"); }
+    default: {
+      MOZ_CRASH("unexpected state in QuotaInitRunnable");
+    }
   }
 
   if (resolver->Resolved()) {
@@ -453,14 +476,15 @@ class Context::ActionRunnable final : public nsIRunnable,
                                       public Action::Resolver,
                                       public Context::Activity {
  public:
-  ActionRunnable(Context* aContext, Data* aData, nsISerialEventTarget* aTarget,
-                 Action* aAction, const QuotaInfo& aQuotaInfo)
-      : mContext(aContext),
+  ActionRunnable(SafeRefPtr<Context> aContext, Data* aData,
+                 nsISerialEventTarget* aTarget, SafeRefPtr<Action> aAction,
+                 const Maybe<CacheDirectoryMetadata>& aDirectoryMetadata)
+      : mContext(std::move(aContext)),
         mData(aData),
         mTarget(aTarget),
-        mAction(aAction),
-        mQuotaInfo(aQuotaInfo),
-        mInitiatingThread(GetCurrentThreadEventTarget()),
+        mAction(std::move(aAction)),
+        mDirectoryMetadata(aDirectoryMetadata),
+        mInitiatingThread(GetCurrentSerialEventTarget()),
         mState(STATE_INIT),
         mResult(NS_OK),
         mExecutingRunOnTarget(false) {
@@ -468,7 +492,7 @@ class Context::ActionRunnable final : public nsIRunnable,
     // mData may be nullptr
     MOZ_DIAGNOSTIC_ASSERT(mTarget);
     MOZ_DIAGNOSTIC_ASSERT(mAction);
-    // mQuotaInfo.mDir may be nullptr if QuotaInitRunnable failed
+    // mDirectoryMetadata.mDir may be nullptr if QuotaInitRunnable failed
     MOZ_DIAGNOSTIC_ASSERT(mInitiatingThread);
   }
 
@@ -530,7 +554,7 @@ class Context::ActionRunnable final : public nsIRunnable,
     NS_ASSERT_OWNINGTHREAD(ActionRunnable);
     MOZ_DIAGNOSTIC_ASSERT(mContext);
     MOZ_DIAGNOSTIC_ASSERT(mAction);
-    mContext->RemoveActivity(this);
+    mContext->RemoveActivity(*this);
     mContext = nullptr;
     mAction = nullptr;
   }
@@ -544,11 +568,11 @@ class Context::ActionRunnable final : public nsIRunnable,
     STATE_COMPLETE
   };
 
-  RefPtr<Context> mContext;
+  SafeRefPtr<Context> mContext;
   RefPtr<Data> mData;
   nsCOMPtr<nsISerialEventTarget> mTarget;
-  RefPtr<Action> mAction;
-  const QuotaInfo mQuotaInfo;
+  SafeRefPtr<Action> mAction;
+  const Maybe<CacheDirectoryMetadata> mDirectoryMetadata;
   nsCOMPtr<nsIEventTarget> mInitiatingThread;
   State mState;
   nsresult mResult;
@@ -610,7 +634,7 @@ Context::ActionRunnable::Run() {
       mExecutingRunOnTarget = true;
 
       mState = STATE_RUNNING;
-      mAction->RunOnTarget(this, mQuotaInfo, mData);
+      mAction->RunOnTarget(SafeRefPtrFromThis(), mDirectoryMetadata, mData);
 
       mData = nullptr;
 
@@ -688,10 +712,10 @@ void Context::ThreadsafeHandle::InvalidateAndAllowToClose() {
                                                    nsIThread::DISPATCH_NORMAL));
 }
 
-Context::ThreadsafeHandle::ThreadsafeHandle(Context* aContext)
-    : mStrongRef(aContext),
-      mWeakRef(aContext),
-      mOwningEventTarget(GetCurrentThreadSerialEventTarget()) {}
+Context::ThreadsafeHandle::ThreadsafeHandle(SafeRefPtr<Context> aContext)
+    : mStrongRef(std::move(aContext)),
+      mWeakRef(mStrongRef.unsafeGetRawPtr()),
+      mOwningEventTarget(GetCurrentSerialEventTarget()) {}
 
 Context::ThreadsafeHandle::~ThreadsafeHandle() {
   // Normally we only touch mStrongRef on the owning thread.  This is safe,
@@ -702,10 +726,11 @@ Context::ThreadsafeHandle::~ThreadsafeHandle() {
     return;
   }
 
-  // Dispatch is guaranteed to succeed here because we block shutdown until
-  // all Contexts have been destroyed.
-  NS_ProxyRelease("Context::ThreadsafeHandle::mStrongRef", mOwningEventTarget,
-                  mStrongRef.forget());
+  // Dispatch in NS_ProxyRelease is guaranteed to succeed here because we block
+  // shutdown until all Contexts have been destroyed. Therefore it is ok to have
+  // MOZ_ALWAYS_SUCCEED here.
+  MOZ_ALWAYS_SUCCEEDS(NS_ProxyRelease("Context::ThreadsafeHandle::mStrongRef",
+                                      mOwningEventTarget, mStrongRef.forget()));
 }
 
 void Context::ThreadsafeHandle::AllowToCloseOnOwningThread() {
@@ -742,51 +767,73 @@ void Context::ThreadsafeHandle::InvalidateAndAllowToCloseOnOwningThread() {
   MOZ_DIAGNOSTIC_ASSERT(!mStrongRef);
 }
 
-void Context::ThreadsafeHandle::ContextDestroyed(Context* aContext) {
+void Context::ThreadsafeHandle::ContextDestroyed(Context& aContext) {
   MOZ_ASSERT(mOwningEventTarget->IsOnCurrentThread());
   MOZ_DIAGNOSTIC_ASSERT(!mStrongRef);
   MOZ_DIAGNOSTIC_ASSERT(mWeakRef);
-  MOZ_DIAGNOSTIC_ASSERT(mWeakRef == aContext);
+  MOZ_DIAGNOSTIC_ASSERT(mWeakRef == &aContext);
   mWeakRef = nullptr;
 }
 
 // static
-already_AddRefed<Context> Context::Create(Manager* aManager,
-                                          nsISerialEventTarget* aTarget,
-                                          Action* aInitAction,
-                                          Context* aOldContext) {
-  RefPtr<Context> context = new Context(aManager, aTarget, aInitAction);
+SafeRefPtr<Context> Context::Create(SafeRefPtr<Manager> aManager,
+                                    nsISerialEventTarget* aTarget,
+                                    SafeRefPtr<Action> aInitAction,
+                                    Maybe<Context&> aOldContext) {
+  auto context = MakeSafeRefPtr<Context>(std::move(aManager), aTarget,
+                                         std::move(aInitAction));
   context->Init(aOldContext);
-  return context.forget();
+  return context;
 }
 
-Context::Context(Manager* aManager, nsISerialEventTarget* aTarget,
-                 Action* aInitAction)
-    : mManager(aManager),
+Context::Context(SafeRefPtr<Manager> aManager, nsISerialEventTarget* aTarget,
+                 SafeRefPtr<Action> aInitAction)
+    : mManager(std::move(aManager)),
       mTarget(aTarget),
       mData(new Data(aTarget)),
       mState(STATE_CONTEXT_PREINIT),
       mOrphanedData(false),
-      mInitAction(aInitAction) {
+      mInitAction(std::move(aInitAction)) {
   MOZ_DIAGNOSTIC_ASSERT(mManager);
   MOZ_DIAGNOSTIC_ASSERT(mTarget);
 }
 
-void Context::Dispatch(Action* aAction) {
+void Context::Dispatch(SafeRefPtr<Action> aAction) {
   NS_ASSERT_OWNINGTHREAD(Context);
   MOZ_DIAGNOSTIC_ASSERT(aAction);
-
   MOZ_DIAGNOSTIC_ASSERT(mState != STATE_CONTEXT_CANCELED);
+
   if (mState == STATE_CONTEXT_CANCELED) {
     return;
-  } else if (mState == STATE_CONTEXT_INIT || mState == STATE_CONTEXT_PREINIT) {
+  }
+
+  if (mState == STATE_CONTEXT_INIT || mState == STATE_CONTEXT_PREINIT) {
     PendingAction* pending = mPendingActions.AppendElement();
-    pending->mAction = aAction;
+    pending->mAction = std::move(aAction);
     return;
   }
 
   MOZ_DIAGNOSTIC_ASSERT(mState == STATE_CONTEXT_READY);
-  DispatchAction(aAction);
+  DispatchAction(std::move(aAction));
+}
+
+Maybe<DirectoryLock&> Context::MaybeDirectoryLockRef() const {
+  NS_ASSERT_OWNINGTHREAD(Context);
+
+  if (mState == STATE_CONTEXT_PREINIT) {
+    MOZ_DIAGNOSTIC_ASSERT(!mInitRunnable);
+    MOZ_DIAGNOSTIC_ASSERT(!mDirectoryLock);
+
+    return Nothing();
+  }
+
+  if (mState == STATE_CONTEXT_INIT) {
+    MOZ_DIAGNOSTIC_ASSERT(!mDirectoryLock);
+
+    return mInitRunnable->MaybeDirectoryLockRef();
+  }
+
+  return ToMaybeRef(mDirectoryLock.get());
 }
 
 void Context::CancelAll() {
@@ -807,11 +854,8 @@ void Context::CancelAll() {
 
   mState = STATE_CONTEXT_CANCELED;
   mPendingActions.Clear();
-  {
-    ActivityList::ForwardIterator iter(mActivityList);
-    while (iter.HasMore()) {
-      iter.GetNext()->Cancel();
-    }
+  for (const auto& activity : mActivityList.ForwardRange()) {
+    activity->Cancel();
   }
   AllowToClose();
 }
@@ -838,16 +882,12 @@ void Context::CancelForCacheId(CacheId aCacheId) {
   NS_ASSERT_OWNINGTHREAD(Context);
 
   // Remove matching pending actions
-  for (int32_t i = mPendingActions.Length() - 1; i >= 0; --i) {
-    if (mPendingActions[i].mAction->MatchesCacheId(aCacheId)) {
-      mPendingActions.RemoveElementAt(i);
-    }
-  }
+  mPendingActions.RemoveElementsBy([aCacheId](const auto& pendingAction) {
+    return pendingAction.mAction->MatchesCacheId(aCacheId);
+  });
 
   // Cancel activities and let them remove themselves
-  ActivityList::ForwardIterator iter(mActivityList);
-  while (iter.HasMore()) {
-    Activity* activity = iter.GetNext();
+  for (const auto& activity : mActivityList.ForwardRange()) {
     if (activity->MatchesCacheId(aCacheId)) {
       activity->Cancel();
     }
@@ -860,14 +900,14 @@ Context::~Context() {
   MOZ_DIAGNOSTIC_ASSERT(!mData);
 
   if (mThreadsafeHandle) {
-    mThreadsafeHandle->ContextDestroyed(this);
+    mThreadsafeHandle->ContextDestroyed(*this);
   }
 
   // Note, this may set the mOrphanedData flag.
-  mManager->RemoveContext(this);
+  mManager->RemoveContext(*this);
 
-  if (mQuotaInfo.mDir && !mOrphanedData) {
-    MOZ_ALWAYS_SUCCEEDS(DeleteMarkerFile(mQuotaInfo));
+  if (mDirectoryMetadata && mDirectoryMetadata->mDir && !mOrphanedData) {
+    MOZ_ALWAYS_SUCCEEDS(DeleteMarkerFile(*mDirectoryMetadata));
   }
 
   if (mNextContext) {
@@ -875,11 +915,11 @@ Context::~Context() {
   }
 }
 
-void Context::Init(Context* aOldContext) {
+void Context::Init(Maybe<Context&> aOldContext) {
   NS_ASSERT_OWNINGTHREAD(Context);
 
   if (aOldContext) {
-    aOldContext->SetNextContext(this);
+    aOldContext->SetNextContext(SafeRefPtrFromThis());
     return;
   }
 
@@ -906,9 +946,8 @@ void Context::Start() {
   MOZ_DIAGNOSTIC_ASSERT(!mInitRunnable);
 
   mInitRunnable =
-      new QuotaInitRunnable(this, mManager, mData, mTarget, mInitAction);
-  mInitAction = nullptr;
-
+      new QuotaInitRunnable(SafeRefPtrFromThis(), mManager.clonePtr(), mData,
+                            mTarget, std::move(mInitAction));
   mState = STATE_CONTEXT_INIT;
 
   nsresult rv = mInitRunnable->Dispatch();
@@ -920,11 +959,12 @@ void Context::Start() {
   }
 }
 
-void Context::DispatchAction(Action* aAction, bool aDoomData) {
+void Context::DispatchAction(SafeRefPtr<Action> aAction, bool aDoomData) {
   NS_ASSERT_OWNINGTHREAD(Context);
 
-  RefPtr<ActionRunnable> runnable =
-      new ActionRunnable(this, mData, mTarget, aAction, mQuotaInfo);
+  auto runnable =
+      MakeSafeRefPtr<ActionRunnable>(SafeRefPtrFromThis(), mData, mTarget,
+                                     std::move(aAction), mDirectoryMetadata);
 
   if (aDoomData) {
     mData = nullptr;
@@ -936,17 +976,20 @@ void Context::DispatchAction(Action* aAction, bool aDoomData) {
     // for this invariant violation.
     MOZ_CRASH("Failed to dispatch ActionRunnable to target thread.");
   }
-  AddActivity(runnable);
+  AddActivity(*runnable);
 }
 
-void Context::OnQuotaInit(nsresult aRv, const QuotaInfo& aQuotaInfo,
-                          already_AddRefed<DirectoryLock> aDirectoryLock) {
+void Context::OnQuotaInit(
+    nsresult aRv, const Maybe<CacheDirectoryMetadata>& aDirectoryMetadata,
+    already_AddRefed<DirectoryLock> aDirectoryLock) {
   NS_ASSERT_OWNINGTHREAD(Context);
 
   MOZ_DIAGNOSTIC_ASSERT(mInitRunnable);
   mInitRunnable = nullptr;
 
-  mQuotaInfo = aQuotaInfo;
+  if (aDirectoryMetadata) {
+    mDirectoryMetadata.emplace(*aDirectoryMetadata);
+  }
 
   // Always save the directory lock to ensure QuotaManager does not shutdown
   // before the Context has gone away.
@@ -975,23 +1018,21 @@ void Context::OnQuotaInit(nsresult aRv, const QuotaInfo& aQuotaInfo,
   mState = STATE_CONTEXT_READY;
 
   for (uint32_t i = 0; i < mPendingActions.Length(); ++i) {
-    DispatchAction(mPendingActions[i].mAction);
+    DispatchAction(std::move(mPendingActions[i].mAction));
   }
   mPendingActions.Clear();
 }
 
-void Context::AddActivity(Activity* aActivity) {
+void Context::AddActivity(Activity& aActivity) {
   NS_ASSERT_OWNINGTHREAD(Context);
-  MOZ_DIAGNOSTIC_ASSERT(aActivity);
-  MOZ_ASSERT(!mActivityList.Contains(aActivity));
-  mActivityList.AppendElement(aActivity);
+  MOZ_ASSERT(!mActivityList.Contains(&aActivity));
+  mActivityList.AppendElement(WrapNotNullUnchecked(&aActivity));
 }
 
-void Context::RemoveActivity(Activity* aActivity) {
+void Context::RemoveActivity(Activity& aActivity) {
   NS_ASSERT_OWNINGTHREAD(Context);
-  MOZ_DIAGNOSTIC_ASSERT(aActivity);
-  MOZ_ALWAYS_TRUE(mActivityList.RemoveElement(aActivity));
-  MOZ_ASSERT(!mActivityList.Contains(aActivity));
+  MOZ_ALWAYS_TRUE(mActivityList.RemoveElement(&aActivity));
+  MOZ_ASSERT(!mActivityList.Contains(&aActivity));
 }
 
 void Context::NoteOrphanedData() {
@@ -1000,20 +1041,19 @@ void Context::NoteOrphanedData() {
   mOrphanedData = true;
 }
 
-already_AddRefed<Context::ThreadsafeHandle> Context::CreateThreadsafeHandle() {
+SafeRefPtr<Context::ThreadsafeHandle> Context::CreateThreadsafeHandle() {
   NS_ASSERT_OWNINGTHREAD(Context);
   if (!mThreadsafeHandle) {
-    mThreadsafeHandle = new ThreadsafeHandle(this);
+    mThreadsafeHandle = MakeSafeRefPtr<ThreadsafeHandle>(SafeRefPtrFromThis());
   }
-  RefPtr<ThreadsafeHandle> ref = mThreadsafeHandle;
-  return ref.forget();
+  return mThreadsafeHandle.clonePtr();
 }
 
-void Context::SetNextContext(Context* aNextContext) {
+void Context::SetNextContext(SafeRefPtr<Context> aNextContext) {
   NS_ASSERT_OWNINGTHREAD(Context);
   MOZ_DIAGNOSTIC_ASSERT(aNextContext);
   MOZ_DIAGNOSTIC_ASSERT(!mNextContext);
-  mNextContext = aNextContext;
+  mNextContext = std::move(aNextContext);
 }
 
 void Context::DoomTargetData() {
@@ -1028,12 +1068,9 @@ void Context::DoomTargetData() {
   // roundtrip to the target thread and back to the owning thread.  The
   // ref to the Data object is cleared on the owning thread after creating
   // the ActionRunnable, but before dispatching it.
-  RefPtr<Action> action = new NullAction();
-  DispatchAction(action, true /* doomed data */);
+  DispatchAction(MakeSafeRefPtr<NullAction>(), true /* doomed data */);
 
   MOZ_DIAGNOSTIC_ASSERT(!mData);
 }
 
-}  // namespace cache
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom::cache
