@@ -6,53 +6,48 @@
 
 #include "mozilla/dom/cache/TypeUtils.h"
 
+#include <algorithm>
+#include "mozilla/StaticPrefs_extensions.h"
 #include "mozilla/Unused.h"
 #include "mozilla/dom/CacheBinding.h"
+#include "mozilla/dom/CacheStorageBinding.h"
+#include "mozilla/dom/FetchTypes.h"
 #include "mozilla/dom/InternalRequest.h"
 #include "mozilla/dom/Request.h"
 #include "mozilla/dom/Response.h"
+#include "mozilla/dom/RootedDictionary.h"
 #include "mozilla/dom/cache/CacheTypes.h"
 #include "mozilla/dom/cache/ReadStream.h"
 #include "mozilla/ipc/BackgroundChild.h"
 #include "mozilla/ipc/IPCStreamUtils.h"
 #include "mozilla/ipc/PBackgroundChild.h"
-#include "mozilla/ipc/PFileDescriptorSetChild.h"
 #include "mozilla/ipc/InputStreamUtils.h"
+#include "nsCharSeparatedTokenizer.h"
 #include "nsCOMPtr.h"
-#include "nsIAsyncInputStream.h"
-#include "nsIAsyncOutputStream.h"
+#include "nsHttp.h"
 #include "nsIIPCSerializableInputStream.h"
-#include "nsQueryObject.h"
 #include "nsPromiseFlatString.h"
+#include "nsQueryObject.h"
 #include "nsStreamUtils.h"
 #include "nsString.h"
 #include "nsURLParsers.h"
-#include "nsCRT.h"
-#include "nsHttp.h"
 
-namespace mozilla {
-namespace dom {
-namespace cache {
+namespace mozilla::dom::cache {
 
-using mozilla::ipc::AutoIPCStream;
 using mozilla::ipc::BackgroundChild;
 using mozilla::ipc::FileDescriptor;
 using mozilla::ipc::PBackgroundChild;
-using mozilla::ipc::PFileDescriptorSetChild;
 
 namespace {
 
 static bool HasVaryStar(mozilla::dom::InternalHeaders* aHeaders) {
   nsCString varyHeaders;
   ErrorResult rv;
-  aHeaders->Get(NS_LITERAL_CSTRING("vary"), varyHeaders, rv);
+  aHeaders->Get("vary"_ns, varyHeaders, rv);
   MOZ_ALWAYS_TRUE(!rv.Failed());
 
-  char* rawBuffer = varyHeaders.BeginWriting();
-  char* token = nsCRT::strtok(rawBuffer, NS_HTTP_HEADER_SEPS, &rawBuffer);
-  for (; token;
-       token = nsCRT::strtok(rawBuffer, NS_HTTP_HEADER_SEPS, &rawBuffer)) {
-    nsDependentCString header(token);
+  for (const nsACString& header :
+       nsCCharSeparatedTokenizer(varyHeaders, NS_HTTP_HEADER_SEP).ToRange()) {
     if (header.EqualsLiteral("*")) {
       return true;
     }
@@ -60,22 +55,25 @@ static bool HasVaryStar(mozilla::dom::InternalHeaders* aHeaders) {
   return false;
 }
 
-void ToHeadersEntryList(nsTArray<HeadersEntry>& aOut,
-                        InternalHeaders* aHeaders) {
+nsTArray<HeadersEntry> ToHeadersEntryList(InternalHeaders* aHeaders) {
   MOZ_DIAGNOSTIC_ASSERT(aHeaders);
 
   AutoTArray<InternalHeaders::Entry, 16> entryList;
   aHeaders->GetEntries(entryList);
 
-  for (uint32_t i = 0; i < entryList.Length(); ++i) {
-    InternalHeaders::Entry& entry = entryList[i];
-    aOut.AppendElement(HeadersEntry(entry.mName, entry.mValue));
-  }
+  nsTArray<HeadersEntry> result;
+  result.SetCapacity(entryList.Length());
+  std::transform(entryList.cbegin(), entryList.cend(), MakeBackInserter(result),
+                 [](const auto& entry) {
+                   return HeadersEntry(entry.mName, entry.mValue);
+                 });
+
+  return result;
 }
 
 }  // namespace
 
-already_AddRefed<InternalRequest> TypeUtils::ToInternalRequest(
+SafeRefPtr<InternalRequest> TypeUtils::ToInternalRequest(
     JSContext* aCx, const RequestOrUSVString& aIn, BodyAction aBodyAction,
     ErrorResult& aRv) {
   if (aIn.IsRequest()) {
@@ -83,7 +81,7 @@ already_AddRefed<InternalRequest> TypeUtils::ToInternalRequest(
 
     // Check and set bodyUsed flag immediately because its on Request
     // instead of InternalRequest.
-    CheckAndSetBodyUsed(aCx, &request, aBodyAction, aRv);
+    CheckAndSetBodyUsed(aCx, request, aBodyAction, aRv);
     if (aRv.Failed()) {
       return nullptr;
     }
@@ -94,11 +92,11 @@ already_AddRefed<InternalRequest> TypeUtils::ToInternalRequest(
   return ToInternalRequest(aIn.GetAsUSVString(), aRv);
 }
 
-already_AddRefed<InternalRequest> TypeUtils::ToInternalRequest(
+SafeRefPtr<InternalRequest> TypeUtils::ToInternalRequest(
     JSContext* aCx, const OwningRequestOrUSVString& aIn, BodyAction aBodyAction,
     ErrorResult& aRv) {
   if (aIn.IsRequest()) {
-    RefPtr<Request> request = aIn.GetAsRequest().get();
+    Request& request = aIn.GetAsRequest();
 
     // Check and set bodyUsed flag immediately because its on Request
     // instead of InternalRequest.
@@ -107,19 +105,17 @@ already_AddRefed<InternalRequest> TypeUtils::ToInternalRequest(
       return nullptr;
     }
 
-    return request->GetInternalRequest();
+    return request.GetInternalRequest();
   }
 
   return ToInternalRequest(aIn.GetAsUSVString(), aRv);
 }
 
-void TypeUtils::ToCacheRequest(
-    CacheRequest& aOut, InternalRequest* aIn, BodyAction aBodyAction,
-    SchemeAction aSchemeAction,
-    nsTArray<UniquePtr<AutoIPCStream>>& aStreamCleanupList, ErrorResult& aRv) {
-  MOZ_DIAGNOSTIC_ASSERT(aIn);
-  aIn->GetMethod(aOut.method());
-  nsCString url(aIn->GetURLWithoutFragment());
+void TypeUtils::ToCacheRequest(CacheRequest& aOut, const InternalRequest& aIn,
+                               BodyAction aBodyAction,
+                               SchemeAction aSchemeAction, ErrorResult& aRv) {
+  aIn.GetMethod(aOut.method());
+  nsCString url(aIn.GetURLWithoutFragment());
   bool schemeValid;
   ProcessURL(url, &schemeValid, &aOut.urlWithoutQuery(), &aOut.urlQuery(), aRv);
   if (aRv.Failed()) {
@@ -127,38 +123,42 @@ void TypeUtils::ToCacheRequest(
   }
   if (!schemeValid) {
     if (aSchemeAction == TypeErrorOnInvalidScheme) {
-      NS_ConvertUTF8toUTF16 urlUTF16(url);
-      aRv.ThrowTypeError<MSG_INVALID_URL_SCHEME>(NS_LITERAL_STRING("Request"),
-                                                 urlUTF16);
+      aRv.ThrowTypeError<MSG_INVALID_URL_SCHEME>("Request", url);
       return;
     }
   }
-  aOut.urlFragment() = aIn->GetFragment();
+  aOut.urlFragment() = aIn.GetFragment();
 
-  aIn->GetReferrer(aOut.referrer());
-  aOut.referrerPolicy() = aIn->ReferrerPolicy_();
-  RefPtr<InternalHeaders> headers = aIn->Headers();
+  aIn.GetReferrer(aOut.referrer());
+  aOut.referrerPolicy() = aIn.ReferrerPolicy_();
+  RefPtr<InternalHeaders> headers = aIn.Headers();
   MOZ_DIAGNOSTIC_ASSERT(headers);
-  ToHeadersEntryList(aOut.headers(), headers);
+  aOut.headers() = ToHeadersEntryList(headers);
   aOut.headersGuard() = headers->Guard();
-  aOut.mode() = aIn->Mode();
-  aOut.credentials() = aIn->GetCredentialsMode();
-  aOut.contentPolicyType() = aIn->ContentPolicyType();
-  aOut.requestCache() = aIn->GetCacheMode();
-  aOut.requestRedirect() = aIn->GetRedirectMode();
+  aOut.mode() = aIn.Mode();
+  aOut.credentials() = aIn.GetCredentialsMode();
+  aOut.contentPolicyType() = aIn.ContentPolicyType();
+  aOut.requestCache() = aIn.GetCacheMode();
+  aOut.requestRedirect() = aIn.GetRedirectMode();
 
-  aOut.integrity() = aIn->GetIntegrity();
+  aOut.integrity() = aIn.GetIntegrity();
+  aOut.loadingEmbedderPolicy() = aIn.GetEmbedderPolicy();
+  const mozilla::UniquePtr<mozilla::ipc::PrincipalInfo>& principalInfo =
+      aIn.GetPrincipalInfo();
+  if (principalInfo) {
+    aOut.principalInfo() = Some(*(principalInfo.get()));
+  }
 
   if (aBodyAction == IgnoreBody) {
-    aOut.body() = void_t();
+    aOut.body() = Nothing();
     return;
   }
 
   // BodyUsed flag is checked and set previously in ToInternalRequest()
 
   nsCOMPtr<nsIInputStream> stream;
-  aIn->GetBody(getter_AddRefs(stream));
-  SerializeCacheStream(stream, &aOut.body(), aStreamCleanupList, aRv);
+  aIn.GetBody(getter_AddRefs(stream));
+  SerializeCacheStream(stream, &aOut.body(), aRv);
   if (NS_WARN_IF(aRv.Failed())) {
     return;
   }
@@ -185,31 +185,30 @@ void TypeUtils::ToCacheResponseWithoutBody(CacheResponse& aOut,
   RefPtr<InternalHeaders> headers = aIn.UnfilteredHeaders();
   MOZ_DIAGNOSTIC_ASSERT(headers);
   if (HasVaryStar(headers)) {
-    aRv.ThrowTypeError<MSG_RESPONSE_HAS_VARY_STAR>();
+    aRv.ThrowTypeError("Invalid Response object with a 'Vary: *' header.");
     return;
   }
-  ToHeadersEntryList(aOut.headers(), headers);
+  aOut.headers() = ToHeadersEntryList(headers);
   aOut.headersGuard() = headers->Guard();
-  aOut.channelInfo() = aIn.GetChannelInfo().AsIPCChannelInfo();
+  aOut.securityInfo() = aIn.GetChannelInfo().SecurityInfo();
   if (aIn.GetPrincipalInfo()) {
-    aOut.principalInfo() = *aIn.GetPrincipalInfo();
+    aOut.principalInfo() = Some(*aIn.GetPrincipalInfo());
   } else {
-    aOut.principalInfo() = void_t();
+    aOut.principalInfo() = Nothing();
   }
 
   aOut.paddingInfo() = aIn.GetPaddingInfo();
   aOut.paddingSize() = aIn.GetPaddingSize();
 }
 
-void TypeUtils::ToCacheResponse(
-    JSContext* aCx, CacheResponse& aOut, Response& aIn,
-    nsTArray<UniquePtr<AutoIPCStream>>& aStreamCleanupList, ErrorResult& aRv) {
+void TypeUtils::ToCacheResponse(JSContext* aCx, CacheResponse& aOut,
+                                Response& aIn, ErrorResult& aRv) {
   if (aIn.BodyUsed()) {
     aRv.ThrowTypeError<MSG_FETCH_BODY_CONSUMED_ERROR>();
     return;
   }
 
-  RefPtr<InternalResponse> ir = aIn.GetInternalResponse();
+  SafeRefPtr<InternalResponse> ir = aIn.GetInternalResponse();
   ToCacheResponseWithoutBody(aOut, *ir, aRv);
   if (NS_WARN_IF(aRv.Failed())) {
     return;
@@ -224,7 +223,7 @@ void TypeUtils::ToCacheResponse(
     }
   }
 
-  SerializeCacheStream(stream, &aOut.body(), aStreamCleanupList, aRv);
+  SerializeCacheStream(stream, &aOut.body(), aRv);
   if (NS_WARN_IF(aRv.Failed())) {
     return;
   }
@@ -236,25 +235,31 @@ void TypeUtils::ToCacheQueryParams(CacheQueryParams& aOut,
   aOut.ignoreSearch() = aIn.mIgnoreSearch;
   aOut.ignoreMethod() = aIn.mIgnoreMethod;
   aOut.ignoreVary() = aIn.mIgnoreVary;
+}
+
+// static
+void TypeUtils::ToCacheQueryParams(CacheQueryParams& aOut,
+                                   const MultiCacheQueryOptions& aIn) {
+  ToCacheQueryParams(aOut, static_cast<const CacheQueryOptions&>(aIn));
   aOut.cacheNameSet() = aIn.mCacheName.WasPassed();
   if (aOut.cacheNameSet()) {
     aOut.cacheName() = aIn.mCacheName.Value();
   } else {
-    aOut.cacheName() = NS_LITERAL_STRING("");
+    aOut.cacheName() = u""_ns;
   }
 }
 
 already_AddRefed<Response> TypeUtils::ToResponse(const CacheResponse& aIn) {
   if (aIn.type() == ResponseType::Error) {
     // We don't bother tracking the internal error code for cached responses...
-    RefPtr<InternalResponse> error =
-        InternalResponse::NetworkError(NS_ERROR_FAILURE);
-    RefPtr<Response> r = new Response(GetGlobalObject(), error, nullptr);
+    RefPtr<Response> r =
+        new Response(GetGlobalObject(),
+                     InternalResponse::NetworkError(NS_ERROR_FAILURE), nullptr);
     return r.forget();
   }
 
-  RefPtr<InternalResponse> ir =
-      new InternalResponse(aIn.status(), aIn.statusText());
+  SafeRefPtr<InternalResponse> ir =
+      MakeSafeRefPtr<InternalResponse>(aIn.status(), aIn.statusText());
   ir->SetURLList(aIn.urlList());
 
   RefPtr<InternalHeaders> internalHeaders =
@@ -268,12 +273,11 @@ already_AddRefed<Response> TypeUtils::ToResponse(const CacheResponse& aIn) {
   ir->Headers()->SetGuard(aIn.headersGuard(), result);
   MOZ_DIAGNOSTIC_ASSERT(!result.Failed());
 
-  ir->InitChannelInfo(aIn.channelInfo());
-  if (aIn.principalInfo().type() ==
-      mozilla::ipc::OptionalPrincipalInfo::TPrincipalInfo) {
-    UniquePtr<mozilla::ipc::PrincipalInfo> info(new mozilla::ipc::PrincipalInfo(
-        aIn.principalInfo().get_PrincipalInfo()));
-    ir->SetPrincipalInfo(Move(info));
+  ir->InitChannelInfo(aIn.securityInfo());
+  if (aIn.principalInfo().isSome()) {
+    UniquePtr<mozilla::ipc::PrincipalInfo> info(
+        new mozilla::ipc::PrincipalInfo(aIn.principalInfo().ref()));
+    ir->SetPrincipalInfo(std::move(info));
   }
 
   nsCOMPtr<nsIInputStream> stream = ReadStream::Create(aIn.body());
@@ -301,15 +305,16 @@ already_AddRefed<Response> TypeUtils::ToResponse(const CacheResponse& aIn) {
 
   ir->SetPaddingSize(aIn.paddingSize());
 
-  RefPtr<Response> ref = new Response(GetGlobalObject(), ir, nullptr);
+  RefPtr<Response> ref =
+      new Response(GetGlobalObject(), std::move(ir), nullptr);
   return ref.forget();
 }
-already_AddRefed<InternalRequest> TypeUtils::ToInternalRequest(
+SafeRefPtr<InternalRequest> TypeUtils::ToInternalRequest(
     const CacheRequest& aIn) {
   nsAutoCString url(aIn.urlWithoutQuery());
   url.Append(aIn.urlQuery());
-  RefPtr<InternalRequest> internalRequest =
-      new InternalRequest(url, aIn.urlFragment());
+  auto internalRequest =
+      MakeSafeRefPtr<InternalRequest>(url, aIn.urlFragment());
   internalRequest->SetMethod(aIn.method());
   internalRequest->SetReferrer(aIn.referrer());
   internalRequest->SetReferrerPolicy(aIn.referrerPolicy());
@@ -336,28 +341,27 @@ already_AddRefed<InternalRequest> TypeUtils::ToInternalRequest(
 
   internalRequest->SetBody(stream, -1);
 
-  return internalRequest.forget();
+  return internalRequest;
 }
 
-already_AddRefed<Request> TypeUtils::ToRequest(const CacheRequest& aIn) {
-  RefPtr<InternalRequest> internalRequest = ToInternalRequest(aIn);
-  RefPtr<Request> request =
-      new Request(GetGlobalObject(), internalRequest, nullptr);
-  return request.forget();
+SafeRefPtr<Request> TypeUtils::ToRequest(const CacheRequest& aIn) {
+  return MakeSafeRefPtr<Request>(GetGlobalObject(), ToInternalRequest(aIn),
+                                 nullptr);
 }
 
 // static
 already_AddRefed<InternalHeaders> TypeUtils::ToInternalHeaders(
     const nsTArray<HeadersEntry>& aHeadersEntryList, HeadersGuardEnum aGuard) {
-  nsTArray<InternalHeaders::Entry> entryList(aHeadersEntryList.Length());
+  nsTArray<InternalHeaders::Entry> entryList;
+  entryList.SetCapacity(aHeadersEntryList.Length());
+  std::transform(aHeadersEntryList.cbegin(), aHeadersEntryList.cend(),
+                 MakeBackInserter(entryList), [](const auto& headersEntry) {
+                   return InternalHeaders::Entry(headersEntry.name(),
+                                                 headersEntry.value());
+                 });
 
-  for (uint32_t i = 0; i < aHeadersEntryList.Length(); ++i) {
-    const HeadersEntry& headersEntry = aHeadersEntryList[i];
-    entryList.AppendElement(
-        InternalHeaders::Entry(headersEntry.name(), headersEntry.value()));
-  }
-
-  RefPtr<InternalHeaders> ref = new InternalHeaders(Move(entryList), aGuard);
+  RefPtr<InternalHeaders> ref =
+      new InternalHeaders(std::move(entryList), aGuard);
   return ref.forget();
 }
 
@@ -386,8 +390,11 @@ void TypeUtils::ProcessURL(nsACString& aUrl, bool* aSchemeValidOut,
 
   if (aSchemeValidOut) {
     nsAutoCString scheme(Substring(flatURL, schemePos, schemeLen));
-    *aSchemeValidOut = scheme.LowerCaseEqualsLiteral("http") ||
-                       scheme.LowerCaseEqualsLiteral("https");
+    *aSchemeValidOut =
+        scheme.LowerCaseEqualsLiteral("http") ||
+        scheme.LowerCaseEqualsLiteral("https") ||
+        (StaticPrefs::extensions_backgroundServiceWorker_enabled_AtStartup() &&
+         scheme.LowerCaseEqualsLiteral("moz-extension"));
   }
 
   uint32_t queryPos;
@@ -408,7 +415,7 @@ void TypeUtils::ProcessURL(nsACString& aUrl, bool* aSchemeValidOut,
 
   if (queryLen < 0) {
     *aUrlWithoutQueryOut = aUrl;
-    *aUrlQueryOut = EmptyCString();
+    aUrlQueryOut->Truncate();
     return;
   }
 
@@ -419,33 +426,31 @@ void TypeUtils::ProcessURL(nsACString& aUrl, bool* aSchemeValidOut,
   *aUrlQueryOut = Substring(aUrl, queryPos - 1, queryLen + 1);
 }
 
-void TypeUtils::CheckAndSetBodyUsed(JSContext* aCx, Request* aRequest,
+void TypeUtils::CheckAndSetBodyUsed(JSContext* aCx, Request& aRequest,
                                     BodyAction aBodyAction, ErrorResult& aRv) {
-  MOZ_DIAGNOSTIC_ASSERT(aRequest);
-
   if (aBodyAction == IgnoreBody) {
     return;
   }
 
-  if (aRequest->BodyUsed()) {
+  if (aRequest.BodyUsed()) {
     aRv.ThrowTypeError<MSG_FETCH_BODY_CONSUMED_ERROR>();
     return;
   }
 
   nsCOMPtr<nsIInputStream> stream;
-  aRequest->GetBody(getter_AddRefs(stream));
+  aRequest.GetBody(getter_AddRefs(stream));
   if (stream) {
-    aRequest->SetBodyUsed(aCx, aRv);
+    aRequest.SetBodyUsed(aCx, aRv);
     if (NS_WARN_IF(aRv.Failed())) {
       return;
     }
   }
 }
 
-already_AddRefed<InternalRequest> TypeUtils::ToInternalRequest(
-    const nsAString& aIn, ErrorResult& aRv) {
+SafeRefPtr<InternalRequest> TypeUtils::ToInternalRequest(const nsAString& aIn,
+                                                         ErrorResult& aRv) {
   RequestOrUSVString requestOrString;
-  requestOrString.SetAsUSVString().Rebind(aIn.Data(), aIn.Length());
+  requestOrString.SetAsUSVString().ShareOrDependUpon(aIn);
 
   // Re-create a GlobalObject stack object so we can use webidl Constructors.
   AutoJSAPI jsapi;
@@ -457,8 +462,9 @@ already_AddRefed<InternalRequest> TypeUtils::ToInternalRequest(
   GlobalObject global(cx, GetGlobalObject()->GetGlobalJSObject());
   MOZ_DIAGNOSTIC_ASSERT(!global.Failed());
 
-  RefPtr<Request> request =
-      Request::Constructor(global, requestOrString, RequestInit(), aRv);
+  RootedDictionary<RequestInit> requestInit(cx);
+  SafeRefPtr<Request> request =
+      Request::Constructor(global, requestOrString, requestInit, aRv);
   if (NS_WARN_IF(aRv.Failed())) {
     return nullptr;
   }
@@ -466,32 +472,28 @@ already_AddRefed<InternalRequest> TypeUtils::ToInternalRequest(
   return request->GetInternalRequest();
 }
 
-void TypeUtils::SerializeCacheStream(
-    nsIInputStream* aStream, CacheReadStreamOrVoid* aStreamOut,
-    nsTArray<UniquePtr<AutoIPCStream>>& aStreamCleanupList, ErrorResult& aRv) {
-  *aStreamOut = void_t();
+void TypeUtils::SerializeCacheStream(nsIInputStream* aStream,
+                                     Maybe<CacheReadStream>* aStreamOut,
+                                     ErrorResult& aRv) {
+  *aStreamOut = Nothing();
   if (!aStream) {
     return;
   }
 
   RefPtr<ReadStream> controlled = do_QueryObject(aStream);
   if (controlled) {
-    controlled->Serialize(aStreamOut, aStreamCleanupList, aRv);
+    controlled->Serialize(aStreamOut, aRv);
     return;
   }
 
-  *aStreamOut = CacheReadStream();
-  CacheReadStream& cacheStream = aStreamOut->get_CacheReadStream();
+  aStreamOut->emplace(CacheReadStream());
+  CacheReadStream& cacheStream = aStreamOut->ref();
 
-  cacheStream.controlChild() = nullptr;
-  cacheStream.controlParent() = nullptr;
+  cacheStream.control() = nullptr;
 
-  UniquePtr<AutoIPCStream> autoStream(new AutoIPCStream(cacheStream.stream()));
-  autoStream->Serialize(aStream, GetIPCManager());
-
-  aStreamCleanupList.AppendElement(Move(autoStream));
+  MOZ_ALWAYS_TRUE(mozilla::ipc::SerializeIPCStream(do_AddRef(aStream),
+                                                   cacheStream.stream(),
+                                                   /* aAllowLazy */ false));
 }
 
-}  // namespace cache
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom::cache
