@@ -1,4 +1,4 @@
-/* -*- Mode: C++; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
+/* -*- Mode: C++; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
 /* This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
@@ -8,15 +8,13 @@
 #include <algorithm>
 #include "GLSLANG/ShaderLang.h"
 #include "CanvasUtils.h"
-#include "gfxPrefs.h"
+#include "gfxEnv.h"
 #include "GLContext.h"
 #include "jsfriendapi.h"
 #include "mozilla/CheckedInt.h"
 #include "mozilla/Preferences.h"
-#include "mozilla/Services.h"
-#include "nsIObserverService.h"
+#include "mozilla/StaticPrefs_webgl.h"
 #include "nsPrintfCString.h"
-#include "WebGLActiveInfo.h"
 #include "WebGLBuffer.h"
 #include "WebGLContextUtils.h"
 #include "WebGLFramebuffer.h"
@@ -25,13 +23,11 @@
 #include "WebGLSampler.h"
 #include "WebGLShader.h"
 #include "WebGLTexture.h"
-#include "WebGLUniformLocation.h"
 #include "WebGLValidateStrings.h"
 #include "WebGLVertexArray.h"
-#include "WebGLVertexAttribData.h"
 
 #if defined(MOZ_WIDGET_COCOA)
-#include "nsCocoaFeatures.h"
+#  include "nsCocoaFeatures.h"
 #endif
 
 ////////////////////
@@ -45,7 +41,6 @@ const uint32_t kMinMaxVertexTextureImageUnits = 0;    // Page 164
 const uint32_t kMinMaxFragmentTextureImageUnits = 8;  // Page 164
 const uint32_t kMinMaxCombinedTextureImageUnits = 8;  // Page 164
 
-const uint32_t kMinMaxColorAttachments = 4;
 const uint32_t kMinMaxDrawBuffers = 4;
 
 // These few deviate from the spec: (The minimum values in the spec are
@@ -142,24 +137,6 @@ bool WebGLContext::ValidateBlendFuncEnumsCompatibility(GLenum sfactor,
   return true;
 }
 
-bool WebGLContext::ValidateComparisonEnum(GLenum target, const char* info) {
-  switch (target) {
-    case LOCAL_GL_NEVER:
-    case LOCAL_GL_LESS:
-    case LOCAL_GL_LEQUAL:
-    case LOCAL_GL_GREATER:
-    case LOCAL_GL_GEQUAL:
-    case LOCAL_GL_EQUAL:
-    case LOCAL_GL_NOTEQUAL:
-    case LOCAL_GL_ALWAYS:
-      return true;
-
-    default:
-      ErrorInvalidEnumInfo(info, target);
-      return false;
-  }
-}
-
 bool WebGLContext::ValidateStencilOpEnum(GLenum action, const char* info) {
   switch (action) {
     case LOCAL_GL_KEEP:
@@ -178,7 +155,7 @@ bool WebGLContext::ValidateStencilOpEnum(GLenum action, const char* info) {
   }
 }
 
-bool WebGLContext::ValidateFaceEnum(GLenum face, const char* info) {
+bool WebGLContext::ValidateFaceEnum(const GLenum face) {
   switch (face) {
     case LOCAL_GL_FRONT:
     case LOCAL_GL_BACK:
@@ -186,173 +163,101 @@ bool WebGLContext::ValidateFaceEnum(GLenum face, const char* info) {
       return true;
 
     default:
-      ErrorInvalidEnumInfo(info, face);
+      ErrorInvalidEnumInfo("face", face);
       return false;
   }
 }
 
-bool WebGLContext::ValidateDrawModeEnum(GLenum mode, const char* info) {
-  switch (mode) {
-    case LOCAL_GL_TRIANGLES:
-    case LOCAL_GL_TRIANGLE_STRIP:
-    case LOCAL_GL_TRIANGLE_FAN:
-    case LOCAL_GL_POINTS:
-    case LOCAL_GL_LINE_STRIP:
-    case LOCAL_GL_LINE_LOOP:
-    case LOCAL_GL_LINES:
-      return true;
-
-    default:
-      ErrorInvalidEnumInfo(info, mode);
-      return false;
-  }
-}
-
-bool WebGLContext::ValidateUniformLocation(WebGLUniformLocation* loc,
-                                           const char* funcName) {
-  /* GLES 2.0.25, p38:
-   *   If the value of location is -1, the Uniform* commands will silently
-   *   ignore the data passed in, and the current uniform values will not be
-   *   changed.
-   */
-  if (!loc) return false;
-
-  if (!ValidateObjectAllowDeleted(funcName, *loc)) return false;
-
-  if (!mCurrentProgram) {
-    ErrorInvalidOperation("%s: No program is currently bound.", funcName);
-    return false;
-  }
-
-  return loc->ValidateForProgram(mCurrentProgram, funcName);
-}
-
-bool WebGLContext::ValidateAttribArraySetter(const char* name,
-                                             uint32_t setterElemSize,
+bool WebGLContext::ValidateAttribArraySetter(uint32_t setterElemSize,
                                              uint32_t arrayLength) {
   if (IsContextLost()) return false;
 
   if (arrayLength < setterElemSize) {
-    ErrorInvalidValue("%s: Array must have >= %d elements.", name,
-                      setterElemSize);
+    ErrorInvalidValue("Array must have >= %d elements.", setterElemSize);
     return false;
   }
 
   return true;
 }
 
-bool WebGLContext::ValidateUniformSetter(WebGLUniformLocation* loc,
-                                         uint8_t setterElemSize,
-                                         GLenum setterType,
-                                         const char* funcName) {
-  if (IsContextLost()) return false;
+// ---------------------
 
-  if (!ValidateUniformLocation(loc, funcName)) return false;
+static webgl::Limits MakeLimits(const WebGLContext& webgl) {
+  webgl::Limits limits;
 
-  if (!loc->ValidateSizeAndType(setterElemSize, setterType, funcName))
-    return false;
+  gl::GLContext& gl = *webgl.GL();
 
-  return true;
-}
+  // -
 
-bool WebGLContext::ValidateUniformArraySetter(
-    WebGLUniformLocation* loc, uint8_t setterElemSize, GLenum setterType,
-    uint32_t setterArraySize, const char* funcName,
-    uint32_t* const out_numElementsToUpload) {
-  if (IsContextLost()) return false;
-
-  if (!ValidateUniformLocation(loc, funcName)) return false;
-
-  if (!loc->ValidateSizeAndType(setterElemSize, setterType, funcName))
-    return false;
-
-  if (!loc->ValidateArrayLength(setterElemSize, setterArraySize, funcName))
-    return false;
-
-  const auto& elemCount = loc->mInfo->mActiveInfo->mElemCount;
-  MOZ_ASSERT(elemCount > loc->mArrayIndex);
-  const uint32_t uniformElemCount = elemCount - loc->mArrayIndex;
-
-  *out_numElementsToUpload =
-      std::min(uniformElemCount, setterArraySize / setterElemSize);
-  return true;
-}
-
-bool WebGLContext::ValidateUniformMatrixArraySetter(
-    WebGLUniformLocation* loc, uint8_t setterCols, uint8_t setterRows,
-    GLenum setterType, uint32_t setterArraySize, bool setterTranspose,
-    const char* funcName, uint32_t* const out_numElementsToUpload) {
-  const uint8_t setterElemSize = setterCols * setterRows;
-
-  if (IsContextLost()) return false;
-
-  if (!ValidateUniformLocation(loc, funcName)) return false;
-
-  if (!loc->ValidateSizeAndType(setterElemSize, setterType, funcName))
-    return false;
-
-  if (!loc->ValidateArrayLength(setterElemSize, setterArraySize, funcName))
-    return false;
-
-  if (setterTranspose && !IsWebGL2()) {
-    ErrorInvalidValue("%s: `transpose` must be false.", funcName);
-    return false;
+  for (const auto i : IntegerRange(UnderlyingValue(WebGLExtensionID::Max))) {
+    const auto ext = WebGLExtensionID(i);
+    limits.supportedExtensions[ext] = webgl.IsExtensionSupported(ext);
   }
 
-  const auto& elemCount = loc->mInfo->mActiveInfo->mElemCount;
-  MOZ_ASSERT(elemCount > loc->mArrayIndex);
-  const uint32_t uniformElemCount = elemCount - loc->mArrayIndex;
+  // -
+  // WebGL 1
 
-  *out_numElementsToUpload =
-      std::min(uniformElemCount, setterArraySize / setterElemSize);
-  return true;
-}
+  // Note: GL_MAX_TEXTURE_UNITS is fixed at 4 for most desktop hardware,
+  // even though the hardware supports much more.  The
+  // GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS value is the accurate value.
+  gl.GetUIntegerv(LOCAL_GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS,
+                  &limits.maxTexUnits);
+  limits.maxTexUnits = std::min(
+      limits.maxTexUnits, uint32_t{UINT8_MAX});  // We want to use uint8_t.
 
-bool WebGLContext::ValidateAttribIndex(GLuint index, const char* info) {
-  bool valid = (index < MaxVertexAttribs());
+  gl.GetUIntegerv(LOCAL_GL_MAX_TEXTURE_SIZE, &limits.maxTex2dSize);
+  gl.GetUIntegerv(LOCAL_GL_MAX_CUBE_MAP_TEXTURE_SIZE, &limits.maxTexCubeSize);
+  gl.GetUIntegerv(LOCAL_GL_MAX_VERTEX_ATTRIBS, &limits.maxVertexAttribs);
 
-  if (!valid) {
-    if (index == GLuint(-1)) {
-      ErrorInvalidValue(
-          "%s: -1 is not a valid `index`. This value"
-          " probably comes from a getAttribLocation()"
-          " call, where this return value -1 means"
-          " that the passed name didn't correspond to"
-          " an active attribute in the specified"
-          " program.",
-          info);
-    } else {
-      ErrorInvalidValue(
-          "%s: `index` must be less than"
-          " MAX_VERTEX_ATTRIBS.",
-          info);
-    }
+  auto dims = std::array<uint32_t, 2>{};
+  gl.GetUIntegerv(LOCAL_GL_MAX_VIEWPORT_DIMS, dims.data());
+  limits.maxViewportDim = std::min(dims[0], dims[1]);
+
+  if (!gl.IsCoreProfile()) {
+    gl.fGetFloatv(LOCAL_GL_ALIASED_LINE_WIDTH_RANGE,
+                  limits.lineWidthRange.data());
   }
 
-  return valid;
-}
-
-bool WebGLContext::ValidateStencilParamsForDrawCall() {
-  const char msg[] =
-      "%s set different front and back stencil %s. Drawing in"
-      " this configuration is not allowed.";
-
-  if (mStencilRefFront != mStencilRefBack) {
-    ErrorInvalidOperation(msg, "stencilFuncSeparate", "reference values");
-    return false;
+  {
+    const GLenum driverPName = gl.IsCoreProfile()
+                                   ? LOCAL_GL_POINT_SIZE_RANGE
+                                   : LOCAL_GL_ALIASED_POINT_SIZE_RANGE;
+    gl.fGetFloatv(driverPName, limits.pointSizeRange.data());
   }
 
-  if (mStencilValueMaskFront != mStencilValueMaskBack) {
-    ErrorInvalidOperation(msg, "stencilFuncSeparate", "value masks");
-    return false;
+  if (webgl.IsWebGL2()) {
+    gl.GetUIntegerv(LOCAL_GL_MAX_ARRAY_TEXTURE_LAYERS,
+                    &limits.maxTexArrayLayers);
+    gl.GetUIntegerv(LOCAL_GL_MAX_3D_TEXTURE_SIZE, &limits.maxTex3dSize);
+    gl.GetUIntegerv(LOCAL_GL_MAX_UNIFORM_BUFFER_BINDINGS,
+                    &limits.maxUniformBufferBindings);
+    gl.GetUIntegerv(LOCAL_GL_UNIFORM_BUFFER_OFFSET_ALIGNMENT,
+                    &limits.uniformBufferOffsetAlignment);
   }
 
-  if (mStencilWriteMaskFront != mStencilWriteMaskBack) {
-    ErrorInvalidOperation(msg, "stencilMaskSeparate", "write masks");
-    return false;
+  if (limits.supportedExtensions
+          [WebGLExtensionID::WEBGL_compressed_texture_astc]) {
+    limits.astcHdr = gl.IsExtensionSupported(
+        gl::GLContext::KHR_texture_compression_astc_hdr);
   }
 
-  return true;
+  if (webgl.IsWebGL2() ||
+      limits.supportedExtensions[WebGLExtensionID::WEBGL_draw_buffers]) {
+    gl.GetUIntegerv(LOCAL_GL_MAX_DRAW_BUFFERS, &limits.maxColorDrawBuffers);
+  }
+
+  if (limits.supportedExtensions[WebGLExtensionID::EXT_disjoint_timer_query]) {
+    gl.fGetQueryiv(LOCAL_GL_TIME_ELAPSED_EXT, LOCAL_GL_QUERY_COUNTER_BITS,
+                   (int32_t*)&limits.queryCounterBitsTimeElapsed);
+    gl.fGetQueryiv(LOCAL_GL_TIMESTAMP_EXT, LOCAL_GL_QUERY_COUNTER_BITS,
+                   (int32_t*)&limits.queryCounterBitsTimestamp);
+  }
+
+  if (limits.supportedExtensions[WebGLExtensionID::OVR_multiview2]) {
+    gl.GetUIntegerv(LOCAL_GL_MAX_VIEWS_OVR, &limits.maxMultiviewLayers);
+  }
+
+  return limits;
 }
 
 bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
@@ -378,26 +283,10 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
     return false;
   }
 
-  mDisableExtensions = gfxPrefs::WebGLDisableExtensions();
-  mLoseContextOnMemoryPressure = gfxPrefs::WebGLLoseContextOnMemoryPressure();
-  mCanLoseContextInForeground = gfxPrefs::WebGLCanLoseContextInForeground();
-  mRestoreWhenVisible = gfxPrefs::WebGLRestoreWhenVisible();
-
-  // These are the default values, see 6.2 State tables in the
-  // OpenGL ES 2.0.25 spec.
-  mColorWriteMask = 0x0f;
-  mDriverColorMask = mColorWriteMask;
-  mColorClearValue[0] = 0.f;
-  mColorClearValue[1] = 0.f;
-  mColorClearValue[2] = 0.f;
-  mColorClearValue[3] = 0.f;
-  mDepthWriteMask = true;
-  mDepthClearValue = 1.f;
-  mStencilClearValue = 0;
-  mStencilRefFront = 0;
-  mStencilRefBack = 0;
-
-  mLineWidth = 1.0;
+  mLoseContextOnMemoryPressure =
+      StaticPrefs::webgl_lose_context_on_memory_pressure();
+  mCanLoseContextInForeground =
+      StaticPrefs::webgl_can_lose_context_in_foreground();
 
   /*
   // Technically, we should be setting mStencil[...] values to
@@ -425,25 +314,7 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
   AssertUintParamCorrect(gl, LOCAL_GL_STENCIL_BACK_WRITEMASK,
                          mStencilWriteMaskBack);
 
-  mDitherEnabled = true;
-  mRasterizerDiscardEnabled = false;
-  mScissorTestEnabled = false;
-
-  mDepthTestEnabled = 0;
-  mDriverDepthTest = false;
-  mStencilTestEnabled = 0;
-  mDriverStencilTest = false;
-
-  mGenerateMipmapHint = LOCAL_GL_DONT_CARE;
-
   // Bindings, etc.
-  mActiveTexture = 0;
-  mDefaultFB_DrawBuffer0 = LOCAL_GL_BACK;
-  mDefaultFB_ReadBuffer = LOCAL_GL_BACK;
-
-  mEmitContextLostErrorOnce = true;
-  mWebGLError = LOCAL_GL_NO_ERROR;
-  mUnderlyingGLError = LOCAL_GL_NO_ERROR;
 
   mBound2DTextures.Clear();
   mBoundCubeMapTextures.Clear();
@@ -456,54 +327,36 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
 
   mBoundDrawFramebuffer = nullptr;
   mBoundReadFramebuffer = nullptr;
-  mBoundRenderbuffer = nullptr;
 
-  gl->GetUIntegerv(LOCAL_GL_MAX_VERTEX_ATTRIBS, &mGLMaxVertexAttribs);
+  // -----------------------
 
-  if (mGLMaxVertexAttribs < 8) {
+  auto limits = MakeLimits(*this);
+
+  // -
+
+  if (limits.maxVertexAttribs < 8) {
     const nsPrintfCString reason("GL_MAX_VERTEX_ATTRIBS: %d is < 8!",
-                                 mGLMaxVertexAttribs);
+                                 limits.maxVertexAttribs);
     *out_failReason = {"FEATURE_FAILURE_WEBGL_V_ATRB", reason};
     return false;
   }
 
-  // Note: GL_MAX_TEXTURE_UNITS is fixed at 4 for most desktop hardware,
-  // even though the hardware supports much more.  The
-  // GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS value is the accurate value.
-  mGLMaxCombinedTextureImageUnits =
-      gl->GetIntAs<GLuint>(LOCAL_GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS);
-  mGLMaxTextureUnits = mGLMaxCombinedTextureImageUnits;
-
-  if (mGLMaxCombinedTextureImageUnits < 8) {
+  if (limits.maxTexUnits < 8) {
     const nsPrintfCString reason(
-        "GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS: %u is < 8!", mGLMaxTextureUnits);
+        "GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS: %u is < 8!", limits.maxTexUnits);
     *out_failReason = {"FEATURE_FAILURE_WEBGL_T_UNIT", reason};
     return false;
   }
 
-  mBound2DTextures.SetLength(mGLMaxTextureUnits);
-  mBoundCubeMapTextures.SetLength(mGLMaxTextureUnits);
-  mBound3DTextures.SetLength(mGLMaxTextureUnits);
-  mBound2DArrayTextures.SetLength(mGLMaxTextureUnits);
-  mBoundSamplers.SetLength(mGLMaxTextureUnits);
-
-  gl->fGetIntegerv(LOCAL_GL_MAX_VIEWPORT_DIMS, (GLint*)mGLMaxViewportDims);
+  mBound2DTextures.SetLength(limits.maxTexUnits);
+  mBoundCubeMapTextures.SetLength(limits.maxTexUnits);
+  mBound3DTextures.SetLength(limits.maxTexUnits);
+  mBound2DArrayTextures.SetLength(limits.maxTexUnits);
+  mBoundSamplers.SetLength(limits.maxTexUnits);
 
   ////////////////
 
-  gl->fGetIntegerv(LOCAL_GL_MAX_TEXTURE_SIZE, (GLint*)&mGLMaxTextureSize);
-  gl->fGetIntegerv(LOCAL_GL_MAX_CUBE_MAP_TEXTURE_SIZE,
-                   (GLint*)&mGLMaxCubeMapTextureSize);
-  gl->fGetIntegerv(LOCAL_GL_MAX_RENDERBUFFER_SIZE,
-                   (GLint*)&mGLMaxRenderbufferSize);
-
-  if (!gl->GetPotentialInteger(LOCAL_GL_MAX_3D_TEXTURE_SIZE,
-                               (GLint*)&mGLMax3DTextureSize))
-    mGLMax3DTextureSize = 0;
-  if (!gl->GetPotentialInteger(LOCAL_GL_MAX_ARRAY_TEXTURE_LAYERS,
-                               (GLint*)&mGLMaxArrayTextureLayers))
-    mGLMaxArrayTextureLayers = 0;
-
+  gl->GetUIntegerv(LOCAL_GL_MAX_RENDERBUFFER_SIZE, &mGLMaxRenderbufferSize);
   gl->GetUIntegerv(LOCAL_GL_MAX_TEXTURE_IMAGE_UNITS,
                    &mGLMaxFragmentTextureImageUnits);
   gl->GetUIntegerv(LOCAL_GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS,
@@ -511,91 +364,65 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
 
   ////////////////
 
-  mGLMaxColorAttachments = 1;
-  mGLMaxDrawBuffers = 1;
-
-  if (IsWebGL2()) {
-    UpdateMaxDrawBuffers();
-  }
-
-  ////////////////
-
-  if (gl->IsSupported(gl::GLFeature::ES2_compatibility)) {
-    gl->GetUIntegerv(LOCAL_GL_MAX_FRAGMENT_UNIFORM_VECTORS,
-                     &mGLMaxFragmentUniformVectors);
-    gl->GetUIntegerv(LOCAL_GL_MAX_VERTEX_UNIFORM_VECTORS,
-                     &mGLMaxVertexUniformVectors);
-    gl->GetUIntegerv(LOCAL_GL_MAX_VARYING_VECTORS, &mGLMaxVaryingVectors);
-  } else {
-    gl->GetUIntegerv(LOCAL_GL_MAX_FRAGMENT_UNIFORM_COMPONENTS,
-                     &mGLMaxFragmentUniformVectors);
-    mGLMaxFragmentUniformVectors /= 4;
-    gl->GetUIntegerv(LOCAL_GL_MAX_VERTEX_UNIFORM_COMPONENTS,
-                     &mGLMaxVertexUniformVectors);
-    mGLMaxVertexUniformVectors /= 4;
-
-    /* We are now going to try to read GL_MAX_VERTEX_OUTPUT_COMPONENTS
-     * and GL_MAX_FRAGMENT_INPUT_COMPONENTS, however these constants
-     * only entered the OpenGL standard at OpenGL 3.2. So we will try
-     * reading, and check OpenGL error for INVALID_ENUM.
-     *
-     * On the public_webgl list, "problematic GetParameter pnames"
-     * thread, the following formula was given:
-     *   maxVaryingVectors = min(GL_MAX_VERTEX_OUTPUT_COMPONENTS,
-     *                           GL_MAX_FRAGMENT_INPUT_COMPONENTS) / 4
-     */
-    uint32_t maxVertexOutputComponents = 0;
-    uint32_t maxFragmentInputComponents = 0;
-    bool ok = true;
-    ok &= gl->GetPotentialInteger(LOCAL_GL_MAX_VERTEX_OUTPUT_COMPONENTS,
-                                  (GLint*)&maxVertexOutputComponents);
-    ok &= gl->GetPotentialInteger(LOCAL_GL_MAX_FRAGMENT_INPUT_COMPONENTS,
-                                  (GLint*)&maxFragmentInputComponents);
-    if (ok) {
-      mGLMaxVaryingVectors =
-          std::min(maxVertexOutputComponents, maxFragmentInputComponents) / 4;
+  if (gl->IsGLES()) {
+    mGLMaxFragmentUniformVectors =
+        gl->GetIntAs<uint32_t>(LOCAL_GL_MAX_FRAGMENT_UNIFORM_VECTORS);
+    mGLMaxVertexUniformVectors =
+        gl->GetIntAs<uint32_t>(LOCAL_GL_MAX_VERTEX_UNIFORM_VECTORS);
+    if (gl->Version() >= 300) {
+      mGLMaxVertexOutputVectors =
+          gl->GetIntAs<uint32_t>(LOCAL_GL_MAX_VERTEX_OUTPUT_COMPONENTS) / 4;
+      mGLMaxFragmentInputVectors =
+          gl->GetIntAs<uint32_t>(LOCAL_GL_MAX_FRAGMENT_INPUT_COMPONENTS) / 4;
     } else {
-      mGLMaxVaryingVectors = 16;
-      // 16 = 64/4, and 64 is the min value for
-      // maxVertexOutputComponents in the OpenGL 3.2 spec.
+      mGLMaxFragmentInputVectors =
+          gl->GetIntAs<uint32_t>(LOCAL_GL_MAX_VARYING_VECTORS);
+      mGLMaxVertexOutputVectors = mGLMaxFragmentInputVectors;
+    }
+  } else {
+    mGLMaxFragmentUniformVectors =
+        gl->GetIntAs<uint32_t>(LOCAL_GL_MAX_FRAGMENT_UNIFORM_COMPONENTS) / 4;
+    mGLMaxVertexUniformVectors =
+        gl->GetIntAs<uint32_t>(LOCAL_GL_MAX_VERTEX_UNIFORM_COMPONENTS) / 4;
+
+    if (gl->Version() >= 320) {
+      mGLMaxVertexOutputVectors =
+          gl->GetIntAs<uint32_t>(LOCAL_GL_MAX_VERTEX_OUTPUT_COMPONENTS) / 4;
+      mGLMaxFragmentInputVectors =
+          gl->GetIntAs<uint32_t>(LOCAL_GL_MAX_FRAGMENT_INPUT_COMPONENTS) / 4;
+    } else {
+      // Same enum val as GL2's GL_MAX_VARYING_FLOATS.
+      mGLMaxFragmentInputVectors =
+          gl->GetIntAs<uint32_t>(LOCAL_GL_MAX_VARYING_COMPONENTS) / 4;
+      mGLMaxVertexOutputVectors = mGLMaxFragmentInputVectors;
     }
   }
 
   ////////////////
 
-  gl->fGetFloatv(LOCAL_GL_ALIASED_LINE_WIDTH_RANGE, mGLAliasedLineWidthRange);
-
-  const GLenum driverPName = gl->IsCoreProfile()
-                                 ? LOCAL_GL_POINT_SIZE_RANGE
-                                 : LOCAL_GL_ALIASED_POINT_SIZE_RANGE;
-  gl->fGetFloatv(driverPName, mGLAliasedPointSizeRange);
-
-  ////////////////
-
-  if (gfxPrefs::WebGLMinCapabilityMode()) {
+  if (StaticPrefs::webgl_min_capability_mode()) {
     bool ok = true;
 
     ok &= RestrictCap(&mGLMaxVertexTextureImageUnits,
                       kMinMaxVertexTextureImageUnits);
     ok &= RestrictCap(&mGLMaxFragmentTextureImageUnits,
                       kMinMaxFragmentTextureImageUnits);
-    ok &= RestrictCap(&mGLMaxCombinedTextureImageUnits,
-                      kMinMaxCombinedTextureImageUnits);
+    ok &= RestrictCap(&limits.maxTexUnits, kMinMaxCombinedTextureImageUnits);
 
-    ok &= RestrictCap(&mGLMaxVertexAttribs, kMinMaxVertexAttribs);
+    ok &= RestrictCap(&limits.maxVertexAttribs, kMinMaxVertexAttribs);
     ok &= RestrictCap(&mGLMaxVertexUniformVectors, kMinMaxVertexUniformVectors);
     ok &= RestrictCap(&mGLMaxFragmentUniformVectors,
                       kMinMaxFragmentUniformVectors);
-    ok &= RestrictCap(&mGLMaxVaryingVectors, kMinMaxVaryingVectors);
+    ok &= RestrictCap(&mGLMaxVertexOutputVectors, kMinMaxVaryingVectors);
+    ok &= RestrictCap(&mGLMaxFragmentInputVectors, kMinMaxVaryingVectors);
 
-    ok &= RestrictCap(&mGLMaxColorAttachments, kMinMaxColorAttachments);
-    ok &= RestrictCap(&mGLMaxDrawBuffers, kMinMaxDrawBuffers);
+    ok &= RestrictCap(&limits.maxColorDrawBuffers, kMinMaxDrawBuffers);
 
-    ok &= RestrictCap(&mGLMaxTextureSize, kMinMaxTextureSize);
-    ok &= RestrictCap(&mGLMaxCubeMapTextureSize, kMinMaxCubeMapTextureSize);
-    ok &= RestrictCap(&mGLMax3DTextureSize, kMinMax3DTextureSize);
+    ok &= RestrictCap(&limits.maxTex2dSize, kMinMaxTextureSize);
+    ok &= RestrictCap(&limits.maxTexCubeSize, kMinMaxCubeMapTextureSize);
+    ok &= RestrictCap(&limits.maxTex3dSize, kMinMax3DTextureSize);
 
-    ok &= RestrictCap(&mGLMaxArrayTextureLayers, kMinMaxArrayTextureLayers);
+    ok &= RestrictCap(&limits.maxTexArrayLayers, kMinMaxArrayTextureLayers);
     ok &= RestrictCap(&mGLMaxRenderbufferSize, kMinMaxRenderbufferSize);
 
     if (!ok) {
@@ -604,38 +431,43 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
     }
 
     mDisableFragHighP = true;
-  } else if (nsContentUtils::ShouldResistFingerprinting()) {
+  } else if (mResistFingerprinting) {
     bool ok = true;
 
-    ok &= RestrictCap(&mGLMaxTextureSize, kCommonMaxTextureSize);
-    ok &= RestrictCap(&mGLMaxCubeMapTextureSize, kCommonMaxCubeMapTextureSize);
+    ok &= RestrictCap(&limits.maxTex2dSize, kCommonMaxTextureSize);
+    ok &= RestrictCap(&limits.maxTexCubeSize, kCommonMaxCubeMapTextureSize);
     ok &= RestrictCap(&mGLMaxRenderbufferSize, kCommonMaxRenderbufferSize);
 
     ok &= RestrictCap(&mGLMaxVertexTextureImageUnits,
                       kCommonMaxVertexTextureImageUnits);
     ok &= RestrictCap(&mGLMaxFragmentTextureImageUnits,
                       kCommonMaxFragmentTextureImageUnits);
-    ok &= RestrictCap(&mGLMaxCombinedTextureImageUnits,
-                      kCommonMaxCombinedTextureImageUnits);
+    ok &= RestrictCap(&limits.maxTexUnits, kCommonMaxCombinedTextureImageUnits);
 
-    ok &= RestrictCap(&mGLMaxVertexAttribs, kCommonMaxVertexAttribs);
+    ok &= RestrictCap(&limits.maxVertexAttribs, kCommonMaxVertexAttribs);
     ok &= RestrictCap(&mGLMaxVertexUniformVectors,
                       kCommonMaxVertexUniformVectors);
     ok &= RestrictCap(&mGLMaxFragmentUniformVectors,
                       kCommonMaxFragmentUniformVectors);
-    ok &= RestrictCap(&mGLMaxVaryingVectors, kCommonMaxVaryingVectors);
+    ok &= RestrictCap(&mGLMaxVertexOutputVectors, kCommonMaxVaryingVectors);
+    ok &= RestrictCap(&mGLMaxFragmentInputVectors, kCommonMaxVaryingVectors);
 
-    ok &= RestrictCap(&mGLAliasedLineWidthRange[0],
-                      kCommonAliasedLineWidthRangeMin);
-    ok &= RestrictCap(&mGLAliasedLineWidthRange[1],
-                      kCommonAliasedLineWidthRangeMax);
-    ok &= RestrictCap(&mGLAliasedPointSizeRange[0],
-                      kCommonAliasedPointSizeRangeMin);
-    ok &= RestrictCap(&mGLAliasedPointSizeRange[1],
-                      kCommonAliasedPointSizeRangeMax);
+    if (limits.lineWidthRange[0] <= kCommonAliasedLineWidthRangeMin) {
+      limits.lineWidthRange[0] = kCommonAliasedLineWidthRangeMin;
+    } else {
+      ok = false;
+    }
+    if (limits.pointSizeRange[0] <= kCommonAliasedPointSizeRangeMin) {
+      limits.pointSizeRange[0] = kCommonAliasedPointSizeRangeMin;
+    } else {
+      ok = false;
+    }
 
-    ok &= RestrictCap(&mGLMaxViewportDims[0], kCommonMaxViewportDims);
-    ok &= RestrictCap(&mGLMaxViewportDims[1], kCommonMaxViewportDims);
+    ok &=
+        RestrictCap(&limits.lineWidthRange[1], kCommonAliasedLineWidthRangeMax);
+    ok &=
+        RestrictCap(&limits.pointSizeRange[1], kCommonAliasedPointSizeRangeMax);
+    ok &= RestrictCap(&limits.maxViewportDim, kCommonMaxViewportDims);
 
     if (!ok) {
       GenerateWarning(
@@ -644,6 +476,8 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
     }
   }
 
+  mLimits = Some(limits);
+
   ////////////////
 
   if (gl->IsCompatibilityProfile()) {
@@ -668,9 +502,6 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
     gl->fEnable(LOCAL_GL_TEXTURE_CUBE_MAP_SEAMLESS);
   }
 
-  // Check the shader validator pref
-  mBypassShaderValidation = gfxPrefs::WebGLBypassShaderValidator();
-
   // initialize shader translator
   if (!sh::Initialize()) {
     *out_failReason = {"FEATURE_FAILURE_WEBGL_GLSL",
@@ -701,16 +532,6 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
     return false;
   }
 
-  if (!gl->IsSupported(GLFeature::vertex_array_object)) {
-    *out_failReason = {"FEATURE_FAILURE_WEBGL_VAOS",
-                       "Requires vertex_array_object."};
-    return false;
-  }
-
-  mDefaultVertexArray = WebGLVertexArray::Create(this);
-  mDefaultVertexArray->mAttribs.SetLength(mGLMaxVertexAttribs);
-  mBoundVertexArray = mDefaultVertexArray;
-
   // OpenGL core profiles remove the default VAO object from version
   // 4.0.0. We create a default VAO for all core profiles,
   // regardless of version.
@@ -719,34 +540,15 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
   // (https://www.opengl.org/registry/doc/glspec40.core.20100311.pdf)
   // in Section E.2.2 "Removed Features", pg 397: "[...] The default
   // vertex array object (the name zero) is also deprecated. [...]"
-
-  if (gl->IsCoreProfile()) {
-    mDefaultVertexArray->GenVertexArray();
-    mDefaultVertexArray->BindVertexArray();
-  }
-
-  mPixelStore_FlipY = false;
-  mPixelStore_PremultiplyAlpha = false;
-  mPixelStore_ColorspaceConversion = BROWSER_DEFAULT_WEBGL;
-  mPixelStore_RequireFastPath = false;
-
-  // GLES 3.0.4, p259:
-  mPixelStore_UnpackImageHeight = 0;
-  mPixelStore_UnpackSkipImages = 0;
-  mPixelStore_UnpackRowLength = 0;
-  mPixelStore_UnpackSkipRows = 0;
-  mPixelStore_UnpackSkipPixels = 0;
-  mPixelStore_UnpackAlignment = 4;
-  mPixelStore_PackRowLength = 0;
-  mPixelStore_PackSkipRows = 0;
-  mPixelStore_PackSkipPixels = 0;
-  mPixelStore_PackAlignment = 4;
+  mDefaultVertexArray = WebGLVertexArray::Create(this);
+  mDefaultVertexArray->BindVertexArray();
 
   mPrimRestartTypeBytes = 0;
 
-  mGenericVertexAttribTypes.reset(new GLenum[mGLMaxVertexAttribs]);
-  std::fill_n(mGenericVertexAttribTypes.get(), mGLMaxVertexAttribs,
-              LOCAL_GL_FLOAT);
+  // -
+
+  mGenericVertexAttribTypes.assign(limits.maxVertexAttribs,
+                                   webgl::AttribBaseType::Float);
   mGenericVertexAttribTypeInvalidator.InvalidateCaches();
 
   static const float kDefaultGenericVertexAttribData[4] = {0, 0, 0, 1};
@@ -755,9 +557,36 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
 
   mFakeVertexAttrib0BufferObject = 0;
 
+  mNeedsLegacyVertexAttrib0Handling = gl->IsCompatibilityProfile();
+  if (gl->WorkAroundDriverBugs() && kIsMacOS) {
+    // Failures in conformance/attribs/gl-disabled-vertex-attrib.
+    // Even in Core profiles on NV. Sigh.
+    mNeedsLegacyVertexAttrib0Handling |= (gl->Vendor() == gl::GLVendor::NVIDIA);
+
+    mBug_DrawArraysInstancedUserAttribFetchAffectedByFirst |=
+        (gl->Vendor() == gl::GLVendor::Intel);
+
+    // Failures for programs with no attribs:
+    // conformance/attribs/gl-vertex-attrib-unconsumed-out-of-bounds.html
+    mMaybeNeedsLegacyVertexAttrib0Handling = true;
+  }
+  mMaybeNeedsLegacyVertexAttrib0Handling |= mNeedsLegacyVertexAttrib0Handling;
+
+  if (const auto& env =
+          gfxEnv::MOZ_WEBGL_WORKAROUND_FIRST_AFFECTS_INSTANCE_ID()) {
+    const auto was = mBug_DrawArraysInstancedUserAttribFetchAffectedByFirst;
+    mBug_DrawArraysInstancedUserAttribFetchAffectedByFirst =
+        (env.as_str != "0");
+    printf_stderr(
+        "mBug_DrawArraysInstancedUserAttribFetchAffectedByFirst: %i -> %i\n",
+        int(was), int(mBug_DrawArraysInstancedUserAttribFetchAffectedByFirst));
+  }
+
+  // -
+
   mNeedsIndexValidation =
       !gl->IsSupported(gl::GLFeature::robust_buffer_access_behavior);
-  switch (gfxPrefs::WebGLForceIndexValidation()) {
+  switch (StaticPrefs::webgl_force_index_validation()) {
     case -1:
       mNeedsIndexValidation = false;
       break;
@@ -765,15 +594,18 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
       mNeedsIndexValidation = true;
       break;
     default:
-      MOZ_ASSERT(gfxPrefs::WebGLForceIndexValidation() == 0);
+      MOZ_ASSERT(StaticPrefs::webgl_force_index_validation() == 0);
       break;
   }
 
+  for (auto& cur : mExtensions) {
+    cur = {};
+  }
+
   return true;
 }
 
-bool WebGLContext::ValidateFramebufferTarget(GLenum target,
-                                             const char* const info) {
+bool WebGLContext::ValidateFramebufferTarget(GLenum target) const {
   bool isValid = true;
   switch (target) {
     case LOCAL_GL_FRAMEBUFFER:
@@ -793,7 +625,7 @@ bool WebGLContext::ValidateFramebufferTarget(GLenum target,
     return true;
   }
 
-  ErrorInvalidEnumArg(info, "target", target);
+  ErrorInvalidEnumArg("target", target);
   return false;
 }
 
