@@ -1,22 +1,61 @@
-/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
+/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
 /* This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "WebGLFormats.h"
 
-#include "gfxPrefs.h"
 #include "GLContext.h"
 #include "GLDefs.h"
 #include "mozilla/gfx/Logging.h"
 #include "mozilla/StaticMutex.h"
 
-#ifdef FOO
-#error FOO is already defined! We use FOO() macros to keep things succinct in this file.
-#endif
+namespace mozilla::webgl {
 
-namespace mozilla {
-namespace webgl {
+const char* ToString(const ComponentType type) {
+  switch (type) {
+    case ComponentType::Int:
+      return "Int";
+    case ComponentType::UInt:
+      return "UInt";
+    case ComponentType::NormInt:
+      return "NormInt";
+    case ComponentType::NormUInt:
+      return "NormUInt";
+    case ComponentType::Float:
+      return "Float";
+  }
+  MOZ_CRASH("pacify gcc6 warning");
+}
+
+static TextureBaseType ToBaseType(const ComponentType type) {
+  switch (type) {
+    case ComponentType::Int:
+      return TextureBaseType::Int;
+    case ComponentType::UInt:
+      return TextureBaseType::UInt;
+    case ComponentType::NormInt:
+    case ComponentType::NormUInt:
+    case ComponentType::Float:
+      // case ComponentType::Depth:
+      return TextureBaseType::Float;
+  }
+  MOZ_CRASH("pacify gcc6 warning");
+}
+
+const char* ToString(const TextureBaseType x) {
+  switch (x) {
+    case webgl::TextureBaseType::Float:
+      return "FLOAT";
+    case webgl::TextureBaseType::Int:
+      return "INT";
+    case webgl::TextureBaseType::UInt:
+      return "UINT";
+  }
+  MOZ_CRASH("pacify gcc6 warning");
+}
+
+// -
 
 template <typename K, typename V, typename K2, typename V2>
 static inline void AlwaysInsert(std::map<K, V>& dest, const K2& key,
@@ -35,10 +74,11 @@ static inline V* FindOrNull(const std::map<K, V*>& dest, const K2& key) {
 }
 
 // Returns a pointer to the in-place value for `key`.
-template <typename K, typename V, typename K2>
-static inline V* FindPtrOrNull(std::map<K, V>& dest, const K2& key) {
-  auto itr = dest.find(key);
-  if (itr == dest.end()) return nullptr;
+template <typename C, typename K2>
+static inline auto FindPtrOrNull(C& container, const K2& key) {
+  auto itr = container.find(key);
+  using R = decltype(&(itr->second));
+  if (itr == container.end()) return R{nullptr};
 
   return &(itr->second);
 }
@@ -66,10 +106,9 @@ static void AddCompressedFormatInfo(EffectiveFormat format,
                                     uint8_t blockHeight,
                                     CompressionFamily family) {
   MOZ_ASSERT(bitsPerBlock % 8 == 0);
-  uint16_t bytesPerBlock =
-      bitsPerBlock / 8;  // The specs always state these in bits,
-                         // but it's only ever useful to us as
-                         // bytes.
+  uint16_t bytesPerBlock = bitsPerBlock / 8;  // The specs always state these in
+                                              // bits, but it's only ever useful
+                                              // to us as bytes.
   MOZ_ASSERT(bytesPerBlock <= 255);
 
   const CompressedFormatInfo info = {format, uint8_t(bytesPerBlock), blockWidth,
@@ -93,10 +132,17 @@ static void InitCompressedFormatInfo() {
     AddCompressedFormatInfo(EffectiveFormat::COMPRESSED_RGB8_PUNCHTHROUGH_ALPHA1_ETC2 ,  64, 4, 4, CompressionFamily::ES3);
     AddCompressedFormatInfo(EffectiveFormat::COMPRESSED_SRGB8_PUNCHTHROUGH_ALPHA1_ETC2,  64, 4, 4, CompressionFamily::ES3);
 
-    // AMD_compressed_ATC_texture
-    AddCompressedFormatInfo(EffectiveFormat::ATC_RGB_AMD                    ,  64, 4, 4, CompressionFamily::ATC);
-    AddCompressedFormatInfo(EffectiveFormat::ATC_RGBA_EXPLICIT_ALPHA_AMD    , 128, 4, 4, CompressionFamily::ATC);
-    AddCompressedFormatInfo(EffectiveFormat::ATC_RGBA_INTERPOLATED_ALPHA_AMD, 128, 4, 4, CompressionFamily::ATC);
+    // EXT_texture_compression_bptc
+    AddCompressedFormatInfo(EffectiveFormat::COMPRESSED_RGBA_BPTC_UNORM        , 16*8, 4, 4, CompressionFamily::BPTC);
+    AddCompressedFormatInfo(EffectiveFormat::COMPRESSED_SRGB_ALPHA_BPTC_UNORM  , 16*8, 4, 4, CompressionFamily::BPTC);
+    AddCompressedFormatInfo(EffectiveFormat::COMPRESSED_RGB_BPTC_SIGNED_FLOAT  , 16*8, 4, 4, CompressionFamily::BPTC);
+    AddCompressedFormatInfo(EffectiveFormat::COMPRESSED_RGB_BPTC_UNSIGNED_FLOAT, 16*8, 4, 4, CompressionFamily::BPTC);
+
+    // EXT_texture_compression_rgtc
+    AddCompressedFormatInfo(EffectiveFormat::COMPRESSED_RED_RGTC1       ,  8*8, 4, 4, CompressionFamily::RGTC);
+    AddCompressedFormatInfo(EffectiveFormat::COMPRESSED_SIGNED_RED_RGTC1,  8*8, 4, 4, CompressionFamily::RGTC);
+    AddCompressedFormatInfo(EffectiveFormat::COMPRESSED_RG_RGTC2        , 16*8, 4, 4, CompressionFamily::RGTC);
+    AddCompressedFormatInfo(EffectiveFormat::COMPRESSED_SIGNED_RG_RGTC2 , 16*8, 4, 4, CompressionFamily::RGTC);
 
     // EXT_texture_compression_s3tc
     AddCompressedFormatInfo(EffectiveFormat::COMPRESSED_RGB_S3TC_DXT1_EXT ,  64, 4, 4, CompressionFamily::S3TC);
@@ -225,6 +271,7 @@ static void AddFormatInfo(EffectiveFormat format, const char* name,
                            sizedFormat,
                            unsizedFormat,
                            componentType,
+                           ToBaseType(componentType),
                            isSRGB,
                            compressedFormatInfo,
                            bytesPerPixel,
@@ -238,8 +285,8 @@ static void AddFormatInfo(EffectiveFormat format, const char* name,
 }
 
 static void InitFormatInfo() {
-// This function is full of expressive formatting, so:
-// clang-format off
+  // This function is full of expressive formatting, so:
+  // clang-format off
 
 #define FOO(x) EffectiveFormat::x, #x, LOCAL_GL_ ## x
 
@@ -305,11 +352,12 @@ static void InitFormatInfo() {
     AddFormatInfo(FOO(DEPTH_COMPONENT16 ), 2, 0,0,0,0, 16,0, UnsizedFormat::D , false, ComponentType::NormUInt);
     AddFormatInfo(FOO(DEPTH_COMPONENT24 ), 3, 0,0,0,0, 24,0, UnsizedFormat::D , false, ComponentType::NormUInt);
     AddFormatInfo(FOO(DEPTH_COMPONENT32F), 4, 0,0,0,0, 32,0, UnsizedFormat::D , false, ComponentType::Float);
-    AddFormatInfo(FOO(DEPTH24_STENCIL8  ), 4, 0,0,0,0, 24,8, UnsizedFormat::DEPTH_STENCIL, false, ComponentType::Special);
-    AddFormatInfo(FOO(DEPTH32F_STENCIL8 ), 5, 0,0,0,0, 32,8, UnsizedFormat::DEPTH_STENCIL, false, ComponentType::Special);
+    // DEPTH_STENCIL types are sampled as their depth component.
+    AddFormatInfo(FOO(DEPTH24_STENCIL8  ), 4, 0,0,0,0, 24,8, UnsizedFormat::DEPTH_STENCIL, false, ComponentType::NormUInt);
+    AddFormatInfo(FOO(DEPTH32F_STENCIL8 ), 5, 0,0,0,0, 32,8, UnsizedFormat::DEPTH_STENCIL, false, ComponentType::Float);
 
     // GLES 3.0.4, p205-206, "Required Renderbuffer Formats"
-    AddFormatInfo(FOO(STENCIL_INDEX8), 1, 0,0,0,0, 0,8, UnsizedFormat::S, false, ComponentType::UInt);
+    AddFormatInfo(FOO(STENCIL_INDEX8), 1, 0,0,0,0, 0,8, UnsizedFormat::S, false, ComponentType::Int);
 
     // GLES 3.0.4, p147, table 3.19
     // GLES 3.0.4  p286+  $C.1 "ETC Compressed Texture Image Formats"
@@ -324,10 +372,17 @@ static void InitFormatInfo() {
     AddFormatInfo(FOO(COMPRESSED_RGB8_PUNCHTHROUGH_ALPHA1_ETC2 ), 0, 1,1,1,1, 0,0, UnsizedFormat::RGBA, false, ComponentType::NormUInt);
     AddFormatInfo(FOO(COMPRESSED_SRGB8_PUNCHTHROUGH_ALPHA1_ETC2), 0, 1,1,1,1, 0,0, UnsizedFormat::RGBA, true , ComponentType::NormUInt);
 
-    // AMD_compressed_ATC_texture
-    AddFormatInfo(FOO(ATC_RGB_AMD                    ), 0, 1,1,1,0, 0,0, UnsizedFormat::RGB , false, ComponentType::NormUInt);
-    AddFormatInfo(FOO(ATC_RGBA_EXPLICIT_ALPHA_AMD    ), 0, 1,1,1,1, 0,0, UnsizedFormat::RGBA, false, ComponentType::NormUInt);
-    AddFormatInfo(FOO(ATC_RGBA_INTERPOLATED_ALPHA_AMD), 0, 1,1,1,1, 0,0, UnsizedFormat::RGBA, false, ComponentType::NormUInt);
+    // EXT_texture_compression_bptc
+    AddFormatInfo(FOO(COMPRESSED_RGBA_BPTC_UNORM        ), 0, 1,1,1,1, 0,0, UnsizedFormat::RGBA, false, ComponentType::NormUInt);
+    AddFormatInfo(FOO(COMPRESSED_SRGB_ALPHA_BPTC_UNORM  ), 0, 1,1,1,1, 0,0, UnsizedFormat::RGBA, true , ComponentType::NormUInt);
+    AddFormatInfo(FOO(COMPRESSED_RGB_BPTC_SIGNED_FLOAT  ), 0, 1,1,1,0, 0,0, UnsizedFormat::RGB , false, ComponentType::Float   );
+    AddFormatInfo(FOO(COMPRESSED_RGB_BPTC_UNSIGNED_FLOAT), 0, 1,1,1,0, 0,0, UnsizedFormat::RGB , false, ComponentType::Float   );
+
+    // EXT_texture_compression_rgtc
+    AddFormatInfo(FOO(COMPRESSED_RED_RGTC1       ), 0, 1,0,0,0, 0,0, UnsizedFormat::R , false, ComponentType::NormUInt);
+    AddFormatInfo(FOO(COMPRESSED_SIGNED_RED_RGTC1), 0, 1,0,0,0, 0,0, UnsizedFormat::R , false, ComponentType::NormInt );
+    AddFormatInfo(FOO(COMPRESSED_RG_RGTC2        ), 0, 1,1,0,0, 0,0, UnsizedFormat::RG, false, ComponentType::NormUInt);
+    AddFormatInfo(FOO(COMPRESSED_SIGNED_RG_RGTC2 ), 0, 1,1,0,0, 0,0, UnsizedFormat::RG, false, ComponentType::NormInt );
 
     // EXT_texture_compression_s3tc
     AddFormatInfo(FOO(COMPRESSED_RGB_S3TC_DXT1_EXT ), 0, 1,1,1,0, 0,0, UnsizedFormat::RGB , false, ComponentType::NormUInt);
@@ -381,6 +436,17 @@ static void InitFormatInfo() {
     // OES_compressed_ETC1_RGB8_texture
     AddFormatInfo(FOO(ETC1_RGB8_OES), 0, 1,1,1,0, 0,0, UnsizedFormat::RGB, false, ComponentType::NormUInt);
 
+    // EXT_texture_norm16
+    AddFormatInfo(FOO(R16   ), 2, 16, 0, 0, 0, 0,0, UnsizedFormat::R   , false, ComponentType::NormUInt);
+    AddFormatInfo(FOO(RG16  ), 4, 16,16, 0, 0, 0,0, UnsizedFormat::RG  , false, ComponentType::NormUInt);
+    AddFormatInfo(FOO(RGB16 ), 6, 16,16,16, 0, 0,0, UnsizedFormat::RGB , false, ComponentType::NormUInt);
+    AddFormatInfo(FOO(RGBA16), 8, 16,16,16,16, 0,0, UnsizedFormat::RGBA, false, ComponentType::NormUInt);
+
+    AddFormatInfo(FOO(R16_SNORM   ), 2, 16, 0, 0, 0, 0,0, UnsizedFormat::R   , false, ComponentType::NormInt);
+    AddFormatInfo(FOO(RG16_SNORM  ), 4, 16,16, 0, 0, 0,0, UnsizedFormat::RG  , false, ComponentType::NormInt);
+    AddFormatInfo(FOO(RGB16_SNORM ), 6, 16,16,16, 0, 0,0, UnsizedFormat::RGB , false, ComponentType::NormInt);
+    AddFormatInfo(FOO(RGBA16_SNORM), 8, 16,16,16,16, 0,0, UnsizedFormat::RGBA, false, ComponentType::NormInt);
+
 #undef FOO
 
     // 'Virtual' effective formats have no sizedFormat.
@@ -461,6 +527,7 @@ static void InitFormatInfo() {
     SET_BY_SUFFIX(8I)
     SET_BY_SUFFIX(8UI)
 
+    SET_BY_SUFFIX(16)
     SET_BY_SUFFIX(16I)
     SET_BY_SUFFIX(16UI)
 
@@ -520,96 +587,86 @@ const FormatInfo* FormatInfo::GetCopyDecayFormat(UnsizedFormat uf) const {
   return FindOrNull(this->copyDecayFormats, uf);
 }
 
-bool GetBytesPerPixel(const PackingInfo& packing, uint8_t* const out_bytes) {
-  uint8_t bytesPerChannel;
+Maybe<PackingInfoInfo> PackingInfoInfo::For(const PackingInfo& pi) {
+  PackingInfoInfo ret{};
 
-  switch (packing.type) {
+  switch (pi.type) {
     case LOCAL_GL_UNSIGNED_SHORT_4_4_4_4:
     case LOCAL_GL_UNSIGNED_SHORT_5_5_5_1:
     case LOCAL_GL_UNSIGNED_SHORT_5_6_5:
-      *out_bytes = 2;
-      return true;
+      ret = {2, 1, true};
+      break;
 
     case LOCAL_GL_UNSIGNED_INT_10F_11F_11F_REV:
     case LOCAL_GL_UNSIGNED_INT_2_10_10_10_REV:
     case LOCAL_GL_UNSIGNED_INT_24_8:
     case LOCAL_GL_UNSIGNED_INT_5_9_9_9_REV:
-      *out_bytes = 4;
-      return true;
+      ret = {4, 1, true};
+      break;
 
     case LOCAL_GL_FLOAT_32_UNSIGNED_INT_24_8_REV:
-      *out_bytes = 8;
-      return true;
+      ret = {8, 1, true};
+      break;
 
       // Alright, that's all the fixed-size unpackTypes.
 
     case LOCAL_GL_BYTE:
     case LOCAL_GL_UNSIGNED_BYTE:
-      bytesPerChannel = 1;
+      ret = {1, 0, false};
       break;
 
     case LOCAL_GL_SHORT:
     case LOCAL_GL_UNSIGNED_SHORT:
     case LOCAL_GL_HALF_FLOAT:
     case LOCAL_GL_HALF_FLOAT_OES:
-      bytesPerChannel = 2;
+      ret = {2, 0, false};
       break;
 
     case LOCAL_GL_INT:
     case LOCAL_GL_UNSIGNED_INT:
     case LOCAL_GL_FLOAT:
-      bytesPerChannel = 4;
+      ret = {4, 0, false};
       break;
 
     default:
-      return false;
+      return {};
   }
 
-  uint8_t channels;
+  if (!ret.isPacked) {
+    switch (pi.format) {
+      case LOCAL_GL_RED:
+      case LOCAL_GL_RED_INTEGER:
+      case LOCAL_GL_LUMINANCE:
+      case LOCAL_GL_ALPHA:
+      case LOCAL_GL_DEPTH_COMPONENT:
+        ret.elementsPerPixel = 1;
+        break;
 
-  switch (packing.format) {
-    case LOCAL_GL_RED:
-    case LOCAL_GL_RED_INTEGER:
-    case LOCAL_GL_LUMINANCE:
-    case LOCAL_GL_ALPHA:
-    case LOCAL_GL_DEPTH_COMPONENT:
-      channels = 1;
-      break;
+      case LOCAL_GL_RG:
+      case LOCAL_GL_RG_INTEGER:
+      case LOCAL_GL_LUMINANCE_ALPHA:
+        ret.elementsPerPixel = 2;
+        break;
 
-    case LOCAL_GL_RG:
-    case LOCAL_GL_RG_INTEGER:
-    case LOCAL_GL_LUMINANCE_ALPHA:
-      channels = 2;
-      break;
+      case LOCAL_GL_RGB:
+      case LOCAL_GL_RGB_INTEGER:
+      case LOCAL_GL_SRGB:
+        ret.elementsPerPixel = 3;
+        break;
 
-    case LOCAL_GL_RGB:
-    case LOCAL_GL_RGB_INTEGER:
-    case LOCAL_GL_SRGB:
-      channels = 3;
-      break;
+      case LOCAL_GL_BGRA:
+      case LOCAL_GL_RGBA:
+      case LOCAL_GL_RGBA_INTEGER:
+      case LOCAL_GL_SRGB_ALPHA:
+        ret.elementsPerPixel = 4;
+        break;
 
-    case LOCAL_GL_BGRA:
-    case LOCAL_GL_RGBA:
-    case LOCAL_GL_RGBA_INTEGER:
-    case LOCAL_GL_SRGB_ALPHA:
-      channels = 4;
-      break;
-
-    default:
-      return false;
+      default:
+        return {};
+    }
   }
 
-  *out_bytes = bytesPerChannel * channels;
-  return true;
-}
-
-uint8_t BytesPerPixel(const PackingInfo& packing) {
-  uint8_t ret;
-  if (MOZ_LIKELY(GetBytesPerPixel(packing, &ret))) return ret;
-
-  gfxCriticalError() << "Bad `packing`: " << gfx::hexa(packing.format) << ", "
-                     << gfx::hexa(packing.type);
-  MOZ_CRASH("Bad `packing`.");
+  return Some(ret);
 }
 
 //////////////////////////////////////////////////////////////////////////////////////////
@@ -631,24 +688,21 @@ bool FormatUsageInfo::IsUnpackValid(
   return true;
 }
 
-void FormatUsageInfo::ResolveMaxSamples(gl::GLContext* gl) {
+void FormatUsageInfo::ResolveMaxSamples(gl::GLContext& gl) const {
+  MOZ_ASSERT(gl.IsCurrent());
   MOZ_ASSERT(!this->maxSamplesKnown);
-  MOZ_ASSERT(this->maxSamples == 0);
-  MOZ_ASSERT(gl->IsCurrent());
-
+  MOZ_ASSERT(!this->maxSamples);
   this->maxSamplesKnown = true;
 
   const GLenum internalFormat = this->format->sizedFormat;
   if (!internalFormat) return;
+  if (!gl.IsSupported(gl::GLFeature::internalformat_query)) return;
 
-  if (!gl->IsSupported(gl::GLFeature::internalformat_query))
-    return;  // Leave it at 0.
-
-  GLint maxSamplesGL = 0;
-  gl->fGetInternalformativ(LOCAL_GL_RENDERBUFFER, internalFormat,
-                           LOCAL_GL_SAMPLES, 1, &maxSamplesGL);
-
-  this->maxSamples = maxSamplesGL;
+  // GL_SAMPLES returns a list in descending order, so ask for just one elem to
+  // get the max.
+  gl.fGetInternalformativ(LOCAL_GL_RENDERBUFFER, internalFormat,
+                          LOCAL_GL_SAMPLES, 1,
+                          reinterpret_cast<GLint*>(&this->maxSamples));
 }
 
 ////////////////////////////////////////
@@ -729,8 +783,10 @@ static bool AddUnsizedFormats(FormatUsageAuthority* fua, gl::GLContext* gl) {
   // clang-format on
 }
 
-void FormatUsageInfo::SetRenderable() {
-  this->isRenderable = true;
+void FormatUsageInfo::SetRenderable(const FormatRenderableState& state) {
+  if (!renderableState.IsExplicit()) {
+    renderableState = state;
+  }
 
 #ifdef DEBUG
   const auto format = this->format;
@@ -781,11 +837,13 @@ UniquePtr<FormatUsageAuthority> FormatUsageAuthority::CreateForWebGL1(
   fnSet(EffectiveFormat::Luminance8, false, true);
   fnSet(EffectiveFormat::Alpha8, false, true);
 
-  fnSet(EffectiveFormat::DEPTH_COMPONENT16, true, false);
+  fnSet(EffectiveFormat::DEPTH_COMPONENT16, true, true);
+  fnSet(EffectiveFormat::DEPTH_COMPONENT24, true,
+        true);  // Requires WEBGL_depth_texture.
   fnSet(EffectiveFormat::STENCIL_INDEX8, true, false);
 
   // Added in WebGL 1.0 spec:
-  fnSet(EffectiveFormat::DEPTH24_STENCIL8, true, false);
+  fnSet(EffectiveFormat::DEPTH24_STENCIL8, true, true);
 
   ////////////////////////////////////
   // RB formats
@@ -810,7 +868,7 @@ UniquePtr<FormatUsageAuthority> FormatUsageAuthority::CreateForWebGL1(
 
   if (!AddUnsizedFormats(ptr, gl)) return nullptr;
 
-  return Move(ret);
+  return ret;
 }
 
 UniquePtr<FormatUsageAuthority> FormatUsageAuthority::CreateForWebGL2(
@@ -1010,7 +1068,10 @@ UniquePtr<FormatUsageAuthority> FormatUsageAuthority::CreateForWebGL2(
   fnAllowES3TexFormat(FOO(RGBA32I), true, false);
   fnAllowES3TexFormat(FOO(RGBA32UI), true, false);
 
-  // GLES 3.0.4, p133, table 3.14
+  // Sized depth or depth-stencil formats are not filterable
+  // per GLES 3.0.6 p161.
+  // Specifically, they're texture-incomplete if depth-compare:none and
+  // not NEAREST.
   fnAllowES3TexFormat(FOO(DEPTH_COMPONENT16), true, false);
   fnAllowES3TexFormat(FOO(DEPTH_COMPONENT24), true, false);
   fnAllowES3TexFormat(FOO(DEPTH_COMPONENT32F), true, false);
@@ -1036,21 +1097,9 @@ UniquePtr<FormatUsageAuthority> FormatUsageAuthority::CreateForWebGL2(
   ptr->AllowRBFormat(LOCAL_GL_DEPTH_STENCIL,
                      ptr->GetUsage(EffectiveFormat::DEPTH24_STENCIL8));
 
-  if (gfxPrefs::WebGL2CompatMode()) {
-    AddSimpleUnsized(ptr, LOCAL_GL_RGBA, LOCAL_GL_FLOAT,
-                     EffectiveFormat::RGBA32F);
-    AddSimpleUnsized(ptr, LOCAL_GL_RGB, LOCAL_GL_FLOAT,
-                     EffectiveFormat::RGB32F);
-
-    AddSimpleUnsized(ptr, LOCAL_GL_RGBA, LOCAL_GL_HALF_FLOAT_OES,
-                     EffectiveFormat::RGBA16F);
-    AddSimpleUnsized(ptr, LOCAL_GL_RGB, LOCAL_GL_HALF_FLOAT_OES,
-                     EffectiveFormat::RGB16F);
-  }
-
   ////////////////////////////////////
 
-  return Move(ret);
+  return ret;
 }
 
 //////////////////////////////////////////////////////////////////////////////////////////
@@ -1090,11 +1139,17 @@ bool FormatUsageAuthority::AreUnpackEnumsValid(GLenum unpackFormat,
 ////////////////////
 
 void FormatUsageAuthority::AllowRBFormat(GLenum sizedFormat,
-                                         const FormatUsageInfo* usage) {
+                                         const FormatUsageInfo* usage,
+                                         const bool expectRenderable) {
   MOZ_ASSERT(!usage->format->compression);
   MOZ_ASSERT(usage->format->sizedFormat);
-  MOZ_ASSERT(usage->IsRenderable());
+  MOZ_ASSERT(usage->IsRenderable() || !expectRenderable);
 
+  const auto& found = mRBFormatMap.find(sizedFormat);
+  if (found != mRBFormatMap.end()) {
+    MOZ_ASSERT(found->second == usage);
+    return;
+  }
   AlwaysInsert(mRBFormatMap, sizedFormat, usage);
 }
 
@@ -1103,7 +1158,7 @@ void FormatUsageAuthority::AllowSizedTexFormat(GLenum sizedFormat,
   if (usage->format->compression) {
     MOZ_ASSERT(usage->isFilterable, "Compressed formats should be filterable.");
   } else {
-    MOZ_ASSERT(usage->validUnpacks.size() && usage->idealUnpack,
+    MOZ_ASSERT(!usage->validUnpacks.empty() && usage->idealUnpack,
                "AddTexUnpack() first.");
   }
 
@@ -1115,7 +1170,7 @@ void FormatUsageAuthority::AllowSizedTexFormat(GLenum sizedFormat,
 void FormatUsageAuthority::AllowUnsizedTexFormat(const PackingInfo& pi,
                                                  const FormatUsageInfo* usage) {
   MOZ_ASSERT(!usage->format->compression);
-  MOZ_ASSERT(usage->validUnpacks.size() && usage->idealUnpack,
+  MOZ_ASSERT(!usage->validUnpacks.empty() && usage->idealUnpack,
              "AddTexUnpack() first.");
 
   AlwaysInsert(mUnsizedTexFormatMap, pi, usage);
@@ -1169,5 +1224,4 @@ const FormatUsageInfo* FormatUsageAuthority::GetUsage(
 
 ////////////////////////////////////////////////////////////////////////////////
 
-}  // namespace webgl
-}  // namespace mozilla
+}  // namespace mozilla::webgl
