@@ -14,6 +14,7 @@
 #include "GLScreenBuffer.h"
 #include "MozFramebuffer.h"
 #include "mozilla/dom/WebGLRenderingContextBinding.h"
+#include "mozilla/IntegerRange.h"
 #include "nsPrintfCString.h"
 #include "WebGLContext.h"
 #include "WebGLContextUtils.h"
@@ -38,6 +39,9 @@ static bool ShouldDeferAttachment(const WebGLContext* const webgl,
   }
 }
 
+WebGLFBAttachPoint::WebGLFBAttachPoint() = default;
+WebGLFBAttachPoint::WebGLFBAttachPoint(WebGLFBAttachPoint&) = default;
+
 WebGLFBAttachPoint::WebGLFBAttachPoint(const WebGLContext* const webgl,
                                        const GLenum attachmentPoint)
     : mAttachmentPoint(attachmentPoint),
@@ -97,21 +101,31 @@ bool WebGLFBAttachPoint::IsComplete(WebGLContext* webgl,
   const auto& tex = Texture();
   if (tex) {
     // ES 3.0 spec, pg 213 has giant blocks of text that bake down to requiring
-    // that attached tex images are within the valid mip-levels of the texture.
-    // While it draws distinction to only test non-immutable textures, that's
-    // because immutable textures are *always* texture-complete. We need to
-    // check immutable textures though, because checking completeness is also
-    // when we zero invalidated/no-data tex images.
+    // that attached *non-immutable* tex images are within the valid mip-levels
+    // of the texture. We still need to check immutable textures though, because
+    // checking completeness is also when we zero invalidated/no-data tex
+    // images.
     const auto attachedMipLevel = MipLevel();
 
     const bool withinValidMipLevels = [&]() {
       const bool ensureInit = false;
       const auto texCompleteness = tex->CalcCompletenessInfo(ensureInit);
-      if (!texCompleteness)  // OOM
-        return false;
+      if (!texCompleteness) return false;  // OOM
+
+      if (tex->Immutable()) {
+        // Immutable textures can attach a level that's not valid for sampling.
+        // It still has to exist though!
+        return attachedMipLevel < tex->ImmutableLevelCount();
+      }
+
+      // Base level must be complete.
       if (!texCompleteness->levels) return false;
 
-      const auto baseLevel = tex->BaseMipmapLevel();
+      const auto baseLevel = tex->Es3_level_base();
+      if (attachedMipLevel == baseLevel) return true;
+
+      // If not base level, must be mip-complete and within mips.
+      if (!texCompleteness->mipmapComplete) return false;
       const auto maxLevel = baseLevel + texCompleteness->levels - 1;
       return baseLevel <= attachedMipLevel && attachedMipLevel <= maxLevel;
     }();
@@ -491,7 +505,6 @@ WebGLFramebuffer::WebGLFramebuffer(WebGLContext* webgl,
   CompletenessInfo info;
   info.width = mOpaque->mSize.width;
   info.height = mOpaque->mSize.height;
-  info.hasFloat32 = false;
   info.zLayerCount = 1;
   info.isMultiview = false;
 
@@ -527,7 +540,7 @@ Maybe<WebGLFBAttachPoint*> WebGLFramebuffer::GetColorAttachPoint(
 
   const size_t colorId = attachPoint - LOCAL_GL_COLOR_ATTACHMENT0;
 
-  MOZ_ASSERT(mContext->Limits().maxColorDrawBuffers <= kMaxColorAttachments);
+  MOZ_ASSERT(mContext->Limits().maxColorDrawBuffers <= webgl::kMaxDrawBuffers);
   if (colorId >= mContext->MaxValidDrawBuffers()) return Nothing();
 
   return Some(&mColorAttachments[colorId]);
@@ -999,7 +1012,7 @@ FBStatus WebGLFramebuffer::CheckFramebufferStatus() const {
     ResolveAttachmentData();
 
     // Sweet, let's cache that.
-    auto info = CompletenessInfo{this, UINT32_MAX, UINT32_MAX};
+    auto info = CompletenessInfo{this};
     mCompletenessInfo.ResetInvalidators({});
     mCompletenessInfo.AddInvalidator(*this);
 
@@ -1021,12 +1034,20 @@ FBStatus WebGLFramebuffer::CheckFramebufferStatus() const {
       }
       const auto& imageInfo = cur->GetImageInfo();
       MOZ_ASSERT(imageInfo);
-      info.width = std::min(info.width, imageInfo->mWidth);
-      info.height = std::min(info.height, imageInfo->mHeight);
-      info.hasFloat32 |= fnIsFloat32(*imageInfo->mFormat->format);
+
+      const auto maybeColorId = cur->ColorAttachmentId();
+      if (maybeColorId) {
+        const auto id = *maybeColorId;
+        info.hasAttachment[id] = true;
+        info.isAttachmentF32[id] = fnIsFloat32(*imageInfo->mFormat->format);
+      }
+
+      info.width = imageInfo->mWidth;
+      info.height = imageInfo->mHeight;
       info.zLayerCount = cur->ZLayerCount();
       info.isMultiview = cur->IsMultiview();
     }
+    MOZ_ASSERT(info.width && info.height);
     mCompletenessInfo = Some(std::move(info));
     info.fb = nullptr;  // Don't trigger the invalidation warning.
     return LOCAL_GL_FRAMEBUFFER_COMPLETE;
@@ -1094,6 +1115,7 @@ void WebGLFramebuffer::DrawBuffers(const std::vector<GLenum>& buffers) {
   std::vector<const WebGLFBAttachPoint*> newColorDrawBuffers;
   newColorDrawBuffers.reserve(buffers.size());
 
+  mDrawBufferEnabled.reset();
   for (const auto i : IntegerRange(buffers.size())) {
     // "If the GL is bound to a draw framebuffer object, the `i`th buffer listed
     // in bufs must be COLOR_ATTACHMENTi or NONE. Specifying a buffer out of
@@ -1110,6 +1132,7 @@ void WebGLFramebuffer::DrawBuffers(const std::vector<GLenum>& buffers) {
     if (cur == LOCAL_GL_COLOR_ATTACHMENT0 + i) {
       const auto& attach = mColorAttachments[i];
       newColorDrawBuffers.push_back(&attach);
+      mDrawBufferEnabled[i] = true;
     } else if (cur != LOCAL_GL_NONE) {
       const bool isColorEnum = (cur >= LOCAL_GL_COLOR_ATTACHMENT0 &&
                                 cur < mContext->LastColorAttachmentEnum());
@@ -1131,16 +1154,6 @@ void WebGLFramebuffer::DrawBuffers(const std::vector<GLenum>& buffers) {
   RefreshDrawBuffers();  // Calls glDrawBuffers.
 }
 
-bool WebGLFramebuffer::IsDrawBufferEnabled(const uint32_t slotId) const {
-  const auto attachEnum = LOCAL_GL_COLOR_ATTACHMENT0 + slotId;
-  for (const auto& cur : mColorDrawBuffers) {
-    if (cur->mAttachmentPoint == attachEnum) {
-      return true;
-    }
-  }
-  return false;
-}
-
 void WebGLFramebuffer::ReadBuffer(GLenum attachPoint) {
   const auto& maybeAttach = GetColorAttachPoint(attachPoint);
   if (!maybeAttach) {
@@ -1516,8 +1529,9 @@ void WebGLFramebuffer::BlitFramebuffer(WebGLContext* webgl, GLint _srcX0,
   // -
   // Mutually constrain src and dst rects for eldritch blits.
 
-  [&]{
-    using fvec2 = avec2<float>; // Switch to float, because there's no perfect solution anyway.
+  [&] {
+    using fvec2 = avec2<float>;  // Switch to float, because there's no perfect
+                                 // solution anyway.
 
     const auto zero2f = fvec2{0, 0};
     const auto srcSizef = AsVec(srcSize).StaticCast<fvec2>();
@@ -1533,7 +1547,7 @@ void WebGLFramebuffer::BlitFramebuffer(WebGLContext* webgl, GLint _srcX0,
 
     // Skip if zero-sized.
     if (!srcRectDiff.x || !srcRectDiff.y || !dstRectDiff.x || !dstRectDiff.y) {
-      srcP0 = srcP1 = dstP0 = dstP1 = {0,0};
+      srcP0 = srcP1 = dstP0 = dstP1 = {0, 0};
       return;
     }
 
@@ -1573,12 +1587,13 @@ void WebGLFramebuffer::BlitFramebuffer(WebGLContext* webgl, GLint _srcX0,
   }();
 
   bool inBounds = true;
-  inBounds &= ( srcP0 == srcP0.Clamp({0,0}, AsVec(srcSize)) );
-  inBounds &= ( srcP1 == srcP1.Clamp({0,0}, AsVec(srcSize)) );
-  inBounds &= ( dstP0 == dstP0.Clamp({0,0}, AsVec(dstSize)) );
-  inBounds &= ( dstP1 == dstP1.Clamp({0,0}, AsVec(dstSize)) );
+  inBounds &= (srcP0 == srcP0.Clamp({0, 0}, AsVec(srcSize)));
+  inBounds &= (srcP1 == srcP1.Clamp({0, 0}, AsVec(srcSize)));
+  inBounds &= (dstP0 == dstP0.Clamp({0, 0}, AsVec(dstSize)));
+  inBounds &= (dstP1 == dstP1.Clamp({0, 0}, AsVec(dstSize)));
   if (!inBounds) {
-    webgl->ErrorImplementationBug("Subrects still not within src and dst after constraining.");
+    webgl->ErrorImplementationBug(
+        "Subrects still not within src and dst after constraining.");
     return;
   }
 
@@ -1642,9 +1657,9 @@ void WebGLFramebuffer::BlitFramebuffer(WebGLContext* webgl, GLint _srcX0,
                              LOCAL_GL_COLOR_BUFFER_BIT, filter);
       }
 
-      const auto& blitHelper = *gl->BlitHelper();
       gl->fBindFramebuffer(LOCAL_GL_DRAW_FRAMEBUFFER, fbC->mFB);
-      blitHelper.DrawBlitTextureToFramebuffer(fbB->ColorTex(), sizeBC, sizeBC);
+      gl->BlitHelper()->DrawBlitTextureToFramebuffer(fbB->ColorTex(), sizeBC,
+                                                     sizeBC);
     }
 
     {
@@ -1669,13 +1684,6 @@ void WebGLFramebuffer::BlitFramebuffer(WebGLContext* webgl, GLint _srcX0,
   // glBlitFramebuffer ignores glColorMask!
 
   if (!webgl->mBoundDrawFramebuffer && webgl->mNeedsFakeNoAlpha) {
-    if (!webgl->mScissorTestEnabled) {
-      gl->fEnable(LOCAL_GL_SCISSOR_TEST);
-    }
-    if (webgl->mRasterizerDiscardEnabled) {
-      gl->fDisable(LOCAL_GL_RASTERIZER_DISCARD);
-    }
-
     const auto dstRectMin = MinExtents(dstP0, dstP1);
     const auto dstRectMax = MaxExtents(dstP0, dstP1);
     const auto dstRectSize = dstRectMax - dstRectMin;
@@ -1683,19 +1691,13 @@ void WebGLFramebuffer::BlitFramebuffer(WebGLContext* webgl, GLint _srcX0,
                                                dstRectSize.x, dstRectSize.y};
     dstRect.Apply(*gl);
 
+    const auto forClear = webgl::ScopedPrepForResourceClear{*webgl};
+
     gl->fClearColor(0, 0, 0, 1);
-    webgl->DoColorMask(1 << 3);
+    webgl->DoColorMask(Some(0), 0b1000);  // Only alpha.
     gl->fClear(LOCAL_GL_COLOR_BUFFER_BIT);
 
-    if (!webgl->mScissorTestEnabled) {
-      gl->fDisable(LOCAL_GL_SCISSOR_TEST);
-    }
-    if (webgl->mRasterizerDiscardEnabled) {
-      gl->fEnable(LOCAL_GL_RASTERIZER_DISCARD);
-    }
     webgl->mScissorRect.Apply(*gl);
-    gl->fClearColor(webgl->mColorClearValue[0], webgl->mColorClearValue[1],
-                    webgl->mColorClearValue[2], webgl->mColorClearValue[3]);
   }
 }
 
