@@ -6,6 +6,7 @@
 
 #include "Clients.h"
 
+#include "Client.h"
 #include "ClientDOMUtil.h"
 #include "mozilla/dom/ClientIPCTypes.h"
 #include "mozilla/dom/ClientManager.h"
@@ -13,14 +14,17 @@
 #include "mozilla/dom/Promise.h"
 #include "mozilla/dom/ServiceWorkerDescriptor.h"
 #include "mozilla/dom/ServiceWorkerManager.h"
-#include "mozilla/dom/WorkerPrivate.h"
-#include "mozilla/SystemGroup.h"
+#include "mozilla/dom/WorkerScope.h"
+#include "mozilla/ipc/BackgroundUtils.h"
+#include "mozilla/SchedulerGroup.h"
+#include "mozilla/StaticPrefs_privacy.h"
+#include "mozilla/StorageAccess.h"
 #include "nsIGlobalObject.h"
 #include "nsString.h"
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
+using mozilla::ipc::CSPInfo;
 using mozilla::ipc::PrincipalInfo;
 
 NS_IMPL_CYCLE_COLLECTING_ADDREF(Clients);
@@ -38,7 +42,7 @@ Clients::Clients(nsIGlobalObject* aGlobal) : mGlobal(aGlobal) {
 
 JSObject* Clients::WrapObject(JSContext* aCx,
                               JS::Handle<JSObject*> aGivenProto) {
-  return ClientsBinding::Wrap(aCx, this, aGivenProto);
+  return Clients_Binding::Wrap(aCx, this, aGivenProto);
 }
 
 nsIGlobalObject* Clients::GetParentObject() const { return mGlobal; }
@@ -75,31 +79,38 @@ already_AddRefed<Promise> Clients::Get(const nsAString& aClientID,
   RefPtr<ClientOpPromise> innerPromise = ClientManager::GetInfoAndState(
       ClientGetInfoAndStateArgs(id, principalInfo), target);
 
-  nsCOMPtr<nsIGlobalObject> global = mGlobal;
   nsCString scope = workerPrivate->ServiceWorkerScope();
+  auto holder =
+      MakeRefPtr<DOMMozPromiseRequestHolder<ClientOpPromise>>(mGlobal);
 
-  innerPromise->Then(
-      target, __func__,
-      [outerPromise, global, scope](const ClientOpResult& aResult) {
-        RefPtr<Client> client =
-            new Client(global, aResult.get_ClientInfoAndState());
-        if (client->GetStorageAccess() ==
-            nsContentUtils::StorageAccess::eAllow) {
-          outerPromise->MaybeResolve(Move(client));
-          return;
-        }
-        nsCOMPtr<nsIRunnable> r = NS_NewRunnableFunction(
-            "Clients::MatchAll() storage denied", [scope] {
-              ServiceWorkerManager::LocalizeAndReportToAllClients(
-                  scope, "ServiceWorkerGetClientStorageError",
-                  nsTArray<nsString>());
-            });
-        SystemGroup::Dispatch(TaskCategory::Other, r.forget());
-        outerPromise->MaybeResolveWithUndefined();
-      },
-      [outerPromise](nsresult aResult) {
-        outerPromise->MaybeResolveWithUndefined();
-      });
+  innerPromise
+      ->Then(
+          target, __func__,
+          [outerPromise, holder, scope](const ClientOpResult& aResult) {
+            holder->Complete();
+            NS_ENSURE_TRUE_VOID(holder->GetParentObject());
+            RefPtr<Client> client = new Client(
+                holder->GetParentObject(), aResult.get_ClientInfoAndState());
+            if (client->GetStorageAccess() == StorageAccess::eAllow ||
+                (StaticPrefs::privacy_partition_serviceWorkers() &&
+                 ShouldPartitionStorage(client->GetStorageAccess()))) {
+              outerPromise->MaybeResolve(std::move(client));
+              return;
+            }
+            nsCOMPtr<nsIRunnable> r = NS_NewRunnableFunction(
+                "Clients::Get() storage denied", [scope] {
+                  ServiceWorkerManager::LocalizeAndReportToAllClients(
+                      scope, "ServiceWorkerGetClientStorageError",
+                      nsTArray<nsString>());
+                });
+            SchedulerGroup::Dispatch(TaskCategory::Other, r.forget());
+            outerPromise->MaybeResolveWithUndefined();
+          },
+          [outerPromise, holder](const CopyableErrorResult& aResult) {
+            holder->Complete();
+            outerPromise->MaybeResolveWithUndefined();
+          })
+      ->Track(*holder);
 
   return outerPromise.forget();
 }
@@ -155,20 +166,20 @@ already_AddRefed<Promise> Clients::MatchAll(const ClientQueryOptions& aOptions,
   ClientMatchAllArgs args(workerPrivate->GetServiceWorkerDescriptor().ToIPC(),
                           aOptions.mType, aOptions.mIncludeUncontrolled);
   StartClientManagerOp(
-      &ClientManager::MatchAll, args,
-      mGlobal->EventTargetFor(TaskCategory::Other),
+      &ClientManager::MatchAll, args, mGlobal,
       [outerPromise, global, scope](const ClientOpResult& aResult) {
         nsTArray<RefPtr<Client>> clientList;
         bool storageDenied = false;
         for (const ClientInfoAndState& value :
              aResult.get_ClientList().values()) {
           RefPtr<Client> client = new Client(global, value);
-          if (client->GetStorageAccess() !=
-              nsContentUtils::StorageAccess::eAllow) {
+          if (client->GetStorageAccess() != StorageAccess::eAllow &&
+              (!StaticPrefs::privacy_partition_serviceWorkers() ||
+               !ShouldPartitionStorage(client->GetStorageAccess()))) {
             storageDenied = true;
             continue;
           }
-          clientList.AppendElement(Move(client));
+          clientList.AppendElement(std::move(client));
         }
         if (storageDenied) {
           nsCOMPtr<nsIRunnable> r = NS_NewRunnableFunction(
@@ -177,12 +188,15 @@ already_AddRefed<Promise> Clients::MatchAll(const ClientQueryOptions& aOptions,
                     scope, "ServiceWorkerGetClientStorageError",
                     nsTArray<nsString>());
               });
-          SystemGroup::Dispatch(TaskCategory::Other, r.forget());
+          SchedulerGroup::Dispatch(TaskCategory::Other, r.forget());
         }
         clientList.Sort(MatchAllComparator());
         outerPromise->MaybeResolve(clientList);
       },
-      [outerPromise](nsresult aResult) { outerPromise->MaybeReject(aResult); });
+      [outerPromise](const CopyableErrorResult& aResult) {
+        // MaybeReject needs a non-const-ref result, so make a copy.
+        outerPromise->MaybeReject(CopyableErrorResult(aResult));
+      });
 
   return outerPromise.forget();
 }
@@ -201,8 +215,10 @@ already_AddRefed<Promise> Clients::OpenWindow(const nsAString& aURL,
   }
 
   if (aURL.EqualsLiteral("about:blank")) {
-    // TODO: Improve this error in bug 1412856.
-    outerPromise->MaybeReject(NS_ERROR_DOM_TYPE_ERR);
+    CopyableErrorResult rv;
+    rv.ThrowTypeError(
+        "Passing \"about:blank\" to Clients.openWindow is not allowed");
+    outerPromise->MaybeReject(std::move(rv));
     return outerPromise.forget();
   }
 
@@ -212,15 +228,16 @@ already_AddRefed<Promise> Clients::OpenWindow(const nsAString& aURL,
   }
 
   const PrincipalInfo& principalInfo = workerPrivate->GetPrincipalInfo();
+  const CSPInfo& cspInfo = workerPrivate->GetCSPInfo();
   nsCString baseURL = workerPrivate->GetLocationInfo().mHref;
-  ClientOpenWindowArgs args(principalInfo, NS_ConvertUTF16toUTF8(aURL),
-                            baseURL);
+
+  ClientOpenWindowArgs args(principalInfo, Some(cspInfo),
+                            NS_ConvertUTF16toUTF8(aURL), baseURL);
 
   nsCOMPtr<nsIGlobalObject> global = mGlobal;
 
   StartClientManagerOp(
-      &ClientManager::OpenWindow, args,
-      mGlobal->EventTargetFor(TaskCategory::Other),
+      &ClientManager::OpenWindow, args, mGlobal,
       [outerPromise, global](const ClientOpResult& aResult) {
         if (aResult.type() != ClientOpResult::TClientInfoAndState) {
           outerPromise->MaybeResolve(JS::NullHandleValue);
@@ -230,10 +247,9 @@ already_AddRefed<Promise> Clients::OpenWindow(const nsAString& aURL,
             new Client(global, aResult.get_ClientInfoAndState());
         outerPromise->MaybeResolve(client);
       },
-      [outerPromise](nsresult aResult) {
-        // TODO: Improve this error in bug 1412856.  Ideally we should throw
-        //       the TypeError in the child process and pass it back to here.
-        outerPromise->MaybeReject(NS_ERROR_TYPE_ERR);
+      [outerPromise](const CopyableErrorResult& aResult) {
+        // MaybeReject needs a non-const-ref result, so make a copy.
+        outerPromise->MaybeReject(CopyableErrorResult(aResult));
       });
 
   return outerPromise.forget();
@@ -256,20 +272,21 @@ already_AddRefed<Promise> Clients::Claim(ErrorResult& aRv) {
 
   if (serviceWorker.State() != ServiceWorkerState::Activating &&
       serviceWorker.State() != ServiceWorkerState::Activated) {
-    aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
+    aRv.ThrowInvalidStateError("Service worker is not active");
     return outerPromise.forget();
   }
 
   StartClientManagerOp(
-      &ClientManager::Claim, ClientClaimArgs(serviceWorker.ToIPC()),
-      mGlobal->EventTargetFor(TaskCategory::Other),
+      &ClientManager::Claim, ClientClaimArgs(serviceWorker.ToIPC()), mGlobal,
       [outerPromise](const ClientOpResult& aResult) {
         outerPromise->MaybeResolveWithUndefined();
       },
-      [outerPromise](nsresult aResult) { outerPromise->MaybeReject(aResult); });
+      [outerPromise](const CopyableErrorResult& aResult) {
+        // MaybeReject needs a non-const-ref result, so make a copy.
+        outerPromise->MaybeReject(CopyableErrorResult(aResult));
+      });
 
   return outerPromise.forget();
 }
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
