@@ -9,17 +9,15 @@
 #include "ClientHandle.h"
 #include "ClientManagerChild.h"
 #include "ClientManagerOpChild.h"
-#include "ClientPrefs.h"
 #include "ClientSource.h"
-#include "mozilla/dom/WorkerHolderToken.h"
 #include "mozilla/dom/WorkerPrivate.h"
 #include "mozilla/ipc/BackgroundChild.h"
 #include "mozilla/ipc/PBackgroundChild.h"
-#include "nsContentUtils.h"
+#include "mozilla/ClearOnShutdown.h"  // PastShutdownPhase
+#include "mozilla/StaticPrefs_dom.h"
 #include "prthread.h"
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 using mozilla::ipc::BackgroundChild;
 using mozilla::ipc::PBackgroundChild;
@@ -27,9 +25,23 @@ using mozilla::ipc::PrincipalInfo;
 
 namespace {
 
-uint32_t kBadThreadLocalIndex = -1;
+const uint32_t kBadThreadLocalIndex = -1;
+#ifdef MOZ_DIAGNOSTIC_ASSERT_ENABLED
+const uint32_t kThreadLocalMagic1 = 0x8d57eea6;
+const uint32_t kThreadLocalMagic2 = 0x59f375c9;
+#endif
+
+#ifdef MOZ_DIAGNOSTIC_ASSERT_ENABLED
+uint32_t sClientManagerThreadLocalMagic1 = kThreadLocalMagic1;
+#endif
+
 uint32_t sClientManagerThreadLocalIndex = kBadThreadLocalIndex;
 
+#ifdef MOZ_DIAGNOSTIC_ASSERT_ENABLED
+uint32_t sClientManagerThreadLocalMagic2 = kThreadLocalMagic2;
+uint32_t sClientManagerThreadLocalIndexDuplicate = kBadThreadLocalIndex;
+#endif
+
 }  // anonymous namespace
 
 ClientManager::ClientManager() {
@@ -40,20 +52,12 @@ ClientManager::ClientManager() {
     return;
   }
 
-  RefPtr<WorkerHolderToken> workerHolderToken;
-  if (!NS_IsMainThread()) {
-    WorkerPrivate* workerPrivate = GetCurrentThreadWorkerPrivate();
-    MOZ_DIAGNOSTIC_ASSERT(workerPrivate);
-
-    workerHolderToken = WorkerHolderToken::Create(
-        workerPrivate, Closing, WorkerHolderToken::AllowIdleShutdownStart);
-    if (NS_WARN_IF(!workerHolderToken)) {
-      Shutdown();
-      return;
-    }
+  ClientManagerChild* actor = ClientManagerChild::Create();
+  if (NS_WARN_IF(!actor)) {
+    Shutdown();
+    return;
   }
 
-  ClientManagerChild* actor = new ClientManagerChild(workerHolderToken);
   PClientManagerChild* sentActor =
       parentActor->SendPClientManagerConstructor(actor);
   if (NS_WARN_IF(!sentActor)) {
@@ -70,6 +74,11 @@ ClientManager::~ClientManager() {
 
   Shutdown();
 
+  MOZ_DIAGNOSTIC_ASSERT(sClientManagerThreadLocalMagic1 == kThreadLocalMagic1);
+  MOZ_DIAGNOSTIC_ASSERT(sClientManagerThreadLocalMagic2 == kThreadLocalMagic2);
+  MOZ_DIAGNOSTIC_ASSERT(sClientManagerThreadLocalIndex != kBadThreadLocalIndex);
+  MOZ_DIAGNOSTIC_ASSERT(sClientManagerThreadLocalIndex ==
+                        sClientManagerThreadLocalIndexDuplicate);
   MOZ_DIAGNOSTIC_ASSERT(this ==
                         PR_GetThreadPrivate(sClientManagerThreadLocalIndex));
 
@@ -96,7 +105,7 @@ UniquePtr<ClientSource> ClientManager::CreateSourceInternal(
   NS_ASSERT_OWNINGTHREAD(ClientManager);
 
   nsID id;
-  nsresult rv = nsContentUtils::GenerateUUIDInPlace(id);
+  nsresult rv = nsID::GenerateUUIDInPlace(id);
   MOZ_DIAGNOSTIC_ASSERT(NS_SUCCEEDED(rv));
   if (NS_WARN_IF(NS_FAILED(rv))) {
     // If we can't even get a UUID, at least make sure not to use a garbage
@@ -106,7 +115,7 @@ UniquePtr<ClientSource> ClientManager::CreateSourceInternal(
     ClientSourceConstructorArgs args(id, aType, aPrincipal, TimeStamp::Now());
     UniquePtr<ClientSource> source(new ClientSource(this, aEventTarget, args));
     source->Shutdown();
-    return Move(source);
+    return source;
   }
 
   ClientSourceConstructorArgs args(id, aType, aPrincipal, TimeStamp::Now());
@@ -114,12 +123,31 @@ UniquePtr<ClientSource> ClientManager::CreateSourceInternal(
 
   if (IsShutdown()) {
     source->Shutdown();
-    return Move(source);
+    return source;
   }
 
   source->Activate(GetActor());
 
-  return Move(source);
+  return source;
+}
+
+UniquePtr<ClientSource> ClientManager::CreateSourceInternal(
+    const ClientInfo& aClientInfo, nsISerialEventTarget* aEventTarget) {
+  NS_ASSERT_OWNINGTHREAD(ClientManager);
+
+  ClientSourceConstructorArgs args(aClientInfo.Id(), aClientInfo.Type(),
+                                   aClientInfo.PrincipalInfo(),
+                                   aClientInfo.CreationTime());
+  UniquePtr<ClientSource> source(new ClientSource(this, aEventTarget, args));
+
+  if (IsShutdown()) {
+    source->Shutdown();
+    return source;
+  }
+
+  source->Activate(GetActor());
+
+  return source;
 }
 
 already_AddRefed<ClientHandle> ClientManager::CreateHandleInternal(
@@ -140,7 +168,7 @@ already_AddRefed<ClientHandle> ClientManager::CreateHandleInternal(
   return handle.forget();
 }
 
-already_AddRefed<ClientOpPromise> ClientManager::StartOp(
+RefPtr<ClientOpPromise> ClientManager::StartOp(
     const ClientOpConstructorArgs& aArgs,
     nsISerialEventTarget* aSerialEventTarget) {
   RefPtr<ClientOpPromise::Private> promise =
@@ -150,27 +178,32 @@ already_AddRefed<ClientOpPromise> ClientManager::StartOp(
   // the ClientHandle might get de-refed and teardown the actor before we
   // get an answer.
   RefPtr<ClientManager> kungFuGrip = this;
-  promise->Then(aSerialEventTarget, __func__,
-                [kungFuGrip](const ClientOpResult&) {},
-                [kungFuGrip](nsresult) {});
 
   MaybeExecute(
-      [aArgs, promise](ClientManagerChild* aActor) {
-        ClientManagerOpChild* actor = new ClientManagerOpChild(aArgs, promise);
+      [&aArgs, promise, kungFuGrip](ClientManagerChild* aActor) {
+        ClientManagerOpChild* actor =
+            new ClientManagerOpChild(kungFuGrip, aArgs, promise);
         if (!aActor->SendPClientManagerOpConstructor(actor, aArgs)) {
           // Constructor failure will reject promise via ActorDestroy()
           return;
         }
       },
-      [promise] { promise->Reject(NS_ERROR_DOM_INVALID_STATE_ERR, __func__); });
+      [promise] {
+        CopyableErrorResult rv;
+        rv.ThrowInvalidStateError("Client has been destroyed");
+        promise->Reject(rv, __func__);
+      });
 
-  RefPtr<ClientOpPromise> ref = promise.get();
-  return ref.forget();
+  return promise;
 }
 
 // static
 already_AddRefed<ClientManager> ClientManager::GetOrCreateForCurrentThread() {
+  MOZ_DIAGNOSTIC_ASSERT(sClientManagerThreadLocalMagic1 == kThreadLocalMagic1);
+  MOZ_DIAGNOSTIC_ASSERT(sClientManagerThreadLocalMagic2 == kThreadLocalMagic2);
   MOZ_DIAGNOSTIC_ASSERT(sClientManagerThreadLocalIndex != kBadThreadLocalIndex);
+  MOZ_DIAGNOSTIC_ASSERT(sClientManagerThreadLocalIndex ==
+                        sClientManagerThreadLocalIndexDuplicate);
   RefPtr<ClientManager> cm = static_cast<ClientManager*>(
       PR_GetThreadPrivate(sClientManagerThreadLocalIndex));
 
@@ -184,7 +217,16 @@ already_AddRefed<ClientManager> ClientManager::GetOrCreateForCurrentThread() {
     MOZ_DIAGNOSTIC_ASSERT(status == PR_SUCCESS);
   }
 
-  MOZ_ASSERT(cm);
+  MOZ_DIAGNOSTIC_ASSERT(cm);
+
+  if (StaticPrefs::dom_workers_testing_enabled()) {
+    // Check that the ClientManager instance associated to the current thread
+    // has not been kept alive when it was expected to have been already
+    // deallocated (e.g. due to a leak ClientManager's mShutdown can have ben
+    // set to true from its RevokeActor method but never fully deallocated and
+    // unset from the thread locals).
+    MOZ_DIAGNOSTIC_ASSERT(!cm->IsShutdown());
+  }
   return cm.forget();
 }
 
@@ -194,16 +236,62 @@ WorkerPrivate* ClientManager::GetWorkerPrivate() const {
   return GetActor()->GetWorkerPrivate();
 }
 
+// Used to share logic between ExpectFutureSource and ForgetFutureSource.
+/* static */ bool ClientManager::ExpectOrForgetFutureSource(
+    const ClientInfo& aClientInfo,
+    bool (PClientManagerChild::*aMethod)(const IPCClientInfo&)) {
+  // Return earlier if called late in the XPCOM shutdown path,
+  // ClientManager would be already shutdown at the point.
+  if (NS_WARN_IF(PastShutdownPhase(ShutdownPhase::XPCOMShutdown))) {
+    return false;
+  }
+
+  bool rv = true;
+
+  RefPtr<ClientManager> mgr = ClientManager::GetOrCreateForCurrentThread();
+  mgr->MaybeExecute(
+      [&](ClientManagerChild* aActor) {
+        if (!(aActor->*aMethod)(aClientInfo.ToIPC())) {
+          rv = false;
+        }
+      },
+      [&] { rv = false; });
+
+  return rv;
+}
+
+/* static */ bool ClientManager::ExpectFutureSource(
+    const ClientInfo& aClientInfo) {
+  return ExpectOrForgetFutureSource(
+      aClientInfo, &PClientManagerChild::SendExpectFutureClientSource);
+}
+
+/* static */ bool ClientManager::ForgetFutureSource(
+    const ClientInfo& aClientInfo) {
+  return ExpectOrForgetFutureSource(
+      aClientInfo, &PClientManagerChild::SendForgetFutureClientSource);
+}
+
 // static
 void ClientManager::Startup() {
   MOZ_ASSERT(NS_IsMainThread());
+
+  MOZ_DIAGNOSTIC_ASSERT(sClientManagerThreadLocalMagic1 == kThreadLocalMagic1);
+  MOZ_DIAGNOSTIC_ASSERT(sClientManagerThreadLocalMagic2 == kThreadLocalMagic2);
+  MOZ_DIAGNOSTIC_ASSERT(sClientManagerThreadLocalIndex == kBadThreadLocalIndex);
+  MOZ_DIAGNOSTIC_ASSERT(sClientManagerThreadLocalIndex ==
+                        sClientManagerThreadLocalIndexDuplicate);
+
 #ifdef MOZ_DIAGNOSTIC_ASSERT_ENABLED
   PRStatus status =
 #endif
       PR_NewThreadPrivateIndex(&sClientManagerThreadLocalIndex, nullptr);
   MOZ_DIAGNOSTIC_ASSERT(status == PR_SUCCESS);
 
-  ClientPrefsInit();
+  MOZ_DIAGNOSTIC_ASSERT(sClientManagerThreadLocalIndex != kBadThreadLocalIndex);
+#ifdef MOZ_DIAGNOSTIC_ASSERT_ENABLED
+  sClientManagerThreadLocalIndexDuplicate = sClientManagerThreadLocalIndex;
+#endif
 }
 
 // static
@@ -231,6 +319,34 @@ UniquePtr<ClientSource> ClientManager::CreateSource(
   return mgr->CreateSourceInternal(aType, aEventTarget, aPrincipal);
 }
 
+// static
+UniquePtr<ClientSource> ClientManager::CreateSourceFromInfo(
+    const ClientInfo& aClientInfo, nsISerialEventTarget* aEventTarget) {
+  RefPtr<ClientManager> mgr = GetOrCreateForCurrentThread();
+  return mgr->CreateSourceInternal(aClientInfo, aEventTarget);
+}
+
+Maybe<ClientInfo> ClientManager::CreateInfo(ClientType aType,
+                                            nsIPrincipal* aPrincipal) {
+  MOZ_ASSERT(NS_IsMainThread());
+  MOZ_ASSERT(aPrincipal);
+
+  PrincipalInfo principalInfo;
+  nsresult rv = PrincipalToPrincipalInfo(aPrincipal, &principalInfo);
+  if (NS_WARN_IF(NS_FAILED(rv))) {
+    MOZ_CRASH("ClientManager::CreateSource() cannot serialize bad principal");
+  }
+
+  nsID id;
+  rv = nsID::GenerateUUIDInPlace(id);
+  MOZ_DIAGNOSTIC_ASSERT(NS_SUCCEEDED(rv));
+  if (NS_WARN_IF(NS_FAILED(rv))) {
+    return Nothing();
+  }
+
+  return Some(ClientInfo(id, aType, principalInfo, TimeStamp::Now()));
+}
+
 // static
 already_AddRefed<ClientHandle> ClientManager::CreateHandle(
     const ClientInfo& aClientInfo, nsISerialEventTarget* aSerialEventTarget) {
@@ -275,5 +391,4 @@ RefPtr<ClientOpPromise> ClientManager::OpenWindow(
   return mgr->StartOp(aArgs, aSerialEventTarget);
 }
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
