@@ -12,7 +12,6 @@
 #include "nsCOMPtr.h"
 #include "nsIFrame.h"
 #include "nsINode.h"
-#include "nsISelectionController.h"
 
 class nsPresContext;
 class nsRange;
@@ -21,6 +20,11 @@ struct nsRect;
 
 namespace mozilla {
 
+namespace dom {
+class Element;
+class Text;
+}  // namespace dom
+
 enum LineBreakType { LINE_BREAK_TYPE_NATIVE, LINE_BREAK_TYPE_XP };
 
 /*
@@ -40,7 +44,7 @@ class MOZ_STACK_CLASS ContentEventHandler {
    */
   class MOZ_STACK_CLASS RawRange final {
    public:
-    RawRange() {}
+    RawRange() = default;
 
     void Clear() {
       mRoot = nullptr;
@@ -52,8 +56,13 @@ class MOZ_STACK_CLASS ContentEventHandler {
     bool Collapsed() const { return mStart == mEnd && IsPositioned(); }
     nsINode* GetStartContainer() const { return mStart.Container(); }
     nsINode* GetEndContainer() const { return mEnd.Container(); }
-    uint32_t StartOffset() const { return mStart.Offset(); }
-    uint32_t EndOffset() const { return mEnd.Offset(); }
+    uint32_t StartOffset() const {
+      return *mStart.Offset(
+          RangeBoundary::OffsetFilter::kValidOrInvalidOffsets);
+    }
+    uint32_t EndOffset() const {
+      return *mEnd.Offset(RangeBoundary::OffsetFilter::kValidOrInvalidOffsets);
+    }
     nsIContent* StartRef() const { return mStart.Ref(); }
     nsIContent* EndRef() const { return mEnd.Ref(); }
 
@@ -81,10 +90,9 @@ class MOZ_STACK_CLASS ContentEventHandler {
     nsresult SetStartAndEnd(const RawRangeBoundary& aStart,
                             const RawRangeBoundary& aEnd);
 
-    nsresult SelectNodeContents(nsINode* aNodeToSelectContents);
+    nsresult SelectNodeContents(const nsINode* aNodeToSelectContents);
 
    private:
-    nsINode* IsValidBoundary(nsINode* aNode) const;
     inline void AssertStartIsBeforeOrEqualToEnd();
 
     nsCOMPtr<nsINode> mRoot;
@@ -94,39 +102,49 @@ class MOZ_STACK_CLASS ContentEventHandler {
   };
 
  public:
-  typedef dom::Selection Selection;
+  using Element = dom::Element;
+  using Selection = dom::Selection;
 
   explicit ContentEventHandler(nsPresContext* aPresContext);
 
   // Handle aEvent in the current process.
-  nsresult HandleQueryContentEvent(WidgetQueryContentEvent* aEvent);
+  MOZ_CAN_RUN_SCRIPT nsresult
+  HandleQueryContentEvent(WidgetQueryContentEvent* aEvent);
 
   // eQuerySelectedText event handler
-  nsresult OnQuerySelectedText(WidgetQueryContentEvent* aEvent);
+  MOZ_CAN_RUN_SCRIPT nsresult
+  OnQuerySelectedText(WidgetQueryContentEvent* aEvent);
   // eQueryTextContent event handler
-  nsresult OnQueryTextContent(WidgetQueryContentEvent* aEvent);
+  MOZ_CAN_RUN_SCRIPT nsresult
+  OnQueryTextContent(WidgetQueryContentEvent* aEvent);
   // eQueryCaretRect event handler
-  nsresult OnQueryCaretRect(WidgetQueryContentEvent* aEvent);
+  MOZ_CAN_RUN_SCRIPT nsresult OnQueryCaretRect(WidgetQueryContentEvent* aEvent);
   // eQueryTextRect event handler
-  nsresult OnQueryTextRect(WidgetQueryContentEvent* aEvent);
+  MOZ_CAN_RUN_SCRIPT nsresult OnQueryTextRect(WidgetQueryContentEvent* aEvent);
   // eQueryTextRectArray event handler
-  nsresult OnQueryTextRectArray(WidgetQueryContentEvent* aEvent);
+  MOZ_CAN_RUN_SCRIPT nsresult
+  OnQueryTextRectArray(WidgetQueryContentEvent* aEvent);
   // eQueryEditorRect event handler
-  nsresult OnQueryEditorRect(WidgetQueryContentEvent* aEvent);
+  MOZ_CAN_RUN_SCRIPT nsresult
+  OnQueryEditorRect(WidgetQueryContentEvent* aEvent);
   // eQueryContentState event handler
-  nsresult OnQueryContentState(WidgetQueryContentEvent* aEvent);
+  MOZ_CAN_RUN_SCRIPT nsresult
+  OnQueryContentState(WidgetQueryContentEvent* aEvent);
   // eQuerySelectionAsTransferable event handler
-  nsresult OnQuerySelectionAsTransferable(WidgetQueryContentEvent* aEvent);
+  MOZ_CAN_RUN_SCRIPT nsresult
+  OnQuerySelectionAsTransferable(WidgetQueryContentEvent* aEvent);
   // eQueryCharacterAtPoint event handler
-  nsresult OnQueryCharacterAtPoint(WidgetQueryContentEvent* aEvent);
+  MOZ_CAN_RUN_SCRIPT nsresult
+  OnQueryCharacterAtPoint(WidgetQueryContentEvent* aEvent);
   // eQueryDOMWidgetHittest event handler
-  nsresult OnQueryDOMWidgetHittest(WidgetQueryContentEvent* aEvent);
+  MOZ_CAN_RUN_SCRIPT nsresult
+  OnQueryDOMWidgetHittest(WidgetQueryContentEvent* aEvent);
 
   // NS_SELECTION_* event
-  nsresult OnSelectionEvent(WidgetSelectionEvent* aEvent);
+  MOZ_CAN_RUN_SCRIPT nsresult OnSelectionEvent(WidgetSelectionEvent* aEvent);
 
  protected:
-  nsCOMPtr<nsIDocument> mDocument;
+  RefPtr<dom::Document> mDocument;
   // mSelection is typically normal selection but if OnQuerySelectedText()
   // is called, i.e., handling eQuerySelectedText, it's the specified selection
   // by WidgetQueryContentEvent::mInput::mSelectionType.
@@ -134,20 +152,24 @@ class MOZ_STACK_CLASS ContentEventHandler {
   // mFirstSelectedRawRange is initialized from the first range of mSelection,
   // if it exists.  Otherwise, it is reset by Clear().
   RawRange mFirstSelectedRawRange;
-  nsCOMPtr<nsIContent> mRootContent;
+  RefPtr<Element> mRootElement;
 
-  nsresult Init(WidgetQueryContentEvent* aEvent);
-  nsresult Init(WidgetSelectionEvent* aEvent);
+  MOZ_CAN_RUN_SCRIPT nsresult Init(WidgetQueryContentEvent* aEvent);
+  MOZ_CAN_RUN_SCRIPT nsresult Init(WidgetSelectionEvent* aEvent);
 
-  nsresult InitBasic();
-  nsresult InitCommon(SelectionType aSelectionType = SelectionType::eNormal);
+  nsresult InitBasic(bool aRequireFlush = true);
+  MOZ_CAN_RUN_SCRIPT nsresult
+  InitCommon(EventMessage aEventMessage,
+             SelectionType aSelectionType = SelectionType::eNormal,
+             bool aRequireFlush = true);
   /**
    * InitRootContent() computes the root content of current focused editor.
    *
    * @param aNormalSelection    This must be a Selection instance whose type is
    *                            SelectionType::eNormal.
    */
-  nsresult InitRootContent(Selection* aNormalSelection);
+  MOZ_CAN_RUN_SCRIPT nsresult
+  InitRootContent(const Selection& aNormalSelection);
 
  public:
   // FlatText means the text that is generated from DOM tree. The BR elements
@@ -164,9 +186,9 @@ class MOZ_STACK_CLASS ContentEventHandler {
     // referred.
     bool mAfterOpenTag = true;
 
-    NodePosition() : RangeBoundary() {}
+    NodePosition() = default;
 
-    NodePosition(nsINode* aContainer, int32_t aOffset)
+    NodePosition(nsINode* aContainer, uint32_t aOffset)
         : RangeBoundary(aContainer, aOffset) {}
 
     NodePosition(nsINode* aContainer, nsIContent* aRef)
@@ -193,7 +215,7 @@ class MOZ_STACK_CLASS ContentEventHandler {
   // Offset() is not 0, though, when Container() is an element node and mOffset
   // is 0, this is treated as before the open tag of Container().
   struct NodePositionBefore final : public NodePosition {
-    NodePositionBefore(nsINode* aContainer, int32_t aOffset)
+    NodePositionBefore(nsINode* aContainer, uint32_t aOffset)
         : NodePosition(aContainer, aOffset) {
       mAfterOpenTag = false;
     }
@@ -207,8 +229,8 @@ class MOZ_STACK_CLASS ContentEventHandler {
   // Get the flatten text length in the range.
   // @param aStartPosition      Start node and offset in the node of the range.
   // @param aEndPosition        End node and offset in the node of the range.
-  // @param aRootContent        The root content of the editor or document.
-  //                            aRootContent won't cause any text including
+  // @param aRootElement        The root element of the editor or document.
+  //                            aRootElement won't cause any text including
   //                            line breaks.
   // @param aLength             The result of the flatten text length of the
   //                            range.
@@ -224,40 +246,38 @@ class MOZ_STACK_CLASS ContentEventHandler {
   //                            be number of the children of mNode.
   static nsresult GetFlatTextLengthInRange(const NodePosition& aStartPosition,
                                            const NodePosition& aEndPosition,
-                                           nsIContent* aRootContent,
+                                           const Element* aRootElement,
                                            uint32_t* aLength,
                                            LineBreakType aLineBreakType,
                                            bool aIsRemovingNode = false);
   // Computes the native text length between aStartOffset and aEndOffset of
-  // aContent.  aContent must be a text node.
-  static uint32_t GetNativeTextLength(nsIContent* aContent,
+  // aTextNode.
+  static uint32_t GetNativeTextLength(const dom::Text& aTextNode,
                                       uint32_t aStartOffset,
                                       uint32_t aEndOffset);
-  // Get the native text length of aContent.  aContent must be a text node.
-  static uint32_t GetNativeTextLength(nsIContent* aContent,
+  // Get the native text length of aTextNode.
+  static uint32_t GetNativeTextLength(const dom::Text& aTextNode,
                                       uint32_t aMaxLength = UINT32_MAX);
-  // Get the native text length which is inserted before aContent.
-  // aContent should be an element.
-  static uint32_t GetNativeTextLengthBefore(nsIContent* aContent,
-                                            nsINode* aRootNode);
+
+  static uint32_t GetNativeTextLength(const nsAString& aText);
 
  protected:
-  // Get the text length of aContent.  aContent must be a text node.
-  static uint32_t GetTextLength(nsIContent* aContent,
+  // Get the text length of aTextNode.
+  static uint32_t GetTextLength(const dom::Text& aTextNode,
                                 LineBreakType aLineBreakType,
                                 uint32_t aMaxLength = UINT32_MAX);
   // Get the text length of a given range of a content node in
   // the given line break type.
-  static uint32_t GetTextLengthInRange(nsIContent* aContent,
+  static uint32_t GetTextLengthInRange(const dom::Text& aTextNode,
                                        uint32_t aXPStartOffset,
                                        uint32_t aXPEndOffset,
                                        LineBreakType aLineBreakType);
-  // Get the contents in aContent (meaning all children of aContent) as plain
-  // text.  E.g., specifying mRootContent gets whole text in it.
+  // Get the contents in aElement (meaning all children of aElement) as plain
+  // text.  E.g., specifying mRootElement gets whole text in it.
   // Note that the result is not same as .textContent.  The result is
   // optimized for native IMEs.  For example, <br> element and some block
   // elements causes "\n" (or "\r\n"), see also ShouldBreakLineBefore().
-  nsresult GenerateFlatTextContent(nsIContent* aContent, nsString& aString,
+  nsresult GenerateFlatTextContent(const Element* aElement, nsString& aString,
                                    LineBreakType aLineBreakType);
   // Get the contents of aRange as plain text.
   nsresult GenerateFlatTextContent(const RawRange& aRawRange, nsString& aString,
@@ -270,7 +290,8 @@ class MOZ_STACK_CLASS ContentEventHandler {
   // Check if we should insert a line break before aContent.
   // This should return false only when aContent is an html element which
   // is typically used in a paragraph like <em>.
-  static bool ShouldBreakLineBefore(nsIContent* aContent, nsINode* aRootNode);
+  static bool ShouldBreakLineBefore(const nsIContent& aContent,
+                                    const Element* aRootElement);
   // Get the line breaker length.
   static inline uint32_t GetBRLength(LineBreakType aLineBreakType);
   static LineBreakType GetLineBreakType(WidgetQueryContentEvent* aEvent);
@@ -278,8 +299,6 @@ class MOZ_STACK_CLASS ContentEventHandler {
   static LineBreakType GetLineBreakType(bool aUseNativeLineBreak);
   // Returns focused content (including its descendant documents).
   nsIContent* GetFocusedContent();
-  // Returns true if the content is a plugin host.
-  bool IsPlugin(nsIContent* aContent);
   // QueryContentRect() sets the rect of aContent's frame(s) to aEvent.
   nsresult QueryContentRect(nsIContent* aContent,
                             WidgetQueryContentEvent* aEvent);
@@ -291,7 +310,7 @@ class MOZ_STACK_CLASS ContentEventHandler {
                                          LineBreakType aLineBreakType,
                                          bool aExpandToClusterBoundaries,
                                          uint32_t* aNewOffset = nullptr,
-                                         nsIContent** aLastTextNode = nullptr);
+                                         dom::Text** aLastTextNode = nullptr);
   // If the aCollapsedRawRange isn't in text node but next to a text node,
   // this method modifies it in the text node.  Otherwise, not modified.
   nsresult AdjustCollapsedRangeMaybeIntoTextNode(RawRange& aCollapsedRawRange);
@@ -300,12 +319,14 @@ class MOZ_STACK_CLASS ContentEventHandler {
   nsresult ConvertToRootRelativeOffset(nsIFrame* aFrame, nsRect& aRect);
   // Expand aXPOffset to the nearest offset in cluster boundary. aForward is
   // true, it is expanded to forward.
-  nsresult ExpandToClusterBoundary(nsIContent* aContent, bool aForward,
-                                   uint32_t* aXPOffset);
+  // FYI: Due to `nsFrameSelection::GetFrameForNodeOffset()`, this cannot
+  //      take `const dom::Text&`.
+  nsresult ExpandToClusterBoundary(dom::Text& aTextNode, bool aForward,
+                                   uint32_t* aXPOffset) const;
 
-  typedef nsTArray<mozilla::FontRange> FontRangeArray;
+  using FontRangeArray = nsTArray<mozilla::FontRange>;
   static void AppendFontRanges(FontRangeArray& aFontRanges,
-                               nsIContent* aContent, uint32_t aBaseOffset,
+                               const dom::Text& aTextNode, uint32_t aBaseOffset,
                                uint32_t aXPStartOffset, uint32_t aXPEndOffset,
                                LineBreakType aLineBreakType);
   nsresult GenerateFlatFontRanges(const RawRange& aRawRange,
@@ -316,11 +337,6 @@ class MOZ_STACK_CLASS ContentEventHandler {
                                 LayoutDeviceIntRect& aRect,
                                 WritingMode& aWritingMode);
 
-  // Returns a node and position in the node for computing text rect.
-  NodePosition GetNodePositionHavingFlatText(const NodePosition& aNodePosition);
-  NodePosition GetNodePositionHavingFlatText(nsINode* aNode,
-                                             int32_t aNodeOffset);
-
   struct MOZ_STACK_CLASS FrameAndNodeOffset final {
     // mFrame is safe since this can live in only stack class and
     // ContentEventHandler doesn't modify layout after
@@ -378,14 +394,14 @@ class MOZ_STACK_CLASS ContentEventHandler {
   // doesn't check if aFrame should cause line break in non-debug build.
   FrameRelativeRect GetLineBreakerRectBefore(nsIFrame* aFrame);
 
-  // Returns a line breaker rect after aTextContent as there is a line breaker
-  // immediately after aTextContent.  This is useful when following block
+  // Returns a line breaker rect after aTextNode as there is a line breaker
+  // immediately after aTextNode.  This is useful when following block
   // element causes a line break before it and it needs to compute the line
   // breaker's rect.  For example, if there is |<p>abc</p><p>def</p>|, the
   // rect of 2nd <p>'s line breaker should be at right of "c" in the first
   // <p>, not the start of 2nd <p>.  The result is relative to the last text
-  // frame which represents the last character of aTextContent.
-  FrameRelativeRect GuessLineBreakerRectAfter(nsIContent* aTextContent);
+  // frame which represents the last character of aTextNode.
+  FrameRelativeRect GuessLineBreakerRectAfter(const dom::Text& aTextNode);
 
   // Returns a guessed first rect.  I.e., it may be different from actual
   // caret when selection is collapsed at start of aFrame.  For example, this
@@ -408,6 +424,19 @@ class MOZ_STACK_CLASS ContentEventHandler {
   // to 2 pixels because x and y may not be aligned to device pixels.
   void EnsureNonEmptyRect(nsRect& aRect) const;
   void EnsureNonEmptyRect(LayoutDeviceIntRect& aRect) const;
+
+  /**
+   * Compute caret rect before or after a character rect.
+   */
+  static LayoutDeviceIntRect GetCaretRectBefore(
+      const LayoutDeviceIntRect& aCharRect, const WritingMode& aWritingMode);
+  static LayoutDeviceIntRect GetCaretRectAfter(
+      const LayoutDeviceIntRect& aCharRect, const WritingMode& aWritingMode);
+  static nsRect GetCaretRectBefore(const nsRect& aCharRect,
+                                   const WritingMode& aWritingMode);
+  static nsRect GetCaretRectAfter(nsPresContext& aPresContext,
+                                  const nsRect& aCharRect,
+                                  const WritingMode& aWritingMode);
 };
 
 }  // namespace mozilla
