@@ -7,20 +7,30 @@
 #include "DataTransferItem.h"
 #include "DataTransferItemList.h"
 
+#include "mozilla/Attributes.h"
+#include "mozilla/BasePrincipal.h"
 #include "mozilla/ContentEvents.h"
 #include "mozilla/EventForwards.h"
+#include "mozilla/StaticPrefs_dom.h"
+#include "mozilla/dom/BlobImpl.h"
 #include "mozilla/dom/DataTransferItemBinding.h"
 #include "mozilla/dom/Directory.h"
 #include "mozilla/dom/Event.h"
 #include "mozilla/dom/FileSystem.h"
 #include "mozilla/dom/FileSystemDirectoryEntry.h"
 #include "mozilla/dom/FileSystemFileEntry.h"
+#include "imgIContainer.h"
+#include "imgITools.h"
+#include "nsComponentManagerUtils.h"
 #include "nsIClipboard.h"
+#include "nsIFile.h"
+#include "nsIInputStream.h"
 #include "nsISupportsPrimitives.h"
 #include "nsIScriptObjectPrincipal.h"
 #include "nsNetUtil.h"
 #include "nsQueryObject.h"
 #include "nsContentUtils.h"
+#include "nsThreadUtils.h"
 #include "nsVariant.h"
 
 namespace {
@@ -37,8 +47,7 @@ FileMimeNameData kFileMimeNameMap[] = {
 
 }  // anonymous namespace
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 NS_IMPL_CYCLE_COLLECTION_WRAPPERCACHE(DataTransferItem, mData, mPrincipal,
                                       mDataTransfer, mCachedFile)
@@ -52,7 +61,7 @@ NS_INTERFACE_MAP_END
 
 JSObject* DataTransferItem::WrapObject(JSContext* aCx,
                                        JS::Handle<JSObject*> aGivenProto) {
-  return DataTransferItemBinding::Wrap(aCx, this, aGivenProto);
+  return DataTransferItem_Binding::Wrap(aCx, this, aGivenProto);
 }
 
 already_AddRefed<DataTransferItem> DataTransferItem::Clone(
@@ -80,6 +89,8 @@ void DataTransferItem::SetData(nsIVariant* aData) {
     // These are provided by the system, and have guaranteed properties about
     // their kind based on their type.
     MOZ_ASSERT(!mType.IsEmpty());
+    // This type should not be provided by the OS.
+    MOZ_ASSERT(!mType.EqualsASCII(kNativeImageMime));
 
     mKind = KIND_STRING;
     for (uint32_t i = 0; i < ArrayLength(kFileMimeNameMap); ++i) {
@@ -103,11 +114,19 @@ void DataTransferItem::SetData(nsIVariant* aData) {
   nsresult rv = aData->GetAsISupports(getter_AddRefs(supports));
   if (NS_SUCCEEDED(rv) && supports) {
     // Check if we have one of the supported file data formats
-    if (nsCOMPtr<nsIDOMBlob>(do_QueryInterface(supports)) ||
+    if (RefPtr<Blob>(do_QueryObject(supports)) ||
         nsCOMPtr<BlobImpl>(do_QueryInterface(supports)) ||
         nsCOMPtr<nsIFile>(do_QueryInterface(supports))) {
       return KIND_FILE;
     }
+
+    if (StaticPrefs::dom_events_dataTransfer_imageAsFile_enabled()) {
+      // Firefox internally uses imgIContainer to represent images being
+      // copied/dragged. These need to be encoded to PNG files.
+      if (nsCOMPtr<imgIContainer>(do_QueryInterface(supports))) {
+        return KIND_FILE;
+      }
+    }
   }
 
   nsAutoString string;
@@ -131,49 +150,48 @@ void DataTransferItem::FillInExternalData() {
 
   NS_ConvertUTF16toUTF8 utf8format(mType);
   const char* format = utf8format.get();
-  if (strcmp(format, "text/plain") == 0) {
-    format = kUnicodeMime;
-  } else if (strcmp(format, "text/uri-list") == 0) {
+  if (strcmp(format, "text/uri-list") == 0) {
     format = kURLDataMime;
   }
 
-  nsCOMPtr<nsITransferable> trans =
-      do_CreateInstance("@mozilla.org/widget/transferable;1");
-  if (NS_WARN_IF(!trans)) {
-    return;
-  }
-
-  trans->Init(nullptr);
-  trans->AddDataFlavor(format);
-
-  if (mDataTransfer->GetEventMessage() == ePaste) {
-    MOZ_ASSERT(mIndex == 0, "index in clipboard must be 0");
-
-    nsCOMPtr<nsIClipboard> clipboard =
-        do_GetService("@mozilla.org/widget/clipboard;1");
-    if (!clipboard || mDataTransfer->ClipboardType() < 0) {
+  nsCOMPtr<nsITransferable> trans = mDataTransfer->GetTransferable();
+  if (!trans) {
+    trans = do_CreateInstance("@mozilla.org/widget/transferable;1");
+    if (NS_WARN_IF(!trans)) {
       return;
     }
 
-    nsresult rv = clipboard->GetData(trans, mDataTransfer->ClipboardType());
-    if (NS_WARN_IF(NS_FAILED(rv))) {
-      return;
-    }
-  } else {
-    nsCOMPtr<nsIDragSession> dragSession = nsContentUtils::GetDragSession();
-    if (!dragSession) {
-      return;
-    }
+    trans->Init(nullptr);
+    trans->AddDataFlavor(format);
 
-    nsresult rv = dragSession->GetData(trans, mIndex);
-    if (NS_WARN_IF(NS_FAILED(rv))) {
-      return;
+    if (mDataTransfer->GetEventMessage() == ePaste) {
+      MOZ_ASSERT(mIndex == 0, "index in clipboard must be 0");
+
+      nsCOMPtr<nsIClipboard> clipboard =
+          do_GetService("@mozilla.org/widget/clipboard;1");
+      if (!clipboard || mDataTransfer->ClipboardType() < 0) {
+        return;
+      }
+
+      nsresult rv = clipboard->GetData(trans, mDataTransfer->ClipboardType());
+      if (NS_WARN_IF(NS_FAILED(rv))) {
+        return;
+      }
+    } else {
+      nsCOMPtr<nsIDragSession> dragSession = nsContentUtils::GetDragSession();
+      if (!dragSession) {
+        return;
+      }
+
+      nsresult rv = dragSession->GetData(trans, mIndex);
+      if (NS_WARN_IF(NS_FAILED(rv))) {
+        return;
+      }
     }
   }
 
-  uint32_t length = 0;
   nsCOMPtr<nsISupports> data;
-  nsresult rv = trans->GetTransferData(format, getter_AddRefs(data), &length);
+  nsresult rv = trans->GetTransferData(format, getter_AddRefs(data));
   if (NS_WARN_IF(NS_FAILED(rv) || !data)) {
     return;
   }
@@ -194,7 +212,6 @@ void DataTransferItem::FillInExternalData() {
       }
       data = do_QueryObject(file);
     }
-
     variant->SetAsISupports(data);
   } else {
     // We have an external piece of string data. Extract it and store it in the
@@ -282,17 +299,45 @@ already_AddRefed<File> DataTransferItem::GetAsFile(
       return nullptr;
     }
 
-    if (nsCOMPtr<nsIDOMBlob> domBlob = do_QueryInterface(supports)) {
-      Blob* blob = static_cast<Blob*>(domBlob.get());
+    if (RefPtr<Blob> blob = do_QueryObject(supports)) {
       mCachedFile = blob->ToFile();
-    } else if (nsCOMPtr<BlobImpl> blobImpl = do_QueryInterface(supports)) {
-      MOZ_ASSERT(blobImpl->IsFile());
-      mCachedFile = File::Create(mDataTransfer, blobImpl);
-    } else if (nsCOMPtr<nsIFile> ifile = do_QueryInterface(supports)) {
-      mCachedFile = File::CreateFromFile(mDataTransfer, ifile);
     } else {
-      MOZ_ASSERT(false, "One of the above code paths should be taken");
-      return nullptr;
+      nsCOMPtr<nsIGlobalObject> global = GetGlobalFromDataTransfer();
+      if (NS_WARN_IF(!global)) {
+        return nullptr;
+      }
+
+      if (nsCOMPtr<BlobImpl> blobImpl = do_QueryInterface(supports)) {
+        MOZ_ASSERT(blobImpl->IsFile());
+        mCachedFile = File::Create(global, blobImpl);
+        if (NS_WARN_IF(!mCachedFile)) {
+          return nullptr;
+        }
+      } else if (nsCOMPtr<nsIFile> ifile = do_QueryInterface(supports)) {
+        mCachedFile = File::CreateFromFile(global, ifile);
+        if (NS_WARN_IF(!mCachedFile)) {
+          return nullptr;
+        }
+      } else if (nsCOMPtr<imgIContainer> img = do_QueryInterface(supports)) {
+        nsCOMPtr<imgITools> imgTools =
+            do_CreateInstance("@mozilla.org/image/tools;1");
+
+        nsCOMPtr<nsIInputStream> inputStream;
+        nsresult rv = imgTools->EncodeImage(img, "image/png"_ns, u""_ns,
+                                            getter_AddRefs(inputStream));
+        if (NS_WARN_IF(NS_FAILED(rv))) {
+          return nullptr;
+        }
+
+        mCachedFile = CreateFileFromInputStream(
+            inputStream, "GenericImageNamePNG", u"image/png"_ns);
+        if (NS_WARN_IF(!mCachedFile)) {
+          return nullptr;
+        }
+      } else {
+        MOZ_ASSERT(false, "One of the above code paths should be taken");
+        return nullptr;
+      }
     }
   }
 
@@ -307,21 +352,8 @@ already_AddRefed<FileSystemEntry> DataTransferItem::GetAsEntry(
     return nullptr;
   }
 
-  nsCOMPtr<nsIGlobalObject> global;
-  // This is annoying, but DataTransfer may have various things as parent.
-  nsCOMPtr<EventTarget> target =
-      do_QueryInterface(mDataTransfer->GetParentObject());
-  if (target) {
-    global = target->GetOwnerGlobal();
-  } else {
-    nsCOMPtr<nsIDOMEvent> event =
-        do_QueryInterface(mDataTransfer->GetParentObject());
-    if (event) {
-      global = event->InternalDOMEvent()->GetParentObject();
-    }
-  }
-
-  if (!global) {
+  nsCOMPtr<nsIGlobalObject> global = GetGlobalFromDataTransfer();
+  if (NS_WARN_IF(!global)) {
     return nullptr;
   }
 
@@ -376,9 +408,15 @@ already_AddRefed<File> DataTransferItem::CreateFileFromInputStream(
     key = "GenericFileName";
   }
 
+  return CreateFileFromInputStream(aStream, key, mType);
+}
+
+already_AddRefed<File> DataTransferItem::CreateFileFromInputStream(
+    nsIInputStream* aStream, const char* aFileNameKey,
+    const nsAString& aContentType) {
   nsAutoString fileName;
   nsresult rv = nsContentUtils::GetLocalizedString(
-      nsContentUtils::eDOM_PROPERTIES, key, fileName);
+      nsContentUtils::eDOM_PROPERTIES, aFileNameKey, fileName);
   if (NS_WARN_IF(NS_FAILED(rv))) {
     return nullptr;
   }
@@ -390,8 +428,13 @@ already_AddRefed<File> DataTransferItem::CreateFileFromInputStream(
     return nullptr;
   }
 
-  return File::CreateMemoryFile(mDataTransfer, data, available, fileName, mType,
-                                PR_Now());
+  nsCOMPtr<nsIGlobalObject> global = GetGlobalFromDataTransfer();
+  if (NS_WARN_IF(!global)) {
+    return nullptr;
+  }
+
+  return File::CreateMemoryFileWithLastModifiedNow(global, data, available,
+                                                   fileName, aContentType);
 }
 
 void DataTransferItem::GetAsString(FunctionStringCallback* aCallback,
@@ -431,6 +474,9 @@ void DataTransferItem::GetAsString(FunctionStringCallback* aCallback,
           mCallback(aCallback),
           mStringData(aStringData) {}
 
+    // MOZ_CAN_RUN_SCRIPT_BOUNDARY until runnables are opted into
+    // MOZ_CAN_RUN_SCRIPT.  See bug 1535398.
+    MOZ_CAN_RUN_SCRIPT_BOUNDARY
     NS_IMETHOD Run() override {
       ErrorResult rv;
       mCallback->Call(mStringData, rv);
@@ -439,7 +485,7 @@ void DataTransferItem::GetAsString(FunctionStringCallback* aCallback,
     }
 
    private:
-    RefPtr<FunctionStringCallback> mCallback;
+    const RefPtr<FunctionStringCallback> mCallback;
     nsString mStringData;
   };
 
@@ -452,8 +498,8 @@ void DataTransferItem::GetAsString(FunctionStringCallback* aCallback,
   if (parent && !global) {
     if (nsCOMPtr<dom::EventTarget> target = do_QueryInterface(parent)) {
       global = target->GetOwnerGlobal();
-    } else if (nsCOMPtr<nsIDOMEvent> event = do_QueryInterface(parent)) {
-      global = event->InternalDOMEvent()->GetParentObject();
+    } else if (RefPtr<Event> event = do_QueryObject(parent)) {
+      global = event->GetParentObject();
     }
   }
   if (global) {
@@ -482,7 +528,7 @@ already_AddRefed<nsIVariant> DataTransferItem::Data(nsIPrincipal* aPrincipal,
 
   // If the inbound principal is system, we can skip the below checks, as
   // they will trivially succeed.
-  if (nsContentUtils::IsSystemPrincipal(aPrincipal)) {
+  if (aPrincipal->IsSystemPrincipal()) {
     return DataNoSecurityCheck();
   }
 
@@ -503,7 +549,8 @@ already_AddRefed<nsIVariant> DataTransferItem::Data(nsIPrincipal* aPrincipal,
 
   bool checkItemPrincipal = mDataTransfer->IsCrossDomainSubFrameDrop() ||
                             (mDataTransfer->GetEventMessage() != eDrop &&
-                             mDataTransfer->GetEventMessage() != ePaste);
+                             mDataTransfer->GetEventMessage() != ePaste &&
+                             mDataTransfer->GetEventMessage() != eEditorInput);
 
   // Check if the caller is allowed to access the drag data. Callers with
   // chrome privileges can always read the data. During the
@@ -529,12 +576,7 @@ already_AddRefed<nsIVariant> DataTransferItem::Data(nsIPrincipal* aPrincipal,
   if (NS_SUCCEEDED(rv) && data) {
     nsCOMPtr<EventTarget> pt = do_QueryInterface(data);
     if (pt) {
-      nsIScriptContext* c = pt->GetContextForEventHandlers(&rv);
-      if (NS_WARN_IF(NS_FAILED(rv) || !c)) {
-        return nullptr;
-      }
-
-      nsIGlobalObject* go = c->GetGlobalObject();
+      nsIGlobalObject* go = pt->GetOwnerGlobal();
       if (NS_WARN_IF(!go)) {
         return nullptr;
       }
@@ -552,5 +594,22 @@ already_AddRefed<nsIVariant> DataTransferItem::Data(nsIPrincipal* aPrincipal,
   return variant.forget();
 }
 
-}  // namespace dom
-}  // namespace mozilla
+already_AddRefed<nsIGlobalObject>
+DataTransferItem::GetGlobalFromDataTransfer() {
+  nsCOMPtr<nsIGlobalObject> global;
+  // This is annoying, but DataTransfer may have various things as parent.
+  nsCOMPtr<EventTarget> target =
+      do_QueryInterface(mDataTransfer->GetParentObject());
+  if (target) {
+    global = target->GetOwnerGlobal();
+  } else {
+    RefPtr<Event> event = do_QueryObject(mDataTransfer->GetParentObject());
+    if (event) {
+      global = event->GetParentObject();
+    }
+  }
+
+  return global.forget();
+}
+
+}  // namespace mozilla::dom
