@@ -7,32 +7,28 @@
 #ifndef mozilla_dom_DragEvent_h_
 #define mozilla_dom_DragEvent_h_
 
-#include "nsIDOMDragEvent.h"
 #include "mozilla/dom/MouseEvent.h"
 #include "mozilla/dom/DragEventBinding.h"
 #include "mozilla/EventForwards.h"
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 class DataTransfer;
 
-class DragEvent : public MouseEvent, public nsIDOMDragEvent {
+class DragEvent : public MouseEvent {
  public:
   DragEvent(EventTarget* aOwner, nsPresContext* aPresContext,
             WidgetDragEvent* aEvent);
 
-  NS_DECL_ISUPPORTS_INHERITED
-
-  NS_DECL_NSIDOMDRAGEVENT
-
-  NS_FORWARD_TO_MOUSEEVENT
+  NS_INLINE_DECL_REFCOUNTING_INHERITED(DragEvent, MouseEvent)
 
   virtual JSObject* WrapObjectInternal(
       JSContext* aCx, JS::Handle<JSObject*> aGivenProto) override {
-    return DragEventBinding::Wrap(aCx, this, aGivenProto);
+    return DragEvent_Binding::Wrap(aCx, this, aGivenProto);
   }
 
+  DragEvent* AsDragEvent() override { return this; }
+
   DataTransfer* GetDataTransfer();
 
   void InitDragEvent(const nsAString& aType, bool aCanBubble, bool aCancelable,
@@ -44,15 +40,13 @@ class DragEvent : public MouseEvent, public nsIDOMDragEvent {
 
   static already_AddRefed<DragEvent> Constructor(const GlobalObject& aGlobal,
                                                  const nsAString& aType,
-                                                 const DragEventInit& aParam,
-                                                 ErrorResult& aRv);
+                                                 const DragEventInit& aParam);
 
  protected:
-  ~DragEvent() {}
+  ~DragEvent() = default;
 };
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
 
 already_AddRefed<mozilla::dom::DragEvent> NS_NewDOMDragEvent(
     mozilla::dom::EventTarget* aOwner, nsPresContext* aPresContext,
