@@ -10,8 +10,7 @@
 #include "mozilla/dom/MouseEvent.h"
 #include "mozilla/dom/MouseScrollEventBinding.h"
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 class MouseScrollEvent : public MouseEvent {
  public:
@@ -20,12 +19,9 @@ class MouseScrollEvent : public MouseEvent {
 
   NS_INLINE_DECL_REFCOUNTING_INHERITED(MouseScrollEvent, MouseEvent)
 
-  // Forward to base class
-  NS_FORWARD_TO_MOUSEEVENT
-
   virtual JSObject* WrapObjectInternal(
       JSContext* aCx, JS::Handle<JSObject*> aGivenProto) override {
-    return MouseScrollEventBinding::Wrap(aCx, this, aGivenProto);
+    return MouseScrollEvent_Binding::Wrap(aCx, this, aGivenProto);
   }
 
   int32_t Axis();
@@ -39,11 +35,10 @@ class MouseScrollEvent : public MouseEvent {
                             int32_t aAxis);
 
  protected:
-  ~MouseScrollEvent() {}
+  ~MouseScrollEvent() = default;
 };
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
 
 already_AddRefed<mozilla::dom::MouseScrollEvent> NS_NewDOMMouseScrollEvent(
     mozilla::dom::EventTarget* aOwner, nsPresContext* aPresContext,
