@@ -9,39 +9,41 @@
 #include "IMEStateManager.h"
 #include "nsContentUtils.h"
 #include "nsIContent.h"
-#include "nsIPresShell.h"
+#include "nsIMutationObserver.h"
 #include "nsPresContext.h"
 #include "mozilla/AutoRestore.h"
 #include "mozilla/EditorBase.h"
 #include "mozilla/EventDispatcher.h"
 #include "mozilla/IMEStateManager.h"
+#include "mozilla/IntegerRange.h"
 #include "mozilla/MiscEvents.h"
-#include "mozilla/Preferences.h"
+#include "mozilla/PresShell.h"
+#include "mozilla/RangeBoundary.h"
+#include "mozilla/StaticPrefs_dom.h"
+#include "mozilla/StaticPrefs_intl.h"
 #include "mozilla/TextComposition.h"
 #include "mozilla/TextEvents.h"
 #include "mozilla/Unused.h"
-#include "mozilla/dom/TabParent.h"
+#include "mozilla/dom/BrowserParent.h"
 
 #ifdef XP_MACOSX
 // Some defiens will be conflict with OSX SDK
-#define TextRange _TextRange
-#define TextRangeArray _TextRangeArray
-#define Comment _Comment
+#  define TextRange _TextRange
+#  define TextRangeArray _TextRangeArray
+#  define Comment _Comment
 #endif
 
-#include "nsPluginInstanceOwner.h"
-
 #ifdef XP_MACOSX
-#undef TextRange
-#undef TextRangeArray
-#undef Comment
+#  undef TextRange
+#  undef TextRangeArray
+#  undef Comment
 #endif
 
 using namespace mozilla::widget;
 
 namespace mozilla {
 
-#define IDEOGRAPHIC_SPACE (NS_LITERAL_STRING(u"\x3000"))
+#define IDEOGRAPHIC_SPACE (u"\x3000"_ns)
 
 /******************************************************************************
  * TextComposition
@@ -50,11 +52,11 @@ namespace mozilla {
 bool TextComposition::sHandlingSelectionEvent = false;
 
 TextComposition::TextComposition(nsPresContext* aPresContext, nsINode* aNode,
-                                 TabParent* aTabParent,
+                                 BrowserParent* aBrowserParent,
                                  WidgetCompositionEvent* aCompositionEvent)
     : mPresContext(aPresContext),
       mNode(aNode),
-      mTabParent(aTabParent),
+      mBrowserParent(aBrowserParent),
       mNativeContext(aCompositionEvent->mNativeIMEContext),
       mCompositionStartOffset(0),
       mTargetClauseOffsetInComposition(0),
@@ -66,10 +68,11 @@ TextComposition::TextComposition(nsPresContext* aPresContext, nsINode* aNode,
       mIsRequestingCommit(false),
       mIsRequestingCancel(false),
       mRequestedToCommitOrCancel(false),
+      mHasDispatchedDOMTextEvent(false),
       mHasReceivedCommitEvent(false),
       mWasNativeCompositionEndEventDiscarded(false),
-      mAllowControlCharacters(Preferences::GetBool(
-          "dom.compositionevent.allow_control_characters", false)),
+      mAllowControlCharacters(
+          StaticPrefs::dom_compositionevent_allow_control_characters()),
       mWasCompositionStringEmpty(true) {
   MOZ_ASSERT(aCompositionEvent->mNativeIMEContext.IsValid());
 }
@@ -77,7 +80,7 @@ TextComposition::TextComposition(nsPresContext* aPresContext, nsINode* aNode,
 void TextComposition::Destroy() {
   mPresContext = nullptr;
   mNode = nullptr;
-  mTabParent = nullptr;
+  mBrowserParent = nullptr;
   mContainerTextNode = nullptr;
   mCompositionStartOffsetInTextNode = UINT32_MAX;
   mCompositionLengthInTextNode = UINT32_MAX;
@@ -85,20 +88,103 @@ void TextComposition::Destroy() {
   //       this being destroyed for cleaning up the stuff.
 }
 
+void TextComposition::OnCharacterDataChanged(
+    Text& aText, const CharacterDataChangeInfo& aInfo) {
+  if (mContainerTextNode != &aText ||
+      mCompositionStartOffsetInTextNode == UINT32_MAX ||
+      mCompositionLengthInTextNode == UINT32_MAX) {
+    return;
+  }
+
+  // Ignore changes after composition string.
+  if (aInfo.mChangeStart >=
+      mCompositionStartOffsetInTextNode + mCompositionLengthInTextNode) {
+    return;
+  }
+
+  // If the change ends before the composition string, we need only to adjust
+  // the start offset.
+  if (aInfo.mChangeEnd <= mCompositionStartOffsetInTextNode) {
+    MOZ_ASSERT(aInfo.LengthOfRemovedText() <=
+               mCompositionStartOffsetInTextNode);
+    mCompositionStartOffsetInTextNode -= aInfo.LengthOfRemovedText();
+    mCompositionStartOffsetInTextNode += aInfo.mReplaceLength;
+    return;
+  }
+
+  // If this is caused by a splitting text node, the composition string
+  // may be split out to the new right node.  In the case,
+  // CompositionTransaction::DoTransaction handles it with warking the
+  // following text nodes.  Therefore, we should NOT shrink the composing
+  // range for avoind breaking the fix of bug 1310912.  Although the handling
+  // looks buggy so that we need to move the handling into here later.
+  if (aInfo.mDetails &&
+      aInfo.mDetails->mType == CharacterDataChangeInfo::Details::eSplit) {
+    return;
+  }
+
+  // If the change removes/replaces the last character of the composition
+  // string, we should shrink the composition range before the change start.
+  // Then, the replace string will be never updated by coming composition
+  // updates.
+  if (aInfo.mChangeEnd >=
+      mCompositionStartOffsetInTextNode + mCompositionLengthInTextNode) {
+    // If deleting the first character of the composition string, collapse IME
+    // selection temporarily.  Updating composition string will insert new
+    // composition string there.
+    if (aInfo.mChangeStart <= mCompositionStartOffsetInTextNode) {
+      mCompositionStartOffsetInTextNode = aInfo.mChangeStart;
+      mCompositionLengthInTextNode = 0u;
+      return;
+    }
+    // If some characters in the composition still stay, composition range
+    // should be shrunken.
+    MOZ_ASSERT(aInfo.mChangeStart > mCompositionStartOffsetInTextNode);
+    mCompositionLengthInTextNode =
+        aInfo.mChangeStart - mCompositionStartOffsetInTextNode;
+    return;
+  }
+
+  // If removed range starts in the composition string, we need only adjust
+  // the length to make composition range contain the replace string.
+  if (aInfo.mChangeStart >= mCompositionStartOffsetInTextNode) {
+    MOZ_ASSERT(aInfo.LengthOfRemovedText() <= mCompositionLengthInTextNode);
+    mCompositionLengthInTextNode -= aInfo.LengthOfRemovedText();
+    mCompositionLengthInTextNode += aInfo.mReplaceLength;
+    return;
+  }
+
+  // If preceding characers of the composition string is also removed,  new
+  // composition start will be there and new composition ends at current
+  // position.
+  const uint32_t removedLengthInCompositionString =
+      aInfo.mChangeEnd - mCompositionStartOffsetInTextNode;
+  mCompositionStartOffsetInTextNode = aInfo.mChangeStart;
+  mCompositionLengthInTextNode -= removedLengthInCompositionString;
+  mCompositionLengthInTextNode += aInfo.mReplaceLength;
+}
+
 bool TextComposition::IsValidStateForComposition(nsIWidget* aWidget) const {
   return !Destroyed() && aWidget && !aWidget->Destroyed() &&
          mPresContext->GetPresShell() &&
-         !mPresContext->GetPresShell()->IsDestroying();
+         !mPresContext->PresShell()->IsDestroying();
 }
 
 bool TextComposition::MaybeDispatchCompositionUpdate(
     const WidgetCompositionEvent* aCompositionEvent) {
-  MOZ_RELEASE_ASSERT(!mTabParent);
+  MOZ_RELEASE_ASSERT(!mBrowserParent);
 
   if (!IsValidStateForComposition(aCompositionEvent->mWidget)) {
     return false;
   }
 
+  // Note that we don't need to dispatch eCompositionUpdate event even if
+  // mHasDispatchedDOMTextEvent is false and eCompositionCommit event is
+  // dispatched with empty string immediately after eCompositionStart
+  // because composition string has never been changed from empty string to
+  // non-empty string in such composition even if selected string was not
+  // empty string (mLastData isn't set to selected text when this receives
+  // eCompositionStart).
   if (mLastData == aCompositionEvent->mData) {
     return true;
   }
@@ -109,14 +195,13 @@ bool TextComposition::MaybeDispatchCompositionUpdate(
 BaseEventFlags TextComposition::CloneAndDispatchAs(
     const WidgetCompositionEvent* aCompositionEvent, EventMessage aMessage,
     nsEventStatus* aStatus, EventDispatchingCallback* aCallBack) {
-  MOZ_RELEASE_ASSERT(!mTabParent);
+  MOZ_RELEASE_ASSERT(!mBrowserParent);
 
   MOZ_ASSERT(IsValidStateForComposition(aCompositionEvent->mWidget),
              "Should be called only when it's safe to dispatch an event");
 
   WidgetCompositionEvent compositionEvent(aCompositionEvent->IsTrusted(),
                                           aMessage, aCompositionEvent->mWidget);
-  compositionEvent.mTime = aCompositionEvent->mTime;
   compositionEvent.mTimeStamp = aCompositionEvent->mTimeStamp;
   compositionEvent.mData = aCompositionEvent->mData;
   compositionEvent.mNativeIMEContext = aCompositionEvent->mNativeIMEContext;
@@ -139,10 +224,13 @@ void TextComposition::DispatchEvent(
     WidgetCompositionEvent* aDispatchEvent, nsEventStatus* aStatus,
     EventDispatchingCallback* aCallBack,
     const WidgetCompositionEvent* aOriginalEvent) {
-  nsPluginInstanceOwner::GeneratePluginEvent(aOriginalEvent, aDispatchEvent);
-
-  EventDispatcher::Dispatch(mNode, mPresContext, aDispatchEvent, nullptr,
-                            aStatus, aCallBack);
+  if (aDispatchEvent->mMessage == eCompositionChange) {
+    aDispatchEvent->mFlags.mOnlySystemGroupDispatchInContent = true;
+  }
+  RefPtr<nsINode> node = mNode;
+  RefPtr<nsPresContext> presContext = mPresContext;
+  EventDispatcher::Dispatch(node, presContext, aDispatchEvent, nullptr, aStatus,
+                            aCallBack);
 
   OnCompositionEventDispatched(aDispatchEvent);
 }
@@ -155,9 +243,9 @@ void TextComposition::OnCompositionEventDiscarded(
   MOZ_ASSERT(aCompositionEvent->IsTrusted(),
              "Shouldn't be called with untrusted event");
 
-  if (mTabParent) {
+  if (mBrowserParent) {
     // The composition event should be discarded in the child process too.
-    Unused << mTabParent->SendCompositionEvent(*aCompositionEvent);
+    Unused << mBrowserParent->SendCompositionEvent(*aCompositionEvent);
   }
 
   // XXX If composition events are discarded, should we dispatch them with
@@ -220,6 +308,18 @@ static void RemoveControlCharactersFrom(nsAString& aStr,
   aStr.SetLength(curDest - dest);
 }
 
+nsString TextComposition::CommitStringIfCommittedAsIs() const {
+  nsString result(mLastData);
+  if (!mAllowControlCharacters) {
+    RemoveControlCharactersFrom(result, nullptr);
+  }
+  if (StaticPrefs::intl_ime_remove_placeholder_character_at_commit() &&
+      mLastData == IDEOGRAPHIC_SPACE) {
+    return EmptyString();
+  }
+  return result;
+}
+
 void TextComposition::DispatchCompositionEvent(
     WidgetCompositionEvent* aCompositionEvent, nsEventStatus* aStatus,
     EventDispatchingCallback* aCallBack, bool aIsSynthesized) {
@@ -240,10 +340,10 @@ void TextComposition::DispatchCompositionEvent(
     return;
   }
 
-  // If the content is a container of TabParent, composition should be in the
-  // remote process.
-  if (mTabParent) {
-    Unused << mTabParent->SendCompositionEvent(*aCompositionEvent);
+  // If the content is a container of BrowserParent, composition should be in
+  // the remote process.
+  if (mBrowserParent) {
+    Unused << mBrowserParent->SendCompositionEvent(*aCompositionEvent);
     aCompositionEvent->StopPropagation();
     if (aCompositionEvent->CausesDOMTextEvent()) {
       mLastData = aCompositionEvent->mData;
@@ -266,9 +366,8 @@ void TextComposition::DispatchCompositionEvent(
     aCompositionEvent->mRanges = nullptr;
     NS_ASSERTION(aCompositionEvent->mData.IsEmpty(),
                  "mData of eCompositionCommitAsIs should be empty string");
-    bool removePlaceholderCharacter = Preferences::GetBool(
-        "intl.ime.remove_placeholder_character_at_commit", false);
-    if (removePlaceholderCharacter && mLastData == IDEOGRAPHIC_SPACE) {
+    if (StaticPrefs::intl_ime_remove_placeholder_character_at_commit() &&
+        mLastData == IDEOGRAPHIC_SPACE) {
       // If the last data is an ideographic space (FullWidth space), it might be
       // a placeholder character of some Chinese IME.  So, committing with
       // this data might not be expected by users.  Let's use empty string.
@@ -329,10 +428,14 @@ void TextComposition::DispatchCompositionEvent(
   // When mIsComposing is false but the committing string is different from
   // the last data (E.g., previous eCompositionChange event made the
   // composition string empty or didn't have clause information), we don't
-  // need to dispatch redundant DOM text event.
+  // need to dispatch redundant DOM text event.  (But note that we need to
+  // dispatch eCompositionChange event if we have not dispatched
+  // eCompositionChange event yet and commit string replaces selected string
+  // with empty string since selected string hasn't been replaced with empty
+  // string yet.)
   if (dispatchDOMTextEvent &&
       aCompositionEvent->mMessage != eCompositionChange && !mIsComposing &&
-      mLastData == aCompositionEvent->mData) {
+      mHasDispatchedDOMTextEvent && mLastData == aCompositionEvent->mData) {
     dispatchEvent = dispatchDOMTextEvent = false;
   }
 
@@ -360,9 +463,13 @@ void TextComposition::DispatchCompositionEvent(
     // we cannot map multiple event messages to a DOM event type.
     if (dispatchDOMTextEvent &&
         aCompositionEvent->mMessage != eCompositionChange) {
+      mHasDispatchedDOMTextEvent = true;
       aCompositionEvent->mFlags = CloneAndDispatchAs(
           aCompositionEvent, eCompositionChange, aStatus, aCallBack);
     } else {
+      if (aCompositionEvent->mMessage == eCompositionChange) {
+        mHasDispatchedDOMTextEvent = true;
+      }
       DispatchEvent(aCompositionEvent, aStatus, aCallBack);
     }
   } else {
@@ -394,19 +501,26 @@ void TextComposition::DispatchCompositionEvent(
 
 // static
 void TextComposition::HandleSelectionEvent(
-    nsPresContext* aPresContext, TabParent* aTabParent,
+    nsPresContext* aPresContext, BrowserParent* aBrowserParent,
     WidgetSelectionEvent* aSelectionEvent) {
-  // If the content is a container of TabParent, composition should be in the
-  // remote process.
-  if (aTabParent) {
-    Unused << aTabParent->SendSelectionEvent(*aSelectionEvent);
+  // If the content is a container of BrowserParent, composition should be in
+  // the remote process.
+  if (aBrowserParent) {
+    Unused << aBrowserParent->SendSelectionEvent(*aSelectionEvent);
     aSelectionEvent->StopPropagation();
     return;
   }
 
-  ContentEventHandler handler(aPresContext);
   AutoRestore<bool> saveHandlingSelectionEvent(sHandlingSelectionEvent);
   sHandlingSelectionEvent = true;
+
+  if (RefPtr<IMEContentObserver> contentObserver =
+          IMEStateManager::GetActiveContentObserver()) {
+    contentObserver->MaybeHandleSelectionEvent(aPresContext, aSelectionEvent);
+    return;
+  }
+
+  ContentEventHandler handler(aPresContext);
   // XXX During setting selection, a selection listener may change selection
   //     again.  In such case, sHandlingSelectionEvent doesn't indicate if
   //     the selection change is caused by a selection event.  However, it
@@ -416,18 +530,19 @@ void TextComposition::HandleSelectionEvent(
 
 uint32_t TextComposition::GetSelectionStartOffset() {
   nsCOMPtr<nsIWidget> widget = mPresContext->GetRootWidget();
-  WidgetQueryContentEvent selectedTextEvent(true, eQuerySelectedText, widget);
+  WidgetQueryContentEvent querySelectedTextEvent(true, eQuerySelectedText,
+                                                 widget);
   // Due to a bug of widget, mRanges may not be nullptr even though composition
   // string is empty.  So, we need to check it here for avoiding to return
   // odd start offset.
   if (!mLastData.IsEmpty() && mRanges && mRanges->HasClauses()) {
-    selectedTextEvent.InitForQuerySelectedText(
+    querySelectedTextEvent.InitForQuerySelectedText(
         ToSelectionType(mRanges->GetFirstClause()->mRangeType));
   } else {
     NS_WARNING_ASSERTION(
         !mLastData.IsEmpty() || !mRanges || !mRanges->HasClauses(),
         "Shouldn't have empty clause info when composition string is empty");
-    selectedTextEvent.InitForQuerySelectedText(SelectionType::eNormal);
+    querySelectedTextEvent.InitForQuerySelectedText(SelectionType::eNormal);
   }
 
   // The editor which has this composition is observed by active
@@ -436,9 +551,9 @@ uint32_t TextComposition::GetSelectionStartOffset() {
       IMEStateManager::GetActiveContentObserver();
   bool doQuerySelection = true;
   if (contentObserver) {
-    if (contentObserver->IsManaging(this)) {
+    if (contentObserver->IsManaging(*this)) {
       doQuerySelection = false;
-      contentObserver->HandleQueryContentEvent(&selectedTextEvent);
+      contentObserver->HandleQueryContentEvent(&querySelectedTextEvent);
     }
     // If another editor already has focus, we cannot retrieve selection
     // in the editor which has this composition...
@@ -451,18 +566,18 @@ uint32_t TextComposition::GetSelectionStartOffset() {
   // ContentEventHandler)
   if (doQuerySelection) {
     ContentEventHandler handler(mPresContext);
-    handler.HandleQueryContentEvent(&selectedTextEvent);
+    handler.HandleQueryContentEvent(&querySelectedTextEvent);
   }
 
-  if (NS_WARN_IF(!selectedTextEvent.mSucceeded)) {
+  if (NS_WARN_IF(querySelectedTextEvent.DidNotFindSelection())) {
     return 0;  // XXX Is this okay?
   }
-  return selectedTextEvent.mReply.mOffset;
+  return querySelectedTextEvent.mReply->AnchorOffset();
 }
 
 void TextComposition::OnCompositionEventDispatched(
     const WidgetCompositionEvent* aCompositionEvent) {
-  MOZ_RELEASE_ASSERT(!mTabParent);
+  MOZ_RELEASE_ASSERT(!mBrowserParent);
 
   if (!IsValidStateForComposition(aCompositionEvent->mWidget)) {
     return;
@@ -512,7 +627,7 @@ void TextComposition::MaybeNotifyIMEOfCompositionEventHandled(
   //     event handled.  Although, this is a bug but it should be okay since
   //     destroying IMEContentObserver notifies IME of blur.  So, native IME
   //     handler can treat it as this notification too.
-  if (contentObserver && contentObserver->IsManaging(this)) {
+  if (contentObserver && contentObserver->IsManaging(*this)) {
     contentObserver->MaybeNotifyCompositionEventHandled();
     return;
   }
@@ -542,7 +657,7 @@ nsresult TextComposition::RequestToCommit(nsIWidget* aWidget, bool aDiscard) {
   RefPtr<TextComposition> kungFuDeathGrip(this);
   const nsAutoString lastData(mLastData);
 
-  {
+  if (IMEStateManager::CanSendNotificationToWidget()) {
     AutoRestore<bool> saveRequestingCancel(mIsRequestingCancel);
     AutoRestore<bool> saveRequestingCommit(mIsRequestingCommit);
     if (aDiscard) {
@@ -572,8 +687,7 @@ nsresult TextComposition::RequestToCommit(nsIWidget* aWidget, bool aDiscard) {
   // Otherwise, synthesize the commit in content.
   nsAutoString data(aDiscard ? EmptyString() : lastData);
   if (data == mLastData) {
-    DispatchCompositionEventRunnable(eCompositionCommitAsIs, EmptyString(),
-                                     true);
+    DispatchCompositionEventRunnable(eCompositionCommitAsIs, u""_ns, true);
   } else {
     DispatchCompositionEventRunnable(eCompositionCommit, data, true);
   }
@@ -582,7 +696,7 @@ nsresult TextComposition::RequestToCommit(nsIWidget* aWidget, bool aDiscard) {
 
 nsresult TextComposition::NotifyIME(IMEMessage aMessage) {
   NS_ENSURE_TRUE(mPresContext, NS_ERROR_NOT_AVAILABLE);
-  return IMEStateManager::NotifyIME(aMessage, mPresContext, mTabParent);
+  return IMEStateManager::NotifyIME(aMessage, mPresContext, mBrowserParent);
 }
 
 void TextComposition::EditorWillHandleCompositionChangeEvent(
@@ -598,7 +712,7 @@ void TextComposition::EditorWillHandleCompositionChangeEvent(
 }
 
 void TextComposition::OnEditorDestroyed() {
-  MOZ_RELEASE_ASSERT(!mTabParent);
+  MOZ_RELEASE_ASSERT(!mBrowserParent);
 
   MOZ_ASSERT(!mIsEditorHandlingEvent,
              "The editor should have stopped listening events");
@@ -618,18 +732,18 @@ void TextComposition::EditorDidHandleCompositionChangeEvent() {
 }
 
 void TextComposition::StartHandlingComposition(EditorBase* aEditorBase) {
-  MOZ_RELEASE_ASSERT(!mTabParent);
+  MOZ_RELEASE_ASSERT(!mBrowserParent);
 
   MOZ_ASSERT(!HasEditor(), "There is a handling editor already");
   mEditorBaseWeak = do_GetWeakReference(static_cast<nsIEditor*>(aEditorBase));
 }
 
 void TextComposition::EndHandlingComposition(EditorBase* aEditorBase) {
-  MOZ_RELEASE_ASSERT(!mTabParent);
+  MOZ_RELEASE_ASSERT(!mBrowserParent);
 
 #ifdef DEBUG
   RefPtr<EditorBase> editorBase = GetEditorBase();
-  MOZ_ASSERT(editorBase == aEditorBase,
+  MOZ_ASSERT(!editorBase || editorBase == aEditorBase,
              "Another editor handled the composition?");
 #endif  // #ifdef DEBUG
   mEditorBaseWeak = nullptr;
@@ -645,6 +759,128 @@ bool TextComposition::HasEditor() const {
   return mEditorBaseWeak && mEditorBaseWeak->IsAlive();
 }
 
+RawRangeBoundary TextComposition::FirstIMESelectionStartRef() const {
+  RefPtr<EditorBase> editorBase = GetEditorBase();
+  if (!editorBase) {
+    return RawRangeBoundary();
+  }
+
+  nsISelectionController* selectionController =
+      editorBase->GetSelectionController();
+  if (NS_WARN_IF(!selectionController)) {
+    return RawRangeBoundary();
+  }
+
+  const nsRange* firstRange = nullptr;
+  static const SelectionType kIMESelectionTypes[] = {
+      SelectionType::eIMERawClause, SelectionType::eIMESelectedRawClause,
+      SelectionType::eIMEConvertedClause, SelectionType::eIMESelectedClause};
+  for (auto selectionType : kIMESelectionTypes) {
+    dom::Selection* selection =
+        selectionController->GetSelection(ToRawSelectionType(selectionType));
+    if (!selection) {
+      continue;
+    }
+    const uint32_t rangeCount = selection->RangeCount();
+    for (const uint32_t i : IntegerRange(rangeCount)) {
+      MOZ_ASSERT(selection->RangeCount() == rangeCount);
+      const nsRange* range = selection->GetRangeAt(i);
+      MOZ_ASSERT(range);
+      if (MOZ_UNLIKELY(NS_WARN_IF(!range)) ||
+          MOZ_UNLIKELY(NS_WARN_IF(!range->GetStartContainer()))) {
+        continue;
+      }
+      if (!firstRange) {
+        firstRange = range;
+        continue;
+      }
+      // In most cases, all composition string should be in same text node.
+      if (firstRange->GetStartContainer() == range->GetStartContainer()) {
+        if (firstRange->StartOffset() > range->StartOffset()) {
+          firstRange = range;
+        }
+        continue;
+      }
+      // However, if web apps have inserted different nodes in composition
+      // string, composition string may span 2 or more nodes.
+      if (firstRange->GetStartContainer()->GetNextSibling() ==
+          range->GetStartContainer()) {
+        // Fast path for some known applications like Google Keep.
+        firstRange = range;
+        continue;
+      }
+      // Unfortunately, really slow path.
+      // The ranges should always have a common ancestor, hence, be comparable.
+      if (*nsContentUtils::ComparePoints(range->StartRef(),
+                                         firstRange->StartRef()) == -1) {
+        firstRange = range;
+      }
+    }
+  }
+  return firstRange ? firstRange->StartRef().AsRaw() : RawRangeBoundary();
+}
+
+RawRangeBoundary TextComposition::LastIMESelectionEndRef() const {
+  RefPtr<EditorBase> editorBase = GetEditorBase();
+  if (!editorBase) {
+    return RawRangeBoundary();
+  }
+
+  nsISelectionController* selectionController =
+      editorBase->GetSelectionController();
+  if (NS_WARN_IF(!selectionController)) {
+    return RawRangeBoundary();
+  }
+
+  const nsRange* lastRange = nullptr;
+  static const SelectionType kIMESelectionTypes[] = {
+      SelectionType::eIMERawClause, SelectionType::eIMESelectedRawClause,
+      SelectionType::eIMEConvertedClause, SelectionType::eIMESelectedClause};
+  for (auto selectionType : kIMESelectionTypes) {
+    dom::Selection* selection =
+        selectionController->GetSelection(ToRawSelectionType(selectionType));
+    if (!selection) {
+      continue;
+    }
+    const uint32_t rangeCount = selection->RangeCount();
+    for (const uint32_t i : IntegerRange(rangeCount)) {
+      MOZ_ASSERT(selection->RangeCount() == rangeCount);
+      const nsRange* range = selection->GetRangeAt(i);
+      MOZ_ASSERT(range);
+      if (MOZ_UNLIKELY(NS_WARN_IF(!range)) ||
+          MOZ_UNLIKELY(NS_WARN_IF(!range->GetEndContainer()))) {
+        continue;
+      }
+      if (!lastRange) {
+        lastRange = range;
+        continue;
+      }
+      // In most cases, all composition string should be in same text node.
+      if (lastRange->GetEndContainer() == range->GetEndContainer()) {
+        if (lastRange->EndOffset() < range->EndOffset()) {
+          lastRange = range;
+        }
+        continue;
+      }
+      // However, if web apps have inserted different nodes in composition
+      // string, composition string may span 2 or more nodes.
+      if (lastRange->GetEndContainer() ==
+          range->GetEndContainer()->GetNextSibling()) {
+        // Fast path for some known applications like Google Keep.
+        lastRange = range;
+        continue;
+      }
+      // Unfortunately, really slow path.
+      // The ranges should always have a common ancestor, hence, be comparable.
+      if (*nsContentUtils::ComparePoints(lastRange->EndRef(),
+                                         range->EndRef()) == -1) {
+        lastRange = range;
+      }
+    }
+  }
+  return lastRange ? lastRange->EndRef().AsRaw() : RawRangeBoundary();
+}
+
 /******************************************************************************
  * TextComposition::CompositionEventDispatcher
  ******************************************************************************/
@@ -673,21 +909,27 @@ TextComposition::CompositionEventDispatcher::Run() {
   }
 
   RefPtr<nsPresContext> presContext = mTextComposition->mPresContext;
+  nsCOMPtr<nsINode> eventTarget = mEventTarget;
+  RefPtr<BrowserParent> browserParent = mTextComposition->mBrowserParent;
   nsEventStatus status = nsEventStatus_eIgnore;
   switch (mEventMessage) {
     case eCompositionStart: {
       WidgetCompositionEvent compStart(true, eCompositionStart, widget);
       compStart.mNativeIMEContext = mTextComposition->mNativeContext;
-      WidgetQueryContentEvent selectedText(true, eQuerySelectedText, widget);
+      WidgetQueryContentEvent querySelectedTextEvent(true, eQuerySelectedText,
+                                                     widget);
       ContentEventHandler handler(presContext);
-      handler.OnQuerySelectedText(&selectedText);
-      NS_ASSERTION(selectedText.mSucceeded, "Failed to get selected text");
-      compStart.mData = selectedText.mReply.mString;
+      handler.OnQuerySelectedText(&querySelectedTextEvent);
+      NS_ASSERTION(querySelectedTextEvent.Succeeded(),
+                   "Failed to get selected text");
+      if (querySelectedTextEvent.FoundSelection()) {
+        compStart.mData = querySelectedTextEvent.mReply->DataRef();
+      }
       compStart.mFlags.mIsSynthesizedForTests =
           mTextComposition->IsSynthesizedForTests();
-      IMEStateManager::DispatchCompositionEvent(mEventTarget, presContext,
-                                                &compStart, &status, nullptr,
-                                                mIsSynthesizedEvent);
+      IMEStateManager::DispatchCompositionEvent(
+          eventTarget, presContext, browserParent, &compStart, &status, nullptr,
+          mIsSynthesizedEvent);
       break;
     }
     case eCompositionChange:
@@ -700,9 +942,9 @@ TextComposition::CompositionEventDispatcher::Run() {
       }
       compEvent.mFlags.mIsSynthesizedForTests =
           mTextComposition->IsSynthesizedForTests();
-      IMEStateManager::DispatchCompositionEvent(mEventTarget, presContext,
-                                                &compEvent, &status, nullptr,
-                                                mIsSynthesizedEvent);
+      IMEStateManager::DispatchCompositionEvent(
+          eventTarget, presContext, browserParent, &compEvent, &status, nullptr,
+          mIsSynthesizedEvent);
       break;
     }
     default:
@@ -793,7 +1035,7 @@ TextComposition* TextCompositionArray::GetCompositionInContent(
   // There should be only one composition per content object.
   for (index_type i = Length(); i > 0; --i) {
     nsINode* node = ElementAt(i - 1)->GetEventTargetNode();
-    if (node && nsContentUtils::ContentIsDescendantOf(node, aContent)) {
+    if (node && node->IsInclusiveDescendantOf(aContent)) {
       return ElementAt(i - 1);
     }
   }
