@@ -9,11 +9,11 @@
 
 #include "nsCOMPtr.h"
 #include "nsINode.h"
-#include "nsIWeakReference.h"
 #include "nsIWidget.h"
 #include "nsTArray.h"
 #include "nsThreadUtils.h"
 #include "nsPresContext.h"
+#include "mozilla/AlreadyAddRefed.h"
 #include "mozilla/Attributes.h"
 #include "mozilla/EventForwards.h"
 #include "mozilla/RangeBoundary.h"
@@ -21,6 +21,10 @@
 #include "mozilla/dom/BrowserParent.h"
 #include "mozilla/dom/Text.h"
 
+class nsRange;
+
+struct CharacterDataChangeInfo;
+
 namespace mozilla {
 
 class EditorBase;
@@ -56,6 +60,8 @@ class TextComposition final {
   // The latest CompositionEvent.data value except compositionstart event.
   // This value is modified at dispatching compositionupdate.
   const nsString& LastData() const { return mLastData; }
+  // Returns commit string if it'll be commited as-is.
+  nsString CommitStringIfCommittedAsIs() const;
   // The composition string which is already handled by the focused editor.
   // I.e., this value must be same as the composition string on the focused
   // editor.  This value is modified at a call of
@@ -74,8 +80,11 @@ class TextComposition final {
   //     error due to inaccessible Release() method.
   TextRangeArray* GetRanges() const { return mRanges; }
   // Returns the widget which is proper to call NotifyIME().
-  nsIWidget* GetWidget() const {
-    return mPresContext ? mPresContext->GetRootWidget() : nullptr;
+  already_AddRefed<nsIWidget> GetWidget() const {
+    if (!mPresContext) {
+      return nullptr;
+    }
+    return do_AddRef(mPresContext->GetRootWidget());
   }
   // Returns the tab parent which has this composition in its remote process.
   BrowserParent* GetBrowserParent() const { return mBrowserParent; }
@@ -133,8 +142,8 @@ class TextComposition final {
    * app.  If there is no composition string the DOM tree, these return
    * unset range boundaries.
    */
-  RawRangeBoundary GetStartRef() const;
-  RawRangeBoundary GetEndRef() const;
+  RawRangeBoundary FirstIMESelectionStartRef() const;
+  RawRangeBoundary LastIMESelectionEndRef() const;
 
   /**
    * The offset of composition string in the text node.  If composition string
@@ -229,9 +238,8 @@ class TextComposition final {
   };
 
   /**
-   * OnCreateCompositionTransaction() is called by
-   * CompositionTransaction::Create() immediately after creating
-   * new CompositionTransaction instance.
+   * OnUpdateCompositionInEditor() is called when editor updates composition
+   * string in the DOM tree.
    *
    * @param aStringToInsert     The string to insert the text node actually.
    *                            This may be different from the data of
@@ -241,20 +249,12 @@ class TextComposition final {
    * @param aTextNode           The text node which includes composition string.
    * @param aOffset             The offset of composition string in aTextNode.
    */
-  void OnCreateCompositionTransaction(const nsAString& aStringToInsert,
-                                      Text* aTextNode, uint32_t aOffset) {
-    if (!mContainerTextNode) {
-      mContainerTextNode = aTextNode;
-      mCompositionStartOffsetInTextNode = aOffset;
-      NS_WARNING_ASSERTION(mCompositionStartOffsetInTextNode != UINT32_MAX,
-                           "The text node is really too long.");
-    }
-#ifdef DEBUG
-    else {
-      MOZ_ASSERT(aTextNode == mContainerTextNode);
-      MOZ_ASSERT(aOffset == mCompositionStartOffsetInTextNode);
-    }
-#endif  // #ifdef DEBUG
+  void OnUpdateCompositionInEditor(const nsAString& aStringToInsert,
+                                   Text& aTextNode, uint32_t aOffset) {
+    mContainerTextNode = &aTextNode;
+    mCompositionStartOffsetInTextNode = aOffset;
+    NS_WARNING_ASSERTION(mCompositionStartOffsetInTextNode != UINT32_MAX,
+                         "The text node is really too long.");
     mCompositionLengthInTextNode = aStringToInsert.Length();
     NS_WARNING_ASSERTION(mCompositionLengthInTextNode != UINT32_MAX,
                          "The string to insert is really too long.");
@@ -272,6 +272,13 @@ class TextComposition final {
     // composition in new text node.
   }
 
+  /**
+   * OnCharacterDataChanged() is called when IMEContentObserver receives
+   * character data change notifications.
+   */
+  void OnCharacterDataChanged(Text& aText,
+                              const CharacterDataChangeInfo& aInfo);
+
  private:
   // Private destructor, to discourage deletion outside of Release():
   ~TextComposition() {
@@ -287,7 +294,7 @@ class TextComposition final {
   // IMEStateManager::OnDestroyPresContext(), and then, it destroy
   // this instance.
   nsPresContext* mPresContext;
-  nsCOMPtr<nsINode> mNode;
+  RefPtr<nsINode> mNode;
   RefPtr<BrowserParent> mBrowserParent;
 
   // The text node which includes the composition string.
@@ -449,19 +456,18 @@ class TextComposition final {
    * DispatchCompositionEvent() dispatches the aCompositionEvent to the mContent
    * synchronously. The caller must ensure that it's safe to dispatch the event.
    */
-  void DispatchCompositionEvent(WidgetCompositionEvent* aCompositionEvent,
-                                nsEventStatus* aStatus,
-                                EventDispatchingCallback* aCallBack,
-                                bool aIsSynthesized);
+  MOZ_CAN_RUN_SCRIPT void DispatchCompositionEvent(
+      WidgetCompositionEvent* aCompositionEvent, nsEventStatus* aStatus,
+      EventDispatchingCallback* aCallBack, bool aIsSynthesized);
 
   /**
    * Simply calling EventDispatcher::Dispatch() with plugin event.
    * If dispatching event has no orginal clone, aOriginalEvent can be null.
    */
-  void DispatchEvent(WidgetCompositionEvent* aDispatchEvent,
-                     nsEventStatus* aStatus,
-                     EventDispatchingCallback* aCallback,
-                     const WidgetCompositionEvent* aOriginalEvent = nullptr);
+  MOZ_CAN_RUN_SCRIPT void DispatchEvent(
+      WidgetCompositionEvent* aDispatchEvent, nsEventStatus* aStatus,
+      EventDispatchingCallback* aCallback,
+      const WidgetCompositionEvent* aOriginalEvent = nullptr);
 
   /**
    * HandleSelectionEvent() sends the selection event to ContentEventHandler
@@ -484,7 +490,7 @@ class TextComposition final {
    * @return Returns false if dispatching the compositionupdate event caused
    *         destroying this composition.
    */
-  bool MaybeDispatchCompositionUpdate(
+  MOZ_CAN_RUN_SCRIPT bool MaybeDispatchCompositionUpdate(
       const WidgetCompositionEvent* aCompositionEvent);
 
   /**
@@ -493,10 +499,10 @@ class TextComposition final {
    *
    * @return Returns BaseEventFlags which is the result of dispatched event.
    */
-  BaseEventFlags CloneAndDispatchAs(
-      const WidgetCompositionEvent* aCompositionEvent, EventMessage aMessage,
-      nsEventStatus* aStatus = nullptr,
-      EventDispatchingCallback* aCallBack = nullptr);
+  MOZ_CAN_RUN_SCRIPT BaseEventFlags
+  CloneAndDispatchAs(const WidgetCompositionEvent* aCompositionEvent,
+                     EventMessage aMessage, nsEventStatus* aStatus = nullptr,
+                     EventDispatchingCallback* aCallBack = nullptr);
 
   /**
    * If IME has already dispatched compositionend event but it was discarded
@@ -517,7 +523,7 @@ class TextComposition final {
    * OnCompositionEventDispatched() is called after a composition event is
    * dispatched.
    */
-  void OnCompositionEventDispatched(
+  MOZ_CAN_RUN_SCRIPT void OnCompositionEventDispatched(
       const WidgetCompositionEvent* aDispatchEvent);
 
   /**
@@ -533,7 +539,7 @@ class TextComposition final {
    * editor which has this composition.
    * If it failed or lost focus, this would return 0.
    */
-  uint32_t GetSelectionStartOffset();
+  MOZ_CAN_RUN_SCRIPT uint32_t GetSelectionStartOffset();
 
   /**
    * OnStartOffsetUpdatedInChild() is called when composition start offset
@@ -555,7 +561,7 @@ class TextComposition final {
                                EventMessage aEventMessage,
                                const nsAString& aData,
                                bool aIsSynthesizedEvent = false);
-    NS_IMETHOD Run() override;
+    MOZ_CAN_RUN_SCRIPT_BOUNDARY NS_IMETHOD Run() override;
 
    private:
     RefPtr<TextComposition> mTextComposition;
