@@ -9,18 +9,19 @@
 
 #include "mozilla/Attributes.h"
 #include "mozilla/dom/Event.h"
+#include "mozilla/dom/Nullable.h"
 #include "mozilla/dom/UIEventBinding.h"
+#include "mozilla/dom/WindowProxyHolder.h"
 #include "nsDeviceContext.h"
-#include "nsIDOMUIEvent.h"
-#include "nsLayoutUtils.h"
+#include "nsDocShell.h"
+#include "nsIContent.h"
 #include "nsPresContext.h"
 
 class nsINode;
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
-class UIEvent : public Event, public nsIDOMUIEvent {
+class UIEvent : public Event {
  public:
   UIEvent(EventTarget* aOwner, nsPresContext* aPresContext,
           WidgetGUIEvent* aEvent);
@@ -28,33 +29,32 @@ class UIEvent : public Event, public nsIDOMUIEvent {
   NS_DECL_ISUPPORTS_INHERITED
   NS_DECL_CYCLE_COLLECTION_CLASS_INHERITED(UIEvent, Event)
 
-  // nsIDOMUIEvent Interface
-  NS_DECL_NSIDOMUIEVENT
-
-  // Forward to Event
-  NS_FORWARD_TO_EVENT_NO_SERIALIZATION_NO_DUPLICATION
-  using Event::GetCurrentTarget;  // Because the forwarding thing shadows it.
-  NS_IMETHOD DuplicatePrivateData() override;
-  NS_IMETHOD_(void)
-  Serialize(IPC::Message* aMsg, bool aSerializeInterfaceType) override;
-  NS_IMETHOD_(bool)
-  Deserialize(const IPC::Message* aMsg, PickleIterator* aIter) override;
+  void DuplicatePrivateData() override;
+  void Serialize(IPC::MessageWriter* aWriter,
+                 bool aSerializeInterfaceType) override;
+  bool Deserialize(IPC::MessageReader* aReader) override;
 
   static already_AddRefed<UIEvent> Constructor(const GlobalObject& aGlobal,
                                                const nsAString& aType,
-                                               const UIEventInit& aParam,
-                                               ErrorResult& aRv);
+                                               const UIEventInit& aParam);
 
   virtual JSObject* WrapObjectInternal(
       JSContext* aCx, JS::Handle<JSObject*> aGivenProto) override {
-    return UIEventBinding::Wrap(aCx, this, aGivenProto);
+    return UIEvent_Binding::Wrap(aCx, this, aGivenProto);
   }
 
+  UIEvent* AsUIEvent() override { return this; }
+
   void InitUIEvent(const nsAString& typeArg, bool canBubbleArg,
                    bool cancelableArg, nsGlobalWindowInner* viewArg,
                    int32_t detailArg);
 
-  nsPIDOMWindowOuter* GetView() const { return mView; }
+  Nullable<WindowProxyHolder> GetView() const {
+    if (!mView) {
+      return nullptr;
+    }
+    return WindowProxyHolder(mView->GetBrowsingContext());
+  }
 
   int32_t Detail() const { return mDetail; }
 
@@ -62,9 +62,6 @@ class UIEvent : public Event, public nsIDOMUIEvent {
 
   int32_t LayerY() const { return GetLayerPoint().y; }
 
-  int32_t PageX() const;
-  int32_t PageY() const;
-
   virtual uint32_t Which(CallerType aCallerType = CallerType::System) {
     MOZ_ASSERT(mEvent->mClass != eKeyboardEventClass,
                "Key events should override Which()");
@@ -73,12 +70,31 @@ class UIEvent : public Event, public nsIDOMUIEvent {
     return 0;
   }
 
-  already_AddRefed<nsINode> GetRangeParent();
+  /**
+   * GetRangeParent() should be used only by JS.  C++ callers should use
+   * GetRangeParentContent() or GetRangeParentContentAndOffset() instead.
+   */
+  MOZ_CAN_RUN_SCRIPT already_AddRefed<nsINode> GetRangeParent() {
+    return GetRangeParentContent();
+  }
+  MOZ_CAN_RUN_SCRIPT already_AddRefed<nsIContent> GetRangeParentContent() {
+    return GetRangeParentContentAndOffset(nullptr);
+  }
+  /**
+   * aOffset is optional (i.e., can be nullptr), but when you call this with
+   * nullptr, you should use GetRangeParentContent() instead.
+   */
+  MOZ_CAN_RUN_SCRIPT already_AddRefed<nsIContent>
+  GetRangeParentContentAndOffset(int32_t* aOffset) const;
 
-  int32_t RangeOffset() const;
+  /**
+   * If you also need to compute range parent in C++ code, you should use
+   * GetRangeParentContentAndOffset() instead.
+   */
+  MOZ_CAN_RUN_SCRIPT int32_t RangeOffset() const;
 
  protected:
-  ~UIEvent() {}
+  ~UIEvent() = default;
 
   // Internal helper functions
   nsIntPoint GetMovementPoint();
@@ -86,7 +102,7 @@ class UIEvent : public Event, public nsIDOMUIEvent {
 
   nsCOMPtr<nsPIDOMWindowOuter> mView;
   int32_t mDetail;
-  CSSIntPoint mClientPoint;
+  CSSIntPoint mDefaultClientPoint;
   // Screenpoint is mEvent->mRefPoint.
   nsIntPoint mLayerPoint;
   CSSIntPoint mPagePoint;
@@ -99,24 +115,7 @@ class UIEvent : public Event, public nsIDOMUIEvent {
   void InitModifiers(const EventModifierInit& aParam);
 };
 
-}  // namespace dom
-}  // namespace mozilla
-
-#define NS_FORWARD_TO_UIEVENT                                             \
-  NS_FORWARD_NSIDOMUIEVENT(UIEvent::)                                     \
-  NS_FORWARD_TO_EVENT_NO_SERIALIZATION_NO_DUPLICATION                     \
-  using Event::GetCurrentTarget; /* Forwarding shadows */                 \
-  NS_IMETHOD DuplicatePrivateData() override {                            \
-    return UIEvent::DuplicatePrivateData();                               \
-  }                                                                       \
-  NS_IMETHOD_(void)                                                       \
-  Serialize(IPC::Message* aMsg, bool aSerializeInterfaceType) override {  \
-    UIEvent::Serialize(aMsg, aSerializeInterfaceType);                    \
-  }                                                                       \
-  NS_IMETHOD_(bool)                                                       \
-  Deserialize(const IPC::Message* aMsg, PickleIterator* aIter) override { \
-    return UIEvent::Deserialize(aMsg, aIter);                             \
-  }
+}  // namespace mozilla::dom
 
 already_AddRefed<mozilla::dom::UIEvent> NS_NewDOMUIEvent(
     mozilla::dom::EventTarget* aOwner, nsPresContext* aPresContext,
