@@ -5,25 +5,16 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "MemoryBlobImpl.h"
+#include "mozilla/ipc/InputStreamUtils.h"
 #include "mozilla/IntegerPrintfMacros.h"
 #include "mozilla/SHA1.h"
+#include "nsIMemoryReporter.h"
 #include "nsPrintfCString.h"
+#include "nsRFPService.h"
+#include "nsStringStream.h"
+#include "prtime.h"
 
-namespace mozilla {
-namespace dom {
-
-NS_IMPL_ADDREF(MemoryBlobImpl::DataOwnerAdapter)
-NS_IMPL_RELEASE(MemoryBlobImpl::DataOwnerAdapter)
-
-NS_INTERFACE_MAP_BEGIN(MemoryBlobImpl::DataOwnerAdapter)
-  NS_INTERFACE_MAP_ENTRY(nsIInputStream)
-  NS_INTERFACE_MAP_ENTRY(nsISeekableStream)
-  NS_INTERFACE_MAP_ENTRY(nsITellableStream)
-  NS_INTERFACE_MAP_ENTRY(nsICloneableInputStream)
-  NS_INTERFACE_MAP_ENTRY_CONDITIONAL(nsIIPCSerializableInputStream,
-                                     mSerializableInputStream)
-  NS_INTERFACE_MAP_ENTRY_AMBIGUOUS(nsISupports, nsIInputStream)
-NS_INTERFACE_MAP_END
+namespace mozilla::dom {
 
 // static
 already_AddRefed<MemoryBlobImpl> MemoryBlobImpl::CreateWithCustomLastModified(
@@ -37,46 +28,33 @@ already_AddRefed<MemoryBlobImpl> MemoryBlobImpl::CreateWithCustomLastModified(
 // static
 already_AddRefed<MemoryBlobImpl> MemoryBlobImpl::CreateWithLastModifiedNow(
     void* aMemoryBuffer, uint64_t aLength, const nsAString& aName,
-    const nsAString& aContentType, bool aCrossOriginIsolated) {
-  int64_t lastModificationDate = nsRFPService::ReduceTimePrecisionAsUSecs(
-      PR_Now(), 0,
-      /* aIsSystemPrincipal */ false, aCrossOriginIsolated);
+    const nsAString& aContentType, RTPCallerType aRTPCallerType) {
+  int64_t lastModificationDate =
+      nsRFPService::ReduceTimePrecisionAsUSecs(PR_Now(), 0, aRTPCallerType);
   return CreateWithCustomLastModified(aMemoryBuffer, aLength, aName,
                                       aContentType, lastModificationDate);
 }
 
 nsresult MemoryBlobImpl::DataOwnerAdapter::Create(DataOwner* aDataOwner,
-                                                  uint32_t aStart,
-                                                  uint32_t aLength,
+                                                  size_t aStart, size_t aLength,
                                                   nsIInputStream** _retval) {
-  nsresult rv;
   MOZ_ASSERT(aDataOwner, "Uh ...");
-
-  nsCOMPtr<nsIInputStream> stream;
-
-  rv = NS_NewByteInputStream(
-      getter_AddRefs(stream),
-      MakeSpan(static_cast<const char*>(aDataOwner->mData) + aStart, aLength),
-      NS_ASSIGNMENT_DEPEND);
-  NS_ENSURE_SUCCESS(rv, rv);
-
-  NS_ADDREF(*_retval =
-                new MemoryBlobImpl::DataOwnerAdapter(aDataOwner, stream));
-
-  return NS_OK;
+  Span data{static_cast<const char*>(aDataOwner->mData) + aStart, aLength};
+  RefPtr adapter = new MemoryBlobImpl::DataOwnerAdapter(aDataOwner, data);
+  return NS_NewByteInputStream(_retval, adapter);
 }
 
 already_AddRefed<BlobImpl> MemoryBlobImpl::CreateSlice(
     uint64_t aStart, uint64_t aLength, const nsAString& aContentType,
-    ErrorResult& aRv) {
+    ErrorResult& aRv) const {
   RefPtr<BlobImpl> impl =
       new MemoryBlobImpl(this, aStart, aLength, aContentType);
   return impl.forget();
 }
 
 void MemoryBlobImpl::CreateInputStream(nsIInputStream** aStream,
-                                       ErrorResult& aRv) {
-  if (mLength > INT32_MAX) {
+                                       ErrorResult& aRv) const {
+  if (mLength >= INT32_MAX) {
     aRv.Throw(NS_ERROR_FAILURE);
     return;
   }
@@ -104,7 +82,7 @@ class MemoryBlobImplDataOwnerMemoryReporter final : public nsIMemoryReporter {
 
   NS_IMETHOD CollectReports(nsIHandleReportCallback* aHandleReport,
                             nsISupports* aData, bool aAnonymize) override {
-    typedef MemoryBlobImpl::DataOwner DataOwner;
+    using DataOwner = MemoryBlobImpl::DataOwner;
 
     StaticMutexAutoLock lock(DataOwner::sDataOwnerMutex);
 
@@ -133,7 +111,7 @@ class MemoryBlobImplDataOwnerMemoryReporter final : public nsIMemoryReporter {
         }
 
         aHandleReport->Callback(
-            /* process */ NS_LITERAL_CSTRING(""),
+            /* process */ ""_ns,
             nsPrintfCString(
                 "explicit/dom/memory-file-data/large/file(length=%" PRIu64
                 ", sha1=%s)",
@@ -156,9 +134,8 @@ class MemoryBlobImplDataOwnerMemoryReporter final : public nsIMemoryReporter {
 
     if (smallObjectsTotal > 0) {
       aHandleReport->Callback(
-          /* process */ NS_LITERAL_CSTRING(""),
-          NS_LITERAL_CSTRING("explicit/dom/memory-file-data/small"), KIND_HEAP,
-          UNITS_BYTES, smallObjectsTotal,
+          /* process */ ""_ns, "explicit/dom/memory-file-data/small"_ns,
+          KIND_HEAP, UNITS_BYTES, smallObjectsTotal,
           nsPrintfCString(
               "Memory used to back small memory files (i.e. those taking up "
               "less "
@@ -188,5 +165,4 @@ void MemoryBlobImpl::DataOwner::EnsureMemoryReporterRegistered() {
   sMemoryReporterRegistered = true;
 }
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
