@@ -12,22 +12,19 @@
 #include "mozilla/dom/GamepadEvent.h"
 #include "mozilla/dom/GamepadEventChannelChild.h"
 #include "mozilla/dom/GamepadMonitoring.h"
+#include "mozilla/dom/Promise.h"
 
 #include "mozilla/ipc/BackgroundChild.h"
 #include "mozilla/ipc/PBackgroundChild.h"
 #include "mozilla/ClearOnShutdown.h"
 #include "mozilla/Preferences.h"
+#include "mozilla/StaticPrefs_dom.h"
 #include "mozilla/StaticPtr.h"
 
-#include "nsAutoPtr.h"
 #include "nsContentUtils.h"
 #include "nsGlobalWindow.h"
-#include "nsIDOMEvent.h"
-#include "nsIDOMDocument.h"
-#include "nsIDOMWindow.h"
 #include "nsIObserver.h"
 #include "nsIObserverService.h"
-#include "nsIServiceManager.h"
 #include "nsThreadUtils.h"
 #include "VRManagerChild.h"
 #include "mozilla/Services.h"
@@ -37,22 +34,20 @@
 
 using namespace mozilla::ipc;
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 namespace {
 
-const char* kGamepadEnabledPref = "dom.gamepad.enabled";
-const char* kGamepadEventsEnabledPref =
-    "dom.gamepad.non_standard_events.enabled";
-
 const nsTArray<RefPtr<nsGlobalWindowInner>>::index_type NoIndex =
     nsTArray<RefPtr<nsGlobalWindowInner>>::NoIndex;
 
 bool sShutdown = false;
 
 StaticRefPtr<GamepadManager> gGamepadManagerSingleton;
-const uint32_t VR_GAMEPAD_IDX_OFFSET = 0x01 << 16;
+
+// A threshold value of axis move to determine the first
+// intent.
+const float AXIS_FIRST_INTENT_THRESHOLD_VALUE = 0.1f;
 
 }  // namespace
 
@@ -65,9 +60,9 @@ GamepadManager::GamepadManager()
       mPromiseID(0) {}
 
 nsresult GamepadManager::Init() {
-  mEnabled = IsAPIEnabled();
+  mEnabled = StaticPrefs::dom_gamepad_enabled();
   mNonstandardEventsEnabled =
-      Preferences::GetBool(kGamepadEventsEnabledPref, false);
+      StaticPrefs::dom_gamepad_non_standard_events_enabled();
   nsCOMPtr<nsIObserverService> observerService =
       mozilla::services::GetObserverService();
 
@@ -99,14 +94,14 @@ GamepadManager::Observe(nsISupports* aSubject, const char* aTopic,
 }
 
 void GamepadManager::StopMonitoring() {
-  for (uint32_t i = 0; i < mChannelChildren.Length(); ++i) {
-    mChannelChildren[i]->SendGamepadListenerRemoved();
+  if (mChannelChild) {
+    PGamepadEventChannelChild::Send__delete__(mChannelChild);
+    mChannelChild = nullptr;
   }
   if (gfx::VRManagerChild::IsCreated()) {
     gfx::VRManagerChild* vm = gfx::VRManagerChild::Get();
     vm->SendControllerListenerRemoved();
   }
-  mChannelChildren.Clear();
   mGamepads.Clear();
 }
 
@@ -126,24 +121,20 @@ void GamepadManager::AddListener(nsGlobalWindowInner* aWindow) {
   MOZ_ASSERT(NS_IsMainThread());
 
   // IPDL child has not been created
-  if (mChannelChildren.IsEmpty()) {
+  if (!mChannelChild) {
     PBackgroundChild* actor = BackgroundChild::GetOrCreateForCurrentThread();
     if (NS_WARN_IF(!actor)) {
       // We are probably shutting down.
       return;
     }
 
-    GamepadEventChannelChild* child = new GamepadEventChannelChild();
-    PGamepadEventChannelChild* initedChild =
-        actor->SendPGamepadEventChannelConstructor(child);
-    if (NS_WARN_IF(!initedChild)) {
+    RefPtr<GamepadEventChannelChild> child(GamepadEventChannelChild::Create());
+    if (!actor->SendPGamepadEventChannelConstructor(child.get())) {
       // We are probably shutting down.
       return;
     }
 
-    MOZ_ASSERT(initedChild == child);
-    child->SendGamepadListenerAdded();
-    mChannelChildren.AppendElement(child);
+    mChannelChild = child;
 
     if (gfx::VRManagerChild::IsCreated()) {
       // Construct VRManagerChannel and ask adding the connected
@@ -153,8 +144,7 @@ void GamepadManager::AddListener(nsGlobalWindowInner* aWindow) {
     }
   }
 
-  if (!mEnabled || mShuttingDown ||
-      nsContentUtils::ShouldResistFingerprinting()) {
+  if (!mEnabled || mShuttingDown || aWindow->ShouldResistFingerprinting()) {
     return;
   }
 
@@ -178,8 +168,8 @@ void GamepadManager::RemoveListener(nsGlobalWindowInner* aWindow) {
     return;  // doesn't exist
   }
 
-  for (auto iter = mGamepads.Iter(); !iter.Done(); iter.Next()) {
-    aWindow->RemoveGamepad(iter.Key());
+  for (const auto& key : mGamepads.Keys()) {
+    aWindow->RemoveGamepad(key);
   }
 
   mListeners.RemoveElement(aWindow);
@@ -189,73 +179,51 @@ void GamepadManager::RemoveListener(nsGlobalWindowInner* aWindow) {
   }
 }
 
-already_AddRefed<Gamepad> GamepadManager::GetGamepad(uint32_t aIndex) const {
+already_AddRefed<Gamepad> GamepadManager::GetGamepad(
+    GamepadHandle aHandle) const {
   RefPtr<Gamepad> gamepad;
-  if (mGamepads.Get(aIndex, getter_AddRefs(gamepad))) {
+  if (mGamepads.Get(aHandle, getter_AddRefs(gamepad))) {
     return gamepad.forget();
   }
 
   return nullptr;
 }
 
-uint32_t GamepadManager::GetGamepadIndexWithServiceType(
-    uint32_t aIndex, GamepadServiceType aServiceType) {
-  uint32_t newIndex = 0;
-
-  switch (aServiceType) {
-    case GamepadServiceType::Standard:
-      MOZ_ASSERT(aIndex <= VR_GAMEPAD_IDX_OFFSET);
-      newIndex = aIndex;
-      break;
-    case GamepadServiceType::VR:
-      newIndex = aIndex + VR_GAMEPAD_IDX_OFFSET;
-      break;
-    default:
-      MOZ_ASSERT(false);
-      break;
-  }
-
-  return newIndex;
-}
-
-void GamepadManager::AddGamepad(uint32_t aIndex, const nsAString& aId,
+void GamepadManager::AddGamepad(GamepadHandle aHandle, const nsAString& aId,
                                 GamepadMappingType aMapping, GamepadHand aHand,
-                                GamepadServiceType aServiceType,
                                 uint32_t aDisplayID, uint32_t aNumButtons,
-                                uint32_t aNumAxes, uint32_t aNumHaptics) {
-  uint32_t newIndex = GetGamepadIndexWithServiceType(aIndex, aServiceType);
-
+                                uint32_t aNumAxes, uint32_t aNumHaptics,
+                                uint32_t aNumLightIndicator,
+                                uint32_t aNumTouchEvents) {
   // TODO: bug 852258: get initial button/axis state
-  RefPtr<Gamepad> gamepad = new Gamepad(nullptr, aId,
-                                        0,  // index is set by global window
-                                        newIndex, aMapping, aHand, aDisplayID,
-                                        aNumButtons, aNumAxes, aNumHaptics);
+  RefPtr<Gamepad> newGamepad =
+      new Gamepad(nullptr, aId,
+                  0,  // index is set by global window
+                  aHandle, aMapping, aHand, aDisplayID, aNumButtons, aNumAxes,
+                  aNumHaptics, aNumLightIndicator, aNumTouchEvents);
 
   // We store the gamepad related to its index given by the parent process,
   // and no duplicate index is allowed.
-  MOZ_ASSERT(!mGamepads.Get(newIndex, nullptr));
-  mGamepads.Put(newIndex, gamepad);
-  NewConnectionEvent(newIndex, true);
+  MOZ_ASSERT(!mGamepads.Contains(aHandle));
+  mGamepads.InsertOrUpdate(aHandle, std::move(newGamepad));
+  NewConnectionEvent(aHandle, true);
 }
 
-void GamepadManager::RemoveGamepad(uint32_t aIndex,
-                                   GamepadServiceType aServiceType) {
-  uint32_t newIndex = GetGamepadIndexWithServiceType(aIndex, aServiceType);
-
-  RefPtr<Gamepad> gamepad = GetGamepad(newIndex);
+void GamepadManager::RemoveGamepad(GamepadHandle aHandle) {
+  RefPtr<Gamepad> gamepad = GetGamepad(aHandle);
   if (!gamepad) {
     NS_WARNING("Trying to delete gamepad with invalid index");
     return;
   }
   gamepad->SetConnected(false);
-  NewConnectionEvent(newIndex, false);
-  mGamepads.Remove(newIndex);
+  NewConnectionEvent(aHandle, false);
+  mGamepads.Remove(aHandle);
 }
 
 void GamepadManager::FireButtonEvent(EventTarget* aTarget, Gamepad* aGamepad,
                                      uint32_t aButton, double aValue) {
-  nsString name = aValue == 1.0L ? NS_LITERAL_STRING("gamepadbuttondown")
-                                 : NS_LITERAL_STRING("gamepadbuttonup");
+  nsString name =
+      aValue == 1.0L ? u"gamepadbuttondown"_ns : u"gamepadbuttonup"_ns;
   GamepadButtonEventInit init;
   init.mBubbles = false;
   init.mCancelable = false;
@@ -266,8 +234,7 @@ void GamepadManager::FireButtonEvent(EventTarget* aTarget, Gamepad* aGamepad,
 
   event->SetTrusted(true);
 
-  bool defaultActionEnabled = true;
-  aTarget->DispatchEvent(event, &defaultActionEnabled);
+  aTarget->DispatchEvent(*event);
 }
 
 void GamepadManager::FireAxisMoveEvent(EventTarget* aTarget, Gamepad* aGamepad,
@@ -278,39 +245,46 @@ void GamepadManager::FireAxisMoveEvent(EventTarget* aTarget, Gamepad* aGamepad,
   init.mGamepad = aGamepad;
   init.mAxis = aAxis;
   init.mValue = aValue;
-  RefPtr<GamepadAxisMoveEvent> event = GamepadAxisMoveEvent::Constructor(
-      aTarget, NS_LITERAL_STRING("gamepadaxismove"), init);
+  RefPtr<GamepadAxisMoveEvent> event =
+      GamepadAxisMoveEvent::Constructor(aTarget, u"gamepadaxismove"_ns, init);
 
   event->SetTrusted(true);
 
-  bool defaultActionEnabled = true;
-  aTarget->DispatchEvent(event, &defaultActionEnabled);
+  aTarget->DispatchEvent(*event);
 }
 
-void GamepadManager::NewConnectionEvent(uint32_t aIndex, bool aConnected) {
-  // Do not fire gamepadconnected and gamepaddisconnected events when
-  // privacy.resistFingerprinting is true.
-  if (nsContentUtils::ShouldResistFingerprinting()) {
-    return;
-  }
-
+void GamepadManager::NewConnectionEvent(GamepadHandle aHandle,
+                                        bool aConnected) {
   if (mShuttingDown) {
     return;
   }
 
-  RefPtr<Gamepad> gamepad = GetGamepad(aIndex);
+  RefPtr<Gamepad> gamepad = GetGamepad(aHandle);
   if (!gamepad) {
     return;
   }
 
   // Hold on to listeners in a separate array because firing events
   // can mutate the mListeners array.
-  nsTArray<RefPtr<nsGlobalWindowInner>> listeners(mListeners);
+  nsTArray<RefPtr<nsGlobalWindowInner>> listeners(mListeners.Clone());
 
   if (aConnected) {
     for (uint32_t i = 0; i < listeners.Length(); i++) {
+#ifdef NIGHTLY_BUILD
+      // Don't fire a gamepadconnected event unless it's a secure context
+      if (!listeners[i]->IsSecureContext()) {
+        continue;
+      }
+#endif
+
+      // Do not fire gamepadconnected and gamepaddisconnected events when
+      // privacy.resistFingerprinting is true.
+      if (listeners[i]->ShouldResistFingerprinting()) {
+        continue;
+      }
+
       // Only send events to non-background windows
-      if (!listeners[i]->AsInner()->IsCurrentInnerWindow() ||
+      if (!listeners[i]->IsCurrentInnerWindow() ||
           listeners[i]->GetOuterWindow()->IsBackground()) {
         continue;
       }
@@ -321,9 +295,9 @@ void GamepadManager::NewConnectionEvent(uint32_t aIndex, bool aConnected) {
         continue;
       }
 
-      SetWindowHasSeenGamepad(listeners[i], aIndex);
+      SetWindowHasSeenGamepad(listeners[i], aHandle);
 
-      RefPtr<Gamepad> listenerGamepad = listeners[i]->GetGamepad(aIndex);
+      RefPtr<Gamepad> listenerGamepad = listeners[i]->GetGamepad(aHandle);
       if (listenerGamepad) {
         // Fire event
         FireConnectionEvent(listeners[i], listenerGamepad, aConnected);
@@ -336,13 +310,19 @@ void GamepadManager::NewConnectionEvent(uint32_t aIndex, bool aConnected) {
       // Even background windows get these events, so we don't have to
       // deal with the hassle of syncing the state of removed gamepads.
 
-      if (WindowHasSeenGamepad(listeners[i], aIndex)) {
-        RefPtr<Gamepad> listenerGamepad = listeners[i]->GetGamepad(aIndex);
+      // Do not fire gamepadconnected and gamepaddisconnected events when
+      // privacy.resistFingerprinting is true.
+      if (listeners[i]->ShouldResistFingerprinting()) {
+        continue;
+      }
+
+      if (WindowHasSeenGamepad(listeners[i], aHandle)) {
+        RefPtr<Gamepad> listenerGamepad = listeners[i]->GetGamepad(aHandle);
         if (listenerGamepad) {
           listenerGamepad->SetConnected(false);
           // Fire event
           FireConnectionEvent(listeners[i], listenerGamepad, false);
-          listeners[i]->RemoveGamepad(aIndex);
+          listeners[i]->RemoveGamepad(aHandle);
         }
       }
     }
@@ -351,8 +331,8 @@ void GamepadManager::NewConnectionEvent(uint32_t aIndex, bool aConnected) {
 
 void GamepadManager::FireConnectionEvent(EventTarget* aTarget,
                                          Gamepad* aGamepad, bool aConnected) {
-  nsString name = aConnected ? NS_LITERAL_STRING("gamepadconnected")
-                             : NS_LITERAL_STRING("gamepaddisconnected");
+  nsString name =
+      aConnected ? u"gamepadconnected"_ns : u"gamepaddisconnected"_ns;
   GamepadEventInit init;
   init.mBubbles = false;
   init.mCancelable = false;
@@ -361,17 +341,17 @@ void GamepadManager::FireConnectionEvent(EventTarget* aTarget,
 
   event->SetTrusted(true);
 
-  bool defaultActionEnabled = true;
-  aTarget->DispatchEvent(event, &defaultActionEnabled);
+  aTarget->DispatchEvent(*event);
 }
 
-void GamepadManager::SyncGamepadState(uint32_t aIndex, Gamepad* aGamepad) {
-  if (mShuttingDown || !mEnabled ||
-      nsContentUtils::ShouldResistFingerprinting()) {
+void GamepadManager::SyncGamepadState(GamepadHandle aHandle,
+                                      nsGlobalWindowInner* aWindow,
+                                      Gamepad* aGamepad) {
+  if (mShuttingDown || !mEnabled || aWindow->ShouldResistFingerprinting()) {
     return;
   }
 
-  RefPtr<Gamepad> gamepad = GetGamepad(aIndex);
+  RefPtr<Gamepad> gamepad = GetGamepad(aHandle);
   if (!gamepad) {
     return;
   }
@@ -402,30 +382,42 @@ already_AddRefed<GamepadManager> GamepadManager::GetService() {
   return service.forget();
 }
 
-// static
-bool GamepadManager::IsAPIEnabled() {
-  return Preferences::GetBool(kGamepadEnabledPref, false);
+bool GamepadManager::AxisMoveIsFirstIntent(nsGlobalWindowInner* aWindow,
+                                           GamepadHandle aHandle,
+                                           const GamepadChangeEvent& aEvent) {
+  const GamepadChangeEventBody& body = aEvent.body();
+  if (!WindowHasSeenGamepad(aWindow, aHandle) &&
+      body.type() == GamepadChangeEventBody::TGamepadAxisInformation) {
+    // Some controllers would send small axis values even they are just idle.
+    // To avoid controllers be activated without its first intent.
+    const GamepadAxisInformation& a = body.get_GamepadAxisInformation();
+    if (abs(a.value()) < AXIS_FIRST_INTENT_THRESHOLD_VALUE) {
+      return false;
+    }
+  }
+  return true;
 }
 
 bool GamepadManager::MaybeWindowHasSeenGamepad(nsGlobalWindowInner* aWindow,
-                                               uint32_t aIndex) {
-  if (!WindowHasSeenGamepad(aWindow, aIndex)) {
+                                               GamepadHandle aHandle) {
+  if (!WindowHasSeenGamepad(aWindow, aHandle)) {
     // This window hasn't seen this gamepad before, so
     // send a connection event first.
-    SetWindowHasSeenGamepad(aWindow, aIndex);
-    return true;
+    SetWindowHasSeenGamepad(aWindow, aHandle);
+    return false;
   }
-  return false;
+  return true;
 }
 
 bool GamepadManager::WindowHasSeenGamepad(nsGlobalWindowInner* aWindow,
-                                          uint32_t aIndex) const {
-  RefPtr<Gamepad> gamepad = aWindow->GetGamepad(aIndex);
+                                          GamepadHandle aHandle) const {
+  RefPtr<Gamepad> gamepad = aWindow->GetGamepad(aHandle);
   return gamepad != nullptr;
 }
 
 void GamepadManager::SetWindowHasSeenGamepad(nsGlobalWindowInner* aWindow,
-                                             uint32_t aIndex, bool aHasSeen) {
+                                             GamepadHandle aHandle,
+                                             bool aHasSeen) {
   MOZ_ASSERT(aWindow);
 
   if (mListeners.IndexOf(aWindow) == NoIndex) {
@@ -436,36 +428,36 @@ void GamepadManager::SetWindowHasSeenGamepad(nsGlobalWindowInner* aWindow,
   if (aHasSeen) {
     aWindow->SetHasSeenGamepadInput(true);
     nsCOMPtr<nsISupports> window = ToSupports(aWindow);
-    RefPtr<Gamepad> gamepad = GetGamepad(aIndex);
+    RefPtr<Gamepad> gamepad = GetGamepad(aHandle);
     if (!gamepad) {
       return;
     }
     RefPtr<Gamepad> clonedGamepad = gamepad->Clone(window);
-    aWindow->AddGamepad(aIndex, clonedGamepad);
+    aWindow->AddGamepad(aHandle, clonedGamepad);
   } else {
-    aWindow->RemoveGamepad(aIndex);
+    aWindow->RemoveGamepad(aHandle);
   }
 }
 
 void GamepadManager::Update(const GamepadChangeEvent& aEvent) {
-  if (!mEnabled || mShuttingDown ||
-      nsContentUtils::ShouldResistFingerprinting()) {
+  if (!mEnabled || mShuttingDown) {
     return;
   }
 
-  const uint32_t index = aEvent.index();
-  GamepadServiceType serviceType = aEvent.service_type();
+  const GamepadHandle handle = aEvent.handle();
+
   GamepadChangeEventBody body = aEvent.body();
 
   if (body.type() == GamepadChangeEventBody::TGamepadAdded) {
     const GamepadAdded& a = body.get_GamepadAdded();
-    AddGamepad(index, a.id(), static_cast<GamepadMappingType>(a.mapping()),
-               static_cast<GamepadHand>(a.hand()), serviceType, a.display_id(),
-               a.num_buttons(), a.num_axes(), a.num_haptics());
+    AddGamepad(handle, a.id(), static_cast<GamepadMappingType>(a.mapping()),
+               static_cast<GamepadHand>(a.hand()), a.display_id(),
+               a.num_buttons(), a.num_axes(), a.num_haptics(), a.num_lights(),
+               a.num_touches());
     return;
   }
   if (body.type() == GamepadChangeEventBody::TGamepadRemoved) {
-    RemoveGamepad(index, serviceType);
+    RemoveGamepad(handle);
     return;
   }
 
@@ -475,12 +467,13 @@ void GamepadManager::Update(const GamepadChangeEvent& aEvent) {
 
   // Hold on to listeners in a separate array because firing events
   // can mutate the mListeners array.
-  nsTArray<RefPtr<nsGlobalWindowInner>> listeners(mListeners);
+  nsTArray<RefPtr<nsGlobalWindowInner>> listeners(mListeners.Clone());
 
   for (uint32_t i = 0; i < listeners.Length(); i++) {
     // Only send events to non-background windows
-    if (!listeners[i]->AsInner()->IsCurrentInnerWindow() ||
-        listeners[i]->GetOuterWindow()->IsBackground()) {
+    if (!listeners[i]->IsCurrentInnerWindow() ||
+        listeners[i]->GetOuterWindow()->IsBackground() ||
+        listeners[i]->ShouldResistFingerprinting()) {
       continue;
     }
 
@@ -497,9 +490,9 @@ void GamepadManager::MaybeConvertToNonstandardGamepadEvent(
     return;
   }
 
-  const uint32_t index =
-      GetGamepadIndexWithServiceType(aEvent.index(), aEvent.service_type());
-  RefPtr<Gamepad> gamepad = aWindow->GetGamepad(index);
+  GamepadHandle handle = aEvent.handle();
+
+  RefPtr<Gamepad> gamepad = aWindow->GetGamepad(handle);
   const GamepadChangeEventBody& body = aEvent.body();
 
   if (gamepad) {
@@ -524,14 +517,18 @@ bool GamepadManager::SetGamepadByEvent(const GamepadChangeEvent& aEvent,
                                        nsGlobalWindowInner* aWindow) {
   bool ret = false;
   bool firstTime = false;
-  const uint32_t index =
-      GetGamepadIndexWithServiceType(aEvent.index(), aEvent.service_type());
+
+  GamepadHandle handle = aEvent.handle();
+
   if (aWindow) {
-    firstTime = MaybeWindowHasSeenGamepad(aWindow, index);
+    if (!AxisMoveIsFirstIntent(aWindow, handle, aEvent)) {
+      return false;
+    }
+    firstTime = !MaybeWindowHasSeenGamepad(aWindow, handle);
   }
 
   RefPtr<Gamepad> gamepad =
-      aWindow ? aWindow->GetGamepad(index) : GetGamepad(index);
+      aWindow ? aWindow->GetGamepad(handle) : GetGamepad(handle);
   const GamepadChangeEventBody& body = aEvent.body();
 
   if (gamepad) {
@@ -551,6 +548,25 @@ bool GamepadManager::SetGamepadByEvent(const GamepadChangeEvent& aEvent,
         gamepad->SetPose(a.pose_state());
         break;
       }
+      case GamepadChangeEventBody::TGamepadLightIndicatorTypeInformation: {
+        const GamepadLightIndicatorTypeInformation& a =
+            body.get_GamepadLightIndicatorTypeInformation();
+        gamepad->SetLightIndicatorType(a.light(), a.type());
+        break;
+      }
+      case GamepadChangeEventBody::TGamepadTouchInformation: {
+        // Avoid GamepadTouch's touchId be accessed in cross-origin tracking.
+        for (uint32_t i = 0; i < mListeners.Length(); i++) {
+          RefPtr<Gamepad> listenerGamepad = mListeners[i]->GetGamepad(handle);
+          if (listenerGamepad && mListeners[i]->IsCurrentInnerWindow() &&
+              !mListeners[i]->GetOuterWindow()->IsBackground()) {
+            const GamepadTouchInformation& a =
+                body.get_GamepadTouchInformation();
+            listenerGamepad->SetTouchEvent(a.index(), a.touch_state());
+          }
+        }
+        break;
+      }
       case GamepadChangeEventBody::TGamepadHandInformation: {
         const GamepadHandInformation& a = body.get_GamepadHandInformation();
         gamepad->SetHand(a.hand());
@@ -571,28 +587,26 @@ bool GamepadManager::SetGamepadByEvent(const GamepadChangeEvent& aEvent,
 }
 
 already_AddRefed<Promise> GamepadManager::VibrateHaptic(
-    uint32_t aControllerIdx, uint32_t aHapticIndex, double aIntensity,
+    GamepadHandle aHandle, uint32_t aHapticIndex, double aIntensity,
     double aDuration, nsIGlobalObject* aGlobal, ErrorResult& aRv) {
-  const char* kGamepadHapticEnabledPref = "dom.gamepad.haptic_feedback.enabled";
   RefPtr<Promise> promise = Promise::Create(aGlobal, aRv);
   if (NS_WARN_IF(aRv.Failed())) {
     aRv.Throw(NS_ERROR_FAILURE);
     return nullptr;
   }
-  if (Preferences::GetBool(kGamepadHapticEnabledPref)) {
-    if (aControllerIdx >= VR_GAMEPAD_IDX_OFFSET) {
+  if (StaticPrefs::dom_gamepad_haptic_feedback_enabled()) {
+    if (aHandle.GetKind() == GamepadHandleKind::VR) {
       if (gfx::VRManagerChild::IsCreated()) {
-        const uint32_t index = aControllerIdx - VR_GAMEPAD_IDX_OFFSET;
         gfx::VRManagerChild* vm = gfx::VRManagerChild::Get();
         vm->AddPromise(mPromiseID, promise);
-        vm->SendVibrateHaptic(index, aHapticIndex, aIntensity, aDuration,
+        vm->SendVibrateHaptic(aHandle, aHapticIndex, aIntensity, aDuration,
                               mPromiseID);
       }
     } else {
-      for (const auto& channelChild : mChannelChildren) {
-        channelChild->AddPromise(mPromiseID, promise);
-        channelChild->SendVibrateHaptic(aControllerIdx, aHapticIndex,
-                                        aIntensity, aDuration, mPromiseID);
+      if (mChannelChild) {
+        mChannelChild->AddPromise(mPromiseID, promise);
+        mChannelChild->SendVibrateHaptic(aHandle, aHapticIndex, aIntensity,
+                                         aDuration, mPromiseID);
       }
     }
   }
@@ -602,26 +616,45 @@ already_AddRefed<Promise> GamepadManager::VibrateHaptic(
 }
 
 void GamepadManager::StopHaptics() {
-  const char* kGamepadHapticEnabledPref = "dom.gamepad.haptic_feedback.enabled";
-  if (!Preferences::GetBool(kGamepadHapticEnabledPref)) {
+  if (!StaticPrefs::dom_gamepad_haptic_feedback_enabled()) {
     return;
   }
 
-  for (auto iter = mGamepads.Iter(); !iter.Done(); iter.Next()) {
-    const uint32_t gamepadIndex = iter.UserData()->HashKey();
-    if (gamepadIndex >= VR_GAMEPAD_IDX_OFFSET) {
+  for (const auto& entry : mGamepads) {
+    const GamepadHandle handle = entry.GetWeak()->GetHandle();
+    if (handle.GetKind() == GamepadHandleKind::VR) {
       if (gfx::VRManagerChild::IsCreated()) {
-        const uint32_t index = gamepadIndex - VR_GAMEPAD_IDX_OFFSET;
         gfx::VRManagerChild* vm = gfx::VRManagerChild::Get();
-        vm->SendStopVibrateHaptic(index);
+        vm->SendStopVibrateHaptic(handle);
       }
     } else {
-      for (auto& channelChild : mChannelChildren) {
-        channelChild->SendStopVibrateHaptic(gamepadIndex);
+      if (mChannelChild) {
+        mChannelChild->SendStopVibrateHaptic(handle);
       }
     }
   }
 }
 
-}  // namespace dom
-}  // namespace mozilla
+already_AddRefed<Promise> GamepadManager::SetLightIndicatorColor(
+    GamepadHandle aHandle, uint32_t aLightColorIndex, uint8_t aRed,
+    uint8_t aGreen, uint8_t aBlue, nsIGlobalObject* aGlobal, ErrorResult& aRv) {
+  RefPtr<Promise> promise = Promise::Create(aGlobal, aRv);
+  if (NS_WARN_IF(aRv.Failed())) {
+    aRv.Throw(NS_ERROR_FAILURE);
+    return nullptr;
+  }
+  if (StaticPrefs::dom_gamepad_extensions_lightindicator()) {
+    MOZ_RELEASE_ASSERT(aHandle.GetKind() != GamepadHandleKind::VR,
+                       "We don't support light indicator in VR.");
+
+    if (mChannelChild) {
+      mChannelChild->AddPromise(mPromiseID, promise);
+      mChannelChild->SendLightIndicatorColor(aHandle, aLightColorIndex, aRed,
+                                             aGreen, aBlue, mPromiseID);
+    }
+  }
+
+  ++mPromiseID;
+  return promise.forget();
+}
+}  // namespace mozilla::dom
