@@ -6,19 +6,54 @@
 
 #include "LSObject.h"
 
+// Local includes
 #include "ActorsChild.h"
-#include "IPCBlobInputStreamThread.h"
-#include "LocalStorageCommon.h"
+#include "LSDatabase.h"
+#include "LSObserver.h"
+
+// Global includes
+#include <utility>
+#include "MainThreadUtils.h"
 #include "mozilla/BasePrincipal.h"
-#include "mozilla/StaticPrefs_dom.h"
-#include "mozilla/ThreadEventQueue.h"
+#include "mozilla/ErrorResult.h"
+#include "mozilla/MacroForEach.h"
+#include "mozilla/Monitor.h"
+#include "mozilla/OriginAttributes.h"
+#include "mozilla/Preferences.h"
+#include "mozilla/RemoteLazyInputStreamThread.h"
+#include "mozilla/ScopeExit.h"
+#include "mozilla/SpinEventLoopUntil.h"
+#include "mozilla/StaticMutex.h"
+#include "mozilla/StorageAccess.h"
+#include "mozilla/Unused.h"
+#include "mozilla/dom/ClientInfo.h"
+#include "mozilla/dom/Document.h"
+#include "mozilla/dom/LocalStorageCommon.h"
+#include "mozilla/dom/PBackgroundLSRequest.h"
+#include "mozilla/dom/PBackgroundLSSharedTypes.h"
 #include "mozilla/dom/quota/QuotaManager.h"
 #include "mozilla/ipc/BackgroundChild.h"
 #include "mozilla/ipc/BackgroundUtils.h"
 #include "mozilla/ipc/PBackgroundChild.h"
+#include "mozilla/ipc/ProcessChild.h"
+#include "nsCOMPtr.h"
 #include "nsContentUtils.h"
+#include "nsDebug.h"
+#include "nsError.h"
+#include "nsIEventTarget.h"
+#include "nsIPrincipal.h"
+#include "nsIRunnable.h"
 #include "nsIScriptObjectPrincipal.h"
+#include "nsISerialEventTarget.h"
+#include "nsITimer.h"
+#include "nsPIDOMWindow.h"
+#include "nsString.h"
+#include "nsTArray.h"
+#include "nsTStringRepr.h"
 #include "nsThread.h"
+#include "nsThreadUtils.h"
+#include "nsXULAppAPI.h"
+#include "nscore.h"
 
 /**
  * Automatically cancel and abort synchronous LocalStorage requests (for example
@@ -32,70 +67,38 @@
  */
 #define FAILSAFE_CANCEL_SYNC_OP_MS 50000
 
-namespace mozilla {
-namespace dom {
+/**
+ * Interval with which to wake up while waiting for the sync op to complete to
+ * check ExpectingShutdown().
+ */
+#define SYNC_OP_WAKE_INTERVAL_MS 500
+
+namespace mozilla::dom {
 
 namespace {
 
 class RequestHelper;
 
-StaticMutex gRequestHelperMutex;
-nsISerialEventTarget* gSyncLoopEventTarget = nullptr;
-/**
- * Tracks whether a sync message has been received to the main thread but not
- * yet processed.  Used by RequestHelper logic to abort effectively synchronous
- * calls if a sync IPC message is received which could result in deadlock.
- * This is a boolean because, by definition, the parent can only send one sync
- * message to the child at a time.
- */
-bool gPendingSyncMessage = false;
-
-/*
- * Wrapper for the pushed event queue.  The wrapper automatically dispatches
- * runnables to the main thread when pushed event queue is no longer active.
- * This exists because the event loop spinning can be aborted.
- */
-class NestedEventTargetWrapper final : public nsISerialEventTarget {
-  nsCOMPtr<nsISerialEventTarget> mNestedEventTarget;
-  bool mDisconnected;
-
- public:
-  explicit NestedEventTargetWrapper(nsISerialEventTarget* aNestedEventTarget)
-      : mNestedEventTarget(aNestedEventTarget), mDisconnected(false) {}
-
- private:
-  ~NestedEventTargetWrapper() = default;
-
-  NS_DECL_THREADSAFE_ISUPPORTS
-
-  NS_DECL_NSIEVENTTARGET_FULL
-};
-
 /**
  * Main-thread helper that implements the blocking logic required by
- * LocalStorage's synchronous semantics.  StartAndReturnResponse pushes an
- * event queue which is a new event target and spins its nested event loop until
- * a result is received or an abort is necessary due to a PContent-managed sync
- * IPC message being received.  Note that because the event queue is its own
- * event target, there is no re-entrancy.  Normal main-thread runnables will not
- * get a chance to run.  See StartAndReturnResponse() for info on this choice.
+ * LocalStorage's synchronous semantics.  StartAndReturnResponse blocks on a
+ * monitor until a result is received. See StartAndReturnResponse() for info on
+ * this choice.
  *
  * The normal life-cycle of this method looks like:
  * - Main Thread: LSObject::DoRequestSynchronously creates a RequestHelper and
- *   invokes StartAndReturnResponse().  It pushes the event queue and Dispatches
- *   the RequestHelper to the DOM File Thread.
- * - DOM File Thread: RequestHelper::Run is called, invoking Start() which
- *   invokes LSObject::StartRequest, which gets-or-creates the PBackground actor
- *   if necessary (which may dispatch a runnable to the nested event queue on
- *   the main thread), sends LSRequest constructor which is provided with a
- *   callback reference to the RequestHelper. State advances to ResponsePending.
- * - DOM File Thread:: LSRequestChild::Recv__delete__ is received, which invokes
- *   RequestHelepr::OnResponse, advancing the state to Finishing and dispatching
- *   RequestHelper to its own nested event target.
- * - Main Thread: RequestHelper::Run is called, invoking Finish() which advances
- *   the state to Complete and sets mWaiting to false, allowing the nested event
- *   loop being spun by StartAndReturnResponse to cease spinning and return the
- *   received response.
+ *   invokes StartAndReturnResponse().  It Dispatches the RequestHelper to the
+ *   RemoteLazyInputStream thread, and waits on mMonitor.
+ * - RemoteLazyInputStream Thread: RequestHelper::Run is called, invoking
+ *   Start() which invokes LSObject::StartRequest, which gets-or-creates the
+ *   PBackground actor if necessary, sends LSRequest constructor which is
+ *   provided with a callback reference to the RequestHelper. State advances to
+ *   ResponsePending.
+ * - RemoteLazyInputStreamThread: LSRequestChild::Recv__delete__ is received,
+ *   which invokes RequestHelepr::OnResponse, advancing the state to Complete
+ *   and notifying mMonitor.
+ * - Main Thread: The main thread wakes up after waiting on the monitor,
+ *   returning the received response.
  *
  * See LocalStorageCommon.h for high-level context and method comments for
  * low-level details.
@@ -103,24 +106,27 @@ class NestedEventTargetWrapper final : public nsISerialEventTarget {
 class RequestHelper final : public Runnable, public LSRequestChildCallback {
   enum class State {
     /**
-     * The RequestHelper has been created and dispatched to the DOM File Thread.
+     * The RequestHelper has been created and dispatched to the
+     * RemoteLazyInputStream Thread.
      */
     Initial,
     /**
-     * Start() has been invoked on the DOM File Thread and
+     * Start() has been invoked on the RemoteLazyInputStream Thread and
      * LSObject::StartRequest has been invoked from there, sending an IPC
      * message to PBackground to service the request.  We stay in this state
-     * until a response is received.
+     * until a response is received or a timeout occurs.
      */
     ResponsePending,
     /**
-     * A response has been received and RequestHelper has been dispatched back
-     * to the nested event loop to call Finish().
+     * The request timed out, or failed in some fashion, and needs to be
+     * cancelled. A runnable has been dispatched to the DOM File thread to
+     * notify the parent actor, and the main thread will continue to block until
+     * we receive a reponse.
      */
-    Finishing,
+    Canceling,
     /**
-     * Finish() has been called on the main thread.  The nested event loop will
-     * terminate imminently and the received response returned to the caller of
+     * The request is complete, either successfully or due to being cancelled.
+     * The main thread can stop waiting and immediately return to the caller of
      * StartAndReturnResponse.
      */
     Complete
@@ -133,37 +139,26 @@ class RequestHelper final : public Runnable, public LSRequestChildCallback {
   // The thread the RequestHelper was created on.  This should be the main
   // thread.
   nsCOMPtr<nsIEventTarget> mOwningEventTarget;
-  // The pushed event queue that we use to spin the event loop without
-  // processing any of the events dispatched at the mOwningEventTarget (which
-  // would result in re-entrancy and violate LocalStorage semantics).
-  nsCOMPtr<nsISerialEventTarget> mNestedEventTarget;
-  // The wrapper for the pushed event queue.  The wrapper automatically
-  // dispatches runnables to the main thread when pushed event queue is no
-  // longer active.  This exists because the event loop spinning can be aborted.
-  nsCOMPtr<nsISerialEventTarget> mNestedEventTargetWrapper;
   // The IPC actor handling the request with standard IPC allocation rules.
   // Our reference is nulled in OnResponse which corresponds to the actor's
   // __destroy__ method.
   LSRequestChild* mActor;
   const LSRequestParams mParams;
-  LSRequestResponse mResponse;
-  nsresult mResultCode;
-  State mState;
-  // Control flag for the nested event loop; once set to false, the loop ends.
-  bool mWaiting;
-  bool mCancelled;
+  Monitor mMonitor;
+  LSRequestResponse mResponse MOZ_GUARDED_BY(mMonitor);
+  nsresult mResultCode MOZ_GUARDED_BY(mMonitor);
+  State mState MOZ_GUARDED_BY(mMonitor);
 
  public:
   RequestHelper(LSObject* aObject, const LSRequestParams& aParams)
       : Runnable("dom::RequestHelper"),
         mObject(aObject),
-        mOwningEventTarget(GetCurrentThreadEventTarget()),
+        mOwningEventTarget(GetCurrentSerialEventTarget()),
         mActor(nullptr),
         mParams(aParams),
+        mMonitor("dom::RequestHelper::mMonitor"),
         mResultCode(NS_OK),
-        mState(State::Initial),
-        mWaiting(true),
-        mCancelled(false) {}
+        mState(State::Initial) {}
 
   bool IsOnOwningThread() const {
     MOZ_ASSERT(mOwningEventTarget);
@@ -183,10 +178,6 @@ class RequestHelper final : public Runnable, public LSRequestChildCallback {
  private:
   ~RequestHelper() = default;
 
-  nsresult Start();
-
-  void Finish();
-
   NS_DECL_ISUPPORTS_INHERITED
 
   NS_DECL_NSIRUNNABLE
@@ -195,6 +186,25 @@ class RequestHelper final : public Runnable, public LSRequestChildCallback {
   void OnResponse(const LSRequestResponse& aResponse) override;
 };
 
+void AssertExplicitSnapshotInvariants(const LSObject& aObject) {
+  // Can be only called if the mInExplicitSnapshot flag is true.
+  // An explicit snapshot must have been created.
+  MOZ_ASSERT(aObject.InExplicitSnapshot());
+
+  // If an explicit snapshot has been created then mDatabase must be not null.
+  // DropDatabase could be called in the meatime, but that must be preceded by
+  // Disconnect which sets mInExplicitSnapshot to false. EnsureDatabase could
+  // be called in the meantime too, but that can't set mDatabase to null or to
+  // a new value. See the comment below.
+  MOZ_ASSERT(aObject.DatabaseStrongRef());
+
+  // Existence of a snapshot prevents the database from allowing to close. See
+  // LSDatabase::RequestAllowToClose and LSDatabase::NoteFinishedSnapshot.
+  // If the database is not allowed to close then mDatabase could not have been
+  // nulled out or set to a new value. See EnsureDatabase.
+  MOZ_ASSERT(!aObject.DatabaseStrongRef()->IsAllowedToClose());
+}
+
 }  // namespace
 
 LSObject::LSObject(nsPIDOMWindowInner* aWindow, nsIPrincipal* aPrincipal,
@@ -212,34 +222,6 @@ LSObject::~LSObject() {
   DropObserver();
 }
 
-// static
-void LSObject::Initialize() {
-  MOZ_ASSERT(NS_IsMainThread());
-
-  nsCOMPtr<nsIEventTarget> domFileThread =
-      IPCBlobInputStreamThread::GetOrCreate();
-  if (NS_WARN_IF(!domFileThread)) {
-    return;
-  }
-
-  RefPtr<Runnable> runnable =
-      NS_NewRunnableFunction("LSObject::Initialize", []() {
-        AssertIsOnDOMFileThread();
-
-        PBackgroundChild* backgroundActor =
-            BackgroundChild::GetOrCreateForCurrentThread();
-
-        if (NS_WARN_IF(!backgroundActor)) {
-          return;
-        }
-      });
-
-  if (NS_WARN_IF(
-          NS_FAILED(domFileThread->Dispatch(runnable, NS_DISPATCH_NORMAL)))) {
-    return;
-  }
-}
-
 // static
 nsresult LSObject::CreateForWindow(nsPIDOMWindowInner* aWindow,
                                    Storage** aStorage) {
@@ -295,19 +277,24 @@ nsresult LSObject::CreateForWindow(nsPIDOMWindowInner* aWindow,
   MOZ_ASSERT(storagePrincipalInfo->type() ==
              PrincipalInfo::TContentPrincipalInfo);
 
-  if (NS_WARN_IF(!QuotaManager::IsPrincipalInfoValid(*storagePrincipalInfo))) {
+  if (NS_WARN_IF(
+          !quota::QuotaManager::IsPrincipalInfoValid(*storagePrincipalInfo))) {
     return NS_ERROR_FAILURE;
   }
 
-  nsCString suffix;
-  nsCString origin;
-  rv = QuotaManager::GetInfoFromPrincipal(storagePrincipal.get(), &suffix,
-                                          nullptr, &origin);
-  if (NS_WARN_IF(NS_FAILED(rv))) {
-    return rv;
-  }
+#ifdef DEBUG
+  QM_TRY_INSPECT(
+      const auto& principalMetadata,
+      quota::QuotaManager::GetInfoFromPrincipal(storagePrincipal.get()));
 
-  MOZ_ASSERT(originAttrSuffix == suffix);
+  MOZ_ASSERT(originAttrSuffix == principalMetadata.mSuffix);
+
+  const auto& origin = principalMetadata.mOrigin;
+#else
+  QM_TRY_INSPECT(
+      const auto& origin,
+      quota::QuotaManager::GetOriginFromPrincipal(storagePrincipal.get()));
+#endif
 
   uint32_t privateBrowsingId;
   rv = storagePrincipal->GetPrivateBrowsingId(&privateBrowsingId);
@@ -322,6 +309,9 @@ nsresult LSObject::CreateForWindow(nsPIDOMWindowInner* aWindow,
 
   Maybe<nsID> clientId = Some(clientInfo.ref().Id());
 
+  Maybe<PrincipalInfo> clientPrincipalInfo =
+      Some(clientInfo.ref().PrincipalInfo());
+
   nsString documentURI;
   if (nsCOMPtr<Document> doc = aWindow->GetExtantDoc()) {
     rv = doc->GetDocumentURI(documentURI);
@@ -335,6 +325,7 @@ nsresult LSObject::CreateForWindow(nsPIDOMWindowInner* aWindow,
   object->mStoragePrincipalInfo = std::move(storagePrincipalInfo);
   object->mPrivateBrowsingId = privateBrowsingId;
   object->mClientId = clientId;
+  object->mClientPrincipalInfo = clientPrincipalInfo;
   object->mOrigin = origin;
   object->mOriginKey = originKey;
   object->mDocumentURI = documentURI;
@@ -381,24 +372,39 @@ nsresult LSObject::CreateForPrincipal(nsPIDOMWindowInner* aWindow,
       storagePrincipalInfo->type() == PrincipalInfo::TContentPrincipalInfo ||
       storagePrincipalInfo->type() == PrincipalInfo::TSystemPrincipalInfo);
 
-  if (NS_WARN_IF(!QuotaManager::IsPrincipalInfoValid(*storagePrincipalInfo))) {
+  if (NS_WARN_IF(
+          !quota::QuotaManager::IsPrincipalInfoValid(*storagePrincipalInfo))) {
     return NS_ERROR_FAILURE;
   }
 
-  nsCString suffix;
-  nsCString origin;
+#ifdef DEBUG
+  QM_TRY_INSPECT(
+      const auto& principalMetadata,
+      ([&storagePrincipalInfo,
+        &aPrincipal]() -> Result<quota::PrincipalMetadata, nsresult> {
+        if (storagePrincipalInfo->type() ==
+            PrincipalInfo::TSystemPrincipalInfo) {
+          return quota::QuotaManager::GetInfoForChrome();
+        }
 
-  if (storagePrincipalInfo->type() == PrincipalInfo::TSystemPrincipalInfo) {
-    QuotaManager::GetInfoForChrome(&suffix, nullptr, &origin);
-  } else {
-    rv = QuotaManager::GetInfoFromPrincipal(aPrincipal, &suffix, nullptr,
-                                            &origin);
-    if (NS_WARN_IF(NS_FAILED(rv))) {
-      return rv;
-    }
-  }
+        QM_TRY_RETURN(quota::QuotaManager::GetInfoFromPrincipal(aPrincipal));
+      }()));
 
-  MOZ_ASSERT(originAttrSuffix == suffix);
+  MOZ_ASSERT(originAttrSuffix == principalMetadata.mSuffix);
+
+  const auto& origin = principalMetadata.mOrigin;
+#else
+  QM_TRY_INSPECT(
+      const auto& origin, ([&storagePrincipalInfo,
+                            &aPrincipal]() -> Result<nsAutoCString, nsresult> {
+        if (storagePrincipalInfo->type() ==
+            PrincipalInfo::TSystemPrincipalInfo) {
+          return nsAutoCString{quota::QuotaManager::GetOriginForChrome()};
+        }
+
+        QM_TRY_RETURN(quota::QuotaManager::GetOriginFromPrincipal(aPrincipal));
+      }()));
+#endif
 
   Maybe<nsID> clientId;
   if (aWindow) {
@@ -424,56 +430,14 @@ nsresult LSObject::CreateForPrincipal(nsPIDOMWindowInner* aWindow,
 
   object.forget(aObject);
   return NS_OK;
-}
+}  // namespace dom
 
-// static
-already_AddRefed<nsISerialEventTarget> LSObject::GetSyncLoopEventTarget() {
-  MOZ_ASSERT(XRE_IsParentProcess());
-
-  nsCOMPtr<nsISerialEventTarget> target;
-
-  {
-    StaticMutexAutoLock lock(gRequestHelperMutex);
-    target = gSyncLoopEventTarget;
-  }
-
-  return target.forget();
-}
-
-// static
-void LSObject::OnSyncMessageReceived() {
-  nsCOMPtr<nsISerialEventTarget> target;
-
-  {
-    StaticMutexAutoLock lock(gRequestHelperMutex);
-    target = gSyncLoopEventTarget;
-    gPendingSyncMessage = true;
-  }
-
-  if (target) {
-    RefPtr<Runnable> runnable =
-        NS_NewRunnableFunction("LSObject::CheckFlagRunnable", []() {});
-
-    MOZ_ALWAYS_SUCCEEDS(
-        target->Dispatch(runnable.forget(), NS_DISPATCH_NORMAL));
-  }
-}
-
-// static
-void LSObject::OnSyncMessageHandled() {
-  StaticMutexAutoLock lock(gRequestHelperMutex);
-  gPendingSyncMessage = false;
-}
-
-LSRequestChild* LSObject::StartRequest(nsIEventTarget* aMainEventTarget,
-                                       const LSRequestParams& aParams,
+LSRequestChild* LSObject::StartRequest(const LSRequestParams& aParams,
                                        LSRequestChildCallback* aCallback) {
   AssertIsOnDOMFileThread();
 
-  PBackgroundChild* backgroundActor =
-      XRE_IsParentProcess()
-          ? BackgroundChild::GetOrCreateForCurrentThread(aMainEventTarget)
-          : BackgroundChild::GetForCurrentThread();
+  mozilla::ipc::PBackgroundChild* backgroundActor =
+      mozilla::ipc::BackgroundChild::GetOrCreateForCurrentThread();
   if (NS_WARN_IF(!backgroundActor)) {
     return nullptr;
   }
@@ -523,6 +487,20 @@ int64_t LSObject::GetOriginQuotaUsage() const {
   return 0;
 }
 
+void LSObject::Disconnect() {
+  // Explicit snapshots which were not ended in JS, must be ended here while
+  // IPC is still available. We can't do that in DropDatabase because actors
+  // may have been destroyed already at that point.
+  if (mInExplicitSnapshot) {
+    AssertExplicitSnapshotInvariants(*this);
+
+    nsresult rv = mDatabase->EndExplicitSnapshot();
+    Unused << NS_WARN_IF(NS_FAILED(rv));
+
+    mInExplicitSnapshot = false;
+  }
+}
+
 uint32_t LSObject::GetLength(nsIPrincipal& aSubjectPrincipal,
                              ErrorResult& aError) {
   AssertIsOnOwningThread();
@@ -757,6 +735,29 @@ void LSObject::BeginExplicitSnapshot(nsIPrincipal& aSubjectPrincipal,
   mInExplicitSnapshot = true;
 }
 
+void LSObject::CheckpointExplicitSnapshot(nsIPrincipal& aSubjectPrincipal,
+                                          ErrorResult& aError) {
+  AssertIsOnOwningThread();
+
+  if (!CanUseStorage(aSubjectPrincipal)) {
+    aError.Throw(NS_ERROR_DOM_SECURITY_ERR);
+    return;
+  }
+
+  if (!mInExplicitSnapshot) {
+    aError.Throw(NS_ERROR_NOT_INITIALIZED);
+    return;
+  }
+
+  AssertExplicitSnapshotInvariants(*this);
+
+  nsresult rv = mDatabase->CheckpointExplicitSnapshot();
+  if (NS_WARN_IF(NS_FAILED(rv))) {
+    aError.Throw(rv);
+    return;
+  }
+}
+
 void LSObject::EndExplicitSnapshot(nsIPrincipal& aSubjectPrincipal,
                                    ErrorResult& aError) {
   AssertIsOnOwningThread();
@@ -771,15 +772,38 @@ void LSObject::EndExplicitSnapshot(nsIPrincipal& aSubjectPrincipal,
     return;
   }
 
-  nsresult rv = EndExplicitSnapshotInternal();
+  AssertExplicitSnapshotInvariants(*this);
+
+  nsresult rv = mDatabase->EndExplicitSnapshot();
   if (NS_WARN_IF(NS_FAILED(rv))) {
     aError.Throw(rv);
     return;
   }
+
+  mInExplicitSnapshot = false;
 }
 
-bool LSObject::GetHasActiveSnapshot(nsIPrincipal& aSubjectPrincipal,
-                                    ErrorResult& aError) {
+bool LSObject::GetHasSnapshot(nsIPrincipal& aSubjectPrincipal,
+                              ErrorResult& aError) {
+  AssertIsOnOwningThread();
+
+  if (!CanUseStorage(aSubjectPrincipal)) {
+    aError.Throw(NS_ERROR_DOM_SECURITY_ERR);
+    return false;
+  }
+
+  // We can't call `HasSnapshot` on the database if it's being closed, but we
+  // know that a database which is being closed can't have a snapshot, so we
+  // return false in that case directly here.
+  if (!mDatabase || mDatabase->IsAllowedToClose()) {
+    return false;
+  }
+
+  return mDatabase->HasSnapshot();
+}
+
+int64_t LSObject::GetSnapshotUsage(nsIPrincipal& aSubjectPrincipal,
+                                   ErrorResult& aError) {
   AssertIsOnOwningThread();
 
   if (!CanUseStorage(aSubjectPrincipal)) {
@@ -787,13 +811,17 @@ bool LSObject::GetHasActiveSnapshot(nsIPrincipal& aSubjectPrincipal,
     return 0;
   }
 
-  if (mDatabase && mDatabase->HasActiveSnapshot()) {
-    MOZ_ASSERT(!mDatabase->IsAllowedToClose());
-
-    return true;
+  if (!mDatabase || mDatabase->IsAllowedToClose()) {
+    aError.Throw(NS_ERROR_NOT_AVAILABLE);
+    return 0;
   }
 
-  return false;
+  if (!mDatabase->HasSnapshot()) {
+    aError.Throw(NS_ERROR_NOT_AVAILABLE);
+    return 0;
+  }
+
+  return mDatabase->GetSnapshotUsage();
 }
 
 NS_IMPL_ADDREF_INHERITED(LSObject, Storage)
@@ -819,17 +847,17 @@ nsresult LSObject::DoRequestSynchronously(const LSRequestParams& aParams,
   // too late to initialize PBackground child on the owning thread, because
   // it can fail and parent would keep an extra strong ref to the datastore or
   // observer.
-  PBackgroundChild* backgroundActor =
-      BackgroundChild::GetOrCreateForCurrentThread();
+  mozilla::ipc::PBackgroundChild* backgroundActor =
+      mozilla::ipc::BackgroundChild::GetOrCreateForCurrentThread();
   if (NS_WARN_IF(!backgroundActor)) {
     return NS_ERROR_FAILURE;
   }
 
   RefPtr<RequestHelper> helper = new RequestHelper(this, aParams);
 
-  // This will start and finish the request on the DOM File thread.
+  // This will start and finish the request on the RemoteLazyInputStream thread.
   // The owning thread is synchronously blocked while the request is
-  // asynchronously processed on the DOM File thread.
+  // asynchronously processed on the RemoteLazyInputStream thread.
   nsresult rv = helper->StartAndReturnResponse(aResponse);
   if (NS_WARN_IF(NS_FAILED(rv))) {
     return rv;
@@ -865,8 +893,8 @@ nsresult LSObject::EnsureDatabase() {
   // We don't need this yet, but once the request successfully finishes, it's
   // too late to initialize PBackground child on the owning thread, because
   // it can fail and parent would keep an extra strong ref to the datastore.
-  PBackgroundChild* backgroundActor =
-      BackgroundChild::GetOrCreateForCurrentThread();
+  mozilla::ipc::PBackgroundChild* backgroundActor =
+      mozilla::ipc::BackgroundChild::GetOrCreateForCurrentThread();
   if (NS_WARN_IF(!backgroundActor)) {
     return NS_ERROR_FAILURE;
   }
@@ -879,6 +907,7 @@ nsresult LSObject::EnsureDatabase() {
   LSRequestPrepareDatastoreParams params;
   params.commonParams() = commonParams;
   params.clientId() = mClientId;
+  params.clientPrincipalInfo() = mClientPrincipalInfo;
 
   LSRequestResponse response;
 
@@ -896,7 +925,7 @@ nsresult LSObject::EnsureDatabase() {
   uint64_t datastoreId = prepareDatastoreResponse.datastoreId();
 
   // The datastore is now ready on the parent side (prepared by the asynchronous
-  // request on the DOM File thread).
+  // request on the RemoteLazyInputStream thread).
   // Let's create a direct connection to the datastore (through a database
   // actor) from the owning thread.
   // Note that we now can't error out, otherwise parent will keep an extra
@@ -919,11 +948,6 @@ nsresult LSObject::EnsureDatabase() {
 void LSObject::DropDatabase() {
   AssertIsOnOwningThread();
 
-  if (mInExplicitSnapshot) {
-    nsresult rv = EndExplicitSnapshotInternal();
-    Unused << NS_WARN_IF(NS_FAILED(rv));
-  }
-
   mDatabase = nullptr;
 }
 
@@ -944,6 +968,7 @@ nsresult LSObject::EnsureObserver() {
   params.principalInfo() = *mPrincipalInfo;
   params.storagePrincipalInfo() = *mStoragePrincipalInfo;
   params.clientId() = mClientId;
+  params.clientPrincipalInfo() = mClientPrincipalInfo;
 
   LSRequestResponse response;
 
@@ -961,13 +986,14 @@ nsresult LSObject::EnsureObserver() {
   uint64_t observerId = prepareObserverResponse.observerId();
 
   // The obsserver is now ready on the parent side (prepared by the asynchronous
-  // request on the DOM File thread).
+  // request on the RemoteLazyInputStream thread).
   // Let's create a direct connection to the observer (through an observer
   // actor) from the owning thread.
   // Note that we now can't error out, otherwise parent will keep an extra
   // strong reference to the observer.
 
-  PBackgroundChild* backgroundActor = BackgroundChild::GetForCurrentThread();
+  mozilla::ipc::PBackgroundChild* backgroundActor =
+      mozilla::ipc::BackgroundChild::GetForCurrentThread();
   MOZ_ASSERT(backgroundActor);
 
   RefPtr<LSObserver> observer = new LSObserver(mOrigin);
@@ -1002,240 +1028,70 @@ void LSObject::OnChange(const nsAString& aKey, const nsAString& aOldValue,
                /* aImmediateDispatch */ false);
 }
 
-nsresult LSObject::EndExplicitSnapshotInternal() {
-  AssertIsOnOwningThread();
-
-  // Can be only called if the mInExplicitSnapshot flag is true.
-  // An explicit snapshot must have been created.
-  MOZ_ASSERT(mInExplicitSnapshot);
-
-  // If an explicit snapshot have been created then mDatabase must be not null.
-  // DropDatabase could be called in the meatime, but that would set
-  // mInExplicitSnapshot to false. EnsureDatabase could be called in the
-  // meantime too, but that can't set mDatabase to null or to a new value. See
-  // the comment below.
-  MOZ_ASSERT(mDatabase);
-
-  // Existence of a snapshot prevents the database from allowing to close. See
-  // LSDatabase::RequestAllowToClose and LSDatabase::NoteFinishedSnapshot.
-  // If the database is not allowed to close then mDatabase could not have been
-  // nulled out or set to a new value. See EnsureDatabase.
-  MOZ_ASSERT(!mDatabase->IsAllowedToClose());
-
-  nsresult rv = mDatabase->EndExplicitSnapshot(this);
-  if (NS_WARN_IF(NS_FAILED(rv))) {
-    return rv;
-  }
-
-  mInExplicitSnapshot = false;
-
-  return NS_OK;
-}
-
 void LSObject::LastRelease() {
   AssertIsOnOwningThread();
 
   DropDatabase();
 }
 
-NS_IMPL_ISUPPORTS(NestedEventTargetWrapper, nsIEventTarget,
-                  nsISerialEventTarget);
-
-NS_IMETHODIMP_(bool)
-NestedEventTargetWrapper::IsOnCurrentThreadInfallible() {
-  MOZ_CRASH(
-      "IsOnCurrentThreadInfallible should never be called on "
-      "NestedEventTargetWrapper");
-}
-
-NS_IMETHODIMP
-NestedEventTargetWrapper::IsOnCurrentThread(bool* aResult) {
-  MOZ_CRASH(
-      "IsOnCurrentThread should never be called on "
-      "NestedEventTargetWrapper");
-}
-
-NS_IMETHODIMP
-NestedEventTargetWrapper::Dispatch(already_AddRefed<nsIRunnable> aEvent,
-                                   uint32_t aFlags) {
-  MOZ_ASSERT(mNestedEventTarget);
-
-  if (mDisconnected) {
-    MOZ_ALWAYS_SUCCEEDS(NS_DispatchToMainThread(std::move(aEvent), aFlags));
-    return NS_OK;
-  }
-
-  nsCOMPtr<nsIRunnable> event(aEvent);
-
-  nsresult rv = mNestedEventTarget->Dispatch(event, aFlags);
-  if (rv == NS_ERROR_UNEXPECTED) {
-    mDisconnected = true;
-
-    // Dispatch leaked the event object on the NS_ERROR_UNEXPECTED failure, so
-    // we explicitly release this object once for that.
-    event.get()->Release();
-
-    MOZ_ALWAYS_SUCCEEDS(NS_DispatchToMainThread(event.forget(), aFlags));
-  } else if (NS_WARN_IF(NS_FAILED(rv))) {
-    return rv;
-  }
-
-  return NS_OK;
-}
-
-NS_IMETHODIMP
-NestedEventTargetWrapper::DispatchFromScript(nsIRunnable* aEvent,
-                                             uint32_t aFlags) {
-  MOZ_CRASH(
-      "DispatchFromScript should never be called on "
-      "NestedEventTargetWrapper");
-}
-
-NS_IMETHODIMP
-NestedEventTargetWrapper::DelayedDispatch(already_AddRefed<nsIRunnable> aEvent,
-                                          uint32_t aDelayMs) {
-  MOZ_CRASH(
-      "DelayedDispatch should never be called on "
-      "NestedEventTargetWrapper");
-}
-
 nsresult RequestHelper::StartAndReturnResponse(LSRequestResponse& aResponse) {
   AssertIsOnOwningThread();
 
-  // Normally, we would use the standard way of blocking the thread using
-  // a monitor.
-  // The problem is that BackgroundChild::GetOrCreateForCurrentThread()
-  // called on the DOM File thread may dispatch a runnable to the main
-  // thread to finish initialization of PBackground. A monitor would block
-  // the main thread and the runnable would never get executed causing the
-  // helper to be stuck in a wait loop.
-  // However, BackgroundChild::GetOrCreateForCurrentThread() supports passing
-  // a custom main event target, so we can create a nested event target and
-  // spin the event loop. Nothing can dispatch to the nested event target
-  // except BackgroundChild::GetOrCreateForCurrentThread(), so spinning of the
-  // event loop can't fire any other events.
-  // This way the thread is synchronously blocked in a safe manner and the
-  // runnable gets executed.
-  {
-    auto thread = static_cast<nsThread*>(NS_GetCurrentThread());
-
-    const nsLocalExecutionGuard localExecution(thread->EnterLocalExecution());
-    mNestedEventTarget = localExecution.GetEventTarget();
-    MOZ_ASSERT(mNestedEventTarget);
-
-    mNestedEventTargetWrapper =
-        new NestedEventTargetWrapper(mNestedEventTarget);
-
-    nsCOMPtr<nsIEventTarget> domFileThread =
-        XRE_IsParentProcess() ? IPCBlobInputStreamThread::GetOrCreate()
-                              : IPCBlobInputStreamThread::Get();
-    if (NS_WARN_IF(!domFileThread)) {
-      return NS_ERROR_FAILURE;
-    }
-
-    nsresult rv;
-
-    {
-      {
-        StaticMutexAutoLock lock(gRequestHelperMutex);
-
-        if (StaticPrefs::dom_storage_abort_on_sync_parent_to_child_messages() &&
-            NS_WARN_IF(gPendingSyncMessage)) {
-          return NS_ERROR_FAILURE;
-        }
-
-        gSyncLoopEventTarget = mNestedEventTargetWrapper;
-      }
-
-      auto autoClearSyncLoopEventTarget = mozilla::MakeScopeExit([&] {
-        StaticMutexAutoLock lock(gRequestHelperMutex);
-        gSyncLoopEventTarget = nullptr;
-      });
-
-      rv = domFileThread->Dispatch(this, NS_DISPATCH_NORMAL);
-      if (NS_WARN_IF(NS_FAILED(rv))) {
-        return rv;
-      }
-
-      nsCOMPtr<nsITimer> timer = NS_NewTimer();
-
-      MOZ_ALWAYS_SUCCEEDS(timer->SetTarget(mNestedEventTarget));
-
-      MOZ_ALWAYS_SUCCEEDS(timer->InitWithNamedFuncCallback(
-          [](nsITimer* aTimer, void* aClosure) {
-            auto helper = static_cast<RequestHelper*>(aClosure);
-
-            helper->mCancelled = true;
-          },
-          this, FAILSAFE_CANCEL_SYNC_OP_MS, nsITimer::TYPE_ONE_SHOT,
-          "RequestHelper::StartAndReturnResponse::SpinEventLoopTimer"));
-
-      MOZ_ALWAYS_TRUE(SpinEventLoopUntil(
-          [&]() {
-            if (mCancelled) {
-              return true;
-            }
-
-            if (!mWaiting) {
-              return true;
-            }
-
-            {
-              StaticMutexAutoLock lock(gRequestHelperMutex);
-              if (StaticPrefs::
-                      dom_storage_abort_on_sync_parent_to_child_messages() &&
-                  NS_WARN_IF(gPendingSyncMessage)) {
-                return true;
-              }
-            }
-
-            return false;
-          },
-          thread));
-
-      MOZ_ALWAYS_SUCCEEDS(timer->Cancel());
-    }
-
-    // If mWaiting is still set to true, it means that the event loop spinning
-    // was aborted and we need to cancel the request in the parent since we
-    // don't care about the result anymore.
-    // We can check mWaiting here because it's only ever touched on the main
-    // thread.
-    if (NS_WARN_IF(mWaiting)) {
-      // Don't touch mResponse, mResultCode or mState here! The DOM File Thread
-      // may be accessing them at the same moment.
-
-      RefPtr<RequestHelper> self = this;
-
-      RefPtr<Runnable> runnable =
-          NS_NewRunnableFunction("RequestHelper::SendCancelRunnable", [self]() {
-            LSRequestChild* actor = self->mActor;
-
-            // Start() could fail or it hasn't had a chance to run yet, so we
-            // need to check if actor is not null.
-            // The actor can also be in the final (finishing) state, in that
-            // case we are not allowed to send the cancel message and it
-            // wouldn't make sense because the request is about to be destroyed
-            // anyway.
-            if (actor && !actor->Finishing()) {
-              actor->SendCancel();
-            }
-          });
-
-      rv = domFileThread->Dispatch(runnable, NS_DISPATCH_NORMAL);
-      if (NS_WARN_IF(NS_FAILED(rv))) {
-        return rv;
-      }
-
-      return NS_ERROR_FAILURE;
-    }
-
-    // localExecution will be destructed when we leave this scope. If the event
-    // loop spinning was aborted and other threads dispatched new runnables to
-    // the nested event queue, they will be moved to the main event queue here
-    // and later asynchronusly processed.  So nothing will be lost.
+  nsCOMPtr<nsIEventTarget> domFileThread =
+      RemoteLazyInputStreamThread::GetOrCreate();
+  if (NS_WARN_IF(!domFileThread)) {
+    return NS_ERROR_ILLEGAL_DURING_SHUTDOWN;
   }
 
+  nsresult rv = domFileThread->Dispatch(this, NS_DISPATCH_NORMAL);
+  if (NS_WARN_IF(NS_FAILED(rv))) {
+    return rv;
+  }
+
+  TimeStamp deadline = TimeStamp::Now() + TimeDuration::FromMilliseconds(
+                                              FAILSAFE_CANCEL_SYNC_OP_MS);
+
+  MonitorAutoLock lock(mMonitor);
+  while (mState != State::Complete) {
+    TimeStamp now = TimeStamp::Now();
+    // If we are expecting shutdown or have passed our deadline, immediately
+    // dispatch ourselves to the DOM File thread to cancel the operation.  We
+    // don't abort until the cancellation has gone through, as otherwise we
+    // could race with the DOM File thread.
+    if (mozilla::ipc::ProcessChild::ExpectingShutdown() || now >= deadline) {
+      switch (mState) {
+        case State::Initial:
+          // The DOM File thread never even woke before ExpectingShutdown() or a
+          // timeout - skip even creating the actor and just report an error.
+          mResultCode = NS_ERROR_FAILURE;
+          mState = State::Complete;
+          continue;
+        case State::ResponsePending:
+          // The DOM File thread is currently waiting for a reply, switch to a
+          // canceling state, and notify it to cancel by dispatching a runnable.
+          mState = State::Canceling;
+          MOZ_ALWAYS_SUCCEEDS(
+              domFileThread->Dispatch(this, NS_DISPATCH_NORMAL));
+          [[fallthrough]];
+        case State::Canceling:
+          // We've cancelled the request, so just need to wait indefinitely for
+          // it to complete.
+          lock.Wait();
+          continue;
+        default:
+          MOZ_ASSERT_UNREACHABLE("unexpected state");
+      }
+    }
+
+    // Wait until either we reach out deadline or for SYNC_OP_WAIT_INTERVAL_MS.
+    lock.Wait(TimeDuration::Min(
+        TimeDuration::FromMilliseconds(SYNC_OP_WAKE_INTERVAL_MS),
+        deadline - now));
+  }
+
+  // The operation is complete, clear our reference to the LSObject.
+  mObject = nullptr;
+
   if (NS_WARN_IF(NS_FAILED(mResultCode))) {
     return mResultCode;
   }
@@ -1244,84 +1100,68 @@ nsresult RequestHelper::StartAndReturnResponse(LSRequestResponse& aResponse) {
   return NS_OK;
 }
 
-nsresult RequestHelper::Start() {
-  AssertIsOnDOMFileThread();
-  MOZ_ASSERT(mState == State::Initial);
-
-  mState = State::ResponsePending;
-
-  LSRequestChild* actor =
-      mObject->StartRequest(mNestedEventTargetWrapper, mParams, this);
-  if (NS_WARN_IF(!actor)) {
-    return NS_ERROR_FAILURE;
-  }
-
-  mActor = actor;
-
-  return NS_OK;
-}
-
-void RequestHelper::Finish() {
-  AssertIsOnOwningThread();
-  MOZ_ASSERT(mState == State::Finishing);
-
-  mObject = nullptr;
-
-  mWaiting = false;
-
-  mState = State::Complete;
-}
-
 NS_IMPL_ISUPPORTS_INHERITED0(RequestHelper, Runnable)
 
 NS_IMETHODIMP
 RequestHelper::Run() {
-  nsresult rv;
+  AssertIsOnDOMFileThread();
+
+  MonitorAutoLock lock(mMonitor);
 
   switch (mState) {
-    case State::Initial:
-      rv = Start();
-      break;
-
-    case State::Finishing:
-      Finish();
+    case State::Initial: {
+      mState = State::ResponsePending;
+      {
+        MonitorAutoUnlock unlock(mMonitor);
+        mActor = mObject->StartRequest(mParams, this);
+      }
+      if (NS_WARN_IF(!mActor) && mState != State::Complete) {
+        // If we fail to even create the actor, instantly fail and notify our
+        // caller of the error. Otherwise we'll notify from OnResponse as called
+        // by the actor.
+        mResultCode = NS_ERROR_FAILURE;
+        mState = State::Complete;
+        lock.Notify();
+      }
       return NS_OK;
+    }
+
+    case State::Canceling: {
+      // StartRequest() could fail or OnResponse was already called, so we need
+      // to check if actor is not null. The actor can also be in the final
+      // (finishing) state, in that case we are not allowed to send the cancel
+      // message and it wouldn't make any sense because the request is about to
+      // be destroyed anyway.
+      if (mActor && !mActor->Finishing()) {
+        mActor->SendCancel();
+      }
+      return NS_OK;
+    }
+
+    case State::Complete: {
+      // The operation was cancelled before we ran, do nothing.
+      return NS_OK;
+    }
 
     default:
       MOZ_CRASH("Bad state!");
   }
-
-  if (NS_WARN_IF(NS_FAILED(rv)) && mState != State::Finishing) {
-    if (NS_SUCCEEDED(mResultCode)) {
-      mResultCode = rv;
-    }
-
-    mState = State::Finishing;
-
-    if (IsOnOwningThread()) {
-      Finish();
-    } else {
-      MOZ_ALWAYS_SUCCEEDS(
-          mNestedEventTargetWrapper->Dispatch(this, NS_DISPATCH_NORMAL));
-    }
-  }
-
-  return NS_OK;
 }
 
 void RequestHelper::OnResponse(const LSRequestResponse& aResponse) {
   AssertIsOnDOMFileThread();
-  MOZ_ASSERT(mState == State::ResponsePending);
+
+  MonitorAutoLock lock(mMonitor);
+
+  MOZ_ASSERT(mState == State::ResponsePending || mState == State::Canceling);
 
   mActor = nullptr;
 
   mResponse = aResponse;
 
-  mState = State::Finishing;
+  mState = State::Complete;
 
-  MOZ_ALWAYS_SUCCEEDS(
-      mNestedEventTargetWrapper->Dispatch(this, NS_DISPATCH_NORMAL));
+  lock.Notify();
 }
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
