@@ -4,23 +4,49 @@
  * License, v. 2.0. If a copy of the MPL was not distributed with this file,
  * You can obtain one at http://mozilla.org/MPL/2.0/. */
 
-#include "LSSnapshot.h"
+#include "mozilla/dom/LSSnapshot.h"
 
+// Local includes
 #include "ActorsChild.h"
-#include "LocalStorageCommon.h"
 #include "LSDatabase.h"
 #include "LSWriteOptimizer.h"
+#include "LSWriteOptimizerImpl.h"
+#include "LocalStorageCommon.h"
+
+// Global includes
+#include <cstdint>
+#include <cstdlib>
+#include <new>
+#include <type_traits>
+#include <utility>
+#include "ErrorList.h"
+#include "mozilla/DebugOnly.h"
+#include "mozilla/MacroForEach.h"
+#include "mozilla/Maybe.h"
+#include "mozilla/Preferences.h"
+#include "mozilla/RefPtr.h"
+#include "mozilla/ScopeExit.h"
+#include "mozilla/StaticPrefs_dom.h"
+#include "mozilla/UniquePtr.h"
+#include "mozilla/dom/BindingDeclarations.h"
+#include "mozilla/dom/LSValue.h"
+#include "mozilla/dom/PBackgroundLSDatabase.h"
+#include "mozilla/dom/PBackgroundLSSharedTypes.h"
 #include "mozilla/dom/PBackgroundLSSnapshot.h"
+#include "nsBaseHashtable.h"
+#include "nsCOMPtr.h"
 #include "nsContentUtils.h"
+#include "nsDebug.h"
+#include "nsError.h"
+#include "nsITimer.h"
+#include "nsString.h"
+#include "nsStringFlags.h"
+#include "nsStringFwd.h"
+#include "nsTArray.h"
+#include "nsTStringRepr.h"
+#include "nscore.h"
 
-namespace mozilla {
-namespace dom {
-
-namespace {
-
-const uint32_t kSnapshotTimeoutMs = 20000;
-
-}  // namespace
+namespace mozilla::dom {
 
 /**
  * Coalescing manipulation queue used by `LSSnapshot`.  Used by `LSSnapshot` to
@@ -40,7 +66,7 @@ void SnapshotWriteOptimizer::Enumerate(nsTArray<LSWriteInfo>& aWriteInfos) {
   // an arbitrary order, which means write infos need to be sorted before being
   // processed.
 
-  nsTArray<WriteInfo*> writeInfos;
+  nsTArray<NotNull<WriteInfo*>> writeInfos;
   GetSortedWriteInfos(writeInfos);
 
   for (WriteInfo* writeInfo : writeInfos) {
@@ -109,13 +135,14 @@ LSSnapshot::LSSnapshot(LSDatabase* aDatabase)
       mActor(nullptr),
       mInitLength(0),
       mLength(0),
-      mExactUsage(0),
+      mUsage(0),
       mPeakUsage(0),
       mLoadState(LoadState::Initial),
+      mHasOtherProcessDatabases(false),
       mHasOtherProcessObservers(false),
       mExplicit(false),
       mHasPendingStableStateCallback(false),
-      mHasPendingTimerCallback(false),
+      mHasPendingIdleTimerCallback(false),
       mDirty(false)
 #ifdef DEBUG
       ,
@@ -130,7 +157,7 @@ LSSnapshot::~LSSnapshot() {
   AssertIsOnOwningThread();
   MOZ_ASSERT(mDatabase);
   MOZ_ASSERT(!mHasPendingStableStateCallback);
-  MOZ_ASSERT(!mHasPendingTimerCallback);
+  MOZ_ASSERT(!mHasPendingIdleTimerCallback);
   MOZ_ASSERT_IF(mInitialized, mSentFinish);
 
   if (mActor) {
@@ -167,15 +194,15 @@ nsresult LSSnapshot::Init(const nsAString& aKey,
     const LSValue& value = itemInfo.value();
 
     if (loadState != LoadState::AllOrderedItems && !value.IsVoid()) {
-      mLoadedItems.PutEntry(itemInfo.key());
+      mLoadedItems.Insert(itemInfo.key());
     }
 
-    mValues.Put(itemInfo.key(), value.AsString());
+    mValues.InsertOrUpdate(itemInfo.key(), value.AsString());
   }
 
   if (loadState == LoadState::Partial) {
     if (aInitInfo.addKeyToUnknownItems()) {
-      mUnknownItems.PutEntry(aKey);
+      mUnknownItems.Insert(aKey);
     }
     mInitLength = aInitInfo.totalLength();
     mLength = mInitLength;
@@ -185,11 +212,12 @@ nsresult LSSnapshot::Init(const nsAString& aKey,
     MOZ_ASSERT(loadState == LoadState::AllOrderedItems);
   }
 
-  mExactUsage = aInitInfo.initialUsage();
+  mUsage = aInitInfo.usage();
   mPeakUsage = aInitInfo.peakUsage();
 
   mLoadState = aInitInfo.loadState();
 
+  mHasOtherProcessDatabases = aInitInfo.hasOtherProcessDatabases();
   mHasOtherProcessObservers = aInitInfo.hasOtherProcessObservers();
 
   mExplicit = aExplicit;
@@ -205,8 +233,8 @@ nsresult LSSnapshot::Init(const nsAString& aKey,
   }
 
   if (!mExplicit) {
-    mTimer = NS_NewTimer();
-    MOZ_ASSERT(mTimer);
+    mIdleTimer = NS_NewTimer();
+    MOZ_ASSERT(mIdleTimer);
 
     ScheduleStableStateCallback();
   }
@@ -287,9 +315,7 @@ nsresult LSSnapshot::GetKeys(nsTArray<nsString>& aKeys) {
     return rv;
   }
 
-  for (auto iter = mValues.ConstIter(); !iter.Done(); iter.Next()) {
-    aKeys.AppendElement(iter.Key());
-  }
+  AppendToArray(aKeys, mValues.Keys());
 
   return NS_OK;
 }
@@ -320,7 +346,7 @@ nsresult LSSnapshot::SetItem(const nsAString& aKey, const nsAString& aValue,
       if (oldValue.IsVoid()) {
         mValues.Remove(aKey);
       } else {
-        mValues.Put(aKey, oldValue);
+        mValues.InsertOrUpdate(aKey, oldValue);
       }
     });
 
@@ -409,7 +435,7 @@ nsresult LSSnapshot::RemoveItem(const nsAString& aKey,
 
     auto autoRevertValue = MakeScopeExit([&] {
       MOZ_ASSERT(!oldValue.IsVoid());
-      mValues.Put(aKey, oldValue);
+      mValues.InsertOrUpdate(aKey, oldValue);
     });
 
     // Anything that can fail must be done early before we start modifying the
@@ -486,7 +512,16 @@ nsresult LSSnapshot::Clear(LSNotifyInfo& aNotifyInfo) {
   } else {
     changed = true;
 
-    DebugOnly<nsresult> rv = UpdateUsage(-mExactUsage);
+    int64_t delta = 0;
+    for (const auto& entry : mValues) {
+      const nsAString& key = entry.GetKey();
+      const nsString& value = entry.GetData();
+
+      delta += -static_cast<int64_t>(key.Length()) -
+               static_cast<int64_t>(value.Length());
+    }
+
+    DebugOnly<nsresult> rv = UpdateUsage(delta);
     MOZ_ASSERT(NS_SUCCEEDED(rv));
 
     mValues.Clear();
@@ -522,22 +557,39 @@ void LSSnapshot::MarkDirty() {
   mDirty = true;
 
   if (!mExplicit && !mHasPendingStableStateCallback) {
-    CancelTimer();
+    CancelIdleTimer();
 
     MOZ_ALWAYS_SUCCEEDS(Checkpoint());
 
     MOZ_ALWAYS_SUCCEEDS(Finish());
   } else {
-    MOZ_ASSERT(!mHasPendingTimerCallback);
+    MOZ_ASSERT(!mHasPendingIdleTimerCallback);
   }
 }
 
-nsresult LSSnapshot::End() {
+nsresult LSSnapshot::ExplicitCheckpoint() {
   AssertIsOnOwningThread();
   MOZ_ASSERT(mActor);
   MOZ_ASSERT(mExplicit);
   MOZ_ASSERT(!mHasPendingStableStateCallback);
-  MOZ_ASSERT(!mHasPendingTimerCallback);
+  MOZ_ASSERT(!mHasPendingIdleTimerCallback);
+  MOZ_ASSERT(mInitialized);
+  MOZ_ASSERT(!mSentFinish);
+
+  nsresult rv = Checkpoint(/* aSync */ true);
+  if (NS_WARN_IF(NS_FAILED(rv))) {
+    return rv;
+  }
+
+  return NS_OK;
+}
+
+nsresult LSSnapshot::ExplicitEnd() {
+  AssertIsOnOwningThread();
+  MOZ_ASSERT(mActor);
+  MOZ_ASSERT(mExplicit);
+  MOZ_ASSERT(!mHasPendingStableStateCallback);
+  MOZ_ASSERT(!mHasPendingIdleTimerCallback);
   MOZ_ASSERT(mInitialized);
   MOZ_ASSERT(!mSentFinish);
 
@@ -548,25 +600,30 @@ nsresult LSSnapshot::End() {
 
   RefPtr<LSSnapshot> kungFuDeathGrip = this;
 
-  rv = Finish();
+  rv = Finish(/* aSync */ true);
   if (NS_WARN_IF(NS_FAILED(rv))) {
     return rv;
   }
 
-  if (NS_WARN_IF(!mActor->SendPing())) {
-    return NS_ERROR_FAILURE;
-  }
-
   return NS_OK;
 }
 
+int64_t LSSnapshot::GetUsage() const {
+  AssertIsOnOwningThread();
+  MOZ_ASSERT(mActor);
+  MOZ_ASSERT(mInitialized);
+  MOZ_ASSERT(!mSentFinish);
+
+  return mUsage;
+}
+
 void LSSnapshot::ScheduleStableStateCallback() {
   AssertIsOnOwningThread();
-  MOZ_ASSERT(mTimer);
+  MOZ_ASSERT(mIdleTimer);
   MOZ_ASSERT(!mExplicit);
   MOZ_ASSERT(!mHasPendingStableStateCallback);
 
-  CancelTimer();
+  CancelIdleTimer();
 
   nsCOMPtr<nsIRunnable> runnable = this;
   nsContentUtils::RunInStableState(runnable.forget());
@@ -580,7 +637,7 @@ void LSSnapshot::MaybeScheduleStableStateCallback() {
   if (!mExplicit && !mHasPendingStableStateCallback) {
     ScheduleStableStateCallback();
   } else {
-    MOZ_ASSERT(!mHasPendingTimerCallback);
+    MOZ_ASSERT(!mHasPendingIdleTimerCallback);
   }
 }
 
@@ -598,7 +655,7 @@ nsresult LSSnapshot::GetItemInternal(const nsAString& aKey,
     case LoadState::Partial: {
       if (mValues.Get(aKey, &result)) {
         MOZ_ASSERT(!result.IsVoid());
-      } else if (mLoadedItems.GetEntry(aKey) || mUnknownItems.GetEntry(aKey)) {
+      } else if (mLoadedItems.Contains(aKey) || mUnknownItems.Contains(aKey)) {
         result.SetIsVoid(true);
       } else {
         LSValue value;
@@ -611,10 +668,10 @@ nsresult LSSnapshot::GetItemInternal(const nsAString& aKey,
         result = value.AsString();
 
         if (result.IsVoid()) {
-          mUnknownItems.PutEntry(aKey);
+          mUnknownItems.Insert(aKey);
         } else {
-          mLoadedItems.PutEntry(aKey);
-          mValues.Put(aKey, result);
+          mLoadedItems.Insert(aKey);
+          mValues.InsertOrUpdate(aKey, result);
 
           // mLoadedItems.Count()==mInitLength is checked below.
         }
@@ -622,8 +679,8 @@ nsresult LSSnapshot::GetItemInternal(const nsAString& aKey,
         for (uint32_t i = 0; i < itemInfos.Length(); i++) {
           const LSItemInfo& itemInfo = itemInfos[i];
 
-          mLoadedItems.PutEntry(itemInfo.key());
-          mValues.Put(itemInfo.key(), itemInfo.value().AsString());
+          mLoadedItems.Insert(itemInfo.key());
+          mValues.InsertOrUpdate(itemInfo.key(), itemInfo.value().AsString());
         }
 
         if (mLoadedItems.Count() == mInitLength) {
@@ -637,7 +694,7 @@ nsresult LSSnapshot::GetItemInternal(const nsAString& aKey,
       if (aValue.WasPassed()) {
         const nsString& value = aValue.Value();
         if (!value.IsVoid()) {
-          mValues.Put(aKey, value);
+          mValues.InsertOrUpdate(aKey, value);
         } else if (!result.IsVoid()) {
           mValues.Remove(aKey);
         }
@@ -660,16 +717,16 @@ nsresult LSSnapshot::GetItemInternal(const nsAString& aKey,
 
           MOZ_ASSERT(!result.IsVoid());
 
-          mLoadedItems.PutEntry(aKey);
-          mValues.Put(aKey, result);
+          mLoadedItems.Insert(aKey);
+          mValues.InsertOrUpdate(aKey, result);
 
           // mLoadedItems.Count()==mInitLength is checked below.
 
           for (uint32_t i = 0; i < itemInfos.Length(); i++) {
             const LSItemInfo& itemInfo = itemInfos[i];
 
-            mLoadedItems.PutEntry(itemInfo.key());
-            mValues.Put(itemInfo.key(), itemInfo.value().AsString());
+            mLoadedItems.Insert(itemInfo.key());
+            mValues.InsertOrUpdate(itemInfo.key(), itemInfo.value().AsString());
           }
 
           if (mLoadedItems.Count() == mInitLength) {
@@ -685,7 +742,7 @@ nsresult LSSnapshot::GetItemInternal(const nsAString& aKey,
       if (aValue.WasPassed()) {
         const nsString& value = aValue.Value();
         if (!value.IsVoid()) {
-          mValues.Put(aKey, value);
+          mValues.InsertOrUpdate(aKey, value);
         } else if (!result.IsVoid()) {
           mValues.Remove(aKey);
         }
@@ -699,14 +756,14 @@ nsresult LSSnapshot::GetItemInternal(const nsAString& aKey,
       if (aValue.WasPassed()) {
         const nsString& value = aValue.Value();
         if (!value.IsVoid()) {
-          auto entry = mValues.LookupForAdd(aKey);
-          if (entry) {
-            result = entry.Data();
-            entry.Data() = value;
-          } else {
-            result.SetIsVoid(true);
-            entry.OrInsert([value]() { return value; });
-          }
+          mValues.WithEntryHandle(aKey, [&](auto&& entry) {
+            if (entry) {
+              result = std::exchange(entry.Data(), value);
+            } else {
+              result.SetIsVoid(true);
+              entry.Insert(value);
+            }
+          });
         } else {
           if (auto entry = mValues.Lookup(aKey)) {
             result = entry.Data();
@@ -752,10 +809,10 @@ nsresult LSSnapshot::EnsureAllKeys() {
     return NS_ERROR_FAILURE;
   }
 
-  nsDataHashtable<nsStringHashKey, nsString> newValues;
+  nsTHashMap<nsStringHashKey, nsString> newValues;
 
   for (auto key : keys) {
-    newValues.Put(key, VoidString());
+    newValues.InsertOrUpdate(key, VoidString());
   }
 
   if (mHasOtherProcessObservers) {
@@ -769,8 +826,9 @@ nsresult LSSnapshot::EnsureAllKeys() {
 
         switch (writeAndNotifyInfo.type()) {
           case LSWriteAndNotifyInfo::TLSSetItemAndNotifyInfo: {
-            newValues.Put(writeAndNotifyInfo.get_LSSetItemAndNotifyInfo().key(),
-                          VoidString());
+            newValues.InsertOrUpdate(
+                writeAndNotifyInfo.get_LSSetItemAndNotifyInfo().key(),
+                VoidString());
             break;
           }
           case LSWriteAndNotifyInfo::TLSRemoveItemAndNotifyInfo: {
@@ -802,7 +860,8 @@ nsresult LSSnapshot::EnsureAllKeys() {
 
         switch (writeInfo.type()) {
           case LSWriteInfo::TLSSetItemInfo: {
-            newValues.Put(writeInfo.get_LSSetItemInfo().key(), VoidString());
+            newValues.InsertOrUpdate(writeInfo.get_LSSetItemInfo().key(),
+                                     VoidString());
             break;
           }
           case LSWriteInfo::TLSRemoveItemInfo: {
@@ -852,17 +911,16 @@ nsresult LSSnapshot::UpdateUsage(int64_t aDelta) {
   AssertIsOnOwningThread();
   MOZ_ASSERT(mDatabase);
   MOZ_ASSERT(mActor);
-  MOZ_ASSERT(mPeakUsage >= mExactUsage);
+  MOZ_ASSERT(mPeakUsage >= mUsage);
   MOZ_ASSERT(mInitialized);
   MOZ_ASSERT(!mSentFinish);
 
-  int64_t newExactUsage = mExactUsage + aDelta;
-  if (newExactUsage > mPeakUsage) {
-    int64_t minSize = newExactUsage - mPeakUsage;
-    int64_t requestedSize = minSize + 4096;
+  int64_t newUsage = mUsage + aDelta;
+  if (newUsage > mPeakUsage) {
+    const int64_t minSize = newUsage - mPeakUsage;
+
     int64_t size;
-    if (NS_WARN_IF(
-            !mActor->SendIncreasePeakUsage(requestedSize, minSize, &size))) {
+    if (NS_WARN_IF(!mActor->SendIncreasePeakUsage(minSize, &size))) {
       return NS_ERROR_FAILURE;
     }
 
@@ -875,11 +933,11 @@ nsresult LSSnapshot::UpdateUsage(int64_t aDelta) {
     mPeakUsage += size;
   }
 
-  mExactUsage = newExactUsage;
+  mUsage = newUsage;
   return NS_OK;
 }
 
-nsresult LSSnapshot::Checkpoint() {
+nsresult LSSnapshot::Checkpoint(bool aSync) {
   AssertIsOnOwningThread();
   MOZ_ASSERT(mActor);
   MOZ_ASSERT(mInitialized);
@@ -889,7 +947,13 @@ nsresult LSSnapshot::Checkpoint() {
     MOZ_ASSERT(mWriteAndNotifyInfos);
 
     if (!mWriteAndNotifyInfos->IsEmpty()) {
-      MOZ_ALWAYS_TRUE(mActor->SendCheckpointAndNotify(*mWriteAndNotifyInfos));
+      if (aSync) {
+        MOZ_ALWAYS_TRUE(
+            mActor->SendSyncCheckpointAndNotify(*mWriteAndNotifyInfos));
+      } else {
+        MOZ_ALWAYS_TRUE(
+            mActor->SendAsyncCheckpointAndNotify(*mWriteAndNotifyInfos));
+      }
 
       mWriteAndNotifyInfos->Clear();
     }
@@ -902,7 +966,11 @@ nsresult LSSnapshot::Checkpoint() {
 
       MOZ_ASSERT(!writeInfos.IsEmpty());
 
-      MOZ_ALWAYS_TRUE(mActor->SendCheckpoint(writeInfos));
+      if (aSync) {
+        MOZ_ALWAYS_TRUE(mActor->SendSyncCheckpoint(writeInfos));
+      } else {
+        MOZ_ALWAYS_TRUE(mActor->SendAsyncCheckpoint(writeInfos));
+      }
 
       mWriteOptimizer->Reset();
     }
@@ -911,14 +979,18 @@ nsresult LSSnapshot::Checkpoint() {
   return NS_OK;
 }
 
-nsresult LSSnapshot::Finish() {
+nsresult LSSnapshot::Finish(bool aSync) {
   AssertIsOnOwningThread();
   MOZ_ASSERT(mDatabase);
   MOZ_ASSERT(mActor);
   MOZ_ASSERT(mInitialized);
   MOZ_ASSERT(!mSentFinish);
 
-  MOZ_ALWAYS_TRUE(mActor->SendFinish());
+  if (aSync) {
+    MOZ_ALWAYS_TRUE(mActor->SendSyncFinish());
+  } else {
+    MOZ_ALWAYS_TRUE(mActor->SendAsyncFinish());
+  }
 
   mDatabase->NoteFinishedSnapshot(this);
 
@@ -933,28 +1005,28 @@ nsresult LSSnapshot::Finish() {
   return NS_OK;
 }
 
-void LSSnapshot::CancelTimer() {
+void LSSnapshot::CancelIdleTimer() {
   AssertIsOnOwningThread();
-  MOZ_ASSERT(mTimer);
+  MOZ_ASSERT(mIdleTimer);
 
-  if (mHasPendingTimerCallback) {
-    MOZ_ALWAYS_SUCCEEDS(mTimer->Cancel());
-    mHasPendingTimerCallback = false;
+  if (mHasPendingIdleTimerCallback) {
+    MOZ_ALWAYS_SUCCEEDS(mIdleTimer->Cancel());
+    mHasPendingIdleTimerCallback = false;
   }
 }
 
 // static
-void LSSnapshot::TimerCallback(nsITimer* aTimer, void* aClosure) {
+void LSSnapshot::IdleTimerCallback(nsITimer* aTimer, void* aClosure) {
   MOZ_ASSERT(aTimer);
 
   auto* self = static_cast<LSSnapshot*>(aClosure);
   MOZ_ASSERT(self);
-  MOZ_ASSERT(self->mTimer);
-  MOZ_ASSERT(SameCOMIdentity(self->mTimer, aTimer));
+  MOZ_ASSERT(self->mIdleTimer);
+  MOZ_ASSERT(SameCOMIdentity(self->mIdleTimer, aTimer));
   MOZ_ASSERT(!self->mHasPendingStableStateCallback);
-  MOZ_ASSERT(self->mHasPendingTimerCallback);
+  MOZ_ASSERT(self->mHasPendingIdleTimerCallback);
 
-  self->mHasPendingTimerCallback = false;
+  self->mHasPendingIdleTimerCallback = false;
 
   MOZ_ALWAYS_SUCCEEDS(self->Finish());
 }
@@ -966,26 +1038,41 @@ LSSnapshot::Run() {
   AssertIsOnOwningThread();
   MOZ_ASSERT(!mExplicit);
   MOZ_ASSERT(mHasPendingStableStateCallback);
-  MOZ_ASSERT(!mHasPendingTimerCallback);
+  MOZ_ASSERT(!mHasPendingIdleTimerCallback);
 
   mHasPendingStableStateCallback = false;
 
   MOZ_ALWAYS_SUCCEEDS(Checkpoint());
 
-  if (mDirty || !Preferences::GetBool("dom.storage.snapshot_reusing")) {
+  // 1. The unused pre-incremented snapshot peak usage can't be undone when
+  //    there are other snapshots for the same database. We only add a pending
+  //    usage delta when a snapshot finishes and usage deltas are then applied
+  //    when the last database becomes inactive.
+  // 2. If there's a snapshot with pre-incremented peak usage, the next
+  //    snapshot will use that as a base for its usage.
+  // 3. When a task for given snapshot finishes, we try to reuse the snapshot
+  //    by only checkpointing the snapshot and delaying the finish by a timer.
+  // 4. If two or more tabs for the same origin use localStorage periodically
+  //    at the same time the usage gradually grows until it hits the quota
+  //    limit.
+  // 5. We prevent that from happening by finishing the snapshot immediatelly
+  //    if there are databases in other processess.
+
+  if (mDirty || mHasOtherProcessDatabases ||
+      !Preferences::GetBool("dom.storage.snapshot_reusing")) {
     MOZ_ALWAYS_SUCCEEDS(Finish());
-  } else if (!mExplicit) {
-    MOZ_ASSERT(mTimer);
+  } else {
+    MOZ_ASSERT(mIdleTimer);
 
-    MOZ_ALWAYS_SUCCEEDS(mTimer->InitWithNamedFuncCallback(
-        TimerCallback, this, kSnapshotTimeoutMs, nsITimer::TYPE_ONE_SHOT,
-        "LSSnapshot::TimerCallback"));
+    MOZ_ALWAYS_SUCCEEDS(mIdleTimer->InitWithNamedFuncCallback(
+        IdleTimerCallback, this,
+        StaticPrefs::dom_storage_snapshot_idle_timeout_ms(),
+        nsITimer::TYPE_ONE_SHOT, "LSSnapshot::IdleTimerCallback"));
 
-    mHasPendingTimerCallback = true;
+    mHasPendingIdleTimerCallback = true;
   }
 
   return NS_OK;
 }
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
