@@ -7,12 +7,11 @@
 #include "mozilla/dom/AudioTrackList.h"
 #include "mozilla/dom/AudioTrackListBinding.h"
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 JSObject* AudioTrackList::WrapObject(JSContext* aCx,
                                      JS::Handle<JSObject*> aGivenProto) {
-  return AudioTrackListBinding::Wrap(aCx, this, aGivenProto);
+  return AudioTrackList_Binding::Wrap(aCx, this, aGivenProto);
 }
 
 AudioTrack* AudioTrackList::operator[](uint32_t aIndex) {
@@ -30,5 +29,4 @@ AudioTrack* AudioTrackList::GetTrackById(const nsAString& aId) {
   return track ? track->AsAudioTrack() : nullptr;
 }
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
