@@ -6,13 +6,10 @@
 
 #include "MediaSource.h"
 
-#if MOZ_AV1
-#include "AOMDecoder.h"
-#endif
 #include "AsyncEventRunner.h"
-#include "DecoderTraits.h"
 #include "Benchmark.h"
 #include "DecoderDoctorDiagnostics.h"
+#include "DecoderTraits.h"
 #include "MediaContainerType.h"
 #include "MediaResult.h"
 #include "MediaSourceDemuxer.h"
@@ -21,25 +18,26 @@
 #include "SourceBufferList.h"
 #include "mozilla/ErrorResult.h"
 #include "mozilla/FloatingPoint.h"
-#include "mozilla/Preferences.h"
+#include "mozilla/Logging.h"
+#include "mozilla/Sprintf.h"
+#include "mozilla/StaticPrefs_media.h"
 #include "mozilla/dom/BindingDeclarations.h"
 #include "mozilla/dom/HTMLMediaElement.h"
+#include "mozilla/gfx/gfxVars.h"
 #include "mozilla/mozalloc.h"
 #include "nsDebug.h"
 #include "nsError.h"
 #include "nsIRunnable.h"
 #include "nsIScriptObjectPrincipal.h"
-#include "nsPIDOMWindow.h"
 #include "nsMimeTypes.h"
+#include "nsPIDOMWindow.h"
+#include "nsServiceManagerUtils.h"
 #include "nsString.h"
 #include "nsThreadUtils.h"
-#include "mozilla/Logging.h"
-#include "nsServiceManagerUtils.h"
-#include "mozilla/gfx/gfxVars.h"
-#include "mozilla/Sprintf.h"
 
 #ifdef MOZ_WIDGET_ANDROID
-#include "AndroidBridge.h"
+#  include "AndroidBridge.h"
+#  include "mozilla/java/HardwareCodecCapabilityUtilsWrappers.h"
 #endif
 
 struct JSContext;
@@ -76,13 +74,13 @@ namespace mozilla {
 //     optional "Windows Media Feature Pack"
 // 2. If H264 hardware acceleration is not available.
 // 3. The CPU is considered to be fast enough
-static bool IsWebMForced(DecoderDoctorDiagnostics* aDiagnostics) {
+static bool IsVP9Forced(DecoderDoctorDiagnostics* aDiagnostics) {
   bool mp4supported = DecoderTraits::IsMP4SupportedType(
       MediaContainerType(MEDIAMIMETYPE(VIDEO_MP4)), aDiagnostics);
   bool hwsupported = gfx::gfxVars::CanUseHardwareVideoDecoding();
 #ifdef MOZ_WIDGET_ANDROID
   return !mp4supported || !hwsupported || VP9Benchmark::IsVP9DecodeFast() ||
-         java::HardwareCodecCapabilityUtils::HasHWVP9();
+         java::HardwareCodecCapabilityUtils::HasHWVP9(false /* aIsEncoder */);
 #else
   return !mp4supported || !hwsupported || VP9Benchmark::IsVP9DecodeFast();
 #endif
@@ -90,21 +88,66 @@ static bool IsWebMForced(DecoderDoctorDiagnostics* aDiagnostics) {
 
 namespace dom {
 
+static void RecordTypeForTelemetry(const nsAString& aType,
+                                   nsPIDOMWindowInner* aWindow) {
+  Maybe<MediaContainerType> containerType = MakeMediaContainerType(aType);
+  if (!containerType) {
+    return;
+  }
+
+  const MediaMIMEType& mimeType = containerType->Type();
+  if (mimeType == MEDIAMIMETYPE(VIDEO_WEBM)) {
+    AccumulateCategorical(
+        mozilla::Telemetry::LABELS_MSE_SOURCE_BUFFER_TYPE::VideoWebm);
+  } else if (mimeType == MEDIAMIMETYPE(AUDIO_WEBM)) {
+    AccumulateCategorical(
+        mozilla::Telemetry::LABELS_MSE_SOURCE_BUFFER_TYPE::AudioWebm);
+  } else if (mimeType == MEDIAMIMETYPE(VIDEO_MP4)) {
+    AccumulateCategorical(
+        mozilla::Telemetry::LABELS_MSE_SOURCE_BUFFER_TYPE::VideoMp4);
+  } else if (mimeType == MEDIAMIMETYPE(AUDIO_MP4)) {
+    AccumulateCategorical(
+        mozilla::Telemetry::LABELS_MSE_SOURCE_BUFFER_TYPE::AudioMp4);
+  } else if (mimeType == MEDIAMIMETYPE(VIDEO_MPEG_TS)) {
+    AccumulateCategorical(
+        mozilla::Telemetry::LABELS_MSE_SOURCE_BUFFER_TYPE::VideoMp2t);
+  } else if (mimeType == MEDIAMIMETYPE(AUDIO_MPEG_TS)) {
+    AccumulateCategorical(
+        mozilla::Telemetry::LABELS_MSE_SOURCE_BUFFER_TYPE::AudioMp2t);
+  } else if (mimeType == MEDIAMIMETYPE(AUDIO_MP3)) {
+    AccumulateCategorical(
+        mozilla::Telemetry::LABELS_MSE_SOURCE_BUFFER_TYPE::AudioMpeg);
+  } else if (mimeType == MEDIAMIMETYPE(AUDIO_AAC)) {
+    AccumulateCategorical(
+        mozilla::Telemetry::LABELS_MSE_SOURCE_BUFFER_TYPE::AudioAac);
+  }
+}
+
 /* static */
-nsresult MediaSource::IsTypeSupported(const nsAString& aType,
-                                      DecoderDoctorDiagnostics* aDiagnostics) {
+void MediaSource::IsTypeSupported(const nsAString& aType,
+                                  DecoderDoctorDiagnostics* aDiagnostics,
+                                  ErrorResult& aRv) {
   if (aType.IsEmpty()) {
-    return NS_ERROR_DOM_TYPE_ERR;
+    return aRv.ThrowTypeError("Empty type");
   }
 
   Maybe<MediaContainerType> containerType = MakeMediaContainerType(aType);
   if (!containerType) {
-    return NS_ERROR_DOM_NOT_SUPPORTED_ERR;
+    return aRv.ThrowNotSupportedError("Unknown type");
   }
 
   if (DecoderTraits::CanHandleContainerType(*containerType, aDiagnostics) ==
       CANPLAY_NO) {
-    return NS_ERROR_DOM_NOT_SUPPORTED_ERR;
+    return aRv.ThrowNotSupportedError("Can't play type");
+  }
+
+  bool hasVP9 = false;
+  const MediaCodecs& codecs = containerType->ExtendedType().Codecs();
+  for (const auto& codec : codecs.Range()) {
+    if (IsVP9CodecString(codec)) {
+      hasVP9 = true;
+      break;
+    }
   }
 
   // Now we know that this media type could be played.
@@ -112,38 +155,49 @@ nsresult MediaSource::IsTypeSupported(const nsAString& aType,
   const MediaMIMEType& mimeType = containerType->Type();
   if (mimeType == MEDIAMIMETYPE("video/mp4") ||
       mimeType == MEDIAMIMETYPE("audio/mp4")) {
-    if (!Preferences::GetBool("media.mediasource.mp4.enabled", false)) {
-      return NS_ERROR_DOM_NOT_SUPPORTED_ERR;
+    if (!StaticPrefs::media_mediasource_mp4_enabled()) {
+      // Don't leak information about the fact that it's pref-disabled; just act
+      // like we can't play it.  Or should this throw "Unknown type"?
+      return aRv.ThrowNotSupportedError("Can't play type");
     }
-    return NS_OK;
+    if (!StaticPrefs::media_mediasource_vp9_enabled() && hasVP9 &&
+        !IsVP9Forced(aDiagnostics)) {
+      // Don't leak information about the fact that it's pref-disabled; just act
+      // like we can't play it.  Or should this throw "Unknown type"?
+      return aRv.ThrowNotSupportedError("Can't play type");
+    }
+
+    return;
   }
   if (mimeType == MEDIAMIMETYPE("video/webm")) {
-    if (!(Preferences::GetBool("media.mediasource.webm.enabled", false) ||
-          containerType->ExtendedType().Codecs().Contains(
-              NS_LITERAL_STRING("vp8")) ||
-#ifdef MOZ_AV1
-          // FIXME: Temporary comparison with the full codecs attribute.
-          // See bug 1377015.
-          AOMDecoder::IsSupportedCodec(
-              containerType->ExtendedType().Codecs().AsString()) ||
-#endif
-          IsWebMForced(aDiagnostics))) {
-      return NS_ERROR_DOM_NOT_SUPPORTED_ERR;
+    if (!StaticPrefs::media_mediasource_webm_enabled()) {
+      // Don't leak information about the fact that it's pref-disabled; just act
+      // like we can't play it.  Or should this throw "Unknown type"?
+      return aRv.ThrowNotSupportedError("Can't play type");
     }
-    return NS_OK;
+    if (!StaticPrefs::media_mediasource_vp9_enabled() && hasVP9 &&
+        !IsVP9Forced(aDiagnostics)) {
+      // Don't leak information about the fact that it's pref-disabled; just act
+      // like we can't play it.  Or should this throw "Unknown type"?
+      return aRv.ThrowNotSupportedError("Can't play type");
+    }
+    return;
   }
   if (mimeType == MEDIAMIMETYPE("audio/webm")) {
-    if (!(Preferences::GetBool("media.mediasource.webm.enabled", false) ||
-          Preferences::GetBool("media.mediasource.webm.audio.enabled", true))) {
-      return NS_ERROR_DOM_NOT_SUPPORTED_ERR;
+    if (!(StaticPrefs::media_mediasource_webm_enabled() ||
+          StaticPrefs::media_mediasource_webm_audio_enabled())) {
+      // Don't leak information about the fact that it's pref-disabled; just act
+      // like we can't play it.  Or should this throw "Unknown type"?
+      return aRv.ThrowNotSupportedError("Can't play type");
     }
-    return NS_OK;
+    return;
   }
 
-  return NS_ERROR_DOM_NOT_SUPPORTED_ERR;
+  return aRv.ThrowNotSupportedError("Type not supported in MediaSource");
 }
 
-/* static */ already_AddRefed<MediaSource> MediaSource::Constructor(
+/* static */
+already_AddRefed<MediaSource> MediaSource::Constructor(
     const GlobalObject& aGlobal, ErrorResult& aRv) {
   nsCOMPtr<nsPIDOMWindowInner> window =
       do_QueryInterface(aGlobal.GetAsSupports());
@@ -194,22 +248,26 @@ double MediaSource::Duration() {
 
 void MediaSource::SetDuration(double aDuration, ErrorResult& aRv) {
   MOZ_ASSERT(NS_IsMainThread());
-  MSE_API("SetDuration(aDuration=%f, ErrorResult)", aDuration);
-  if (aDuration < 0 || IsNaN(aDuration)) {
-    aRv.Throw(NS_ERROR_DOM_TYPE_ERR);
+  if (aDuration < 0 || std::isnan(aDuration)) {
+    nsPrintfCString error("Invalid duration value %f", aDuration);
+    MSE_API("SetDuration(aDuration=%f, invalid value)", aDuration);
+    aRv.ThrowTypeError(error);
     return;
   }
   if (mReadyState != MediaSourceReadyState::Open ||
       mSourceBuffers->AnyUpdating()) {
+    MSE_API("SetDuration(aDuration=%f, invalid state)", aDuration);
     aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
     return;
   }
   DurationChange(aDuration, aRv);
+  MSE_API("SetDuration(aDuration=%f, errorCode=%d)", aDuration,
+          aRv.ErrorCodeAsInt());
 }
 
-void MediaSource::SetDuration(double aDuration) {
+void MediaSource::SetDuration(const media::TimeUnit& aDuration) {
   MOZ_ASSERT(NS_IsMainThread());
-  MSE_API("SetDuration(aDuration=%f)", aDuration);
+  MSE_API("SetDuration(aDuration=%f)", aDuration.ToSeconds());
   mDecoder->SetMediaSourceDuration(aDuration);
 }
 
@@ -217,18 +275,19 @@ already_AddRefed<SourceBuffer> MediaSource::AddSourceBuffer(
     const nsAString& aType, ErrorResult& aRv) {
   MOZ_ASSERT(NS_IsMainThread());
   DecoderDoctorDiagnostics diagnostics;
-  nsresult rv = IsTypeSupported(aType, &diagnostics);
+  IsTypeSupported(aType, &diagnostics, aRv);
+  RecordTypeForTelemetry(aType, GetOwner());
+  bool supported = !aRv.Failed();
   diagnostics.StoreFormatDiagnostics(
-      GetOwner() ? GetOwner()->GetExtantDoc() : nullptr, aType,
-      NS_SUCCEEDED(rv), __func__);
+      GetOwner() ? GetOwner()->GetExtantDoc() : nullptr, aType, supported,
+      __func__);
   MSE_API("AddSourceBuffer(aType=%s)%s", NS_ConvertUTF16toUTF8(aType).get(),
-          rv == NS_OK ? "" : " [not supported]");
-  if (NS_FAILED(rv)) {
-    aRv.Throw(rv);
+          supported ? "" : " [not supported]");
+  if (!supported) {
     return nullptr;
   }
   if (mSourceBuffers->Length() >= MAX_SOURCE_BUFFERS) {
-    aRv.Throw(NS_ERROR_DOM_QUOTA_EXCEEDED_ERR);
+    aRv.Throw(NS_ERROR_DOM_MEDIA_SOURCE_MAX_BUFFER_QUOTA_EXCEEDED_ERR);
     return nullptr;
   }
   if (mReadyState != MediaSourceReadyState::Open) {
@@ -241,10 +300,6 @@ already_AddRefed<SourceBuffer> MediaSource::AddSourceBuffer(
     return nullptr;
   }
   RefPtr<SourceBuffer> sourceBuffer = new SourceBuffer(this, *containerType);
-  if (!sourceBuffer) {
-    aRv.Throw(NS_ERROR_FAILURE);  // XXX need a better error here
-    return nullptr;
-  }
   mSourceBuffers->Append(sourceBuffer);
   DDLINKCHILD("sourcebuffer[]", sourceBuffer.get());
   MSE_DEBUG("sourceBuffer=%p", sourceBuffer.get());
@@ -280,7 +335,7 @@ RefPtr<MediaSource::ActiveCompletionPromise> MediaSource::SourceBufferIsActive(
   // It will be resolved once the HTMLMediaElement modifies its readyState.
   MozPromiseHolder<ActiveCompletionPromise> holder;
   RefPtr<ActiveCompletionPromise> promise = holder.Ensure(__func__);
-  mCompletionPromises.AppendElement(Move(holder));
+  mCompletionPromises.AppendElement(std::move(holder));
   return promise;
 }
 
@@ -336,7 +391,8 @@ void MediaSource::EndOfStream(
   SetReadyState(MediaSourceReadyState::Ended);
   mSourceBuffers->Ended();
   if (!aError.WasPassed()) {
-    DurationChange(mSourceBuffers->GetHighestBufferedEndTime(), aRv);
+    DurationChange(mSourceBuffers->GetHighestBufferedEndTime().ToBase(1000000),
+                   aRv);
     // Notify reader that all data is now available.
     mDecoder->Ended(true);
     return;
@@ -349,7 +405,10 @@ void MediaSource::EndOfStream(
       mDecoder->DecodeError(NS_ERROR_DOM_MEDIA_FATAL_ERR);
       break;
     default:
-      aRv.Throw(NS_ERROR_DOM_TYPE_ERR);
+      MOZ_ASSERT_UNREACHABLE(
+          "Someone added a MediaSourceReadyState value and didn't handle it "
+          "here");
+      break;
   }
 }
 
@@ -362,24 +421,24 @@ void MediaSource::EndOfStream(const MediaResult& aError) {
   mDecoder->DecodeError(aError);
 }
 
-/* static */ bool MediaSource::IsTypeSupported(const GlobalObject& aOwner,
-                                               const nsAString& aType) {
+/* static */
+bool MediaSource::IsTypeSupported(const GlobalObject& aOwner,
+                                  const nsAString& aType) {
   MOZ_ASSERT(NS_IsMainThread());
   DecoderDoctorDiagnostics diagnostics;
-  nsresult rv = IsTypeSupported(aType, &diagnostics);
+  IgnoredErrorResult rv;
+  IsTypeSupported(aType, &diagnostics, rv);
+  bool supported = !rv.Failed();
   nsCOMPtr<nsPIDOMWindowInner> window =
       do_QueryInterface(aOwner.GetAsSupports());
+  RecordTypeForTelemetry(aType, window);
   diagnostics.StoreFormatDiagnostics(window ? window->GetExtantDoc() : nullptr,
-                                     aType, NS_SUCCEEDED(rv), __func__);
+                                     aType, supported, __func__);
   MOZ_LOG(GetMediaSourceAPILog(), mozilla::LogLevel::Debug,
           ("MediaSource::%s: IsTypeSupported(aType=%s) %s", __func__,
            NS_ConvertUTF16toUTF8(aType).get(),
-           rv == NS_OK ? "OK" : "[not supported]"));
-  return NS_SUCCEEDED(rv);
-}
-
-/* static */ bool MediaSource::Enabled(JSContext* cx, JSObject* aGlobal) {
-  return Preferences::GetBool("media.mediasource.enabled");
+           supported ? "OK" : "[not supported]"));
+  return supported;
 }
 
 void MediaSource::SetLiveSeekableRange(double aStart, double aEnd,
@@ -396,16 +455,14 @@ void MediaSource::SetLiveSeekableRange(double aStart, double aEnd,
   // 2. If start is negative or greater than end, then throw a TypeError
   // exception and abort these steps.
   if (aStart < 0 || aStart > aEnd) {
-    aRv.Throw(NS_ERROR_DOM_TYPE_ERR);
+    aRv.ThrowTypeError("Invalid start value");
     return;
   }
 
   // 3. Set live seekable range to be a new normalized TimeRanges object
   // containing a single range whose start position is start and end position is
   // end.
-  mLiveSeekableRange =
-      Some(media::TimeInterval(media::TimeUnit::FromSeconds(aStart),
-                               media::TimeUnit::FromSeconds(aEnd)));
+  mLiveSeekableRange = Some(media::TimeRanges(media::TimeRange(aStart, aEnd)));
 }
 
 void MediaSource::ClearLiveSeekableRange(ErrorResult& aRv) {
@@ -530,12 +587,13 @@ void MediaSource::QueueAsyncSimpleEvent(const char* aName) {
   mAbstractMainThread->Dispatch(event.forget());
 }
 
-void MediaSource::DurationChange(double aNewDuration, ErrorResult& aRv) {
+void MediaSource::DurationChange(const media::TimeUnit& aNewDuration,
+                                 ErrorResult& aRv) {
   MOZ_ASSERT(NS_IsMainThread());
-  MSE_DEBUG("DurationChange(aNewDuration=%f)", aNewDuration);
+  MSE_DEBUG("DurationChange(aNewDuration=%s)", aNewDuration.ToString().get());
 
   // 1. If the current value of duration is equal to new duration, then return.
-  if (mDecoder->GetDuration() == aNewDuration) {
+  if (mDecoder->GetDuration() == aNewDuration.ToSeconds()) {
     return;
   }
 
@@ -549,27 +607,76 @@ void MediaSource::DurationChange(double aNewDuration, ErrorResult& aRv) {
 
   // 3. Let highest end time be the largest track buffer ranges end time across
   // all the track buffers across all SourceBuffer objects in sourceBuffers.
-  double highestEndTime = mSourceBuffers->HighestEndTime();
+  media::TimeUnit highestEndTime = mSourceBuffers->HighestEndTime();
   // 4. If new duration is less than highest end time, then
   //    4.1 Update new duration to equal highest end time.
-  aNewDuration = std::max(aNewDuration, highestEndTime);
+  media::TimeUnit newDuration = std::max(aNewDuration, highestEndTime);
 
   // 5. Update the media duration to new duration and run the HTMLMediaElement
   // duration change algorithm.
-  mDecoder->SetMediaSourceDuration(aNewDuration);
+  mDecoder->SetMediaSourceDuration(newDuration);
 }
 
-void MediaSource::GetMozDebugReaderData(nsAString& aString) {
-  nsAutoCString result;
-  mDecoder->GetMozDebugReaderData(result);
-  aString = NS_ConvertUTF8toUTF16(result);
+void MediaSource::DurationChange(double aNewDuration, ErrorResult& aRv) {
+  MOZ_ASSERT(NS_IsMainThread());
+  MSE_DEBUG("DurationChange(aNewDuration=%f)", aNewDuration);
+
+  // 1. If the current value of duration is equal to new duration, then return.
+  if (mDecoder->GetDuration() == aNewDuration) {
+    return;
+  }
+
+  // 2. If new duration is less than the highest starting presentation timestamp
+  // of any buffered coded frames for all SourceBuffer objects in sourceBuffers,
+  // then throw an InvalidStateError exception and abort these steps.
+  if (aNewDuration < mSourceBuffers->HighestStartTime().ToSeconds()) {
+    aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
+    return;
+  }
+
+  // 3. Let highest end time be the largest track buffer ranges end time across
+  // all the track buffers across all SourceBuffer objects in sourceBuffers.
+  double highestEndTime = mSourceBuffers->HighestEndTime().ToSeconds();
+  // 4. If new duration is less than highest end time, then
+  //    4.1 Update new duration to equal highest end time.
+  double newDuration = std::max(aNewDuration, highestEndTime);
+
+  // 5. Update the media duration to new duration and run the HTMLMediaElement
+  // duration change algorithm.
+  mDecoder->SetMediaSourceDuration(newDuration);
+}
+
+already_AddRefed<Promise> MediaSource::MozDebugReaderData(ErrorResult& aRv) {
+  // Creating a JS promise
+  nsPIDOMWindowInner* win = GetOwner();
+  if (!win) {
+    aRv.Throw(NS_ERROR_UNEXPECTED);
+    return nullptr;
+  }
+  RefPtr<Promise> domPromise = Promise::Create(win->AsGlobal(), aRv);
+  if (NS_WARN_IF(aRv.Failed())) {
+    return nullptr;
+  }
+  MOZ_ASSERT(domPromise);
+  UniquePtr<MediaSourceDecoderDebugInfo> info =
+      MakeUnique<MediaSourceDecoderDebugInfo>();
+  mDecoder->RequestDebugInfo(*info)->Then(
+      mAbstractMainThread, __func__,
+      [domPromise, infoPtr = std::move(info)] {
+        domPromise->MaybeResolve(infoPtr.get());
+      },
+      [] {
+        MOZ_ASSERT_UNREACHABLE("Unexpected rejection while getting debug data");
+      });
+
+  return domPromise.forget();
 }
 
 nsPIDOMWindowInner* MediaSource::GetParentObject() const { return GetOwner(); }
 
 JSObject* MediaSource::WrapObject(JSContext* aCx,
                                   JS::Handle<JSObject*> aGivenProto) {
-  return MediaSourceBinding::Wrap(aCx, this, aGivenProto);
+  return MediaSource_Binding::Wrap(aCx, this, aGivenProto);
 }
 
 NS_IMPL_CYCLE_COLLECTION_INHERITED(MediaSource, DOMEventTargetHelper,
@@ -580,7 +687,7 @@ NS_IMPL_ADDREF_INHERITED(MediaSource, DOMEventTargetHelper)
 NS_IMPL_RELEASE_INHERITED(MediaSource, DOMEventTargetHelper)
 
 NS_INTERFACE_MAP_BEGIN_CYCLE_COLLECTION(MediaSource)
-  NS_INTERFACE_MAP_ENTRY(mozilla::dom::MediaSource)
+  NS_INTERFACE_MAP_ENTRY_CONCRETE(mozilla::dom::MediaSource)
 NS_INTERFACE_MAP_END_INHERITING(DOMEventTargetHelper)
 
 #undef MSE_DEBUG
