@@ -16,20 +16,12 @@
 #include "mozilla/dom/MediaSourceBinding.h"
 #include "mozilla/dom/TimeRanges.h"
 #include "nsError.h"
-#include "nsIEventTarget.h"
 #include "nsIRunnable.h"
 #include "nsThreadUtils.h"
 #include "mozilla/Logging.h"
 #include <time.h>
 #include "TimeUnits.h"
 
-// GetCurrentTime is defined in winbase.h as zero argument macro forwarding to
-// GetTickCount() and conflicts with MediaDecoder::GetCurrentTime
-// implementation.
-#ifdef GetCurrentTime
-#undef GetCurrentTime
-#endif
-
 struct JSContext;
 class JSObject;
 
@@ -63,7 +55,9 @@ void SourceBuffer::SetMode(SourceBufferAppendMode aMode, ErrorResult& aRv) {
   }
   if (mCurrentAttributes.mGenerateTimestamps &&
       aMode == SourceBufferAppendMode::Segments) {
-    aRv.Throw(NS_ERROR_DOM_TYPE_ERR);
+    aRv.ThrowTypeError(
+        "Can't set mode to \"segments\" when the byte stream generates "
+        "timestamps");
     return;
   }
   MOZ_ASSERT(mMediaSource->ReadyState() != MediaSourceReadyState::Closed);
@@ -108,6 +102,11 @@ void SourceBuffer::SetTimestampOffset(double aTimestampOffset,
   }
 }
 
+media::TimeIntervals SourceBuffer::GetBufferedIntervals() {
+  MOZ_ASSERT(mTrackBuffersManager);
+  return mTrackBuffersManager->Buffered();
+}
+
 TimeRanges* SourceBuffer::GetBuffered(ErrorResult& aRv) {
   MOZ_ASSERT(NS_IsMainThread());
   // http://w3c.github.io/media-source/index.html#widl-SourceBuffer-buffered
@@ -130,13 +129,15 @@ TimeRanges* SourceBuffer::GetBuffered(ErrorResult& aRv) {
   // as the current value of this attribute, then update the current value of
   // this attribute to intersection ranges.
   if (rangeChanged) {
-    mBuffered = new TimeRanges(ToSupports(this), intersection);
+    mBuffered = new TimeRanges(ToSupports(this),
+                               intersection.ToMicrosecondResolution());
   }
   // 6. Return the current value of this attribute.
   return mBuffered;
 }
 
 media::TimeIntervals SourceBuffer::GetTimeIntervals() {
+  MOZ_ASSERT(mTrackBuffersManager);
   return mTrackBuffersManager->Buffered();
 }
 
@@ -151,7 +152,7 @@ void SourceBuffer::SetAppendWindowStart(double aAppendWindowStart,
   }
   if (aAppendWindowStart < 0 ||
       aAppendWindowStart >= mCurrentAttributes.GetAppendWindowEnd()) {
-    aRv.Throw(NS_ERROR_DOM_TYPE_ERR);
+    aRv.ThrowTypeError("Invalid appendWindowStart value");
     return;
   }
   mCurrentAttributes.SetAppendWindowStart(aAppendWindowStart);
@@ -166,9 +167,9 @@ void SourceBuffer::SetAppendWindowEnd(double aAppendWindowEnd,
     aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
     return;
   }
-  if (IsNaN(aAppendWindowEnd) ||
+  if (std::isnan(aAppendWindowEnd) ||
       aAppendWindowEnd <= mCurrentAttributes.GetAppendWindowStart()) {
-    aRv.Throw(NS_ERROR_DOM_TYPE_ERR);
+    aRv.ThrowTypeError("Invalid appendWindowEnd value");
     return;
   }
   mCurrentAttributes.SetAppendWindowEnd(aAppendWindowEnd);
@@ -177,7 +178,7 @@ void SourceBuffer::SetAppendWindowEnd(double aAppendWindowEnd,
 void SourceBuffer::AppendBuffer(const ArrayBuffer& aData, ErrorResult& aRv) {
   MOZ_ASSERT(NS_IsMainThread());
   MSE_API("AppendBuffer(ArrayBuffer)");
-  aData.ComputeLengthAndData();
+  aData.ComputeState();
   DDLOG(DDLogCategory::API, "AppendBuffer", aData.Length());
   AppendData(aData.Data(), aData.Length(), aRv);
 }
@@ -186,11 +187,33 @@ void SourceBuffer::AppendBuffer(const ArrayBufferView& aData,
                                 ErrorResult& aRv) {
   MOZ_ASSERT(NS_IsMainThread());
   MSE_API("AppendBuffer(ArrayBufferView)");
-  aData.ComputeLengthAndData();
+  aData.ComputeState();
   DDLOG(DDLogCategory::API, "AppendBuffer", aData.Length());
   AppendData(aData.Data(), aData.Length(), aRv);
 }
 
+already_AddRefed<Promise> SourceBuffer::AppendBufferAsync(
+    const ArrayBuffer& aData, ErrorResult& aRv) {
+  MOZ_ASSERT(NS_IsMainThread());
+
+  MSE_API("AppendBufferAsync(ArrayBuffer)");
+  aData.ComputeState();
+  DDLOG(DDLogCategory::API, "AppendBufferAsync", aData.Length());
+
+  return AppendDataAsync(aData.Data(), aData.Length(), aRv);
+}
+
+already_AddRefed<Promise> SourceBuffer::AppendBufferAsync(
+    const ArrayBufferView& aData, ErrorResult& aRv) {
+  MOZ_ASSERT(NS_IsMainThread());
+
+  MSE_API("AppendBufferAsync(ArrayBufferView)");
+  aData.ComputeState();
+  DDLOG(DDLogCategory::API, "AppendBufferAsync", aData.Length());
+
+  return AppendDataAsync(aData.Data(), aData.Length(), aRv);
+}
+
 void SourceBuffer::Abort(ErrorResult& aRv) {
   MOZ_ASSERT(NS_IsMainThread());
   MSE_API("Abort()");
@@ -236,6 +259,52 @@ void SourceBuffer::Remove(double aStart, double aEnd, ErrorResult& aRv) {
   MSE_API("Remove(aStart=%f, aEnd=%f)", aStart, aEnd);
   DDLOG(DDLogCategory::API, "Remove-from", aStart);
   DDLOG(DDLogCategory::API, "Remove-until", aEnd);
+
+  PrepareRemove(aStart, aEnd, aRv);
+  if (aRv.Failed()) {
+    return;
+  }
+  RangeRemoval(aStart, aEnd);
+}
+
+already_AddRefed<Promise> SourceBuffer::RemoveAsync(double aStart, double aEnd,
+                                                    ErrorResult& aRv) {
+  MOZ_ASSERT(NS_IsMainThread());
+  MSE_API("RemoveAsync(aStart=%f, aEnd=%f)", aStart, aEnd);
+  DDLOG(DDLogCategory::API, "Remove-from", aStart);
+  DDLOG(DDLogCategory::API, "Remove-until", aEnd);
+
+  if (!IsAttached()) {
+    aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
+    return nullptr;
+  }
+
+  nsCOMPtr<nsIGlobalObject> parentObject =
+      do_QueryInterface(mMediaSource->GetParentObject());
+  if (!parentObject) {
+    aRv.Throw(NS_ERROR_UNEXPECTED);
+    return nullptr;
+  }
+
+  RefPtr<Promise> promise = Promise::Create(parentObject, aRv);
+  if (aRv.Failed()) {
+    return nullptr;
+  }
+
+  PrepareRemove(aStart, aEnd, aRv);
+
+  if (aRv.Failed()) {
+    // The bindings will automatically return a rejected promise.
+    return nullptr;
+  }
+  MOZ_ASSERT(!mDOMPromise, "Can't have a pending operation going");
+  mDOMPromise = promise;
+  RangeRemoval(aStart, aEnd);
+
+  return promise.forget();
+}
+
+void SourceBuffer::PrepareRemove(double aStart, double aEnd, ErrorResult& aRv) {
   if (!IsAttached()) {
     aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
     return;
@@ -244,16 +313,21 @@ void SourceBuffer::Remove(double aStart, double aEnd, ErrorResult& aRv) {
     aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
     return;
   }
-  if (IsNaN(mMediaSource->Duration()) || aStart < 0 ||
-      aStart > mMediaSource->Duration() || aEnd <= aStart || IsNaN(aEnd)) {
-    aRv.Throw(NS_ERROR_DOM_TYPE_ERR);
+  if (std::isnan(mMediaSource->Duration())) {
+    aRv.ThrowTypeError("Duration is NaN");
+    return;
+  }
+  if (aStart < 0 || aStart > mMediaSource->Duration()) {
+    aRv.ThrowTypeError("Invalid start value");
+    return;
+  }
+  if (aEnd <= aStart || std::isnan(aEnd)) {
+    aRv.ThrowTypeError("Invalid end value");
     return;
   }
   if (mMediaSource->ReadyState() == MediaSourceReadyState::Ended) {
     mMediaSource->SetReadyState(MediaSourceReadyState::Open);
   }
-
-  RangeRemoval(aStart, aEnd);
 }
 
 void SourceBuffer::RangeRemoval(double aStart, double aEnd) {
@@ -262,15 +336,94 @@ void SourceBuffer::RangeRemoval(double aStart, double aEnd) {
   RefPtr<SourceBuffer> self = this;
   mTrackBuffersManager
       ->RangeRemoval(TimeUnit::FromSeconds(aStart), TimeUnit::FromSeconds(aEnd))
-      ->Then(mAbstractMainThread, __func__,
-             [self](bool) {
-               self->mPendingRemoval.Complete();
-               self->StopUpdating();
-             },
-             []() { MOZ_ASSERT(false); })
+      ->Then(
+          mAbstractMainThread, __func__,
+          [self](bool) {
+            self->mPendingRemoval.Complete();
+            self->StopUpdating();
+          },
+          []() { MOZ_ASSERT(false); })
       ->Track(mPendingRemoval);
 }
 
+void SourceBuffer::ChangeType(const nsAString& aType, ErrorResult& aRv) {
+  MOZ_ASSERT(NS_IsMainThread());
+
+  // 1. If type is an empty string then throw a TypeError exception and abort
+  //    these steps.
+  if (aType.IsEmpty()) {
+    aRv.ThrowTypeError("Type must not be empty");
+    return;
+  }
+
+  // 2. If this object has been removed from the sourceBuffers attribute of the
+  //    parent media source , then throw an InvalidStateError exception and
+  //    abort these steps.
+  // 3. If the updating attribute equals true, then throw an InvalidStateError
+  //    exception and abort these steps.
+  if (!IsAttached() || mUpdating) {
+    DDLOG(DDLogCategory::API, "ChangeType", NS_ERROR_DOM_INVALID_STATE_ERR);
+    aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
+    return;
+  }
+
+  // 4. If type contains a MIME type that is not supported or contains a MIME
+  //    type that is not supported with the types specified (currently or
+  //    previously) of SourceBuffer objects in the sourceBuffers attribute of
+  //    the parent media source , then throw a NotSupportedError exception and
+  //    abort these steps.
+  DecoderDoctorDiagnostics diagnostics;
+  MediaSource::IsTypeSupported(aType, &diagnostics, aRv);
+  bool supported = !aRv.Failed();
+  diagnostics.StoreFormatDiagnostics(
+      mMediaSource->GetOwner() ? mMediaSource->GetOwner()->GetExtantDoc()
+                               : nullptr,
+      aType, supported, __func__);
+  MSE_API("ChangeType(aType=%s)%s", NS_ConvertUTF16toUTF8(aType).get(),
+          supported ? "" : " [not supported]");
+  if (!supported) {
+    DDLOG(DDLogCategory::API, "ChangeType",
+          static_cast<nsresult>(aRv.ErrorCodeAsInt()));
+    return;
+  }
+
+  // 5. If the readyState attribute of the parent media source is in the "ended"
+  //    state then run the following steps:
+  //    1. Set the readyState attribute of the parent media source to "open"
+  //    2.   Queue a task to fire a simple event named sourceopen at the parent
+  //         media source .
+  MOZ_ASSERT(mMediaSource->ReadyState() != MediaSourceReadyState::Closed);
+  if (mMediaSource->ReadyState() == MediaSourceReadyState::Ended) {
+    mMediaSource->SetReadyState(MediaSourceReadyState::Open);
+  }
+  Maybe<MediaContainerType> containerType = MakeMediaContainerType(aType);
+  MOZ_ASSERT(containerType);
+  mType = *containerType;
+  // 6. Run the reset parser state algorithm .
+  ResetParserState();
+
+  // 7. Update the generate timestamps flag on this SourceBuffer object to the
+  //    value in the "Generate Timestamps Flag" column of the byte stream format
+  //    registry [ MSE-REGISTRY ] entry that is associated with type .
+  if (mType.Type() == MEDIAMIMETYPE("audio/mpeg") ||
+      mType.Type() == MEDIAMIMETYPE("audio/aac")) {
+    mCurrentAttributes.mGenerateTimestamps = true;
+    // 8. If the generate timestamps flag equals true:
+    //    Set the mode attribute on this SourceBuffer object to "sequence" ,
+    //    including running the associated steps for that attribute being set.
+    ErrorResult dummy;
+    SetMode(SourceBufferAppendMode::Sequence, dummy);
+  } else {
+    mCurrentAttributes.mGenerateTimestamps = false;
+    //    Otherwise: Keep the previous value of the mode attribute on this
+    //    SourceBuffer object, without running any associated steps for that
+    //    attribute being set.
+  }
+
+  // 9. Set pending initialization segment for changeType flag to true.
+  mTrackBuffersManager->ChangeType(mType);
+}
+
 void SourceBuffer::Detach() {
   MOZ_ASSERT(NS_IsMainThread());
   MSE_DEBUG("Detach");
@@ -334,7 +487,7 @@ MediaSource* SourceBuffer::GetParentObject() const { return mMediaSource; }
 
 JSObject* SourceBuffer::WrapObject(JSContext* aCx,
                                    JS::Handle<JSObject*> aGivenProto) {
-  return SourceBufferBinding::Wrap(aCx, this, aGivenProto);
+  return SourceBuffer_Binding::Wrap(aCx, this, aGivenProto);
 }
 
 void SourceBuffer::DispatchSimpleEvent(const char* aName) {
@@ -366,6 +519,10 @@ void SourceBuffer::StopUpdating() {
   mUpdating = false;
   QueueAsyncSimpleEvent("update");
   QueueAsyncSimpleEvent("updateend");
+  if (mDOMPromise) {
+    mDOMPromise->MaybeResolveWithUndefined();
+    mDOMPromise = nullptr;
+  }
 }
 
 void SourceBuffer::AbortUpdating() {
@@ -373,20 +530,25 @@ void SourceBuffer::AbortUpdating() {
   mUpdating = false;
   QueueAsyncSimpleEvent("abort");
   QueueAsyncSimpleEvent("updateend");
+  if (mDOMPromise) {
+    mDOMPromise->MaybeReject(NS_ERROR_DOM_MEDIA_ABORT_ERR);
+    mDOMPromise = nullptr;
+  }
 }
 
 void SourceBuffer::CheckEndTime() {
   MOZ_ASSERT(NS_IsMainThread());
   // Check if we need to update mMediaSource duration
-  double endTime = mCurrentAttributes.GetGroupEndTimestamp().ToSeconds();
+  TimeUnit endTime = mCurrentAttributes.GetGroupEndTimestamp();
   double duration = mMediaSource->Duration();
-  if (endTime > duration) {
+  if (!std::isnan(duration) && endTime.ToSeconds() > duration) {
     mMediaSource->SetDuration(endTime);
   }
 }
 
 void SourceBuffer::AppendData(const uint8_t* aData, uint32_t aLength,
                               ErrorResult& aRv) {
+  MOZ_ASSERT(NS_IsMainThread());
   MSE_DEBUG("AppendData(aLength=%u)", aLength);
 
   RefPtr<MediaByteBuffer> data = PrepareAppend(aData, aLength, aRv);
@@ -402,13 +564,47 @@ void SourceBuffer::AppendData(const uint8_t* aData, uint32_t aLength,
       ->Track(mPendingAppend);
 }
 
+already_AddRefed<Promise> SourceBuffer::AppendDataAsync(const uint8_t* aData,
+                                                        uint32_t aLength,
+                                                        ErrorResult& aRv) {
+  MOZ_ASSERT(NS_IsMainThread());
+
+  if (!IsAttached()) {
+    aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
+    return nullptr;
+  }
+
+  nsCOMPtr<nsIGlobalObject> parentObject =
+      do_QueryInterface(mMediaSource->GetParentObject());
+  if (!parentObject) {
+    aRv.Throw(NS_ERROR_UNEXPECTED);
+    return nullptr;
+  }
+
+  RefPtr<Promise> promise = Promise::Create(parentObject, aRv);
+  if (aRv.Failed()) {
+    return nullptr;
+  }
+
+  AppendData(aData, aLength, aRv);
+
+  if (aRv.Failed()) {
+    return nullptr;
+  }
+
+  MOZ_ASSERT(!mDOMPromise, "Can't have a pending operation going");
+  mDOMPromise = promise;
+
+  return promise.forget();
+}
+
 void SourceBuffer::AppendDataCompletedWithSuccess(
     const SourceBufferTask::AppendBufferResult& aResult) {
   MOZ_ASSERT(mUpdating);
   mPendingAppend.Complete();
   DDLOG(DDLogCategory::API, "AppendBuffer-completed", NS_OK);
 
-  if (aResult.first()) {
+  if (aResult.first) {
     if (!mActive) {
       mActive = true;
       MSE_DEBUG("Init segment received");
@@ -429,7 +625,7 @@ void SourceBuffer::AppendDataCompletedWithSuccess(
     mMediaSource->GetDecoder()->NotifyDataArrived();
   }
 
-  mCurrentAttributes = aResult.second();
+  mCurrentAttributes = aResult.second;
 
   CheckEndTime();
 
@@ -467,6 +663,11 @@ void SourceBuffer::AppendError(const MediaResult& aDecodeError) {
   MOZ_ASSERT(NS_FAILED(aDecodeError));
 
   mMediaSource->EndOfStream(aDecodeError);
+
+  if (mDOMPromise) {
+    mDOMPromise->MaybeReject(aDecodeError);
+    mDOMPromise = nullptr;
+  }
 }
 
 already_AddRefed<MediaByteBuffer> SourceBuffer::PrepareAppend(
@@ -502,44 +703,35 @@ already_AddRefed<MediaByteBuffer> SourceBuffer::PrepareAppend(
 
   // See if we have enough free space to append our new data.
   if (evicted == Result::BUFFER_FULL) {
-    aRv.Throw(NS_ERROR_DOM_QUOTA_EXCEEDED_ERR);
+    aRv.Throw(NS_ERROR_DOM_MEDIA_SOURCE_FULL_BUFFER_QUOTA_EXCEEDED_ERR);
     return nullptr;
   }
 
   RefPtr<MediaByteBuffer> data = new MediaByteBuffer();
   if (!data->AppendElements(aData, aLength, fallible)) {
-    aRv.Throw(NS_ERROR_DOM_QUOTA_EXCEEDED_ERR);
+    aRv.Throw(NS_ERROR_DOM_MEDIA_SOURCE_FULL_BUFFER_QUOTA_EXCEEDED_ERR);
     return nullptr;
   }
   return data.forget();
 }
 
-double SourceBuffer::GetBufferedStart() {
+TimeUnit SourceBuffer::GetBufferedEnd() {
   MOZ_ASSERT(NS_IsMainThread());
   ErrorResult dummy;
-  RefPtr<TimeRanges> ranges = GetBuffered(dummy);
-  return ranges->Length() > 0 ? ranges->GetStartTime() : 0;
+  media::TimeIntervals intervals = GetBufferedIntervals();
+  return intervals.GetEnd();
 }
 
-double SourceBuffer::GetBufferedEnd() {
+TimeUnit SourceBuffer::HighestStartTime() {
   MOZ_ASSERT(NS_IsMainThread());
-  ErrorResult dummy;
-  RefPtr<TimeRanges> ranges = GetBuffered(dummy);
-  return ranges->Length() > 0 ? ranges->GetEndTime() : 0;
+  MOZ_ASSERT(mTrackBuffersManager);
+  return mTrackBuffersManager->HighestStartTime();
 }
 
-double SourceBuffer::HighestStartTime() {
+TimeUnit SourceBuffer::HighestEndTime() {
   MOZ_ASSERT(NS_IsMainThread());
-  return mTrackBuffersManager
-             ? mTrackBuffersManager->HighestStartTime().ToSeconds()
-             : 0.0;
-}
-
-double SourceBuffer::HighestEndTime() {
-  MOZ_ASSERT(NS_IsMainThread());
-  return mTrackBuffersManager
-             ? mTrackBuffersManager->HighestEndTime().ToSeconds()
-             : 0.0;
+  MOZ_ASSERT(mTrackBuffersManager);
+  return mTrackBuffersManager->HighestEndTime();
 }
 
 NS_IMPL_CYCLE_COLLECTION_CLASS(SourceBuffer)
@@ -548,12 +740,14 @@ NS_IMPL_CYCLE_COLLECTION_UNLINK_BEGIN(SourceBuffer)
   tmp->Detach();
   NS_IMPL_CYCLE_COLLECTION_UNLINK(mMediaSource)
   NS_IMPL_CYCLE_COLLECTION_UNLINK(mBuffered)
+  NS_IMPL_CYCLE_COLLECTION_UNLINK(mDOMPromise)
 NS_IMPL_CYCLE_COLLECTION_UNLINK_END_INHERITED(DOMEventTargetHelper)
 
 NS_IMPL_CYCLE_COLLECTION_TRAVERSE_BEGIN_INHERITED(SourceBuffer,
                                                   DOMEventTargetHelper)
   NS_IMPL_CYCLE_COLLECTION_TRAVERSE(mMediaSource)
   NS_IMPL_CYCLE_COLLECTION_TRAVERSE(mBuffered)
+  NS_IMPL_CYCLE_COLLECTION_TRAVERSE(mDOMPromise)
 NS_IMPL_CYCLE_COLLECTION_TRAVERSE_END
 
 NS_IMPL_ADDREF_INHERITED(SourceBuffer, DOMEventTargetHelper)
