@@ -5,17 +5,17 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "OpusDecoder.h"
-#include "OpusParser.h"
-#include "TimeUnits.h"
-#include "VorbisUtils.h"
-#include "VorbisDecoder.h"  // For VorbisLayout
-#include "mozilla/EndianUtils.h"
-#include "mozilla/PodOperations.h"
-#include "mozilla/SyncRunnable.h"
-#include "VideoUtils.h"
 
 #include <inttypes.h>  // For PRId64
 
+#include "OpusParser.h"
+#include "TimeUnits.h"
+#include "VideoUtils.h"
+#include "VorbisDecoder.h"  // For VorbisLayout
+#include "VorbisUtils.h"
+#include "mozilla/EndianUtils.h"
+#include "mozilla/PodOperations.h"
+#include "mozilla/SyncRunnable.h"
 #include "opus/opus.h"
 extern "C" {
 #include "opus/opus_multistream.h"
@@ -29,7 +29,6 @@ namespace mozilla {
 
 OpusDataDecoder::OpusDataDecoder(const CreateDecoderParams& aParams)
     : mInfo(aParams.AudioConfig()),
-      mTaskQueue(aParams.mTaskQueue),
       mOpusDecoder(nullptr),
       mSkip(0),
       mDecodedHeader(false),
@@ -47,33 +46,28 @@ OpusDataDecoder::~OpusDataDecoder() {
 }
 
 RefPtr<ShutdownPromise> OpusDataDecoder::Shutdown() {
-  RefPtr<OpusDataDecoder> self = this;
-  return InvokeAsync(mTaskQueue, __func__, [self]() {
-    return ShutdownPromise::CreateAndResolve(true, __func__);
-  });
-}
-
-void OpusDataDecoder::AppendCodecDelay(MediaByteBuffer* config,
-                                       uint64_t codecDelayUS) {
-  uint8_t buffer[sizeof(uint64_t)];
-  BigEndian::writeUint64(buffer, codecDelayUS);
-  config->AppendElements(buffer, sizeof(uint64_t));
+  // mThread may not be set if Init hasn't been called first.
+  MOZ_ASSERT(!mThread || mThread->IsOnCurrentThread());
+  return ShutdownPromise::CreateAndResolve(true, __func__);
 }
 
 RefPtr<MediaDataDecoder::InitPromise> OpusDataDecoder::Init() {
-  size_t length = mInfo.mCodecSpecificConfig->Length();
-  uint8_t* p = mInfo.mCodecSpecificConfig->Elements();
-  if (length < sizeof(uint64_t)) {
-    OPUS_DEBUG("CodecSpecificConfig too short to read codecDelay!");
+  mThread = GetCurrentSerialEventTarget();
+  if (!mInfo.mCodecSpecificConfig.is<OpusCodecSpecificData>()) {
+    MOZ_ASSERT_UNREACHABLE();
+    OPUS_DEBUG("Opus decoder got non-opus codec specific data");
     return InitPromise::CreateAndReject(
         MediaResult(
             NS_ERROR_DOM_MEDIA_FATAL_ERR,
-            RESULT_DETAIL("CodecSpecificConfig too short to read codecDelay!")),
+            RESULT_DETAIL("Opus decoder got non-opus codec specific data!")),
         __func__);
   }
-  int64_t codecDelay = BigEndian::readUint64(p);
-  length -= sizeof(uint64_t);
-  p += sizeof(uint64_t);
+  const OpusCodecSpecificData opusCodecSpecificData =
+      mInfo.mCodecSpecificConfig.as<OpusCodecSpecificData>();
+  RefPtr<MediaByteBuffer> opusHeaderBlob =
+      opusCodecSpecificData.mHeadersBinaryBlob;
+  size_t length = opusHeaderBlob->Length();
+  uint8_t* p = opusHeaderBlob->Elements();
   if (NS_FAILED(DecodeHeader(p, length))) {
     OPUS_DEBUG("Error decoding header!");
     return InitPromise::CreateAndReject(
@@ -109,11 +103,14 @@ RefPtr<MediaDataDecoder::InitPromise> OpusDataDecoder::Init() {
   mSkip = mOpusParser->mPreSkip;
   mPaddingDiscarded = false;
 
-  if (codecDelay !=
+  if (opusCodecSpecificData.mContainerCodecDelayMicroSeconds !=
       FramesToUsecs(mOpusParser->mPreSkip, mOpusParser->mRate).value()) {
-    NS_WARNING("Invalid Opus header: CodecDelay and pre-skip do not match!");
-    return InitPromise::CreateAndReject(NS_ERROR_DOM_MEDIA_FATAL_ERR, __func__);
+    NS_WARNING(
+        "Invalid Opus header: container CodecDelay and Opus pre-skip do not "
+        "match!");
   }
+  OPUS_DEBUG("Opus preskip in extradata: %" PRId64 "us",
+             opusCodecSpecificData.mContainerCodecDelayMicroSeconds);
 
   if (mInfo.mRate != (uint32_t)mOpusParser->mRate) {
     NS_WARNING("Invalid Opus header: container and codec rate do not match!");
@@ -179,12 +176,8 @@ nsresult OpusDataDecoder::DecodeHeader(const unsigned char* aData,
 
 RefPtr<MediaDataDecoder::DecodePromise> OpusDataDecoder::Decode(
     MediaRawData* aSample) {
-  return InvokeAsync<MediaRawData*>(mTaskQueue, this, __func__,
-                                    &OpusDataDecoder::ProcessDecode, aSample);
-}
-
-RefPtr<MediaDataDecoder::DecodePromise> OpusDataDecoder::ProcessDecode(
-    MediaRawData* aSample) {
+  MOZ_ASSERT(mThread->IsOnCurrentThread());
+  PROCESS_DECODE_LOG(aSample);
   uint32_t channels = mOpusParser->mChannels;
 
   if (mPaddingDiscarded) {
@@ -264,16 +257,23 @@ RefPtr<MediaDataDecoder::DecodePromise> OpusDataDecoder::ProcessDecode(
   NS_ASSERTION(ret == frames, "Opus decoded too few audio samples");
   auto startTime = aSample->mTime;
 
+  OPUS_DEBUG("Decoding frames: [%lf, %lf]", aSample->mTime.ToSeconds(),
+             aSample->GetEndTime().ToSeconds());
+
   // Trim the initial frames while the decoder is settling.
   if (mSkip > 0) {
     int32_t skipFrames = std::min<int32_t>(mSkip, frames);
     int32_t keepFrames = frames - skipFrames;
-    OPUS_DEBUG("Opus decoder skipping %d of %d frames", skipFrames, frames);
+    OPUS_DEBUG("Opus decoder trimming %d of %d frames", skipFrames, frames);
     PodMove(buffer.get(), buffer.get() + skipFrames * channels,
             keepFrames * channels);
-    startTime = startTime + FramesToTimeUnit(skipFrames, mOpusParser->mRate);
+    startTime = startTime + media::TimeUnit(skipFrames, mOpusParser->mRate);
     frames = keepFrames;
     mSkip -= skipFrames;
+    aSample->mTime += media::TimeUnit(skipFrames, 48000);
+    aSample->mDuration -= media::TimeUnit(skipFrames, 48000);
+    OPUS_DEBUG("Adjusted frame after trimming pre-roll: [%lf, %lf]",
+               aSample->mTime.ToSeconds(), aSample->GetEndTime().ToSeconds());
   }
 
   if (aSample->mDiscardPadding > 0) {
@@ -315,7 +315,7 @@ RefPtr<MediaDataDecoder::DecodePromise> OpusDataDecoder::ProcessDecode(
   }
 #endif
 
-  auto duration = FramesToTimeUnit(frames, mOpusParser->mRate);
+  auto duration = media::TimeUnit(frames, mOpusParser->mRate);
   if (!duration.IsValid()) {
     return DecodePromise::CreateAndReject(
         MediaResult(NS_ERROR_DOM_MEDIA_OVERFLOW_ERR,
@@ -323,8 +323,8 @@ RefPtr<MediaDataDecoder::DecodePromise> OpusDataDecoder::ProcessDecode(
         __func__);
   }
   auto time = startTime -
-              FramesToTimeUnit(mOpusParser->mPreSkip, mOpusParser->mRate) +
-              FramesToTimeUnit(mFrames, mOpusParser->mRate);
+              media::TimeUnit(mOpusParser->mPreSkip, mOpusParser->mRate) +
+              media::TimeUnit(mFrames, mOpusParser->mRate);
   if (!time.IsValid()) {
     return DecodePromise::CreateAndReject(
         MediaResult(NS_ERROR_DOM_MEDIA_OVERFLOW_ERR,
@@ -333,6 +333,9 @@ RefPtr<MediaDataDecoder::DecodePromise> OpusDataDecoder::ProcessDecode(
   };
 
   mFrames += frames;
+  mTotalFrames += frames;
+
+  OPUS_DEBUG("Total frames so far: %" PRId64, mTotalFrames);
 
   if (!frames) {
     return DecodePromise::CreateAndResolve(DecodedData(), __func__);
@@ -349,29 +352,23 @@ RefPtr<MediaDataDecoder::DecodePromise> OpusDataDecoder::ProcessDecode(
 }
 
 RefPtr<MediaDataDecoder::DecodePromise> OpusDataDecoder::Drain() {
-  RefPtr<OpusDataDecoder> self = this;
-  // InvokeAsync dispatches a task that will be run after any pending decode
-  // completes. As such, once the drain task run, there's nothing more to do.
-  return InvokeAsync(mTaskQueue, __func__, [] {
-    return DecodePromise::CreateAndResolve(DecodedData(), __func__);
-  });
+  MOZ_ASSERT(mThread->IsOnCurrentThread());
+  return DecodePromise::CreateAndResolve(DecodedData(), __func__);
 }
 
 RefPtr<MediaDataDecoder::FlushPromise> OpusDataDecoder::Flush() {
+  MOZ_ASSERT(mThread->IsOnCurrentThread());
   if (!mOpusDecoder) {
     return FlushPromise::CreateAndResolve(true, __func__);
   }
 
-  RefPtr<OpusDataDecoder> self = this;
-  return InvokeAsync(mTaskQueue, __func__, [self, this]() {
-    MOZ_ASSERT(mOpusDecoder);
-    // Reset the decoder.
-    opus_multistream_decoder_ctl(mOpusDecoder, OPUS_RESET_STATE);
-    mSkip = mOpusParser->mPreSkip;
-    mPaddingDiscarded = false;
-    mLastFrameTime.reset();
-    return FlushPromise::CreateAndResolve(true, __func__);
-  });
+  MOZ_ASSERT(mOpusDecoder);
+  // Reset the decoder.
+  opus_multistream_decoder_ctl(mOpusDecoder, OPUS_RESET_STATE);
+  mSkip = mOpusParser->mPreSkip;
+  mPaddingDiscarded = false;
+  mLastFrameTime.reset();
+  return FlushPromise::CreateAndResolve(true, __func__);
 }
 
 /* static */
