@@ -5,14 +5,15 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "AppleATDecoder.h"
+#include "Adts.h"
 #include "AppleUtils.h"
 #include "MP4Decoder.h"
-#include "Adts.h"
 #include "MediaInfo.h"
+#include "VideoUtils.h"
 #include "mozilla/Logging.h"
 #include "mozilla/SyncRunnable.h"
 #include "mozilla/UniquePtr.h"
-#include "VideoUtils.h"
+#include "nsTArray.h"
 
 #define LOG(...) DDMOZ_LOG(sPDMLog, mozilla::LogLevel::Debug, __VA_ARGS__)
 #define LOGEX(_this, ...) \
@@ -22,11 +23,11 @@
 
 namespace mozilla {
 
-AppleATDecoder::AppleATDecoder(const AudioInfo& aConfig, TaskQueue* aTaskQueue)
+AppleATDecoder::AppleATDecoder(const AudioInfo& aConfig)
     : mConfig(aConfig),
       mFileStreamError(false),
-      mTaskQueue(aTaskQueue),
       mConverter(nullptr),
+      mOutputFormat(),
       mStream(nullptr),
       mParsedFramesForAACMagicCookie(0),
       mErrored(false) {
@@ -40,6 +41,15 @@ AppleATDecoder::AppleATDecoder(const AudioInfo& aConfig, TaskQueue* aTaskQueue)
     mFormatID = kAudioFormatMPEGLayer3;
   } else if (mConfig.mMimeType.EqualsLiteral("audio/mp4a-latm")) {
     mFormatID = kAudioFormatMPEG4AAC;
+    if (aConfig.mCodecSpecificConfig.is<AacCodecSpecificData>()) {
+      const AacCodecSpecificData& aacCodecSpecificData =
+          aConfig.mCodecSpecificConfig.as<AacCodecSpecificData>();
+      mEncoderDelay = aacCodecSpecificData.mEncoderDelayFrames;
+      mTotalMediaFrames = aacCodecSpecificData.mMediaFrameCount;
+      LOG("AppleATDecoder (aac), found encoder delay (%" PRIu32
+          ") and total frame count (%" PRIu64 ") in codec-specific side data",
+          mEncoderDelay, mTotalMediaFrames);
+    }
   } else {
     mFormatID = 0;
   }
@@ -57,24 +67,14 @@ RefPtr<MediaDataDecoder::InitPromise> AppleATDecoder::Init() {
                     RESULT_DETAIL("Non recognised format")),
         __func__);
   }
+  mThread = GetCurrentSerialEventTarget();
 
   return InitPromise::CreateAndResolve(TrackType::kAudioTrack, __func__);
 }
 
-RefPtr<MediaDataDecoder::DecodePromise> AppleATDecoder::Decode(
-    MediaRawData* aSample) {
-  LOG("mp4 input sample %p %lld us %lld pts%s %llu bytes audio", aSample,
-      aSample->mDuration.ToMicroseconds(), aSample->mTime.ToMicroseconds(),
-      aSample->mKeyframe ? " keyframe" : "",
-      (unsigned long long)aSample->Size());
-  RefPtr<AppleATDecoder> self = this;
-  RefPtr<MediaRawData> sample = aSample;
-  return InvokeAsync(mTaskQueue, __func__,
-                     [self, sample] { return self->ProcessDecode(sample); });
-}
-
-RefPtr<MediaDataDecoder::FlushPromise> AppleATDecoder::ProcessFlush() {
-  MOZ_ASSERT(mTaskQueue->IsCurrentThreadIn());
+RefPtr<MediaDataDecoder::FlushPromise> AppleATDecoder::Flush() {
+  MOZ_ASSERT(mThread->IsOnCurrentThread());
+  LOG("Flushing AudioToolbox AAC decoder");
   mQueuedSamples.Clear();
   mDecodedSamples.Clear();
 
@@ -93,29 +93,22 @@ RefPtr<MediaDataDecoder::FlushPromise> AppleATDecoder::ProcessFlush() {
   return FlushPromise::CreateAndResolve(true, __func__);
 }
 
-RefPtr<MediaDataDecoder::FlushPromise> AppleATDecoder::Flush() {
-  LOG("Flushing AudioToolbox AAC decoder");
-  return InvokeAsync(mTaskQueue, this, __func__, &AppleATDecoder::ProcessFlush);
-}
-
 RefPtr<MediaDataDecoder::DecodePromise> AppleATDecoder::Drain() {
+  MOZ_ASSERT(mThread->IsOnCurrentThread());
   LOG("Draining AudioToolbox AAC decoder");
-  RefPtr<AppleATDecoder> self = this;
-  return InvokeAsync(mTaskQueue, __func__, [] {
-    return DecodePromise::CreateAndResolve(DecodedData(), __func__);
-  });
+  return DecodePromise::CreateAndResolve(DecodedData(), __func__);
 }
 
 RefPtr<ShutdownPromise> AppleATDecoder::Shutdown() {
-  RefPtr<AppleATDecoder> self = this;
-  return InvokeAsync(mTaskQueue, __func__, [self]() {
-    self->ProcessShutdown();
-    return ShutdownPromise::CreateAndResolve(true, __func__);
-  });
+  // mThread may not be set if Init hasn't been called first.
+  MOZ_ASSERT(!mThread || mThread->IsOnCurrentThread());
+  ProcessShutdown();
+  return ShutdownPromise::CreateAndResolve(true, __func__);
 }
 
 void AppleATDecoder::ProcessShutdown() {
-  MOZ_ASSERT(mTaskQueue->IsCurrentThreadIn());
+  // mThread may not be set if Init hasn't been called first.
+  MOZ_ASSERT(!mThread || mThread->IsOnCurrentThread());
 
   if (mStream) {
     OSStatus rv = AudioFileStreamClose(mStream);
@@ -136,6 +129,17 @@ void AppleATDecoder::ProcessShutdown() {
   }
 }
 
+nsCString AppleATDecoder::GetCodecName() const {
+  switch (mFormatID) {
+    case kAudioFormatMPEGLayer3:
+      return "mp3"_ns;
+    case kAudioFormatMPEG4AAC:
+      return "aac"_ns;
+    default:
+      return "unknown"_ns;
+  }
+}
+
 struct PassthroughUserData {
   UInt32 mChannels;
   UInt32 mDataSize;
@@ -174,9 +178,13 @@ static OSStatus _PassthroughInputDataCallback(
   return noErr;
 }
 
-RefPtr<MediaDataDecoder::DecodePromise> AppleATDecoder::ProcessDecode(
+RefPtr<MediaDataDecoder::DecodePromise> AppleATDecoder::Decode(
     MediaRawData* aSample) {
-  MOZ_ASSERT(mTaskQueue->IsCurrentThreadIn());
+  MOZ_ASSERT(mThread->IsOnCurrentThread());
+  LOG("mp4 input sample pts=%s duration=%s %s %llu bytes audio",
+      aSample->mTime.ToString().get(), aSample->GetEndTime().ToString().get(),
+      aSample->mKeyframe ? " keyframe" : "",
+      (unsigned long long)aSample->Size());
 
   MediaResult rv = NS_OK;
   if (!mConverter) {
@@ -199,18 +207,21 @@ RefPtr<MediaDataDecoder::DecodePromise> AppleATDecoder::ProcessDecode(
     mQueuedSamples.Clear();
   }
 
-  return DecodePromise::CreateAndResolve(Move(mDecodedSamples), __func__);
+  DecodedData results = std::move(mDecodedSamples);
+  mDecodedSamples = DecodedData();
+  return DecodePromise::CreateAndResolve(std::move(results), __func__);
 }
 
 MediaResult AppleATDecoder::DecodeSample(MediaRawData* aSample) {
-  MOZ_ASSERT(mTaskQueue->IsCurrentThreadIn());
+  MOZ_ASSERT(mThread->IsOnCurrentThread());
 
   // Array containing the queued decoded audio frames, about to be output.
   nsTArray<AudioDataValue> outputData;
   UInt32 channels = mOutputFormat.mChannelsPerFrame;
   // Pick a multiple of the frame size close to a power of two
-  // for efficient allocation.
-  const uint32_t MAX_AUDIO_FRAMES = 128;
+  // for efficient allocation. We're mainly using this decoder to decode AAC,
+  // that has packets of 1024 audio frames.
+  const uint32_t MAX_AUDIO_FRAMES = 1024;
   const uint32_t maxDecodedSamples = MAX_AUDIO_FRAMES * channels;
 
   // Descriptions for _decompressed_ audio packets. ignored.
@@ -247,12 +258,13 @@ MediaResult AppleATDecoder::DecodeSample(MediaRawData* aSample) {
       LOG("Error decoding audio sample: %d\n", static_cast<int>(rv));
       return MediaResult(
           NS_ERROR_DOM_MEDIA_DECODE_ERR,
-          RESULT_DETAIL("Error decoding audio sample: %d @ %lld",
-                        static_cast<int>(rv), aSample->mTime.ToMicroseconds()));
+          RESULT_DETAIL("Error decoding audio sample: %d @ %s",
+                        static_cast<int>(rv), aSample->mTime.ToString().get()));
     }
 
     if (numFrames) {
-      outputData.AppendElements(decoded.get(), numFrames * channels);
+      AudioDataValue* outputFrames = decoded.get();
+      outputData.AppendElements(outputFrames, numFrames * channels);
     }
 
     if (rv == kNoMoreDataErr) {
@@ -266,7 +278,7 @@ MediaResult AppleATDecoder::DecodeSample(MediaRawData* aSample) {
 
   size_t numFrames = outputData.Length() / channels;
   int rate = mOutputFormat.mSampleRate;
-  media::TimeUnit duration = FramesToTimeUnit(numFrames, rate);
+  media::TimeUnit duration(numFrames, rate);
   if (!duration.IsValid()) {
     NS_WARNING("Invalid count of accumulated audio samples");
     return MediaResult(
@@ -276,39 +288,38 @@ MediaResult AppleATDecoder::DecodeSample(MediaRawData* aSample) {
             uint64_t(numFrames), rate));
   }
 
-#ifdef LOG_SAMPLE_DECODE
-  LOG("pushed audio at time %lfs; duration %lfs\n",
-      (double)aSample->mTime / USECS_PER_S, duration.ToSeconds());
-#endif
+  LOG("Decoded audio packet [%s, %s] (duration: %s)\n",
+      aSample->mTime.ToString().get(), aSample->GetEndTime().ToString().get(),
+      duration.ToString().get());
 
   AudioSampleBuffer data(outputData.Elements(), outputData.Length());
   if (!data.Data()) {
     return NS_ERROR_OUT_OF_MEMORY;
   }
   if (mChannelLayout && !mAudioConverter) {
-    AudioConfig in(*mChannelLayout.get(), rate);
-    AudioConfig out(channels, rate);
-    if (!in.IsValid() || !out.IsValid()) {
-      return MediaResult(NS_ERROR_DOM_MEDIA_DECODE_ERR,
-                         RESULT_DETAIL("Invalid audio config"));
-    }
+    AudioConfig in(*mChannelLayout, channels, rate);
+    AudioConfig out(AudioConfig::ChannelLayout::SMPTEDefault(*mChannelLayout),
+                    channels, rate);
     mAudioConverter = MakeUnique<AudioConverter>(in, out);
   }
-  if (mAudioConverter) {
+  if (mAudioConverter && mChannelLayout && mChannelLayout->IsValid()) {
     MOZ_ASSERT(mAudioConverter->CanWorkInPlace());
-    data = mAudioConverter->Process(Move(data));
+    data = mAudioConverter->Process(std::move(data));
   }
 
-  RefPtr<AudioData> audio =
-      new AudioData(aSample->mOffset, aSample->mTime, duration, numFrames,
-                    data.Forget(), channels, rate);
-  mDecodedSamples.AppendElement(Move(audio));
+  RefPtr<AudioData> audio = new AudioData(
+      aSample->mOffset, aSample->mTime, data.Forget(), channels, rate,
+      mChannelLayout && mChannelLayout->IsValid()
+          ? mChannelLayout->Map()
+          : AudioConfig::ChannelLayout::UNKNOWN_MAP);
+  MOZ_DIAGNOSTIC_ASSERT(duration == audio->mDuration, "must be equal");
+  mDecodedSamples.AppendElement(std::move(audio));
   return NS_OK;
 }
 
 MediaResult AppleATDecoder::GetInputAudioDescription(
     AudioStreamBasicDescription& aDesc, const nsTArray<uint8_t>& aExtraData) {
-  MOZ_ASSERT(mTaskQueue->IsCurrentThreadIn());
+  MOZ_ASSERT(mThread->IsOnCurrentThread());
 
   // Request the properties from CoreAudio using the codec magic cookie
   AudioFormatInfo formatInfo;
@@ -375,26 +386,25 @@ MediaResult AppleATDecoder::GetInputAudioDescription(
 
 AudioConfig::Channel ConvertChannelLabel(AudioChannelLabel id) {
   switch (id) {
-    case kAudioChannelLabel_Mono:
-      return AudioConfig::CHANNEL_MONO;
     case kAudioChannelLabel_Left:
-      return AudioConfig::CHANNEL_LEFT;
+      return AudioConfig::CHANNEL_FRONT_LEFT;
     case kAudioChannelLabel_Right:
-      return AudioConfig::CHANNEL_RIGHT;
+      return AudioConfig::CHANNEL_FRONT_RIGHT;
+    case kAudioChannelLabel_Mono:
     case kAudioChannelLabel_Center:
-      return AudioConfig::CHANNEL_CENTER;
+      return AudioConfig::CHANNEL_FRONT_CENTER;
     case kAudioChannelLabel_LFEScreen:
       return AudioConfig::CHANNEL_LFE;
     case kAudioChannelLabel_LeftSurround:
-      return AudioConfig::CHANNEL_LS;
+      return AudioConfig::CHANNEL_SIDE_LEFT;
     case kAudioChannelLabel_RightSurround:
-      return AudioConfig::CHANNEL_RS;
+      return AudioConfig::CHANNEL_SIDE_RIGHT;
     case kAudioChannelLabel_CenterSurround:
-      return AudioConfig::CHANNEL_RCENTER;
+      return AudioConfig::CHANNEL_BACK_CENTER;
     case kAudioChannelLabel_RearSurroundLeft:
-      return AudioConfig::CHANNEL_RLS;
+      return AudioConfig::CHANNEL_BACK_LEFT;
     case kAudioChannelLabel_RearSurroundRight:
-      return AudioConfig::CHANNEL_RRS;
+      return AudioConfig::CHANNEL_BACK_RIGHT;
     default:
       return AudioConfig::CHANNEL_INVALID;
   }
@@ -403,7 +413,7 @@ AudioConfig::Channel ConvertChannelLabel(AudioChannelLabel id) {
 // Will set mChannelLayout if a channel layout could properly be identified
 // and is supported.
 nsresult AppleATDecoder::SetupChannelLayout() {
-  MOZ_ASSERT(mTaskQueue->IsCurrentThreadIn());
+  MOZ_ASSERT(mThread->IsOnCurrentThread());
 
   // Determine the channel layout.
   UInt32 propertySize;
@@ -473,26 +483,25 @@ nsresult AppleATDecoder::SetupChannelLayout() {
     layout->mChannelLayoutTag = kAudioChannelLayoutTag_UseChannelDescriptions;
   }
 
-  if (layout->mNumberChannelDescriptions > MAX_AUDIO_CHANNELS ||
-      layout->mNumberChannelDescriptions != mOutputFormat.mChannelsPerFrame) {
-    LOG("Nonsensical channel layout or not matching the original channel "
-        "number");
+  if (layout->mNumberChannelDescriptions != mOutputFormat.mChannelsPerFrame) {
+    LOG("Not matching the original channel number");
     return NS_ERROR_FAILURE;
   }
 
-  AudioConfig::Channel channels[MAX_AUDIO_CHANNELS];
+  AutoTArray<AudioConfig::Channel, 8> channels;
+  channels.SetLength(layout->mNumberChannelDescriptions);
   for (uint32_t i = 0; i < layout->mNumberChannelDescriptions; i++) {
     AudioChannelLabel id = layout->mChannelDescriptions[i].mChannelLabel;
     AudioConfig::Channel channel = ConvertChannelLabel(id);
     channels[i] = channel;
   }
   mChannelLayout = MakeUnique<AudioConfig::ChannelLayout>(
-      mOutputFormat.mChannelsPerFrame, channels);
+      mOutputFormat.mChannelsPerFrame, channels.Elements());
   return NS_OK;
 }
 
 MediaResult AppleATDecoder::SetupDecoder(MediaRawData* aSample) {
-  MOZ_ASSERT(mTaskQueue->IsCurrentThreadIn());
+  MOZ_ASSERT(mThread->IsOnCurrentThread());
   static const uint32_t MAX_FRAMES = 2;
 
   if (mFormatID == kAudioFormatMPEG4AAC && mConfig.mExtendedProfile == 2 &&
@@ -511,8 +520,19 @@ MediaResult AppleATDecoder::SetupDecoder(MediaRawData* aSample) {
 
   LOG("Initializing Apple AudioToolbox decoder");
 
+  // Should we try and use magic cookie data from the AAC data? We do this if
+  // - We have an AAC config &
+  // - We do not aleady have magic cookie data.
+  // Otherwise we just use the existing cookie (which may be empty).
+  bool shouldUseAacMagicCookie =
+      mConfig.mCodecSpecificConfig.is<AacCodecSpecificData>() &&
+      mMagicCookie.IsEmpty();
+
   nsTArray<uint8_t>& magicCookie =
-      mMagicCookie.Length() ? mMagicCookie : *mConfig.mExtraData;
+      shouldUseAacMagicCookie
+          ? *mConfig.mCodecSpecificConfig.as<AacCodecSpecificData>()
+                 .mEsDescriptorBinaryBlob
+          : mMagicCookie;
   AudioStreamBasicDescription inputFormat;
   PodZero(&inputFormat);
 
@@ -532,7 +552,7 @@ MediaResult AppleATDecoder::SetupDecoder(MediaRawData* aSample) {
   mOutputFormat.mBitsPerChannel = 16;
   mOutputFormat.mFormatFlags = kLinearPCMFormatFlagIsSignedInteger | 0;
 #else
-#error Unknown audio sample type
+#  error Unknown audio sample type
 #endif
   // Set up the decoder so it gives us one sample per frame
   mOutputFormat.mFramesPerPacket = 1;
@@ -574,7 +594,7 @@ static void _MetadataCallback(void* aAppleATDecoder, AudioFileStreamID aStream,
                               AudioFileStreamPropertyID aProperty,
                               UInt32* aFlags) {
   AppleATDecoder* decoder = static_cast<AppleATDecoder*>(aAppleATDecoder);
-  MOZ_RELEASE_ASSERT(decoder->mTaskQueue->IsCurrentThreadIn());
+  MOZ_RELEASE_ASSERT(decoder->mThread->IsOnCurrentThread());
 
   LOGEX(decoder, "MetadataCallback receiving: '%s'", FourCC2Str(aProperty));
   if (aProperty == kAudioFileStreamProperty_MagicCookieData) {
@@ -606,7 +626,7 @@ static void _SampleCallback(void* aSBR, UInt32 aNumBytes, UInt32 aNumPackets,
 
 nsresult AppleATDecoder::GetImplicitAACMagicCookie(
     const MediaRawData* aSample) {
-  MOZ_ASSERT(mTaskQueue->IsCurrentThreadIn());
+  MOZ_ASSERT(mThread->IsOnCurrentThread());
 
   // Prepend ADTS header to AAC audio.
   RefPtr<MediaRawData> adtssample(aSample->Clone());
