@@ -27,10 +27,11 @@ extern LazyLogModule sPEMLog;
   MOZ_LOG(sPEMLog, mozilla::LogLevel::Debug, \
           ("[AppleVTEncoder] %s: " fmt, __func__, ##__VA_ARGS__))
 
-static CFDictionaryRef BuildEncoderSpec() {
+static CFDictionaryRef BuildEncoderSpec(const bool aHardwareNotAllowed) {
   const void* keys[] = {
       kVTVideoEncoderSpecification_EnableHardwareAcceleratedVideoEncoder};
-  const void* values[] = {kCFBooleanTrue};
+  const void* values[] = {aHardwareNotAllowed ? kCFBooleanFalse
+                                              : kCFBooleanTrue};
 
   static_assert(ArrayLength(keys) == ArrayLength(values),
                 "Non matching keys/values array size");
@@ -92,7 +93,11 @@ static bool SetProfileLevel(VTCompressionSessionRef& aSession,
 RefPtr<MediaDataEncoder::InitPromise> AppleVTEncoder::Init() {
   MOZ_ASSERT(!mInited, "Cannot initialize encoder again without shutting down");
 
-  AutoCFRelease<CFDictionaryRef> spec(BuildEncoderSpec());
+  if (mConfig.mSize.width == 0 || mConfig.mSize.height == 0) {
+    return InitPromise::CreateAndReject(NS_ERROR_ILLEGAL_VALUE, __func__);
+  }
+
+  AutoCFRelease<CFDictionaryRef> spec(BuildEncoderSpec(mHardwareNotAllowed));
   AutoCFRelease<CFDictionaryRef> srcBufferAttr(
       BuildSourceImageBufferAttributes());
   if (!srcBufferAttr) {
@@ -129,6 +134,23 @@ RefPtr<MediaDataEncoder::InitPromise> AppleVTEncoder::Init() {
         __func__);
   }
 
+  int64_t interval =
+      mConfig.mKeyframeInterval > std::numeric_limits<int64_t>::max()
+          ? std::numeric_limits<int64_t>::max()
+          : mConfig.mKeyframeInterval;
+  AutoCFRelease<CFNumberRef> cf(
+      CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt64Type, &interval));
+  if (VTSessionSetProperty(mSession,
+                           kVTCompressionPropertyKey_MaxKeyFrameInterval,
+                           cf) != noErr) {
+    return InitPromise::CreateAndReject(
+        MediaResult(
+            NS_ERROR_DOM_MEDIA_FATAL_ERR,
+            nsPrintfCString("fail to configurate keyframe interval:%" PRId64,
+                            interval)),
+        __func__);
+  }
+
   if (mConfig.mCodecSpecific) {
     const H264Specific& specific = mConfig.mCodecSpecific.ref();
     if (!SetProfileLevel(mSession, specific.mProfileLevel)) {
@@ -138,20 +160,6 @@ RefPtr<MediaDataEncoder::InitPromise> AppleVTEncoder::Init() {
                                       specific.mProfileLevel)),
           __func__);
     }
-
-    int64_t interval = specific.mKeyframeInterval;
-    AutoCFRelease<CFNumberRef> cf(
-        CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt64Type, &interval));
-    if (VTSessionSetProperty(mSession,
-                             kVTCompressionPropertyKey_MaxKeyFrameInterval,
-                             cf) != noErr) {
-      return InitPromise::CreateAndReject(
-          MediaResult(
-              NS_ERROR_DOM_MEDIA_FATAL_ERR,
-              nsPrintfCString("fail to configurate keyframe interval:%" PRId64,
-                              interval)),
-          __func__);
-    }
   }
 
   CFBooleanRef isUsingHW = nullptr;
@@ -159,6 +167,9 @@ RefPtr<MediaDataEncoder::InitPromise> AppleVTEncoder::Init() {
       mSession, kVTCompressionPropertyKey_UsingHardwareAcceleratedVideoEncoder,
       kCFAllocatorDefault, &isUsingHW);
   mIsHardwareAccelerated = status == noErr && isUsingHW == kCFBooleanTrue;
+  if (isUsingHW) {
+    CFRelease(isUsingHW);
+  }
 
   mError = NS_OK;
   return InitPromise::CreateAndResolve(TrackInfo::TrackType::kVideoTrack,
@@ -508,6 +519,11 @@ CVPixelBufferRef AppleVTEncoder::CreateCVPixelBuffer(const Image* aSource) {
   OSType format = MapPixelFormat(mConfig.mSourcePixelFormat).ref();
   size_t numPlanes = NumberOfPlanes(mConfig.mSourcePixelFormat);
   const PlanarYCbCrImage::Data* yuv = image->GetData();
+  if (!yuv) {
+    return nullptr;
+  }
+  auto ySize = yuv->YDataSize();
+  auto cbcrSize = yuv->CbCrDataSize();
   void* addresses[3] = {};
   size_t widths[3] = {};
   size_t heights[3] = {};
@@ -515,20 +531,20 @@ CVPixelBufferRef AppleVTEncoder::CreateCVPixelBuffer(const Image* aSource) {
   switch (numPlanes) {
     case 3:
       addresses[2] = yuv->mCrChannel;
-      widths[2] = yuv->mCbCrSize.width;
-      heights[2] = yuv->mCbCrSize.height;
+      widths[2] = cbcrSize.width;
+      heights[2] = cbcrSize.height;
       strides[2] = yuv->mCbCrStride;
       [[fallthrough]];
     case 2:
       addresses[1] = yuv->mCbChannel;
-      widths[1] = yuv->mCbCrSize.width;
-      heights[1] = yuv->mCbCrSize.height;
+      widths[1] = cbcrSize.width;
+      heights[1] = cbcrSize.height;
       strides[1] = yuv->mCbCrStride;
       [[fallthrough]];
     case 1:
       addresses[0] = yuv->mYChannel;
-      widths[0] = yuv->mYSize.width;
-      heights[0] = yuv->mYSize.height;
+      widths[0] = ySize.width;
+      heights[0] = ySize.height;
       strides[0] = yuv->mYStride;
       break;
     default:
@@ -538,9 +554,9 @@ CVPixelBufferRef AppleVTEncoder::CreateCVPixelBuffer(const Image* aSource) {
   CVPixelBufferRef buffer = nullptr;
   image->AddRef();  // Grip input buffers.
   CVReturn rv = CVPixelBufferCreateWithPlanarBytes(
-      kCFAllocatorDefault, yuv->mPicSize.width, yuv->mPicSize.height, format,
-      nullptr /* dataPtr */, 0 /* dataSize */, numPlanes, addresses, widths,
-      heights, strides, ReleaseImage /* releaseCallback */,
+      kCFAllocatorDefault, yuv->mPictureRect.width, yuv->mPictureRect.height,
+      format, nullptr /* dataPtr */, 0 /* dataSize */, numPlanes, addresses,
+      widths, heights, strides, ReleaseImage /* releaseCallback */,
       image /* releaseRefCon */, nullptr /* pixelBufferAttributes */, &buffer);
   if (rv == kCVReturnSuccess) {
     return buffer;
