@@ -6,29 +6,32 @@
 
 #include <string.h>
 #ifdef __GNUC__
-#include <unistd.h>
+#  include <unistd.h>
 #endif
 
-#include "FFmpegLog.h"
 #include "FFmpegDataDecoder.h"
+#include "FFmpegLog.h"
+#include "mozilla/StaticPrefs_media.h"
 #include "mozilla/TaskQueue.h"
 #include "prsystem.h"
+#include "VideoUtils.h"
 
 namespace mozilla {
 
-StaticMutex FFmpegDataDecoder<LIBAV_VER>::sMonitor;
+StaticMutex FFmpegDataDecoder<LIBAV_VER>::sMutex;
 
 FFmpegDataDecoder<LIBAV_VER>::FFmpegDataDecoder(FFmpegLibWrapper* aLib,
-                                                TaskQueue* aTaskQueue,
                                                 AVCodecID aCodecID)
     : mLib(aLib),
       mCodecContext(nullptr),
       mCodecParser(nullptr),
-      mFrame(NULL),
+      mFrame(nullptr),
       mExtraData(nullptr),
       mCodecID(aCodecID),
-      mTaskQueue(aTaskQueue),
-      mLastInputDts(media::TimeUnit::FromMicroseconds(INT64_MIN)) {
+      mTaskQueue(TaskQueue::Create(
+          GetMediaThreadPool(MediaThreadType::PLATFORM_DECODER),
+          "FFmpegDataDecoder")),
+      mLastInputDts(media::TimeUnit::FromNegativeInfinity()) {
   MOZ_ASSERT(aLib);
   MOZ_COUNT_CTOR(FFmpegDataDecoder);
 }
@@ -41,33 +44,7 @@ FFmpegDataDecoder<LIBAV_VER>::~FFmpegDataDecoder() {
   }
 }
 
-MediaResult FFmpegDataDecoder<LIBAV_VER>::InitDecoder() {
-  FFMPEG_LOG("Initialising FFmpeg decoder.");
-
-  AVCodec* codec = FindAVCodec(mLib, mCodecID);
-  if (!codec) {
-    return MediaResult(NS_ERROR_DOM_MEDIA_FATAL_ERR,
-                       RESULT_DETAIL("Couldn't find ffmpeg decoder"));
-  }
-
-  StaticMutexAutoLock mon(sMonitor);
-
-  if (!(mCodecContext = mLib->avcodec_alloc_context3(codec))) {
-    return MediaResult(NS_ERROR_OUT_OF_MEMORY,
-                       RESULT_DETAIL("Couldn't init ffmpeg context"));
-  }
-
-  if (NeedParser()) {
-    MOZ_ASSERT(mCodecParser == nullptr);
-    mCodecParser = mLib->av_parser_init(mCodecID);
-    if (mCodecParser) {
-      mCodecParser->flags |= ParserFlags();
-    }
-  }
-  mCodecContext->opaque = this;
-
-  InitCodecContext();
-
+MediaResult FFmpegDataDecoder<LIBAV_VER>::AllocateExtraData() {
   if (mExtraData) {
     mCodecContext->extradata_size = mExtraData->Length();
     // FFmpeg may use SIMD instructions to access the data which reads the
@@ -90,6 +67,57 @@ MediaResult FFmpegDataDecoder<LIBAV_VER>::InitDecoder() {
     mCodecContext->extradata_size = 0;
   }
 
+  return NS_OK;
+}
+
+// Note: This doesn't run on the ffmpeg TaskQueue, it runs on some other media
+// taskqueue
+MediaResult FFmpegDataDecoder<LIBAV_VER>::InitDecoder() {
+  FFMPEG_LOG("Initialising FFmpeg decoder");
+
+  AVCodec* codec = FindAVCodec(mLib, mCodecID);
+  if (!codec) {
+    FFMPEG_LOG("  couldn't find ffmpeg decoder for codec id %d", mCodecID);
+    return MediaResult(NS_ERROR_DOM_MEDIA_FATAL_ERR,
+                       RESULT_DETAIL("unable to find codec"));
+  }
+  // openh264 has broken decoding of some h264 videos so
+  // don't use it unless explicitly allowed for now.
+  if (!strcmp(codec->name, "libopenh264") &&
+      !StaticPrefs::media_ffmpeg_allow_openh264()) {
+    FFMPEG_LOG("  unable to find codec (openh264 disabled by pref)");
+    return MediaResult(
+        NS_ERROR_DOM_MEDIA_FATAL_ERR,
+        RESULT_DETAIL("unable to find codec (openh264 disabled by pref)"));
+  }
+  FFMPEG_LOG("  codec %s : %s", codec->name, codec->long_name);
+
+  StaticMutexAutoLock mon(sMutex);
+
+  if (!(mCodecContext = mLib->avcodec_alloc_context3(codec))) {
+    FFMPEG_LOG("  couldn't allocate ffmpeg context for codec %s", codec->name);
+    return MediaResult(NS_ERROR_OUT_OF_MEMORY,
+                       RESULT_DETAIL("Couldn't init ffmpeg context"));
+  }
+
+  if (NeedParser()) {
+    MOZ_ASSERT(mCodecParser == nullptr);
+    mCodecParser = mLib->av_parser_init(mCodecID);
+    if (mCodecParser) {
+      mCodecParser->flags |= ParserFlags();
+    }
+  }
+  mCodecContext->opaque = this;
+
+  InitCodecContext();
+  MediaResult ret = AllocateExtraData();
+  if (NS_FAILED(ret)) {
+    FFMPEG_LOG("  couldn't allocate ffmpeg extra data for codec %s",
+               codec->name);
+    mLib->av_freep(&mCodecContext);
+    return ret;
+  }
+
 #if LIBAVCODEC_VERSION_MAJOR < 57
   if (codec->capabilities & CODEC_CAP_DR1) {
     mCodecContext->flags |= CODEC_FLAG_EMU_EDGE;
@@ -97,26 +125,22 @@ MediaResult FFmpegDataDecoder<LIBAV_VER>::InitDecoder() {
 #endif
 
   if (mLib->avcodec_open2(mCodecContext, codec, nullptr) < 0) {
-    mLib->avcodec_close(mCodecContext);
     mLib->av_freep(&mCodecContext);
+    FFMPEG_LOG("  Couldn't open avcodec");
     return MediaResult(NS_ERROR_DOM_MEDIA_FATAL_ERR,
-                       RESULT_DETAIL("Couldn't initialise ffmpeg decoder"));
+                       RESULT_DETAIL("Couldn't open avcodec"));
   }
 
-  FFMPEG_LOG("FFmpeg init successful.");
+  FFMPEG_LOG("  FFmpeg decoder init successful.");
   return NS_OK;
 }
 
 RefPtr<ShutdownPromise> FFmpegDataDecoder<LIBAV_VER>::Shutdown() {
-  if (mTaskQueue) {
-    RefPtr<FFmpegDataDecoder<LIBAV_VER>> self = this;
-    return InvokeAsync(mTaskQueue, __func__, [self]() {
-      self->ProcessShutdown();
-      return ShutdownPromise::CreateAndResolve(true, __func__);
-    });
-  }
-  ProcessShutdown();
-  return ShutdownPromise::CreateAndResolve(true, __func__);
+  RefPtr<FFmpegDataDecoder<LIBAV_VER>> self = this;
+  return InvokeAsync(mTaskQueue, __func__, [self]() {
+    self->ProcessShutdown();
+    return self->mTaskQueue->BeginShutdown();
+  });
 }
 
 RefPtr<MediaDataDecoder::DecodePromise> FFmpegDataDecoder<LIBAV_VER>::Decode(
@@ -127,28 +151,32 @@ RefPtr<MediaDataDecoder::DecodePromise> FFmpegDataDecoder<LIBAV_VER>::Decode(
 
 RefPtr<MediaDataDecoder::DecodePromise>
 FFmpegDataDecoder<LIBAV_VER>::ProcessDecode(MediaRawData* aSample) {
+  MOZ_ASSERT(mTaskQueue->IsOnCurrentThread());
+  PROCESS_DECODE_LOG(aSample);
   bool gotFrame = false;
   DecodedData results;
   MediaResult rv = DoDecode(aSample, &gotFrame, results);
   if (NS_FAILED(rv)) {
     return DecodePromise::CreateAndReject(rv, __func__);
   }
-  return DecodePromise::CreateAndResolve(Move(results), __func__);
+  return DecodePromise::CreateAndResolve(std::move(results), __func__);
 }
 
 MediaResult FFmpegDataDecoder<LIBAV_VER>::DoDecode(
     MediaRawData* aSample, bool* aGotFrame,
     MediaDataDecoder::DecodedData& aResults) {
+  MOZ_ASSERT(mTaskQueue->IsOnCurrentThread());
+
   uint8_t* inputData = const_cast<uint8_t*>(aSample->Data());
   size_t inputSize = aSample->Size();
 
   mLastInputDts = aSample->mTimecode;
 
-  if (mCodecParser) {
+  if (inputData && mCodecParser) {  // inputData is null when draining.
     if (aGotFrame) {
       *aGotFrame = false;
     }
-    do {
+    while (inputSize) {
       uint8_t* data = inputData;
       int size = inputSize;
       int len = mLib->av_parser_parse2(
@@ -158,7 +186,7 @@ MediaResult FFmpegDataDecoder<LIBAV_VER>::DoDecode(
       if (size_t(len) > inputSize) {
         return NS_ERROR_DOM_MEDIA_DECODE_ERR;
       }
-      if (size || !inputSize) {
+      if (size) {
         bool gotFrame = false;
         MediaResult rv = DoDecode(aSample, data, size, &gotFrame, aResults);
         if (NS_FAILED(rv)) {
@@ -170,7 +198,7 @@ MediaResult FFmpegDataDecoder<LIBAV_VER>::DoDecode(
       }
       inputData += len;
       inputSize -= len;
-    } while (inputSize > 0);
+    }
     return NS_OK;
   }
   return DoDecode(aSample, inputData, inputSize, aGotFrame, aResults);
@@ -188,22 +216,28 @@ RefPtr<MediaDataDecoder::DecodePromise> FFmpegDataDecoder<LIBAV_VER>::Drain() {
 
 RefPtr<MediaDataDecoder::DecodePromise>
 FFmpegDataDecoder<LIBAV_VER>::ProcessDrain() {
+  MOZ_ASSERT(mTaskQueue->IsOnCurrentThread());
   RefPtr<MediaRawData> empty(new MediaRawData());
   empty->mTimecode = mLastInputDts;
   bool gotFrame = false;
   DecodedData results;
+  // When draining the FFmpeg decoder will return either a single frame at a
+  // time until gotFrame is set to false; or return a block of frames with
+  // NS_ERROR_DOM_MEDIA_END_OF_STREAM
   while (NS_SUCCEEDED(DoDecode(empty, &gotFrame, results)) && gotFrame) {
   }
-  return DecodePromise::CreateAndResolve(Move(results), __func__);
+  return DecodePromise::CreateAndResolve(std::move(results), __func__);
 }
 
 RefPtr<MediaDataDecoder::FlushPromise>
 FFmpegDataDecoder<LIBAV_VER>::ProcessFlush() {
-  MOZ_ASSERT(mTaskQueue->IsCurrentThreadIn());
+  MOZ_ASSERT(mTaskQueue->IsOnCurrentThread());
   if (mCodecContext) {
+    FFMPEG_LOG("FFmpegDataDecoder: flushing buffers");
     mLib->avcodec_flush_buffers(mCodecContext);
   }
   if (mCodecParser) {
+    FFMPEG_LOG("FFmpegDataDecoder: reinitializing parser");
     mLib->av_parser_close(mCodecParser);
     mCodecParser = mLib->av_parser_init(mCodecID);
   }
@@ -211,9 +245,11 @@ FFmpegDataDecoder<LIBAV_VER>::ProcessFlush() {
 }
 
 void FFmpegDataDecoder<LIBAV_VER>::ProcessShutdown() {
-  StaticMutexAutoLock mon(sMonitor);
+  MOZ_ASSERT(mTaskQueue->IsOnCurrentThread());
+  StaticMutexAutoLock mon(sMutex);
 
   if (mCodecContext) {
+    FFMPEG_LOG("FFmpegDataDecoder: shutdown");
     if (mCodecContext->extradata) {
       mLib->av_freep(&mCodecContext->extradata);
     }
@@ -230,7 +266,7 @@ void FFmpegDataDecoder<LIBAV_VER>::ProcessShutdown() {
 }
 
 AVFrame* FFmpegDataDecoder<LIBAV_VER>::PrepareFrame() {
-  MOZ_ASSERT(mTaskQueue->IsCurrentThreadIn());
+  MOZ_ASSERT(mTaskQueue->IsOnCurrentThread());
 #if LIBAVCODEC_VERSION_MAJOR >= 55
   if (mFrame) {
     mLib->av_frame_unref(mFrame);
@@ -255,4 +291,18 @@ AVFrame* FFmpegDataDecoder<LIBAV_VER>::PrepareFrame() {
   return aLib->avcodec_find_decoder(aCodec);
 }
 
+#ifdef MOZ_WAYLAND
+/* static */ AVCodec* FFmpegDataDecoder<LIBAV_VER>::FindHardwareAVCodec(
+    FFmpegLibWrapper* aLib, AVCodecID aCodec) {
+  void* opaque = nullptr;
+  while (AVCodec* codec = aLib->av_codec_iterate(&opaque)) {
+    if (codec->id == aCodec && aLib->av_codec_is_decoder(codec) &&
+        aLib->avcodec_get_hw_config(codec, 0)) {
+      return codec;
+    }
+  }
+  return nullptr;
+}
+#endif
+
 }  // namespace mozilla
