@@ -11,8 +11,8 @@
 
 #include "mozilla/Assertions.h"
 #include "mozilla/EndianUtils.h"
+#include "mozilla/Utf8.h"
 #include "BufferReader.h"
-#include "nsAutoPtr.h"
 #include "VideoUtils.h"
 #include "TimeUnits.h"
 
@@ -65,10 +65,12 @@ WAVTrackDemuxer::WAVTrackDemuxer(MediaResource* aSource)
       mFirstChunkOffset(0),
       mNumParsedChunks(0),
       mChunkIndex(0),
+      mDataLength(0),
       mTotalChunkLen(0),
       mSamplesPerChunk(0),
       mSamplesPerSecond(0),
-      mChannels(0) {
+      mChannels(0),
+      mSampleFormat(0) {
   DDLINKCHILD("source", aSource);
   Reset();
 }
@@ -79,6 +81,8 @@ bool WAVTrackDemuxer::Init() {
 
   if (!mInfo) {
     mInfo = MakeUnique<AudioInfo>();
+    mInfo->mCodecSpecificConfig =
+        AudioCodecSpecificVariant{WaveCodecSpecificData{}};
   }
 
   if (!RIFFParserInit()) {
@@ -122,8 +126,16 @@ bool WAVTrackDemuxer::Init() {
     mHeaderParser.Reset();
   }
 
-  if (mDataLength > StreamLength() - mFirstChunkOffset) {
-    mDataLength = StreamLength() - mFirstChunkOffset;
+  int64_t streamLength = StreamLength();
+  // If the chunk length and the resource length are not equal, use the
+  // resource length as the "real" data chunk length, if it's longer than the
+  // chunk size.
+  if (streamLength != -1) {
+    uint64_t streamLengthPositive = static_cast<uint64_t>(streamLength);
+    if (streamLengthPositive > mFirstChunkOffset &&
+        mDataLength > streamLengthPositive - mFirstChunkOffset) {
+      mDataLength = streamLengthPositive - mFirstChunkOffset;
+    }
   }
 
   mSamplesPerSecond = mFmtParser.FmtChunk().SampleRate();
@@ -228,27 +240,23 @@ bool WAVTrackDemuxer::ListChunkParserInit(uint32_t aChunkSize) {
 
     bytesRead += length;
 
-    if (!IsUTF8(val)) {
+    if (!IsUtf8(val)) {
       mHeaderParser.Reset();
       continue;
     }
 
     switch (id) {
       case 0x49415254:  // IART
-        mInfo->mTags.AppendElement(
-            MetadataTag(NS_LITERAL_CSTRING("artist"), val));
+        mInfo->mTags.AppendElement(MetadataTag("artist"_ns, val));
         break;
       case 0x49434d54:  // ICMT
-        mInfo->mTags.AppendElement(
-            MetadataTag(NS_LITERAL_CSTRING("comments"), val));
+        mInfo->mTags.AppendElement(MetadataTag("comments"_ns, val));
         break;
       case 0x49474e52:  // IGNR
-        mInfo->mTags.AppendElement(
-            MetadataTag(NS_LITERAL_CSTRING("genre"), val));
+        mInfo->mTags.AppendElement(MetadataTag("genre"_ns, val));
         break;
       case 0x494e414d:  // INAM
-        mInfo->mTags.AppendElement(
-            MetadataTag(NS_LITERAL_CSTRING("name"), val));
+        mInfo->mTags.AppendElement(MetadataTag("name"_ns, val));
         break;
     }
 
@@ -288,7 +296,7 @@ TimeUnit WAVTrackDemuxer::FastSeek(const TimeUnit& aTime) {
   mOffset = OffsetFromChunkIndex(mChunkIndex);
 
   if (mOffset > mFirstChunkOffset && StreamLength() > 0) {
-    mOffset = std::min(StreamLength() - 1, mOffset);
+    mOffset = std::min(static_cast<uint64_t>(StreamLength() - 1), mOffset);
   }
 
   return Duration(mChunkIndex);
@@ -317,10 +325,14 @@ RefPtr<WAVTrackDemuxer::SamplesPromise> WAVTrackDemuxer::GetSamples(
     if (!datachunk) {
       break;
     }
-    datachunks->mSamples.AppendElement(datachunk);
+    if (!datachunk->HasValidTime()) {
+      return SamplesPromise::CreateAndReject(NS_ERROR_DOM_MEDIA_DEMUXER_ERR,
+                                             __func__);
+    }
+    datachunks->AppendSample(datachunk);
   }
 
-  if (datachunks->mSamples.IsEmpty()) {
+  if (datachunks->GetSamples().IsEmpty()) {
     return SamplesPromise::CreateAndReject(NS_ERROR_DOM_MEDIA_END_OF_STREAM,
                                            __func__);
   }
@@ -388,9 +400,9 @@ TimeUnit WAVTrackDemuxer::DurationFromBytes(uint32_t aNumBytes) const {
     return TimeUnit();
   }
 
-  int64_t numSamples = aNumBytes * 8 / mChannels / mSampleFormat;
+  uint64_t numSamples = aNumBytes * 8 / mChannels / mSampleFormat;
 
-  int64_t numUSeconds = USECS_PER_S * numSamples / mSamplesPerSecond;
+  uint64_t numUSeconds = USECS_PER_S * numSamples / mSamplesPerSecond;
 
   if (USECS_PER_S * numSamples % mSamplesPerSecond > mSamplesPerSecond / 2) {
     numUSeconds++;
@@ -439,8 +451,8 @@ already_AddRefed<MediaRawData> WAVTrackDemuxer::GetNextChunk(
   RefPtr<MediaRawData> datachunk = new MediaRawData();
   datachunk->mOffset = aRange.mStart;
 
-  nsAutoPtr<MediaRawDataWriter> chunkWriter(datachunk->CreateWriter());
-  if (!chunkWriter->SetSize(aRange.Length())) {
+  UniquePtr<MediaRawDataWriter> chunkWriter(datachunk->CreateWriter());
+  if (!chunkWriter->SetSize(static_cast<uint32_t>(aRange.Length()))) {
     return nullptr;
   }
 
@@ -481,8 +493,8 @@ already_AddRefed<MediaRawData> WAVTrackDemuxer::GetFileHeader(
   RefPtr<MediaRawData> fileHeader = new MediaRawData();
   fileHeader->mOffset = aRange.mStart;
 
-  nsAutoPtr<MediaRawDataWriter> headerWriter(fileHeader->CreateWriter());
-  if (!headerWriter->SetSize(aRange.Length())) {
+  UniquePtr<MediaRawDataWriter> headerWriter(fileHeader->CreateWriter());
+  if (!headerWriter->SetSize(static_cast<uint32_t>(aRange.Length()))) {
     return nullptr;
   }
 
@@ -498,42 +510,38 @@ already_AddRefed<MediaRawData> WAVTrackDemuxer::GetFileHeader(
   return fileHeader.forget();
 }
 
-int64_t WAVTrackDemuxer::OffsetFromChunkIndex(int64_t aChunkIndex) const {
+uint64_t WAVTrackDemuxer::OffsetFromChunkIndex(uint32_t aChunkIndex) const {
   return mFirstChunkOffset + aChunkIndex * DATA_CHUNK_SIZE;
 }
 
-int64_t WAVTrackDemuxer::ChunkIndexFromOffset(int64_t aOffset) const {
-  int64_t chunkIndex = (aOffset - mFirstChunkOffset) / DATA_CHUNK_SIZE;
-  return std::max<int64_t>(0, chunkIndex);
-}
-
-int64_t WAVTrackDemuxer::ChunkIndexFromTime(
+uint64_t WAVTrackDemuxer::ChunkIndexFromTime(
     const media::TimeUnit& aTime) const {
   if (!mSamplesPerChunk || !mSamplesPerSecond) {
     return 0;
   }
-  int64_t chunkIndex =
+  uint64_t chunkIndex =
       (aTime.ToSeconds() * mSamplesPerSecond / mSamplesPerChunk) - 1;
   return chunkIndex;
 }
 
 void WAVTrackDemuxer::UpdateState(const MediaByteRange& aRange) {
   // Full chunk parsed, move offset to its end.
-  mOffset = aRange.mEnd;
-  mTotalChunkLen += aRange.Length();
+  mOffset = static_cast<uint32_t>(aRange.mEnd);
+  mTotalChunkLen += static_cast<uint64_t>(aRange.Length());
 }
 
-int32_t WAVTrackDemuxer::Read(uint8_t* aBuffer, int64_t aOffset,
-                              int32_t aSize) {
+uint32_t WAVTrackDemuxer::Read(uint8_t* aBuffer, int64_t aOffset,
+                               int32_t aSize) {
   const int64_t streamLen = StreamLength();
   if (mInfo && streamLen > 0) {
-    aSize = std::min<int64_t>(aSize, streamLen - aOffset);
+    int64_t max = streamLen > aOffset ? streamLen - aOffset : 0;
+    aSize = std::min<int64_t>(aSize, max);
   }
   uint32_t read = 0;
   const nsresult rv = mSource.ReadAt(aOffset, reinterpret_cast<char*>(aBuffer),
                                      static_cast<uint32_t>(aSize), &read);
   NS_ENSURE_SUCCESS(rv, 0);
-  return static_cast<int32_t>(read);
+  return read;
 }
 
 // RIFFParser
@@ -634,11 +642,13 @@ bool HeaderParser::ChunkHeader::IsValid() const {
 }
 
 uint32_t HeaderParser::ChunkHeader::ChunkName() const {
-  return ((mRaw[0] << 24) | (mRaw[1] << 16) | (mRaw[2] << 8) | (mRaw[3]));
+  return static_cast<uint32_t>(
+      ((mRaw[0] << 24) | (mRaw[1] << 16) | (mRaw[2] << 8) | (mRaw[3])));
 }
 
 uint32_t HeaderParser::ChunkHeader::ChunkSize() const {
-  return ((mRaw[7] << 24) | (mRaw[6] << 16) | (mRaw[5] << 8) | (mRaw[4]));
+  return static_cast<uint32_t>(
+      ((mRaw[7] << 24) | (mRaw[6] << 16) | (mRaw[5] << 8) | (mRaw[4])));
 }
 
 void HeaderParser::ChunkHeader::Update(uint8_t c) {
@@ -686,7 +696,8 @@ uint16_t FormatParser::FormatChunk::Channels() const {
 }
 
 uint32_t FormatParser::FormatChunk::SampleRate() const {
-  return (mRaw[7] << 24) | (mRaw[6] << 16) | (mRaw[5] << 8) | (mRaw[4]);
+  return static_cast<uint32_t>((mRaw[7] << 24) | (mRaw[6] << 16) |
+                               (mRaw[5] << 8) | (mRaw[4]));
 }
 
 uint16_t FormatParser::FormatChunk::FrameSize() const {
@@ -715,7 +726,7 @@ void FormatParser::FormatChunk::Update(uint8_t c) {
 
 // DataParser
 
-DataParser::DataParser() {}
+DataParser::DataParser() = default;
 
 void DataParser::Reset() { mChunk.Reset(); }
 
