@@ -5,27 +5,27 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "nsError.h"
-#include "MediaDecoderStateMachine.h"
 #include "MediaResource.h"
 #ifdef MOZ_AV1
-#include "AOMDecoder.h"
+#  include "AOMDecoder.h"
 #endif
 #include "OpusDecoder.h"
 #include "VPXDecoder.h"
 #include "WebMDemuxer.h"
 #include "WebMBufferedParser.h"
 #include "gfx2DGlue.h"
-#include "mozilla/Atomics.h"
+#include "gfxUtils.h"
 #include "mozilla/EndianUtils.h"
+#include "mozilla/Maybe.h"
 #include "mozilla/SharedThreadPool.h"
 #include "MediaDataDemuxer.h"
-#include "nsAutoPtr.h"
 #include "nsAutoRef.h"
 #include "NesteggPacketHolder.h"
 #include "XiphExtradata.h"
 #include "prprf.h"  // leaving it for PR_vsnprintf()
 #include "mozilla/IntegerPrintfMacros.h"
 #include "mozilla/Sprintf.h"
+#include "VideoUtils.h"
 
 #include <algorithm>
 #include <numeric>
@@ -48,8 +48,6 @@ LazyLogModule gNesteggLog("Nestegg");
 // files encountered appear to have keyframes located < 4s.
 #define MAX_LOOK_AHEAD 10000000
 
-static Atomic<uint32_t> sStreamSourceID(0u);
-
 // Functions for reading and seeking using WebMDemuxer required for
 // nestegg_io. The 'user data' passed to these functions is the
 // demuxer.
@@ -154,7 +152,9 @@ int WebMDemuxer::NestEggContext::Init() {
 WebMDemuxer::WebMDemuxer(MediaResource* aResource)
     : WebMDemuxer(aResource, false) {}
 
-WebMDemuxer::WebMDemuxer(MediaResource* aResource, bool aIsMediaSource)
+WebMDemuxer::WebMDemuxer(
+    MediaResource* aResource, bool aIsMediaSource,
+    Maybe<media::TimeUnit> aFrameEndTimeBeforeRecreateDemuxer)
     : mVideoContext(this, aResource),
       mAudioContext(this, aResource),
       mBufferedState(nullptr),
@@ -173,6 +173,14 @@ WebMDemuxer::WebMDemuxer(MediaResource* aResource, bool aIsMediaSource)
   // Audio/video contexts hold a MediaResourceIndex.
   DDLINKCHILD("video context", mVideoContext.GetResource());
   DDLINKCHILD("audio context", mAudioContext.GetResource());
+
+  MOZ_ASSERT_IF(!aIsMediaSource,
+                aFrameEndTimeBeforeRecreateDemuxer.isNothing());
+  if (aIsMediaSource && aFrameEndTimeBeforeRecreateDemuxer) {
+    mVideoFrameEndTimeBeforeReset = aFrameEndTimeBeforeRecreateDemuxer;
+    WEBM_DEBUG("Set mVideoFrameEndTimeBeforeReset=%" PRId64,
+               mVideoFrameEndTimeBeforeReset->ToMicroseconds());
+  }
 }
 
 WebMDemuxer::~WebMDemuxer() {
@@ -237,21 +245,22 @@ already_AddRefed<MediaTrackDemuxer> WebMDemuxer::GetTrackDemuxer(
   return e.forget();
 }
 
-nsresult WebMDemuxer::Reset(TrackInfo::TrackType aType) {
+void WebMDemuxer::Reset(TrackInfo::TrackType aType) {
   if (aType == TrackInfo::kVideoTrack) {
     mVideoPackets.Reset();
   } else {
     mAudioPackets.Reset();
   }
-  return NS_OK;
 }
 
 nsresult WebMDemuxer::ReadMetadata() {
   int r = mVideoContext.Init();
   if (r == -1) {
+    WEBM_DEBUG("mVideoContext::Init failure");
     return NS_ERROR_FAILURE;
   }
   if (mAudioContext.Init() == -1) {
+    WEBM_DEBUG("mAudioContext::Init failure");
     return NS_ERROR_FAILURE;
   }
 
@@ -263,10 +272,12 @@ nsresult WebMDemuxer::ReadMetadata() {
     // Check how much data nestegg read and force feed it to BufferedState.
     RefPtr<MediaByteBuffer> buffer = resource.MediaReadAt(0, resource.Tell());
     if (!buffer) {
+      WEBM_DEBUG("resource.MediaReadAt error");
       return NS_ERROR_FAILURE;
     }
     mBufferedState->NotifyDataArrived(buffer->Elements(), buffer->Length(), 0);
     if (mBufferedState->GetInitEndOffset() < 0) {
+      WEBM_DEBUG("Couldn't find init end");
       return NS_ERROR_FAILURE;
     }
     MOZ_ASSERT(mBufferedState->GetInitEndOffset() <= resource.Tell());
@@ -274,18 +285,21 @@ nsresult WebMDemuxer::ReadMetadata() {
   mInitData = resource.MediaReadAt(0, mBufferedState->GetInitEndOffset());
   if (!mInitData ||
       mInitData->Length() != size_t(mBufferedState->GetInitEndOffset())) {
+    WEBM_DEBUG("Couldn't read init data");
     return NS_ERROR_FAILURE;
   }
 
   unsigned int ntracks = 0;
   r = nestegg_track_count(context, &ntracks);
   if (r == -1) {
+    WEBM_DEBUG("nestegg_track_count error");
     return NS_ERROR_FAILURE;
   }
 
   for (unsigned int track = 0; track < ntracks; ++track) {
     int id = nestegg_track_codec_id(context, track);
     if (id == -1) {
+      WEBM_DEBUG("nestegg_track_codec_id error");
       return NS_ERROR_FAILURE;
     }
     int type = nestegg_track_type(context, track);
@@ -293,6 +307,7 @@ nsresult WebMDemuxer::ReadMetadata() {
       nestegg_video_params params;
       r = nestegg_track_video_params(context, track, &params);
       if (r == -1) {
+        WEBM_DEBUG("nestegg_track_video_params error");
         return NS_ERROR_FAILURE;
       }
       mVideoCodec = nestegg_track_codec_id(context, track);
@@ -310,6 +325,19 @@ nsresult WebMDemuxer::ReadMetadata() {
           NS_WARNING("Unknown WebM video codec");
           return NS_ERROR_FAILURE;
       }
+
+      mInfo.mVideo.mColorPrimaries = gfxUtils::CicpToColorPrimaries(
+          static_cast<gfx::CICP::ColourPrimaries>(params.primaries),
+          gMediaDemuxerLog);
+
+      // For VPX, this is our only chance to capture the transfer
+      // characteristics, which we can't get from a VPX bitstream later.
+      // We only need this value if the video is using the BT2020
+      // colorspace, which will be determined on a per-frame basis later.
+      mInfo.mVideo.mTransferFunction = gfxUtils::CicpToTransferFunction(
+          static_cast<gfx::CICP::TransferCharacteristics>(
+              params.transfer_characteristics));
+
       // Picture region, taking into account cropping, before scaling
       // to the display size.
       unsigned int cropH = params.crop_right + params.crop_left;
@@ -365,36 +393,54 @@ nsresult WebMDemuxer::ReadMetadata() {
       if (!r) {
         mInfo.mVideo.mDuration = TimeUnit::FromNanoseconds(duration);
       }
+      WEBM_DEBUG("stream duration: %lf\n", mInfo.mVideo.mDuration.ToSeconds());
       mInfo.mVideo.mCrypto = GetTrackCrypto(TrackInfo::kVideoTrack, track);
-      if (mInfo.mVideo.mCrypto.mValid) {
-        mCrypto.AddInitData(NS_LITERAL_STRING("webm"),
-                            mInfo.mVideo.mCrypto.mKeyId);
+      if (mInfo.mVideo.mCrypto.IsEncrypted()) {
+        MOZ_ASSERT(mInfo.mVideo.mCrypto.mCryptoScheme == CryptoScheme::Cenc,
+                   "WebM should only use cenc scheme");
+        mCrypto.AddInitData(u"webm"_ns, mInfo.mVideo.mCrypto.mKeyId);
       }
     } else if (type == NESTEGG_TRACK_AUDIO && !mHasAudio) {
       nestegg_audio_params params;
       r = nestegg_track_audio_params(context, track, &params);
       if (r == -1) {
+        WEBM_DEBUG("nestegg_track_audio_params error");
         return NS_ERROR_FAILURE;
       }
+      if (params.rate >
+              static_cast<decltype(params.rate)>(AudioInfo::MAX_RATE) ||
+          params.rate <= static_cast<decltype(params.rate)>(0) ||
+          params.channels > AudioConfig::ChannelLayout::MAX_CHANNELS) {
+        WEBM_DEBUG("Invalid audio param rate: %lf channel count: %d",
+                   params.rate, params.channels);
+        return NS_ERROR_DOM_MEDIA_METADATA_ERR;
+      }
 
       mAudioTrack = track;
       mHasAudio = true;
       mAudioCodec = nestegg_track_codec_id(context, track);
       if (mAudioCodec == NESTEGG_CODEC_VORBIS) {
+        mInfo.mAudio.mCodecSpecificConfig =
+            AudioCodecSpecificVariant{VorbisCodecSpecificData{}};
         mInfo.mAudio.mMimeType = "audio/vorbis";
       } else if (mAudioCodec == NESTEGG_CODEC_OPUS) {
+        uint64_t codecDelayUs = params.codec_delay / 1000;
         mInfo.mAudio.mMimeType = "audio/opus";
-        OpusDataDecoder::AppendCodecDelay(
-            mInfo.mAudio.mCodecSpecificConfig,
-            TimeUnit::FromNanoseconds(params.codec_delay).ToMicroseconds());
+        OpusCodecSpecificData opusCodecSpecificData;
+        opusCodecSpecificData.mContainerCodecDelayMicroSeconds =
+            AssertedCast<int64_t>(codecDelayUs);
+        mInfo.mAudio.mCodecSpecificConfig =
+            AudioCodecSpecificVariant{std::move(opusCodecSpecificData)};
+        WEBM_DEBUG("Preroll for Opus: %" PRIu64, codecDelayUs);
       }
       mSeekPreroll = params.seek_preroll;
-      mInfo.mAudio.mRate = params.rate;
+      mInfo.mAudio.mRate = AssertedCast<uint32_t>(params.rate);
       mInfo.mAudio.mChannels = params.channels;
 
       unsigned int nheaders = 0;
       r = nestegg_track_codec_data_count(context, track, &nheaders);
       if (r == -1) {
+        WEBM_DEBUG("nestegg_track_codec_data_count error");
         return NS_ERROR_FAILURE;
       }
 
@@ -405,6 +451,7 @@ nsresult WebMDemuxer::ReadMetadata() {
         size_t length = 0;
         r = nestegg_track_codec_data(context, track, header, &data, &length);
         if (r == -1) {
+          WEBM_DEBUG("nestegg_track_codec_data error");
           return NS_ERROR_FAILURE;
         }
         headers.AppendElement(data);
@@ -416,27 +463,33 @@ nsresult WebMDemuxer::ReadMetadata() {
       // TODO: This is already the format WebM stores them in. Would be nice
       // to avoid having libnestegg split them only for us to pack them again,
       // but libnestegg does not give us an API to access this data directly.
+      RefPtr<MediaByteBuffer> audioCodecSpecificBlob =
+          GetAudioCodecSpecificBlob(mInfo.mAudio.mCodecSpecificConfig);
       if (nheaders > 1) {
-        if (!XiphHeadersToExtradata(mInfo.mAudio.mCodecSpecificConfig, headers,
+        if (!XiphHeadersToExtradata(audioCodecSpecificBlob, headers,
                                     headerLens)) {
+          WEBM_DEBUG("Couldn't parse Xiph headers");
           return NS_ERROR_FAILURE;
         }
       } else {
-        mInfo.mAudio.mCodecSpecificConfig->AppendElements(headers[0],
-                                                          headerLens[0]);
+        audioCodecSpecificBlob->AppendElements(headers[0], headerLens[0]);
       }
       uint64_t duration = 0;
       r = nestegg_duration(context, &duration);
       if (!r) {
         mInfo.mAudio.mDuration = TimeUnit::FromNanoseconds(duration);
+        WEBM_DEBUG("audio track duration: %lf",
+                   mInfo.mAudio.mDuration.ToSeconds());
       }
       mInfo.mAudio.mCrypto = GetTrackCrypto(TrackInfo::kAudioTrack, track);
-      if (mInfo.mAudio.mCrypto.mValid) {
-        mCrypto.AddInitData(NS_LITERAL_STRING("webm"),
-                            mInfo.mAudio.mCrypto.mKeyId);
+      if (mInfo.mAudio.mCrypto.IsEncrypted()) {
+        MOZ_ASSERT(mInfo.mAudio.mCrypto.mCryptoScheme == CryptoScheme::Cenc,
+                   "WebM should only use cenc scheme");
+        mCrypto.AddInitData(u"webm"_ns, mInfo.mAudio.mCrypto.mKeyId);
       }
     }
   }
+  WEBM_DEBUG("Read metadata OK");
   return NS_OK;
 }
 
@@ -458,7 +511,7 @@ void WebMDemuxer::EnsureUpToDateIndex() {
       Resource(TrackInfo::kVideoTrack).GetResource());
   MediaByteRangeSet byteRanges;
   nsresult rv = resource->GetCachedRanges(byteRanges);
-  if (NS_FAILED(rv) || !byteRanges.Length()) {
+  if (NS_FAILED(rv) || byteRanges.IsEmpty()) {
     return;
   }
   mBufferedState->UpdateIndex(byteRanges, resource);
@@ -513,10 +566,10 @@ CryptoTrack WebMDemuxer::GetTrackCrypto(TrackInfo::TrackType aType,
   }
 
   if (!initData.IsEmpty()) {
-    crypto.mValid = true;
-    // crypto.mMode is not used for WebMs
+    // Webm only uses a cenc style scheme.
+    crypto.mCryptoScheme = CryptoScheme::Cenc;
     crypto.mIVSize = WEBM_IV_SIZE;
-    crypto.mKeyId = Move(initData);
+    crypto.mKeyId = std::move(initData);
   }
 
   return crypto;
@@ -544,50 +597,64 @@ nsresult WebMDemuxer::GetNextPacket(TrackInfo::TrackType aType,
   }
   int64_t tstamp = holder->Timestamp();
   int64_t duration = holder->Duration();
+  int64_t defaultDuration = holder->DefaultDuration();
+  if (aType == TrackInfo::TrackType::kVideoTrack) {
+    WEBM_DEBUG("GetNextPacket(video): tstamp=%" PRId64 ", duration=%" PRId64
+               ", defaultDuration=%" PRId64,
+               tstamp, duration, defaultDuration);
+  }
 
   // The end time of this frame is the start time of the next frame. Fetch
   // the timestamp of the next packet for this track.  If we've reached the
   // end of the resource, use the file's duration as the end time of this
   // video frame.
+  RefPtr<NesteggPacketHolder> next_holder;
+  rv = NextPacket(aType, next_holder);
+  if (NS_FAILED(rv) && rv != NS_ERROR_DOM_MEDIA_END_OF_STREAM) {
+    return rv;
+  }
+
   int64_t next_tstamp = INT64_MIN;
+  auto calculateNextTimestamp = [&](auto&& pushPacket, auto&& lastFrameTime,
+                                    int64_t trackEndTime) {
+    if (next_holder) {
+      next_tstamp = next_holder->Timestamp();
+      (this->*pushPacket)(next_holder);
+    } else if (duration >= 0) {
+      next_tstamp = tstamp + duration;
+    } else if (lastFrameTime.isSome()) {
+      next_tstamp = tstamp + (tstamp - lastFrameTime.ref());
+    } else if (defaultDuration >= 0) {
+      next_tstamp = tstamp + defaultDuration;
+    } else if (mVideoFrameEndTimeBeforeReset) {
+      WEBM_DEBUG("Setting next timestamp to be %" PRId64 " us",
+                 mVideoFrameEndTimeBeforeReset->ToMicroseconds());
+      next_tstamp = mVideoFrameEndTimeBeforeReset->ToMicroseconds();
+    } else if (mIsMediaSource) {
+      (this->*pushPacket)(holder);
+    } else {
+      // If we can't get frame's duration, it means either we need to wait for
+      // more data for MSE case or this is the last frame for file resource
+      // case.
+      if (tstamp > trackEndTime) {
+        // This shouldn't happen, but some muxers give incorrect durations to
+        // segments, then have samples appear beyond those durations.
+        WEBM_DEBUG("Found tstamp=%" PRIi64 " > trackEndTime=%" PRIi64
+                   " while calculating next timestamp! Indicates a bad mux! "
+                   "Will use tstamp value.",
+                   tstamp, trackEndTime);
+      }
+      next_tstamp = std::max<int64_t>(tstamp, trackEndTime);
+    }
+    lastFrameTime = Some(tstamp);
+  };
+
   if (aType == TrackInfo::kAudioTrack) {
-    RefPtr<NesteggPacketHolder> next_holder;
-    rv = NextPacket(aType, next_holder);
-    if (NS_FAILED(rv) && rv != NS_ERROR_DOM_MEDIA_END_OF_STREAM) {
-      return rv;
-    }
-    if (next_holder) {
-      next_tstamp = next_holder->Timestamp();
-      PushAudioPacket(next_holder);
-    } else if (duration >= 0) {
-      next_tstamp = tstamp + duration;
-    } else if (!mIsMediaSource ||
-               (mIsMediaSource && mLastAudioFrameTime.isSome())) {
-      next_tstamp = tstamp;
-      next_tstamp += tstamp - mLastAudioFrameTime.refOr(0);
-    } else {
-      PushAudioPacket(holder);
-    }
-    mLastAudioFrameTime = Some(tstamp);
-  } else if (aType == TrackInfo::kVideoTrack) {
-    RefPtr<NesteggPacketHolder> next_holder;
-    rv = NextPacket(aType, next_holder);
-    if (NS_FAILED(rv) && rv != NS_ERROR_DOM_MEDIA_END_OF_STREAM) {
-      return rv;
-    }
-    if (next_holder) {
-      next_tstamp = next_holder->Timestamp();
-      PushVideoPacket(next_holder);
-    } else if (duration >= 0) {
-      next_tstamp = tstamp + duration;
-    } else if (!mIsMediaSource ||
-               (mIsMediaSource && mLastVideoFrameTime.isSome())) {
-      next_tstamp = tstamp;
-      next_tstamp += tstamp - mLastVideoFrameTime.refOr(0);
-    } else {
-      PushVideoPacket(holder);
-    }
-    mLastVideoFrameTime = Some(tstamp);
+    calculateNextTimestamp(&WebMDemuxer::PushAudioPacket, mLastAudioFrameTime,
+                           mInfo.mAudio.mDuration.ToMicroseconds());
+  } else {
+    calculateNextTimestamp(&WebMDemuxer::PushVideoPacket, mLastVideoFrameTime,
+                           mInfo.mVideo.mDuration.ToMicroseconds());
   }
 
   if (mIsMediaSource && next_tstamp == INT64_MIN) {
@@ -636,8 +703,8 @@ nsresult WebMDemuxer::GetNextPacket(TrackInfo::TrackType aType,
             packetEncryption == NESTEGG_PACKET_HAS_SIGNAL_BYTE_UNENCRYPTED ||
                 packetEncryption == NESTEGG_PACKET_HAS_SIGNAL_BYTE_FALSE,
             "Unencrypted packet expected");
-        auto sample = MakeSpan(data, length);
-        auto alphaSample = MakeSpan(alphaData, alphaLength);
+        auto sample = Span(data, length);
+        auto alphaSample = Span(alphaData, alphaLength);
 
         switch (mVideoCodec) {
           case NESTEGG_CODEC_VP8:
@@ -666,33 +733,6 @@ nsresult WebMDemuxer::GetNextPacket(TrackInfo::TrackType aType,
             NS_WARNING("Cannot detect keyframes in unknown WebM video codec");
             return NS_ERROR_FAILURE;
         }
-        if (isKeyframe) {
-          // For both VP8 and VP9, we only look for resolution changes
-          // on keyframes. Other resolution changes are invalid.
-          auto dimensions = gfx::IntSize(0, 0);
-          switch (mVideoCodec) {
-            case NESTEGG_CODEC_VP8:
-              dimensions =
-                  VPXDecoder::GetFrameSize(sample, VPXDecoder::Codec::VP8);
-              break;
-            case NESTEGG_CODEC_VP9:
-              dimensions =
-                  VPXDecoder::GetFrameSize(sample, VPXDecoder::Codec::VP9);
-              break;
-#ifdef MOZ_AV1
-            case NESTEGG_CODEC_AV1:
-              dimensions = AOMDecoder::GetFrameSize(sample);
-              break;
-#endif
-          }
-          if (mLastSeenFrameSize.isSome() &&
-              (dimensions != mLastSeenFrameSize.value())) {
-            mInfo.mVideo.mDisplay = dimensions;
-            mSharedVideoTrackInfo =
-                new TrackInfoSharedPtr(mInfo.mVideo, ++sStreamSourceID);
-          }
-          mLastSeenFrameSize = Some(dimensions);
-        }
       }
     }
 
@@ -716,7 +756,9 @@ nsresult WebMDemuxer::GetNextPacket(TrackInfo::TrackType aType,
     }
     sample->mTimecode = TimeUnit::FromMicroseconds(tstamp);
     sample->mTime = TimeUnit::FromMicroseconds(tstamp);
-    sample->mDuration = TimeUnit::FromMicroseconds(next_tstamp - tstamp);
+    if (next_tstamp > tstamp) {
+      sample->mDuration = TimeUnit::FromMicroseconds(next_tstamp - tstamp);
+    }
     sample->mOffset = holder->Offset();
     sample->mKeyframe = isKeyframe;
     if (discardPadding && i == count - 1) {
@@ -738,11 +780,11 @@ nsresult WebMDemuxer::GetNextPacket(TrackInfo::TrackType aType,
 
     if (packetEncryption == NESTEGG_PACKET_HAS_SIGNAL_BYTE_ENCRYPTED ||
         packetEncryption == NESTEGG_PACKET_HAS_SIGNAL_BYTE_PARTITIONED) {
-      nsAutoPtr<MediaRawDataWriter> writer(sample->CreateWriter());
+      UniquePtr<MediaRawDataWriter> writer(sample->CreateWriter());
       unsigned char const* iv;
       size_t ivLength;
       nestegg_packet_iv(holder->Packet(), &iv, &ivLength);
-      writer->mCrypto.mValid = true;
+      writer->mCrypto.mCryptoScheme = CryptoScheme::Cenc;
       writer->mCrypto.mIVSize = ivLength;
       if (ivLength == 0) {
         // Frame is not encrypted. This shouldn't happen as it means the
@@ -797,7 +839,7 @@ nsresult WebMDemuxer::GetNextPacket(TrackInfo::TrackType aType,
             encrypted = !encrypted;
             lastOffset = partition;
 
-            assert(lastOffset <= length);
+            MOZ_ASSERT(lastOffset <= length);
           }
 
           // Add the data between the last offset and the end of the data.
@@ -824,7 +866,7 @@ nsresult WebMDemuxer::GetNextPacket(TrackInfo::TrackType aType,
 
           // Assert that the lengths of the encrypted and plain samples add to
           // the length of the data.
-          assert(
+          MOZ_ASSERT(
               ((size_t)(std::accumulate(writer->mCrypto.mPlainSizes.begin(),
                                         writer->mCrypto.mPlainSizes.end(), 0) +
                         std::accumulate(writer->mCrypto.mEncryptedSizes.begin(),
@@ -833,9 +875,6 @@ nsresult WebMDemuxer::GetNextPacket(TrackInfo::TrackType aType,
         }
       }
     }
-    if (aType == TrackInfo::kVideoTrack) {
-      sample->mTrackInfo = mSharedVideoTrackInfo;
-    }
     aSamples->Push(sample);
   }
   return NS_OK;
@@ -900,7 +939,7 @@ nsresult WebMDemuxer::DemuxPacket(TrackInfo::TrackType aType,
 
   int64_t offset = Resource(aType).Tell();
   RefPtr<NesteggPacketHolder> holder = new NesteggPacketHolder();
-  if (!holder->Init(packet, offset, track, false)) {
+  if (!holder->Init(packet, Context(aType), offset, track, false)) {
     return NS_ERROR_DOM_MEDIA_DEMUXER_ERR;
   }
 
@@ -920,11 +959,11 @@ nsresult WebMDemuxer::SeekInternal(TrackInfo::TrackType aType,
                                    const TimeUnit& aTarget) {
   EnsureUpToDateIndex();
   uint32_t trackToSeek = mHasVideo ? mVideoTrack : mAudioTrack;
-  uint64_t target = aTarget.ToNanoseconds();
+  MOZ_ASSERT(aTarget.ToNanoseconds() >= 0, "Seek time can't be negative");
+  uint64_t target = static_cast<uint64_t>(aTarget.ToNanoseconds());
+  WEBM_DEBUG("Seeking to %lf", aTarget.ToSeconds());
 
-  if (NS_FAILED(Reset(aType))) {
-    return NS_ERROR_FAILURE;
-  }
+  Reset(aType);
 
   if (mSeekPreroll) {
     uint64_t startTime = 0;
@@ -955,7 +994,13 @@ nsresult WebMDemuxer::SeekInternal(TrackInfo::TrackType aType,
       return NS_ERROR_FAILURE;
     }
 
-    r = nestegg_offset_seek(Context(aType), offset);
+    if (offset < 0) {
+      WEBM_DEBUG("Unknow byte offset time for seek target %" PRIu64 "ns",
+                 target);
+      return NS_ERROR_FAILURE;
+    }
+
+    r = nestegg_offset_seek(Context(aType), static_cast<uint64_t>(offset));
     if (r == -1) {
       WEBM_DEBUG("and nestegg_offset_seek to %" PRIu64 " failed", offset);
       return NS_ERROR_FAILURE;
@@ -972,6 +1017,33 @@ nsresult WebMDemuxer::SeekInternal(TrackInfo::TrackType aType,
   return NS_OK;
 }
 
+bool WebMDemuxer::IsBufferedIntervalValid(uint64_t start, uint64_t end) {
+  if (start > end) {
+    // Buffered ranges are clamped to the media's start time and duration. Any
+    // frames with timestamps outside that range are ignored, see bug 1697641
+    // for more info.
+    WEBM_DEBUG("Ignoring range %" PRIu64 "-%" PRIu64
+               ", due to invalid interval (start > end).",
+               start, end);
+    return false;
+  }
+
+  auto startTime = TimeUnit::FromNanoseconds(start);
+  auto endTime = TimeUnit::FromNanoseconds(end);
+
+  if (startTime.IsNegative() || endTime.IsNegative()) {
+    // We can get timestamps that are conceptually valid, but become
+    // negative due to uint64 -> int64 conversion from TimeUnit. We should
+    // not get negative timestamps, so guard against them.
+    WEBM_DEBUG(
+        "Invalid range %f-%f, likely result of uint64 -> int64 conversion.",
+        startTime.ToSeconds(), endTime.ToSeconds());
+    return false;
+  }
+
+  return true;
+}
+
 media::TimeIntervals WebMDemuxer::GetBuffered() {
   EnsureUpToDateIndex();
   AutoPinned<MediaResource> resource(
@@ -1008,8 +1080,15 @@ media::TimeIntervals WebMDemuxer::GetBuffered() {
                    TimeUnit::FromNanoseconds(duration).ToSeconds());
         end = duration;
       }
+
+      if (!IsBufferedIntervalValid(start, end)) {
+        WEBM_DEBUG("Invalid interval, bailing");
+        break;
+      }
+
       auto startTime = TimeUnit::FromNanoseconds(start);
       auto endTime = TimeUnit::FromNanoseconds(end);
+
       WEBM_DEBUG("add range %f-%f", startTime.ToSeconds(), endTime.ToSeconds());
       buffered += media::TimeInterval(startTime, endTime);
     }
@@ -1093,7 +1172,6 @@ RefPtr<WebMTrackDemuxer::SeekPromise> WebMTrackDemuxer::Seek(
 
 nsresult WebMTrackDemuxer::NextSample(RefPtr<MediaRawData>& aData) {
   nsresult rv = NS_ERROR_DOM_MEDIA_END_OF_STREAM;
-  ;
   while (mSamples.GetSize() < 1 &&
          NS_SUCCEEDED((rv = mParent->GetNextPacket(mType, &mSamples)))) {
   }
@@ -1117,18 +1195,28 @@ RefPtr<WebMTrackDemuxer::SamplesPromise> WebMTrackDemuxer::GetSamples(
     if (NS_FAILED(rv)) {
       break;
     }
+    // Ignore empty samples.
+    if (sample->Size() == 0) {
+      WEBM_DEBUG(
+          "0 sized sample encountered while getting samples, skipping it");
+      continue;
+    }
     if (mNeedKeyframe && !sample->mKeyframe) {
       continue;
     }
+    if (!sample->HasValidTime()) {
+      return SamplesPromise::CreateAndReject(NS_ERROR_DOM_MEDIA_DEMUXER_ERR,
+                                             __func__);
+    }
     mNeedKeyframe = false;
-    samples->mSamples.AppendElement(sample);
+    samples->AppendSample(sample);
     aNumSamples--;
   }
 
-  if (samples->mSamples.IsEmpty()) {
+  if (samples->GetSamples().IsEmpty()) {
     return SamplesPromise::CreateAndReject(rv, __func__);
   } else {
-    UpdateSamples(samples->mSamples);
+    UpdateSamples(samples->GetSamples());
     return SamplesPromise::CreateAndResolve(samples, __func__);
   }
 }
@@ -1177,7 +1265,7 @@ void WebMTrackDemuxer::SetNextKeyFrameTime() {
   }
   // We may have demuxed more than intended, so ensure that all frames are kept
   // in the right order.
-  mSamples.PushFront(Move(skipSamplesQueue));
+  mSamples.PushFront(std::move(skipSamplesQueue));
 
   if (frameTime.IsValid()) {
     mNextKeyframeTime.emplace(frameTime);
@@ -1195,7 +1283,7 @@ void WebMTrackDemuxer::Reset() {
   mSamples.Reset();
   media::TimeIntervals buffered = GetBuffered();
   mNeedKeyframe = true;
-  if (buffered.Length()) {
+  if (!buffered.IsEmpty()) {
     WEBM_DEBUG("Seek to start point: %f", buffered.Start(0).ToSeconds());
     mParent->SeekInternal(mType, buffered.Start(0));
     SetNextKeyFrameTime();
@@ -1204,11 +1292,11 @@ void WebMTrackDemuxer::Reset() {
   }
 }
 
-void WebMTrackDemuxer::UpdateSamples(nsTArray<RefPtr<MediaRawData>>& aSamples) {
+void WebMTrackDemuxer::UpdateSamples(
+    const nsTArray<RefPtr<MediaRawData>>& aSamples) {
   for (const auto& sample : aSamples) {
-    if (sample->mCrypto.mValid) {
-      nsAutoPtr<MediaRawDataWriter> writer(sample->CreateWriter());
-      writer->mCrypto.mMode = mInfo->mCrypto.mMode;
+    if (sample->mCrypto.IsEncrypted()) {
+      UniquePtr<MediaRawDataWriter> writer(sample->CreateWriter());
       writer->mCrypto.mIVSize = mInfo->mCrypto.mIVSize;
       writer->mCrypto.mKeyId.AppendElements(mInfo->mCrypto.mKeyId);
     }
@@ -1254,7 +1342,8 @@ WebMTrackDemuxer::SkipToNextRandomAccessPoint(const TimeUnit& aTimeThreshold) {
     return SkipAccessPointPromise::CreateAndResolve(parsed, __func__);
   } else {
     SkipFailureHolder failure(NS_ERROR_DOM_MEDIA_END_OF_STREAM, parsed);
-    return SkipAccessPointPromise::CreateAndReject(Move(failure), __func__);
+    return SkipAccessPointPromise::CreateAndReject(std::move(failure),
+                                                   __func__);
   }
 }
 
@@ -1266,12 +1355,14 @@ void WebMTrackDemuxer::BreakCycles() { mParent = nullptr; }
 
 int64_t WebMTrackDemuxer::GetEvictionOffset(const TimeUnit& aTime) {
   int64_t offset;
-  if (!mParent->GetOffsetForTime(aTime.ToNanoseconds(), &offset)) {
+  int64_t nanos = aTime.ToNanoseconds();
+  if (nanos < 0 ||
+      !mParent->GetOffsetForTime(static_cast<uint64_t>(nanos), &offset)) {
     return 0;
   }
 
   return offset;
 }
+}  // namespace mozilla
 
 #undef WEBM_DEBUG
-}  // namespace mozilla
