@@ -19,15 +19,14 @@
 #include "mozilla/dom/MIDITypes.h"
 #include "mozilla/dom/Promise.h"
 #include "mozilla/dom/PContent.h"
-#include "nsIRunnable.h"
-#include "nsIDocument.h"
+#include "mozilla/dom/Document.h"
 #include "nsPIDOMWindow.h"
 #include "nsContentPermissionHelper.h"
 #include "nsISupportsImpl.h"  // for MOZ_COUNT_CTOR, MOZ_COUNT_DTOR
-#include "IPCMessageUtils.h"
+#include "ipc/IPCMessageUtils.h"
+#include "MIDILog.h"
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 NS_IMPL_CYCLE_COLLECTION_CLASS(MIDIAccess)
 NS_IMPL_CYCLE_COLLECTION_TRACE_BEGIN_INHERITED(MIDIAccess, DOMEventTargetHelper)
@@ -56,20 +55,22 @@ NS_IMPL_ADDREF_INHERITED(MIDIAccess, DOMEventTargetHelper)
 NS_IMPL_RELEASE_INHERITED(MIDIAccess, DOMEventTargetHelper)
 
 MIDIAccess::MIDIAccess(nsPIDOMWindowInner* aWindow, bool aSysexEnabled,
-                       Promise* aPromise)
+                       Promise* aAccessPromise)
     : DOMEventTargetHelper(aWindow),
       mInputMap(new MIDIInputMap(aWindow)),
       mOutputMap(new MIDIOutputMap(aWindow)),
       mSysexEnabled(aSysexEnabled),
-      mAccessPromise(aPromise),
+      mAccessPromise(aAccessPromise),
       mHasShutdown(false) {
   MOZ_ASSERT(aWindow);
-  MOZ_ASSERT(aPromise);
+  MOZ_ASSERT(aAccessPromise);
+  KeepAliveIfHasListenersFor(nsGkAtoms::onstatechange);
 }
 
 MIDIAccess::~MIDIAccess() { Shutdown(); }
 
 void MIDIAccess::Shutdown() {
+  LOG("MIDIAccess::Shutdown");
   if (mHasShutdown) {
     return;
   }
@@ -88,15 +89,18 @@ void MIDIAccess::FireConnectionEvent(MIDIPort* aPort) {
   aPort->GetId(id);
   ErrorResult rv;
   if (aPort->State() == MIDIPortDeviceState::Disconnected) {
-    if (aPort->Type() == MIDIPortType::Input &&
-        MIDIInputMapBinding::MaplikeHelpers::Has(mInputMap, id, rv)) {
-      MIDIInputMapBinding::MaplikeHelpers::Delete(mInputMap, id, rv);
-    } else if (aPort->Type() == MIDIPortType::Output &&
-               MIDIOutputMapBinding::MaplikeHelpers::Has(mOutputMap, id, rv)) {
-      MIDIOutputMapBinding::MaplikeHelpers::Delete(mOutputMap, id, rv);
+    if (aPort->Type() == MIDIPortType::Input && mInputMap->Has(id)) {
+      MIDIInputMap_Binding::MaplikeHelpers::Delete(mInputMap, aPort->StableId(),
+                                                   rv);
+      mInputMap->Remove(id);
+    } else if (aPort->Type() == MIDIPortType::Output && mOutputMap->Has(id)) {
+      MIDIOutputMap_Binding::MaplikeHelpers::Delete(mOutputMap,
+                                                    aPort->StableId(), rv);
+      mOutputMap->Remove(id);
     }
     // Check to make sure Has()/Delete() calls haven't failed.
     if (NS_WARN_IF(rv.Failed())) {
+      LOG("Inconsistency during FireConnectionEvent");
       return;
     }
   } else {
@@ -104,30 +108,35 @@ void MIDIAccess::FireConnectionEvent(MIDIPort* aPort) {
     // this means a port that was disconnected has been reconnected, with the
     // port owner holding the object during that time, and we should add that
     // port object to our maps again.
-    if (aPort->Type() == MIDIPortType::Input &&
-        !MIDIInputMapBinding::MaplikeHelpers::Has(mInputMap, id, rv)) {
+    if (aPort->Type() == MIDIPortType::Input && !mInputMap->Has(id)) {
       if (NS_WARN_IF(rv.Failed())) {
+        LOG("Input port not found");
         return;
       }
-      MIDIInputMapBinding::MaplikeHelpers::Set(
-          mInputMap, id, *(static_cast<MIDIInput*>(aPort)), rv);
+      MIDIInputMap_Binding::MaplikeHelpers::Set(
+          mInputMap, aPort->StableId(), *(static_cast<MIDIInput*>(aPort)), rv);
       if (NS_WARN_IF(rv.Failed())) {
+        LOG("Map Set failed for input port");
         return;
       }
-    } else if (aPort->Type() == MIDIPortType::Output &&
-               !MIDIOutputMapBinding::MaplikeHelpers::Has(mOutputMap, id, rv)) {
+      mInputMap->Insert(id, aPort);
+    } else if (aPort->Type() == MIDIPortType::Output && mOutputMap->Has(id)) {
       if (NS_WARN_IF(rv.Failed())) {
+        LOG("Output port not found");
         return;
       }
-      MIDIOutputMapBinding::MaplikeHelpers::Set(
-          mOutputMap, id, *(static_cast<MIDIOutput*>(aPort)), rv);
+      MIDIOutputMap_Binding::MaplikeHelpers::Set(
+          mOutputMap, aPort->StableId(), *(static_cast<MIDIOutput*>(aPort)),
+          rv);
       if (NS_WARN_IF(rv.Failed())) {
+        LOG("Map set failed for output port");
         return;
       }
+      mOutputMap->Insert(id, aPort);
     }
   }
-  RefPtr<MIDIConnectionEvent> event = MIDIConnectionEvent::Constructor(
-      this, NS_LITERAL_STRING("statechange"), init);
+  RefPtr<MIDIConnectionEvent> event =
+      MIDIConnectionEvent::Constructor(this, u"statechange"_ns, init);
   DispatchTrustedEvent(event);
 }
 
@@ -137,38 +146,43 @@ void MIDIAccess::MaybeCreateMIDIPort(const MIDIPortInfo& aInfo,
   MIDIPortType type = static_cast<MIDIPortType>(aInfo.type());
   RefPtr<MIDIPort> port;
   if (type == MIDIPortType::Input) {
-    bool hasPort = MIDIInputMapBinding::MaplikeHelpers::Has(mInputMap, id, aRv);
-    if (hasPort || NS_WARN_IF(aRv.Failed())) {
+    if (mInputMap->Has(id) || NS_WARN_IF(aRv.Failed())) {
       // We already have the port in our map.
       return;
     }
     port = MIDIInput::Create(GetOwner(), this, aInfo, mSysexEnabled);
     if (NS_WARN_IF(!port)) {
+      LOG("Couldn't create input port");
       aRv.Throw(NS_ERROR_FAILURE);
       return;
     }
-    MIDIInputMapBinding::MaplikeHelpers::Set(
-        mInputMap, id, *(static_cast<MIDIInput*>(port.get())), aRv);
+    MIDIInputMap_Binding::MaplikeHelpers::Set(
+        mInputMap, port->StableId(), *(static_cast<MIDIInput*>(port.get())),
+        aRv);
     if (NS_WARN_IF(aRv.Failed())) {
+      LOG("Coudld't set input port in map");
       return;
     }
+    mInputMap->Insert(id, port);
   } else if (type == MIDIPortType::Output) {
-    bool hasPort =
-        MIDIOutputMapBinding::MaplikeHelpers::Has(mOutputMap, id, aRv);
-    if (hasPort || NS_WARN_IF(aRv.Failed())) {
+    if (mOutputMap->Has(id) || NS_WARN_IF(aRv.Failed())) {
       // We already have the port in our map.
       return;
     }
     port = MIDIOutput::Create(GetOwner(), this, aInfo, mSysexEnabled);
     if (NS_WARN_IF(!port)) {
+      LOG("Couldn't create output port");
       aRv.Throw(NS_ERROR_FAILURE);
       return;
     }
-    MIDIOutputMapBinding::MaplikeHelpers::Set(
-        mOutputMap, id, *(static_cast<MIDIOutput*>(port.get())), aRv);
+    MIDIOutputMap_Binding::MaplikeHelpers::Set(
+        mOutputMap, port->StableId(), *(static_cast<MIDIOutput*>(port.get())),
+        aRv);
     if (NS_WARN_IF(aRv.Failed())) {
+      LOG("Coudld't set output port in map");
       return;
     }
+    mOutputMap->Insert(id, port);
   } else {
     // If we hit this, then we have some port that is neither input nor output.
     // That is bad.
@@ -190,15 +204,21 @@ void MIDIAccess::MaybeCreateMIDIPort(const MIDIPortInfo& aInfo,
 // received, that will be handled by the MIDIPort object itself, and it will
 // request removal from MIDIAccess's maps.
 void MIDIAccess::Notify(const MIDIPortList& aEvent) {
-  ErrorResult rv;
-  for (auto& port : aEvent.ports()) {
+  LOG("MIDIAcess::Notify");
+  if (!GetOwner()) {
+    // Do nothing if we've already been disconnected from the document.
+    return;
+  }
+
+  for (const auto& port : aEvent.ports()) {
     // Something went very wrong. Warn and return.
+    ErrorResult rv;
     MaybeCreateMIDIPort(port, rv);
     if (rv.Failed()) {
       if (!mAccessPromise) {
         return;
       }
-      mAccessPromise->MaybeReject(rv);
+      mAccessPromise->MaybeReject(std::move(rv));
       mAccessPromise = nullptr;
     }
   }
@@ -211,12 +231,18 @@ void MIDIAccess::Notify(const MIDIPortList& aEvent) {
 
 JSObject* MIDIAccess::WrapObject(JSContext* aCx,
                                  JS::Handle<JSObject*> aGivenProto) {
-  return MIDIAccessBinding::Wrap(aCx, this, aGivenProto);
+  return MIDIAccess_Binding::Wrap(aCx, this, aGivenProto);
 }
 
 void MIDIAccess::RemovePortListener(MIDIAccessDestructionObserver* aObs) {
   mDestructionObservers.RemoveObserver(aObs);
 }
 
-}  // namespace dom
-}  // namespace mozilla
+void MIDIAccess::DisconnectFromOwner() {
+  IgnoreKeepAliveIfHasListenersFor(nsGkAtoms::onstatechange);
+
+  DOMEventTargetHelper::DisconnectFromOwner();
+  MIDIAccessManager::Get()->SendRefresh();
+}
+
+}  // namespace mozilla::dom
