@@ -10,25 +10,26 @@
 #include "mozilla/dom/MIDIOutputBinding.h"
 #include "mozilla/dom/MIDIUtils.h"
 #include "nsDOMNavigationTiming.h"
+#include "mozilla/ErrorResult.h"
 #include "mozilla/TimeStamp.h"
+#include "mozilla/dom/Document.h"
 #include "mozilla/dom/Performance.h"
 
 using namespace mozilla;
 using namespace mozilla::dom;
 
-MIDIOutput::MIDIOutput(nsPIDOMWindowInner* aWindow,
-                       MIDIAccess* aMIDIAccessParent)
-    : MIDIPort(aWindow, aMIDIAccessParent) {}
+MIDIOutput::MIDIOutput(nsPIDOMWindowInner* aWindow) : MIDIPort(aWindow) {}
 
 // static
-MIDIOutput* MIDIOutput::Create(nsPIDOMWindowInner* aWindow,
-                               MIDIAccess* aMIDIAccessParent,
-                               const MIDIPortInfo& aPortInfo,
-                               const bool aSysexEnabled) {
+RefPtr<MIDIOutput> MIDIOutput::Create(nsPIDOMWindowInner* aWindow,
+                                      MIDIAccess* aMIDIAccessParent,
+                                      const MIDIPortInfo& aPortInfo,
+                                      const bool aSysexEnabled) {
   MOZ_ASSERT(static_cast<MIDIPortType>(aPortInfo.type()) ==
              MIDIPortType::Output);
-  auto port = new MIDIOutput(aWindow, aMIDIAccessParent);
-  if (NS_WARN_IF(!port->Initialize(aPortInfo, aSysexEnabled))) {
+  RefPtr<MIDIOutput> port = new MIDIOutput(aWindow);
+  if (NS_WARN_IF(
+          !port->Initialize(aPortInfo, aSysexEnabled, aMIDIAccessParent))) {
     return nullptr;
   }
   return port;
@@ -36,12 +37,12 @@ MIDIOutput* MIDIOutput::Create(nsPIDOMWindowInner* aWindow,
 
 JSObject* MIDIOutput::WrapObject(JSContext* aCx,
                                  JS::Handle<JSObject*> aGivenProto) {
-  return MIDIOutputBinding::Wrap(aCx, this, aGivenProto);
+  return MIDIOutput_Binding::Wrap(aCx, this, aGivenProto);
 }
 
 void MIDIOutput::Send(const Sequence<uint8_t>& aData,
                       const Optional<double>& aTimestamp, ErrorResult& aRv) {
-  if (mPort->DeviceState() == MIDIPortDeviceState::Disconnected) {
+  if (Port()->DeviceState() == MIDIPortDeviceState::Disconnected) {
     aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
     return;
   }
@@ -53,7 +54,7 @@ void MIDIOutput::Send(const Sequence<uint8_t>& aData,
   // message ASAP.
   TimeStamp timestamp;
   if (aTimestamp.WasPassed() && aTimestamp.Value() != 0) {
-    nsCOMPtr<nsIDocument> doc = GetOwner()->GetDoc();
+    nsCOMPtr<Document> doc = GetOwner()->GetDoc();
     if (!doc) {
       aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
       return;
@@ -69,11 +70,14 @@ void MIDIOutput::Send(const Sequence<uint8_t>& aData,
   }
 
   nsTArray<MIDIMessage> msgArray;
-  MIDIUtils::ParseMessages(aData, timestamp, msgArray);
-  // Our translation of the spec is that invalid messages in a multi-message
-  // sequence will be thrown out, but that valid messages will still be used.
+  bool ret = MIDIUtils::ParseMessages(aData, timestamp, msgArray);
+  if (!ret) {
+    aRv.ThrowTypeError("Invalid MIDI message");
+    return;
+  }
+
   if (msgArray.IsEmpty()) {
-    aRv.Throw(NS_ERROR_DOM_TYPE_ERR);
+    aRv.ThrowTypeError("Empty message array");
     return;
   }
 
@@ -87,12 +91,12 @@ void MIDIOutput::Send(const Sequence<uint8_t>& aData,
       }
     }
   }
-  mPort->SendSend(msgArray);
+  Port()->SendSend(msgArray);
 }
 
 void MIDIOutput::Clear() {
-  if (mPort->ConnectionState() == MIDIPortConnectionState::Closed) {
+  if (Port()->ConnectionState() == MIDIPortConnectionState::Closed) {
     return;
   }
-  mPort->SendClear();
+  Port()->SendClear();
 }
