@@ -5,40 +5,74 @@
  * You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "StorageManager.h"
+#include "fs/FileSystemRequestHandler.h"
 
-#include "mozilla/dom/DOMPrefs.h"
-#include "mozilla/dom/PromiseWorkerProxy.h"
-#include "mozilla/dom/quota/QuotaManagerService.h"
-#include "mozilla/dom/StorageManagerBinding.h"
-#include "mozilla/dom/WorkerPrivate.h"
+#include <cstdint>
+#include <cstdlib>
+#include <utility>
+#include "ErrorList.h"
+#include "fs/FileSystemRequestHandler.h"
+#include "MainThreadUtils.h"
+#include "js/CallArgs.h"
+#include "js/TypeDecls.h"
+#include "mozilla/Attributes.h"
 #include "mozilla/ErrorResult.h"
-#include "mozilla/EventStateManager.h"
+#include "mozilla/MacroForEach.h"
+#include "mozilla/Maybe.h"
+#include "mozilla/Mutex.h"
+#include "mozilla/RefPtr.h"
 #include "mozilla/Telemetry.h"
+#include "mozilla/TelemetryScalarEnums.h"
+#include "mozilla/dom/BindingDeclarations.h"
+#include "mozilla/dom/Document.h"
+#include "mozilla/dom/FileSystemManager.h"
+#include "mozilla/dom/Promise.h"
+#include "mozilla/dom/PromiseWorkerProxy.h"
+#include "mozilla/dom/StorageManagerBinding.h"
+#include "mozilla/dom/WorkerCommon.h"
+#include "mozilla/dom/WorkerPrivate.h"
+#include "mozilla/dom/WorkerRunnable.h"
+#include "mozilla/dom/WorkerStatus.h"
+#include "mozilla/dom/quota/QuotaManagerService.h"
 #include "nsContentPermissionHelper.h"
+#include "nsContentUtils.h"
+#include "nsDebug.h"
+#include "nsError.h"
+#include "nsIGlobalObject.h"
+#include "nsIPrincipal.h"
 #include "nsIQuotaCallbacks.h"
+#include "nsIQuotaManagerService.h"
 #include "nsIQuotaRequests.h"
+#include "nsIQuotaResults.h"
+#include "nsIVariant.h"
+#include "nsLiteralString.h"
 #include "nsPIDOMWindow.h"
+#include "nsString.h"
+#include "nsStringFlags.h"
+#include "nsTLiteralString.h"
+#include "nscore.h"
+
+class JSObject;
+struct JSContext;
+struct nsID;
 
 namespace mozilla {
-namespace dom {
+class Runnable;
+}
+
+using namespace mozilla::dom::quota;
+
+namespace mozilla::dom {
 
 namespace {
 
 // This class is used to get quota usage, request persist and check persisted
 // status callbacks.
-class RequestResolver final
-  : public nsIQuotaCallback
-  , public nsIQuotaUsageCallback
-{
-public:
-  enum Type
-  {
-    Estimate,
-    Persist,
-    Persisted
-  };
+class RequestResolver final : public nsIQuotaCallback {
+ public:
+  enum Type { Estimate, Persist, Persisted };
 
-private:
+ private:
   class FinishWorkerRunnable;
 
   // If this resolver was created for a window then mPromise must be non-null.
@@ -51,111 +85,87 @@ private:
   const Type mType;
   bool mPersisted;
 
-public:
+ public:
   RequestResolver(Type aType, Promise* aPromise)
-    : mPromise(aPromise)
-    , mResultCode(NS_OK)
-    , mType(aType)
-    , mPersisted(false)
-  {
+      : mPromise(aPromise),
+        mResultCode(NS_OK),
+        mType(aType),
+        mPersisted(false) {
     MOZ_ASSERT(NS_IsMainThread());
     MOZ_ASSERT(aPromise);
   }
 
   RequestResolver(Type aType, PromiseWorkerProxy* aProxy)
-    : mProxy(aProxy)
-    , mResultCode(NS_OK)
-    , mType(aType)
-    , mPersisted(false)
-  {
+      : mProxy(aProxy), mResultCode(NS_OK), mType(aType), mPersisted(false) {
     MOZ_ASSERT(NS_IsMainThread());
     MOZ_ASSERT(aProxy);
   }
 
-  void
-  ResolveOrReject();
+  void ResolveOrReject();
 
   NS_DECL_THREADSAFE_ISUPPORTS
   NS_DECL_NSIQUOTACALLBACK
-  NS_DECL_NSIQUOTAUSAGECALLBACK
 
-private:
-  ~RequestResolver()
-  { }
+ private:
+  ~RequestResolver() = default;
 
-  nsresult
-  GetStorageEstimate(nsIVariant* aResult);
+  nsresult GetStorageEstimate(nsIVariant* aResult);
 
-  nsresult
-  GetPersisted(nsIVariant* aResult);
+  nsresult GetPersisted(nsIVariant* aResult);
 
-  template <typename T>
-  nsresult
-  OnCompleteOrUsageResult(T* aRequest);
+  nsresult OnCompleteInternal(nsIQuotaRequest* aRequest);
 
-  nsresult
-  Finish();
+  nsresult Finish();
 };
 
 // This class is used to return promise on worker thread.
-class RequestResolver::FinishWorkerRunnable final
-  : public WorkerRunnable
-{
+class RequestResolver::FinishWorkerRunnable final : public WorkerRunnable {
   RefPtr<RequestResolver> mResolver;
 
-public:
+ public:
   explicit FinishWorkerRunnable(RequestResolver* aResolver)
-    : WorkerRunnable(aResolver->mProxy->GetWorkerPrivate())
-    , mResolver(aResolver)
-  {
+      : WorkerRunnable(aResolver->mProxy->GetWorkerPrivate()),
+        mResolver(aResolver) {
     MOZ_ASSERT(NS_IsMainThread());
     MOZ_ASSERT(aResolver);
   }
 
-  bool
-  WorkerRun(JSContext* aCx, WorkerPrivate* aWorkerPrivate) override;
+  bool WorkerRun(JSContext* aCx, WorkerPrivate* aWorkerPrivate) override;
 };
 
-class EstimateWorkerMainThreadRunnable final
-  : public WorkerMainThreadRunnable
-{
+class EstimateWorkerMainThreadRunnable final : public WorkerMainThreadRunnable {
   RefPtr<PromiseWorkerProxy> mProxy;
 
-public:
+ public:
   EstimateWorkerMainThreadRunnable(WorkerPrivate* aWorkerPrivate,
                                    PromiseWorkerProxy* aProxy)
-    : WorkerMainThreadRunnable(aWorkerPrivate,
-                               NS_LITERAL_CSTRING("StorageManager :: Estimate"))
-    , mProxy(aProxy)
-  {
+      : WorkerMainThreadRunnable(aWorkerPrivate,
+                                 "StorageManager :: Estimate"_ns),
+        mProxy(aProxy) {
     MOZ_ASSERT(aWorkerPrivate);
     aWorkerPrivate->AssertIsOnWorkerThread();
     MOZ_ASSERT(aProxy);
   }
 
-  bool
-  MainThreadRun() override;
+  bool MainThreadRun() override;
 };
 
 class PersistedWorkerMainThreadRunnable final
-  : public WorkerMainThreadRunnable
-{
+    : public WorkerMainThreadRunnable {
   RefPtr<PromiseWorkerProxy> mProxy;
 
-public:
+ public:
   PersistedWorkerMainThreadRunnable(WorkerPrivate* aWorkerPrivate,
                                     PromiseWorkerProxy* aProxy)
-    : WorkerMainThreadRunnable(aWorkerPrivate,
-                               NS_LITERAL_CSTRING("StorageManager :: Persisted"))
-    , mProxy(aProxy)
-  {
+      : WorkerMainThreadRunnable(aWorkerPrivate,
+                                 "StorageManager :: Persisted"_ns),
+        mProxy(aProxy) {
     MOZ_ASSERT(aWorkerPrivate);
     aWorkerPrivate->AssertIsOnWorkerThread();
     MOZ_ASSERT(aProxy);
   }
 
-  bool
-  MainThreadRun() override;
+  bool MainThreadRun() override;
 };
 
 /*******************************************************************************
@@ -163,73 +173,69 @@ public:
  ******************************************************************************/
 
 class PersistentStoragePermissionRequest final
-  : public nsIContentPermissionRequest
-{
-  nsCOMPtr<nsIPrincipal> mPrincipal;
-  nsCOMPtr<nsPIDOMWindowInner> mWindow;
-  bool mIsHandlingUserInput;
+    : public ContentPermissionRequestBase {
   RefPtr<Promise> mPromise;
-  nsCOMPtr<nsIContentPermissionRequester> mRequester;
 
-public:
+ public:
   PersistentStoragePermissionRequest(nsIPrincipal* aPrincipal,
                                      nsPIDOMWindowInner* aWindow,
-                                     bool aIsHandlingUserInput,
                                      Promise* aPromise)
-    : mPrincipal(aPrincipal)
-    , mWindow(aWindow)
-    , mIsHandlingUserInput(aIsHandlingUserInput)
-    , mPromise(aPromise)
-  {
-    MOZ_ASSERT(aPrincipal);
+      : ContentPermissionRequestBase(aPrincipal, aWindow,
+                                     "dom.storageManager"_ns,
+                                     "persistent-storage"_ns),
+        mPromise(aPromise) {
     MOZ_ASSERT(aWindow);
     MOZ_ASSERT(aPromise);
-
-    mRequester = new nsContentPermissionRequester(mWindow);
   }
 
-  nsresult
-  Start();
+  nsresult Start();
 
-  NS_DECL_CYCLE_COLLECTING_ISUPPORTS
-  NS_DECL_NSICONTENTPERMISSIONREQUEST
-  NS_DECL_CYCLE_COLLECTION_CLASS(PersistentStoragePermissionRequest)
+  NS_DECL_ISUPPORTS_INHERITED
+  NS_DECL_CYCLE_COLLECTION_CLASS_INHERITED(PersistentStoragePermissionRequest,
+                                           ContentPermissionRequestBase)
 
-private:
-  ~PersistentStoragePermissionRequest()
-  { }
+  // nsIContentPermissionRequest
+  NS_IMETHOD Cancel(void) override;
+  NS_IMETHOD Allow(JS::Handle<JS::Value> choices) override;
+
+ private:
+  ~PersistentStoragePermissionRequest() = default;
 };
 
-nsresult
-GetUsageForPrincipal(nsIPrincipal* aPrincipal,
-                     nsIQuotaUsageCallback* aCallback,
-                     nsIQuotaUsageRequest** aRequest)
-{
+nsresult Estimate(nsIPrincipal* aPrincipal, nsIQuotaCallback* aCallback,
+                  nsIQuotaRequest** aRequest) {
   MOZ_ASSERT(aPrincipal);
   MOZ_ASSERT(aCallback);
   MOZ_ASSERT(aRequest);
 
+  // Firefox and Quota Manager have always used the schemeless origin group
+  // (https://storage.spec.whatwg.org/#schemeless-origin-group) for quota limit
+  // purposes. This has been to prevent a site/eTLD+1 from claiming more than
+  // its fair share of storage through the use of sub-domains. Because the limit
+  // is at the group level and the usage needs to make sense in the context of
+  // that limit, we also expose the group usage. Bug 1374970 reflects this
+  // reality and bug 1305665 tracks our plan to eliminate our use of groups for
+  // this.
+
   nsCOMPtr<nsIQuotaManagerService> qms = QuotaManagerService::GetOrCreate();
   if (NS_WARN_IF(!qms)) {
     return NS_ERROR_FAILURE;
   }
 
-  nsresult rv = qms->GetUsageForPrincipal(aPrincipal,
-                                          aCallback,
-                                          true,
-                                          aRequest);
+  nsCOMPtr<nsIQuotaRequest> request;
+  nsresult rv = qms->Estimate(aPrincipal, getter_AddRefs(request));
   if (NS_WARN_IF(NS_FAILED(rv))) {
     return rv;
   }
 
+  MOZ_ALWAYS_SUCCEEDS(request->SetCallback(aCallback));
+
+  request.forget(aRequest);
   return NS_OK;
 };
 
-nsresult
-Persisted(nsIPrincipal* aPrincipal,
-          nsIQuotaCallback* aCallback,
-          nsIQuotaRequest** aRequest)
-{
+nsresult Persisted(nsIPrincipal* aPrincipal, nsIQuotaCallback* aCallback,
+                   nsIQuotaRequest** aRequest) {
   MOZ_ASSERT(aPrincipal);
   MOZ_ASSERT(aCallback);
   MOZ_ASSERT(aRequest);
@@ -256,14 +262,10 @@ Persisted(nsIPrincipal* aPrincipal,
   return NS_OK;
 };
 
-already_AddRefed<Promise>
-ExecuteOpOnMainOrWorkerThread(nsIGlobalObject* aGlobal,
-                              RequestResolver::Type aType,
-                              ErrorResult& aRv)
-{
+already_AddRefed<Promise> ExecuteOpOnMainOrWorkerThread(
+    nsIGlobalObject* aGlobal, RequestResolver::Type aType, ErrorResult& aRv) {
   MOZ_ASSERT(aGlobal);
-  MOZ_ASSERT_IF(aType == RequestResolver::Type::Persist,
-                NS_IsMainThread());
+  MOZ_ASSERT_IF(aType == RequestResolver::Type::Persist, NS_IsMainThread());
 
   RefPtr<Promise> promise = Promise::Create(aGlobal, aRv);
   if (NS_WARN_IF(!promise)) {
@@ -277,7 +279,7 @@ ExecuteOpOnMainOrWorkerThread(nsIGlobalObject* aGlobal,
       return nullptr;
     }
 
-    nsCOMPtr<nsIDocument> doc = window->GetExtantDoc();
+    nsCOMPtr<Document> doc = window->GetExtantDoc();
     if (NS_WARN_IF(!doc)) {
       aRv.Throw(NS_ERROR_FAILURE);
       return nullptr;
@@ -289,14 +291,28 @@ ExecuteOpOnMainOrWorkerThread(nsIGlobalObject* aGlobal,
     // Storage Standard 7. API
     // If origin is an opaque origin, then reject promise with a TypeError.
     if (principal->GetIsNullPrincipal()) {
-      promise->MaybeReject(NS_ERROR_DOM_TYPE_ERR);
+      switch (aType) {
+        case RequestResolver::Type::Persisted:
+          promise->MaybeRejectWithTypeError(
+              "persisted() called for opaque origin");
+          break;
+        case RequestResolver::Type::Persist:
+          promise->MaybeRejectWithTypeError(
+              "persist() called for opaque origin");
+          break;
+        case RequestResolver::Type::Estimate:
+          promise->MaybeRejectWithTypeError(
+              "estimate() called for opaque origin");
+          break;
+      }
+
       return promise.forget();
     }
 
     switch (aType) {
       case RequestResolver::Type::Persisted: {
         RefPtr<RequestResolver> resolver =
-          new RequestResolver(RequestResolver::Type::Persisted, promise);
+            new RequestResolver(RequestResolver::Type::Persisted, promise);
 
         RefPtr<nsIQuotaRequest> request;
         aRv = Persisted(principal, resolver, getter_AddRefs(request));
@@ -306,14 +322,13 @@ ExecuteOpOnMainOrWorkerThread(nsIGlobalObject* aGlobal,
 
       case RequestResolver::Type::Persist: {
         RefPtr<PersistentStoragePermissionRequest> request =
-          new PersistentStoragePermissionRequest(principal,
-                                                 window,
-                                                 EventStateManager::IsHandlingUserInput(),
-                                                 promise);
+            new PersistentStoragePermissionRequest(principal, window, promise);
 
         // In private browsing mode, no permission prompt.
         if (nsContentUtils::IsInPrivateBrowsing(doc)) {
           aRv = request->Cancel();
+        } else if (!request->CheckPermissionDelegate()) {
+          aRv = request->Cancel();
         } else {
           aRv = request->Start();
         }
@@ -323,12 +338,10 @@ ExecuteOpOnMainOrWorkerThread(nsIGlobalObject* aGlobal,
 
       case RequestResolver::Type::Estimate: {
         RefPtr<RequestResolver> resolver =
-          new RequestResolver(RequestResolver::Type::Estimate, promise);
+            new RequestResolver(RequestResolver::Type::Estimate, promise);
 
-        RefPtr<nsIQuotaUsageRequest> request;
-        aRv = GetUsageForPrincipal(principal,
-                                   resolver,
-                                   getter_AddRefs(request));
+        RefPtr<nsIQuotaRequest> request;
+        aRv = Estimate(principal, resolver, getter_AddRefs(request));
 
         break;
       }
@@ -348,26 +361,27 @@ ExecuteOpOnMainOrWorkerThread(nsIGlobalObject* aGlobal,
   MOZ_ASSERT(workerPrivate);
 
   RefPtr<PromiseWorkerProxy> promiseProxy =
-    PromiseWorkerProxy::Create(workerPrivate, promise);
+      PromiseWorkerProxy::Create(workerPrivate, promise);
   if (NS_WARN_IF(!promiseProxy)) {
+    aRv.Throw(NS_ERROR_FAILURE);
     return nullptr;
   }
 
   switch (aType) {
     case RequestResolver::Type::Estimate: {
       RefPtr<EstimateWorkerMainThreadRunnable> runnnable =
-        new EstimateWorkerMainThreadRunnable(promiseProxy->GetWorkerPrivate(),
-                                             promiseProxy);
-      runnnable->Dispatch(Terminating, aRv);
+          new EstimateWorkerMainThreadRunnable(promiseProxy->GetWorkerPrivate(),
+                                               promiseProxy);
+      runnnable->Dispatch(Canceling, aRv);
 
       break;
     }
 
     case RequestResolver::Type::Persisted: {
       RefPtr<PersistedWorkerMainThreadRunnable> runnnable =
-        new PersistedWorkerMainThreadRunnable(promiseProxy->GetWorkerPrivate(),
-                                              promiseProxy);
-      runnnable->Dispatch(Terminating, aRv);
+          new PersistedWorkerMainThreadRunnable(
+              promiseProxy->GetWorkerPrivate(), promiseProxy);
+      runnnable->Dispatch(Canceling, aRv);
 
       break;
     }
@@ -383,28 +397,22 @@ ExecuteOpOnMainOrWorkerThread(nsIGlobalObject* aGlobal,
   return promise.forget();
 };
 
-} // namespace
+}  // namespace
 
 /*******************************************************************************
  * Local class implementations
  ******************************************************************************/
 
-void
-RequestResolver::ResolveOrReject()
-{
-  class MOZ_STACK_CLASS AutoCleanup final
-  {
+void RequestResolver::ResolveOrReject() {
+  class MOZ_STACK_CLASS AutoCleanup final {
     RefPtr<PromiseWorkerProxy> mProxy;
 
-  public:
-    explicit AutoCleanup(PromiseWorkerProxy* aProxy)
-      : mProxy(aProxy)
-    {
+   public:
+    explicit AutoCleanup(PromiseWorkerProxy* aProxy) : mProxy(aProxy) {
       MOZ_ASSERT(aProxy);
     }
 
-    ~AutoCleanup()
-    {
+    ~AutoCleanup() {
       MOZ_ASSERT(mProxy);
 
       mProxy->CleanUp();
@@ -431,7 +439,8 @@ RequestResolver::ResolveOrReject()
     if (NS_SUCCEEDED(mResultCode)) {
       promise->MaybeResolve(mStorageEstimate);
     } else {
-      promise->MaybeReject(NS_ERROR_DOM_TYPE_ERR);
+      promise->MaybeRejectWithTypeError(
+          "Internal error while estimating storage usage");
     }
 
     return;
@@ -446,19 +455,13 @@ RequestResolver::ResolveOrReject()
   }
 }
 
-NS_IMPL_ISUPPORTS(RequestResolver, nsIQuotaUsageCallback, nsIQuotaCallback)
+NS_IMPL_ISUPPORTS(RequestResolver, nsIQuotaCallback)
 
-nsresult
-RequestResolver::GetStorageEstimate(nsIVariant* aResult)
-{
+nsresult RequestResolver::GetStorageEstimate(nsIVariant* aResult) {
   MOZ_ASSERT(aResult);
   MOZ_ASSERT(mType == Type::Estimate);
 
-#ifdef DEBUG
-  uint16_t dataType;
-  MOZ_ALWAYS_SUCCEEDS(aResult->GetDataType(&dataType));
-  MOZ_ASSERT(dataType == nsIDataType::VTYPE_INTERFACE_IS);
-#endif
+  MOZ_ASSERT(aResult->GetDataType() == nsIDataType::VTYPE_INTERFACE_IS);
 
   nsID* iid;
   nsCOMPtr<nsISupports> supports;
@@ -469,28 +472,24 @@ RequestResolver::GetStorageEstimate(nsIVariant* aResult)
 
   free(iid);
 
-  nsCOMPtr<nsIQuotaOriginUsageResult> originUsageResult =
-    do_QueryInterface(supports);
-  MOZ_ASSERT(originUsageResult);
+  nsCOMPtr<nsIQuotaEstimateResult> estimateResult = do_QueryInterface(supports);
+  MOZ_ASSERT(estimateResult);
 
   MOZ_ALWAYS_SUCCEEDS(
-    originUsageResult->GetUsage(&mStorageEstimate.mUsage.Construct()));
+      estimateResult->GetUsage(&mStorageEstimate.mUsage.Construct()));
 
   MOZ_ALWAYS_SUCCEEDS(
-    originUsageResult->GetLimit(&mStorageEstimate.mQuota.Construct()));
+      estimateResult->GetLimit(&mStorageEstimate.mQuota.Construct()));
 
   return NS_OK;
 }
 
-nsresult
-RequestResolver::GetPersisted(nsIVariant* aResult)
-{
+nsresult RequestResolver::GetPersisted(nsIVariant* aResult) {
   MOZ_ASSERT(aResult);
   MOZ_ASSERT(mType == Type::Persist || mType == Type::Persisted);
 
 #ifdef DEBUG
-  uint16_t dataType;
-  MOZ_ALWAYS_SUCCEEDS(aResult->GetDataType(&dataType));
+  uint16_t dataType = aResult->GetDataType();
 #endif
 
   if (mType == Type::Persist) {
@@ -512,10 +511,7 @@ RequestResolver::GetPersisted(nsIVariant* aResult)
   return NS_OK;
 }
 
-template <typename T>
-nsresult
-RequestResolver::OnCompleteOrUsageResult(T* aRequest)
-{
+nsresult RequestResolver::OnCompleteInternal(nsIQuotaRequest* aRequest) {
   MOZ_ASSERT(NS_IsMainThread());
   MOZ_ASSERT(aRequest);
 
@@ -549,9 +545,7 @@ RequestResolver::OnCompleteOrUsageResult(T* aRequest)
   return NS_OK;
 }
 
-nsresult
-RequestResolver::Finish()
-{
+nsresult RequestResolver::Finish() {
   // In a main thread request.
   if (!mProxy) {
     MOZ_ASSERT(mPromise);
@@ -578,12 +572,11 @@ RequestResolver::Finish()
 }
 
 NS_IMETHODIMP
-RequestResolver::OnComplete(nsIQuotaRequest *aRequest)
-{
+RequestResolver::OnComplete(nsIQuotaRequest* aRequest) {
   MOZ_ASSERT(NS_IsMainThread());
   MOZ_ASSERT(aRequest);
 
-  mResultCode = OnCompleteOrUsageResult(aRequest);
+  mResultCode = OnCompleteInternal(aRequest);
 
   nsresult rv = Finish();
   if (NS_WARN_IF(NS_FAILED(rv))) {
@@ -593,26 +586,8 @@ RequestResolver::OnComplete(nsIQuotaRequest *aRequest)
   return NS_OK;
 }
 
-NS_IMETHODIMP
-RequestResolver::OnUsageResult(nsIQuotaUsageRequest *aRequest)
-{
-  MOZ_ASSERT(NS_IsMainThread());
-  MOZ_ASSERT(aRequest);
-
-  mResultCode = OnCompleteOrUsageResult(aRequest);
-
-  nsresult rv = Finish();
-  if (NS_WARN_IF(NS_FAILED(rv))) {
-    return rv;
-  }
-
-  return NS_OK;
-}
-
-bool
-RequestResolver::
-FinishWorkerRunnable::WorkerRun(JSContext* aCx, WorkerPrivate* aWorkerPrivate)
-{
+bool RequestResolver::FinishWorkerRunnable::WorkerRun(
+    JSContext* aCx, WorkerPrivate* aWorkerPrivate) {
   MOZ_ASSERT(aCx);
   MOZ_ASSERT(aWorkerPrivate);
   aWorkerPrivate->AssertIsOnWorkerThread();
@@ -623,9 +598,7 @@ FinishWorkerRunnable::WorkerRun(JSContext* aCx, WorkerPrivate* aWorkerPrivate)
   return true;
 }
 
-bool
-EstimateWorkerMainThreadRunnable::MainThreadRun()
-{
+bool EstimateWorkerMainThreadRunnable::MainThreadRun() {
   MOZ_ASSERT(NS_IsMainThread());
 
   nsCOMPtr<nsIPrincipal> principal;
@@ -641,11 +614,10 @@ EstimateWorkerMainThreadRunnable::MainThreadRun()
   MOZ_ASSERT(principal);
 
   RefPtr<RequestResolver> resolver =
-    new RequestResolver(RequestResolver::Type::Estimate, mProxy);
+      new RequestResolver(RequestResolver::Type::Estimate, mProxy);
 
-  RefPtr<nsIQuotaUsageRequest> request;
-  nsresult rv =
-    GetUsageForPrincipal(principal, resolver, getter_AddRefs(request));
+  RefPtr<nsIQuotaRequest> request;
+  nsresult rv = Estimate(principal, resolver, getter_AddRefs(request));
   if (NS_WARN_IF(NS_FAILED(rv))) {
     return false;
   }
@@ -653,9 +625,7 @@ EstimateWorkerMainThreadRunnable::MainThreadRun()
   return true;
 }
 
-bool
-PersistedWorkerMainThreadRunnable::MainThreadRun()
-{
+bool PersistedWorkerMainThreadRunnable::MainThreadRun() {
   MOZ_ASSERT(NS_IsMainThread());
 
   nsCOMPtr<nsIPrincipal> principal;
@@ -671,7 +641,7 @@ PersistedWorkerMainThreadRunnable::MainThreadRun()
   MOZ_ASSERT(principal);
 
   RefPtr<RequestResolver> resolver =
-    new RequestResolver(RequestResolver::Type::Persisted, mProxy);
+      new RequestResolver(RequestResolver::Type::Persisted, mProxy);
 
   RefPtr<nsIQuotaRequest> request;
   nsresult rv = Persisted(principal, resolver, getter_AddRefs(request));
@@ -682,83 +652,46 @@ PersistedWorkerMainThreadRunnable::MainThreadRun()
   return true;
 }
 
-nsresult
-PersistentStoragePermissionRequest::Start()
-{
+nsresult PersistentStoragePermissionRequest::Start() {
   MOZ_ASSERT(NS_IsMainThread());
 
-  // Grant permission if pref'ed on.
-  if (Preferences::GetBool("dom.storageManager.prompt.testing", false)) {
-    if (Preferences::GetBool("dom.storageManager.prompt.testing.allow",
-                             false)) {
-      return Allow(JS::UndefinedHandleValue);
-    }
-
+  PromptResult pr;
+#ifdef MOZ_WIDGET_ANDROID
+  // on Android calling `ShowPrompt` here calls
+  // `nsContentPermissionUtils::AskPermission` once, and a response of
+  // `PromptResult::Pending` calls it again. This results in multiple requests
+  // for storage access, so we check the prompt prefs only to ensure we only
+  // request it once.
+  pr = CheckPromptPrefs();
+#else
+  nsresult rv = ShowPrompt(pr);
+  if (NS_WARN_IF(NS_FAILED(rv))) {
+    return rv;
+  }
+#endif
+  if (pr == PromptResult::Granted) {
+    return Allow(JS::UndefinedHandleValue);
+  }
+  if (pr == PromptResult::Denied) {
     return Cancel();
   }
 
   return nsContentPermissionUtils::AskPermission(this, mWindow);
 }
 
-NS_IMPL_CYCLE_COLLECTING_ADDREF(PersistentStoragePermissionRequest)
-NS_IMPL_CYCLE_COLLECTING_RELEASE(PersistentStoragePermissionRequest)
+NS_IMPL_ISUPPORTS_CYCLE_COLLECTION_INHERITED_0(
+    PersistentStoragePermissionRequest, ContentPermissionRequestBase)
 
-NS_INTERFACE_MAP_BEGIN_CYCLE_COLLECTION(PersistentStoragePermissionRequest)
-  NS_INTERFACE_MAP_ENTRY(nsIContentPermissionRequest)
-  NS_INTERFACE_MAP_ENTRY(nsISupports)
-NS_INTERFACE_MAP_END
-
-NS_IMPL_CYCLE_COLLECTION(PersistentStoragePermissionRequest, mWindow, mPromise)
+NS_IMPL_CYCLE_COLLECTION_INHERITED(PersistentStoragePermissionRequest,
+                                   ContentPermissionRequestBase, mPromise)
 
 NS_IMETHODIMP
-PersistentStoragePermissionRequest::GetPrincipal(nsIPrincipal** aPrincipal)
-{
-  MOZ_ASSERT(NS_IsMainThread());
-  MOZ_ASSERT(aPrincipal);
-  MOZ_ASSERT(mPrincipal);
-
-  NS_ADDREF(*aPrincipal = mPrincipal);
-
-  return NS_OK;
-}
-
-NS_IMETHODIMP
-PersistentStoragePermissionRequest::GetIsHandlingUserInput(bool* aIsHandlingUserInput)
-{
-  *aIsHandlingUserInput = mIsHandlingUserInput;
-  return NS_OK;
-}
-
-NS_IMETHODIMP
-PersistentStoragePermissionRequest::GetWindow(mozIDOMWindow** aRequestingWindow)
-{
-  MOZ_ASSERT(NS_IsMainThread());
-  MOZ_ASSERT(aRequestingWindow);
-  MOZ_ASSERT(mWindow);
-
-  NS_ADDREF(*aRequestingWindow = mWindow);
-
-  return NS_OK;
-}
-
-NS_IMETHODIMP
-PersistentStoragePermissionRequest::GetElement(nsIDOMElement** aElement)
-{
-  MOZ_ASSERT(NS_IsMainThread());
-  MOZ_ASSERT(aElement);
-
-  *aElement = nullptr;
-  return NS_OK;
-}
-
-NS_IMETHODIMP
-PersistentStoragePermissionRequest::Cancel()
-{
+PersistentStoragePermissionRequest::Cancel() {
   MOZ_ASSERT(NS_IsMainThread());
   MOZ_ASSERT(mPromise);
 
   RefPtr<RequestResolver> resolver =
-    new RequestResolver(RequestResolver::Type::Persisted, mPromise);
+      new RequestResolver(RequestResolver::Type::Persisted, mPromise);
 
   RefPtr<nsIQuotaRequest> request;
 
@@ -766,12 +699,11 @@ PersistentStoragePermissionRequest::Cancel()
 }
 
 NS_IMETHODIMP
-PersistentStoragePermissionRequest::Allow(JS::HandleValue aChoices)
-{
+PersistentStoragePermissionRequest::Allow(JS::Handle<JS::Value> aChoices) {
   MOZ_ASSERT(NS_IsMainThread());
 
   RefPtr<RequestResolver> resolver =
-    new RequestResolver(RequestResolver::Type::Persist, mPromise);
+      new RequestResolver(RequestResolver::Type::Persist, mPromise);
 
   nsCOMPtr<nsIQuotaManagerService> qms = QuotaManagerService::GetOrCreate();
   if (NS_WARN_IF(!qms)) {
@@ -790,96 +722,87 @@ PersistentStoragePermissionRequest::Allow(JS::HandleValue aChoices)
   return NS_OK;
 }
 
-NS_IMETHODIMP
-PersistentStoragePermissionRequest::GetRequester(
-                                     nsIContentPermissionRequester** aRequester)
-{
-  MOZ_ASSERT(NS_IsMainThread());
-  MOZ_ASSERT(aRequester);
-
-  nsCOMPtr<nsIContentPermissionRequester> requester = mRequester;
-  requester.forget(aRequester);
-
-  return NS_OK;
-}
-
-NS_IMETHODIMP
-PersistentStoragePermissionRequest::GetTypes(nsIArray** aTypes)
-{
-  MOZ_ASSERT(aTypes);
-
-  nsTArray<nsString> emptyOptions;
-
-  return nsContentPermissionUtils::CreatePermissionArray(
-                                       NS_LITERAL_CSTRING("persistent-storage"),
-                                       NS_LITERAL_CSTRING("unused"),
-                                       emptyOptions,
-                                       aTypes);
-}
-
 /*******************************************************************************
  * StorageManager
  ******************************************************************************/
 
-StorageManager::StorageManager(nsIGlobalObject* aGlobal)
-  : mOwner(aGlobal)
-{
+StorageManager::StorageManager(nsIGlobalObject* aGlobal) : mOwner(aGlobal) {
   MOZ_ASSERT(aGlobal);
 }
 
-StorageManager::~StorageManager()
-{
-}
-
-already_AddRefed<Promise>
-StorageManager::Persisted(ErrorResult& aRv)
-{
-  MOZ_ASSERT(mOwner);
-
-  return ExecuteOpOnMainOrWorkerThread(mOwner,
-                                       RequestResolver::Type::Persisted,
-                                       aRv);
-}
-
-already_AddRefed<Promise>
-StorageManager::Persist(ErrorResult& aRv)
-{
-  MOZ_ASSERT(mOwner);
-
-  Telemetry::ScalarAdd(Telemetry::ScalarID::NAVIGATOR_STORAGE_PERSIST_COUNT, 1);
-  return ExecuteOpOnMainOrWorkerThread(mOwner,
-                                       RequestResolver::Type::Persist,
-                                       aRv);
-}
-
-already_AddRefed<Promise>
-StorageManager::Estimate(ErrorResult& aRv)
-{
-  MOZ_ASSERT(mOwner);
-
-  Telemetry::ScalarAdd(Telemetry::ScalarID::NAVIGATOR_STORAGE_ESTIMATE_COUNT,
-                       1);
-  return ExecuteOpOnMainOrWorkerThread(mOwner,
-                                       RequestResolver::Type::Estimate,
-                                       aRv);
-}
-
-NS_IMPL_CYCLE_COLLECTION_WRAPPERCACHE(StorageManager, mOwner)
-
-NS_IMPL_CYCLE_COLLECTING_ADDREF(StorageManager)
-NS_IMPL_CYCLE_COLLECTING_RELEASE(StorageManager)
+StorageManager::~StorageManager() { Shutdown(); }
 
 NS_INTERFACE_MAP_BEGIN_CYCLE_COLLECTION(StorageManager)
   NS_WRAPPERCACHE_INTERFACE_MAP_ENTRY
   NS_INTERFACE_MAP_ENTRY(nsISupports)
 NS_INTERFACE_MAP_END
 
-JSObject*
-StorageManager::WrapObject(JSContext* aCx,
-                           JS::Handle<JSObject*> aGivenProto)
-{
-  return StorageManagerBinding::Wrap(aCx, this, aGivenProto);
+NS_IMPL_CYCLE_COLLECTING_ADDREF(StorageManager)
+NS_IMPL_CYCLE_COLLECTING_RELEASE(StorageManager)
+
+NS_IMPL_CYCLE_COLLECTION_WRAPPERCACHE_CLASS(StorageManager)
+NS_IMPL_CYCLE_COLLECTION_UNLINK_BEGIN(StorageManager)
+  tmp->Shutdown();
+  NS_IMPL_CYCLE_COLLECTION_UNLINK(mOwner)
+  NS_IMPL_CYCLE_COLLECTION_UNLINK_PRESERVED_WRAPPER
+NS_IMPL_CYCLE_COLLECTION_UNLINK_END
+NS_IMPL_CYCLE_COLLECTION_TRAVERSE_BEGIN(StorageManager)
+  NS_IMPL_CYCLE_COLLECTION_TRAVERSE(mOwner)
+  NS_IMPL_CYCLE_COLLECTION_TRAVERSE(mFileSystemManager)
+NS_IMPL_CYCLE_COLLECTION_TRAVERSE_END
+
+void StorageManager::Shutdown() {
+  if (mFileSystemManager) {
+    mFileSystemManager->Shutdown();
+    mFileSystemManager = nullptr;
+  }
 }
 
-} // namespace dom
-} // namespace mozilla
+already_AddRefed<FileSystemManager> StorageManager::GetFileSystemManager() {
+  if (!mFileSystemManager) {
+    MOZ_ASSERT(mOwner);
+
+    mFileSystemManager = MakeRefPtr<FileSystemManager>(mOwner, this);
+  }
+
+  return do_AddRef(mFileSystemManager);
+}
+
+// WebIDL Boilerplate
+
+JSObject* StorageManager::WrapObject(JSContext* aCx,
+                                     JS::Handle<JSObject*> aGivenProto) {
+  return StorageManager_Binding::Wrap(aCx, this, aGivenProto);
+}
+
+// WebIDL Interface
+
+already_AddRefed<Promise> StorageManager::Persisted(ErrorResult& aRv) {
+  MOZ_ASSERT(mOwner);
+
+  return ExecuteOpOnMainOrWorkerThread(mOwner, RequestResolver::Type::Persisted,
+                                       aRv);
+}
+
+already_AddRefed<Promise> StorageManager::Persist(ErrorResult& aRv) {
+  MOZ_ASSERT(mOwner);
+
+  Telemetry::ScalarAdd(Telemetry::ScalarID::NAVIGATOR_STORAGE_PERSIST_COUNT, 1);
+  return ExecuteOpOnMainOrWorkerThread(mOwner, RequestResolver::Type::Persist,
+                                       aRv);
+}
+
+already_AddRefed<Promise> StorageManager::Estimate(ErrorResult& aRv) {
+  MOZ_ASSERT(mOwner);
+
+  Telemetry::ScalarAdd(Telemetry::ScalarID::NAVIGATOR_STORAGE_ESTIMATE_COUNT,
+                       1);
+  return ExecuteOpOnMainOrWorkerThread(mOwner, RequestResolver::Type::Estimate,
+                                       aRv);
+}
+
+already_AddRefed<Promise> StorageManager::GetDirectory(ErrorResult& aRv) {
+  return RefPtr(GetFileSystemManager())->GetDirectory(aRv);
+}
+
+}  // namespace mozilla::dom
