@@ -8,29 +8,46 @@
 
 #include "MessageEventRunnable.h"
 #include "mozilla/dom/WorkerBinding.h"
+#include "mozilla/ProfilerLabels.h"
+#include "mozilla/ProfilerMarkers.h"
 #include "mozilla/TimelineConsumers.h"
+#include "mozilla/Unused.h"
 #include "mozilla/WorkerTimelineMarker.h"
 #include "nsContentUtils.h"
+#include "nsGlobalWindowOuter.h"
 #include "WorkerPrivate.h"
 
-namespace mozilla {
-namespace dom {
+#ifdef XP_WIN
+#  undef PostMessage
+#endif
 
-/* static */ already_AddRefed<Worker> Worker::Constructor(
-    const GlobalObject& aGlobal, const nsAString& aScriptURL,
-    const WorkerOptions& aOptions, ErrorResult& aRv) {
+namespace mozilla::dom {
+
+/* static */
+already_AddRefed<Worker> Worker::Constructor(const GlobalObject& aGlobal,
+                                             const nsAString& aScriptURL,
+                                             const WorkerOptions& aOptions,
+                                             ErrorResult& aRv) {
   JSContext* cx = aGlobal.Context();
 
-  RefPtr<WorkerPrivate> workerPrivate = WorkerPrivate::Constructor(
-      cx, aScriptURL, false /* aIsChromeWorker */, WorkerTypeDedicated,
-      aOptions.mName, VoidCString(), nullptr /*aLoadInfo */, aRv);
-  if (NS_WARN_IF(aRv.Failed())) {
-    return nullptr;
-  }
-
   nsCOMPtr<nsIGlobalObject> globalObject =
       do_QueryInterface(aGlobal.GetAsSupports());
 
+  if (globalObject->AsInnerWindow() &&
+      !globalObject->AsInnerWindow()->IsCurrentInnerWindow()) {
+    aRv.ThrowInvalidStateError(
+        "Cannot create worker for a going to be discarded document");
+    return nullptr;
+  }
+
+  RefPtr<WorkerPrivate> workerPrivate = WorkerPrivate::Constructor(
+      cx, aScriptURL, false /* aIsChromeWorker */, WorkerKindDedicated,
+      aOptions.mCredentials, aOptions.mType, aOptions.mName, VoidCString(),
+      nullptr /*aLoadInfo */, aRv);
+  if (NS_WARN_IF(aRv.Failed())) {
+    return nullptr;
+  }
+
   RefPtr<Worker> worker = new Worker(globalObject, workerPrivate.forget());
   return worker.forget();
 }
@@ -38,7 +55,7 @@ namespace dom {
 Worker::Worker(nsIGlobalObject* aGlobalObject,
                already_AddRefed<WorkerPrivate> aWorkerPrivate)
     : DOMEventTargetHelper(aGlobalObject),
-      mWorkerPrivate(Move(aWorkerPrivate)) {
+      mWorkerPrivate(std::move(aWorkerPrivate)) {
   MOZ_ASSERT(mWorkerPrivate);
   mWorkerPrivate->SetParentEventTargetRef(this);
 }
@@ -48,7 +65,7 @@ Worker::~Worker() { Terminate(); }
 JSObject* Worker::WrapObject(JSContext* aCx,
                              JS::Handle<JSObject*> aGivenProto) {
   JS::Rooted<JSObject*> wrapper(aCx,
-                                WorkerBinding::Wrap(aCx, this, aGivenProto));
+                                Worker_Binding::Wrap(aCx, this, aGivenProto));
   if (wrapper) {
     // Most DOM objects don't assume they have a reflector. If they don't have
     // one and need one, they create it. But in workers code, we assume that the
@@ -69,6 +86,8 @@ void Worker::PostMessage(JSContext* aCx, JS::Handle<JS::Value> aMessage,
   if (!mWorkerPrivate || mWorkerPrivate->ParentStatusProtected() > Running) {
     return;
   }
+  RefPtr<WorkerPrivate> workerPrivate = mWorkerPrivate;
+  Unused << workerPrivate;
 
   JS::Rooted<JS::Value> transferable(aCx, JS::UndefinedValue());
 
@@ -78,13 +97,29 @@ void Worker::PostMessage(JSContext* aCx, JS::Handle<JS::Value> aMessage,
     return;
   }
 
+  NS_ConvertUTF16toUTF8 nameOrScriptURL(
+      mWorkerPrivate->WorkerName().IsEmpty()
+          ? Substring(
+                mWorkerPrivate->ScriptURL(), 0,
+                std::min(size_t(1024), mWorkerPrivate->ScriptURL().Length()))
+          : Substring(
+                mWorkerPrivate->WorkerName(), 0,
+                std::min(size_t(1024), mWorkerPrivate->WorkerName().Length())));
+  AUTO_PROFILER_MARKER_TEXT("Worker.postMessage", DOM, {}, nameOrScriptURL);
+  uint32_t flags = uint32_t(js::ProfilingStackFrame::Flags::RELEVANT_FOR_JS);
+  if (mWorkerPrivate->IsChromeWorker()) {
+    flags |= uint32_t(js::ProfilingStackFrame::Flags::NONSENSITIVE);
+  }
+  mozilla::AutoProfilerLabel PROFILER_RAII(
+      "Worker.postMessage", nameOrScriptURL.get(),
+      JS::ProfilingCategoryPair::DOM, flags);
+
   RefPtr<MessageEventRunnable> runnable = new MessageEventRunnable(
       mWorkerPrivate, WorkerRunnable::WorkerThreadModifyBusyCount);
 
   UniquePtr<AbstractTimelineMarker> start;
   UniquePtr<AbstractTimelineMarker> end;
-  RefPtr<TimelineConsumers> timelines = TimelineConsumers::Get();
-  bool isTimelineRecording = timelines && !timelines->IsEmpty();
+  bool isTimelineRecording = !TimelineConsumers::IsEmpty();
 
   if (isTimelineRecording) {
     start = MakeUnique<WorkerTimelineMarker>(
@@ -94,7 +129,27 @@ void Worker::PostMessage(JSContext* aCx, JS::Handle<JS::Value> aMessage,
         MarkerTracingType::START);
   }
 
-  runnable->Write(aCx, aMessage, transferable, JS::CloneDataPolicy(), aRv);
+  JS::CloneDataPolicy clonePolicy;
+  // DedicatedWorkers are always part of the same agent cluster.
+  clonePolicy.allowIntraClusterClonableSharedObjects();
+
+  if (NS_IsMainThread()) {
+    nsGlobalWindowInner* win = nsContentUtils::IncumbentInnerWindow();
+    if (win && win->IsSharedMemoryAllowed()) {
+      clonePolicy.allowSharedMemoryObjects();
+    }
+  } else {
+    WorkerPrivate* worker = GetCurrentThreadWorkerPrivate();
+    if (worker && worker->IsSharedMemoryAllowed()) {
+      clonePolicy.allowSharedMemoryObjects();
+    }
+  }
+
+  runnable->Write(aCx, aMessage, transferable, clonePolicy, aRv);
+
+  if (!mWorkerPrivate || mWorkerPrivate->ParentStatusProtected() > Running) {
+    return;
+  }
 
   if (isTimelineRecording) {
     end = MakeUnique<WorkerTimelineMarker>(
@@ -102,24 +157,31 @@ void Worker::PostMessage(JSContext* aCx, JS::Handle<JS::Value> aMessage,
             ? ProfileTimelineWorkerOperationType::SerializeDataOnMainThread
             : ProfileTimelineWorkerOperationType::SerializeDataOffMainThread,
         MarkerTracingType::END);
-    timelines->AddMarkerForAllObservedDocShells(start);
-    timelines->AddMarkerForAllObservedDocShells(end);
+    TimelineConsumers::AddMarkerForAllObservedDocShells(start);
+    TimelineConsumers::AddMarkerForAllObservedDocShells(end);
   }
 
   if (NS_WARN_IF(aRv.Failed())) {
     return;
   }
 
-  if (!runnable->Dispatch()) {
-    aRv.Throw(NS_ERROR_FAILURE);
-  }
+  // The worker could have closed between the time we entered this function and
+  // checked ParentStatusProtected and now, which could cause the dispatch to
+  // fail.
+  Unused << NS_WARN_IF(!runnable->Dispatch());
+}
+
+void Worker::PostMessage(JSContext* aCx, JS::Handle<JS::Value> aMessage,
+                         const StructuredSerializeOptions& aOptions,
+                         ErrorResult& aRv) {
+  PostMessage(aCx, aMessage, aOptions.mTransfer, aRv);
 }
 
 void Worker::Terminate() {
   NS_ASSERT_OWNINGTHREAD(Worker);
 
   if (mWorkerPrivate) {
-    mWorkerPrivate->Terminate();
+    mWorkerPrivate->Cancel();
     mWorkerPrivate = nullptr;
   }
 }
@@ -134,6 +196,7 @@ NS_IMPL_CYCLE_COLLECTION_TRAVERSE_END
 
 NS_IMPL_CYCLE_COLLECTION_UNLINK_BEGIN_INHERITED(Worker, DOMEventTargetHelper)
   tmp->Terminate();
+  NS_IMPL_CYCLE_COLLECTION_UNLINK_WEAK_PTR
 NS_IMPL_CYCLE_COLLECTION_UNLINK_END
 
 NS_IMPL_CYCLE_COLLECTION_TRACE_BEGIN_INHERITED(Worker, DOMEventTargetHelper)
@@ -145,5 +208,4 @@ NS_INTERFACE_MAP_END_INHERITING(DOMEventTargetHelper)
 NS_IMPL_ADDREF_INHERITED(Worker, DOMEventTargetHelper)
 NS_IMPL_RELEASE_INHERITED(Worker, DOMEventTargetHelper)
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
