@@ -6,25 +6,34 @@
 
 #include "WorkerRunnable.h"
 
-#include "nsGlobalWindow.h"
+#include "WorkerScope.h"
+#include "js/RootingAPI.h"
+#include "jsapi.h"
+#include "jsfriendapi.h"
+#include "mozilla/AlreadyAddRefed.h"
+#include "mozilla/AppShutdown.h"
+#include "mozilla/Assertions.h"
+#include "mozilla/CycleCollectedJSContext.h"
+#include "mozilla/DebugOnly.h"
+#include "mozilla/ErrorResult.h"
+#include "mozilla/Maybe.h"
+#include "mozilla/Telemetry.h"
+#include "mozilla/TelemetryHistogramEnums.h"
+#include "mozilla/TimeStamp.h"
+#include "mozilla/Unused.h"
+#include "mozilla/dom/ScriptSettings.h"
+#include "mozilla/dom/Worker.h"
+#include "mozilla/dom/WorkerCommon.h"
+#include "nsDebug.h"
+#include "nsGlobalWindowInner.h"
+#include "nsID.h"
 #include "nsIEventTarget.h"
 #include "nsIGlobalObject.h"
 #include "nsIRunnable.h"
 #include "nsThreadUtils.h"
+#include "nsWrapperCacheInlines.h"
 
-#include "mozilla/DebugOnly.h"
-#include "mozilla/ErrorResult.h"
-#include "mozilla/dom/ScriptSettings.h"
-#include "mozilla/Telemetry.h"
-
-#include "js/RootingAPI.h"
-#include "js/Value.h"
-
-#include "WorkerPrivate.h"
-#include "WorkerScope.h"
-
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 namespace {
 
@@ -111,6 +120,13 @@ bool WorkerRunnable::DispatchInternal() {
     return NS_SUCCEEDED(parent->Dispatch(runnable.forget()));
   }
 
+  if (IsDebuggeeRunnable()) {
+    RefPtr<WorkerDebuggeeRunnable> debuggeeRunnable =
+        runnable.forget().downcast<WorkerDebuggeeRunnable>();
+    return NS_SUCCEEDED(mWorkerPrivate->DispatchDebuggeeToMainThread(
+        debuggeeRunnable.forget(), NS_DISPATCH_NORMAL));
+  }
+
   return NS_SUCCEEDED(mWorkerPrivate->DispatchToMainThread(runnable.forget()));
 }
 
@@ -209,6 +225,33 @@ WorkerRunnable::Run() {
   bool targetIsWorkerThread = mBehavior == WorkerThreadModifyBusyCount ||
                               mBehavior == WorkerThreadUnchangedBusyCount;
 
+  if (targetIsWorkerThread) {
+    // On a worker thread, a WorkerRunnable should only run when there is an
+    // underlying WorkerThreadPrimaryRunnable active, which means we should
+    // find a CycleCollectedJSContext.
+    if (!CycleCollectedJSContext::Get()) {
+#if (defined(MOZ_COLLECTING_RUNNABLE_TELEMETRY) && defined(DEBUG))
+      // Temporarily set the LogLevel high enough to be certain the messages
+      // are visible.
+      LogModule* module = sWorkerRunnableLog;
+      LogLevel prevLevel = module->Level();
+      if (prevLevel < LogLevel::Error) {
+        module->SetLevel(LogLevel::Error);
+      }
+      MOZ_LOG(sWorkerRunnableLog, LogLevel::Error,
+              ("Runnable '%s' was executed after WorkerThreadPrimaryRunnable "
+               "ended.",
+               "WorkerRunnable"));
+      module->SetLevel(prevLevel);
+#endif
+      MOZ_DIAGNOSTIC_ASSERT(false,
+                            "A WorkerRunnable was executed after "
+                            "WorkerThreadPrimaryRunnable ended.");
+
+      return NS_OK;
+    }
+  }
+
 #ifdef DEBUG
   MOZ_ASSERT_IF(mCallingCancelWithinRun, targetIsWorkerThread);
   if (targetIsWorkerThread) {
@@ -250,7 +293,7 @@ WorkerRunnable::Run() {
                "ScriptExecutorRunnable");
     mWorkerPrivate->AssertIsOnWorkerThread();
     MOZ_ASSERT(!JS_IsExceptionPending(mWorkerPrivate->GetJSContext()));
-    // We can't enter a useful compartment on the JSContext here; just pass it
+    // We can't enter a useful realm on the JSContext here; just pass it
     // in as-is.
     PostRun(mWorkerPrivate->GetJSContext(), mWorkerPrivate, false);
     return NS_ERROR_FAILURE;
@@ -262,16 +305,22 @@ WorkerRunnable::Run() {
   MOZ_ASSERT(isMainThread == NS_IsMainThread());
   RefPtr<WorkerPrivate> kungFuDeathGrip;
   if (targetIsWorkerThread) {
-    JSContext* cx = GetCurrentWorkerThreadJSContext();
-    if (NS_WARN_IF(!cx)) {
-      return NS_ERROR_FAILURE;
-    }
-
-    JSObject* global = JS::CurrentGlobalOrNull(cx);
-    if (global) {
-      globalObject = xpc::NativeGlobal(global);
-    } else {
+    globalObject = mWorkerPrivate->GetCurrentEventLoopGlobal();
+    if (!globalObject) {
       globalObject = DefaultGlobalObject();
+      // Our worker thread may not be in a good state here if there is no
+      // JSContext avaliable.  The way this manifests itself is that
+      // globalObject ends up null (though it's not clear to me how we can be
+      // running runnables at all when DefaultGlobalObject() is returning
+      // false!) and then when we try to init the AutoJSAPI either
+      // CycleCollectedJSContext::Get() returns null or it has a null JSContext.
+      // In any case, we used to have a check for
+      // GetCurrentWorkerThreadJSContext() being non-null here and that seems to
+      // avoid the problem, so let's keep doing that check even if we don't need
+      // the JSContext here at all.
+      if (NS_WARN_IF(!globalObject && !GetCurrentWorkerThreadJSContext())) {
+        return NS_ERROR_FAILURE;
+      }
     }
 
     // We may still not have a globalObject here: in the case of
@@ -322,39 +371,38 @@ WorkerRunnable::Run() {
   MOZ_ASSERT_IF(!targetIsWorkerThread && !isMainThread,
                 mWorkerPrivate->IsDedicatedWorker() && globalObject);
 
-  // If we're on the parent thread we might be in a null compartment in the
+  // If we're on the parent thread we might be in a null realm in the
   // situation described above when globalObject is null.  Make sure to enter
-  // the compartment of the worker's reflector if there is one.  There might
+  // the realm of the worker's reflector if there is one.  There might
   // not be one if we're just starting to compile the script for this worker.
-  Maybe<JSAutoCompartment> ac;
+  Maybe<JSAutoRealm> ar;
   if (!targetIsWorkerThread && mWorkerPrivate->IsDedicatedWorker() &&
       mWorkerPrivate->ParentEventTargetRef()->GetWrapper()) {
     JSObject* wrapper = mWorkerPrivate->ParentEventTargetRef()->GetWrapper();
 
     // If we're on the parent thread and have a reflector and a globalObject,
-    // then the compartments of cx, globalObject, and the worker's reflector
+    // then the realms of cx, globalObject, and the worker's reflector
     // should all match.
-    MOZ_ASSERT_IF(globalObject, js::GetObjectCompartment(wrapper) ==
-                                    js::GetContextCompartment(cx));
-    MOZ_ASSERT_IF(globalObject, js::GetObjectCompartment(wrapper) ==
-                                    js::GetObjectCompartment(
-                                        globalObject->GetGlobalJSObject()));
+    MOZ_ASSERT_IF(globalObject,
+                  js::GetNonCCWObjectRealm(wrapper) == js::GetContextRealm(cx));
+    MOZ_ASSERT_IF(globalObject,
+                  js::GetNonCCWObjectRealm(wrapper) ==
+                      js::GetNonCCWObjectRealm(
+                          globalObject->GetGlobalJSObjectPreserveColor()));
 
     // If we're on the parent thread and have a reflector, then our
-    // JSContext had better be either in the null compartment (and hence
-    // have no globalObject) or in the compartment of our reflector.
-    MOZ_ASSERT(
-        !js::GetContextCompartment(cx) ||
-            js::GetObjectCompartment(wrapper) == js::GetContextCompartment(cx),
-        "Must either be in the null compartment or in our reflector "
-        "compartment");
+    // JSContext had better be either in the null realm (and hence
+    // have no globalObject) or in the realm of our reflector.
+    MOZ_ASSERT(!js::GetContextRealm(cx) ||
+                   js::GetNonCCWObjectRealm(wrapper) == js::GetContextRealm(cx),
+               "Must either be in the null compartment or in our reflector "
+               "compartment");
 
-    ac.emplace(cx, wrapper);
+    ar.emplace(cx, wrapper);
   }
 
   MOZ_ASSERT(!jsapi->HasException());
   result = WorkerRun(cx, mWorkerPrivate);
-  MOZ_ASSERT_IF(result, !jsapi->HasException());
   jsapi->ReportException();
 
   // We can't even assert that this didn't create our global, since in the case
@@ -409,10 +457,9 @@ WorkerSyncRunnable::WorkerSyncRunnable(WorkerPrivate* aWorkerPrivate,
 }
 
 WorkerSyncRunnable::WorkerSyncRunnable(
-    WorkerPrivate* aWorkerPrivate,
-    already_AddRefed<nsIEventTarget>&& aSyncLoopTarget)
+    WorkerPrivate* aWorkerPrivate, nsCOMPtr<nsIEventTarget>&& aSyncLoopTarget)
     : WorkerRunnable(aWorkerPrivate, WorkerThreadUnchangedBusyCount),
-      mSyncLoopTarget(aSyncLoopTarget) {
+      mSyncLoopTarget(std::move(aSyncLoopTarget)) {
 #ifdef DEBUG
   if (mSyncLoopTarget) {
     mWorkerPrivate->AssertValidSyncLoop(mSyncLoopTarget);
@@ -420,7 +467,7 @@ WorkerSyncRunnable::WorkerSyncRunnable(
 #endif
 }
 
-WorkerSyncRunnable::~WorkerSyncRunnable() {}
+WorkerSyncRunnable::~WorkerSyncRunnable() = default;
 
 bool WorkerSyncRunnable::DispatchInternal() {
   if (mSyncLoopTarget) {
@@ -436,9 +483,9 @@ void MainThreadWorkerSyncRunnable::PostDispatch(WorkerPrivate* aWorkerPrivate,
                                                 bool aDispatchResult) {}
 
 MainThreadStopSyncLoopRunnable::MainThreadStopSyncLoopRunnable(
-    WorkerPrivate* aWorkerPrivate,
-    already_AddRefed<nsIEventTarget>&& aSyncLoopTarget, bool aResult)
-    : WorkerSyncRunnable(aWorkerPrivate, Move(aSyncLoopTarget)),
+    WorkerPrivate* aWorkerPrivate, nsCOMPtr<nsIEventTarget>&& aSyncLoopTarget,
+    nsresult aResult)
+    : WorkerSyncRunnable(aWorkerPrivate, std::move(aSyncLoopTarget)),
       mResult(aResult) {
   AssertIsOnMainThread();
 #ifdef DEBUG
@@ -447,13 +494,14 @@ MainThreadStopSyncLoopRunnable::MainThreadStopSyncLoopRunnable(
 }
 
 nsresult MainThreadStopSyncLoopRunnable::Cancel() {
-  nsresult rv = Run();
+  // We need to check first if cancel is called twice
+  nsresult rv = WorkerSyncRunnable::Cancel();
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  rv = Run();
   NS_WARNING_ASSERTION(NS_SUCCEEDED(rv), "Run() failed");
 
-  nsresult rv2 = WorkerSyncRunnable::Cancel();
-  NS_WARNING_ASSERTION(NS_SUCCEEDED(rv2), "Cancel() failed");
-
-  return NS_FAILED(rv) ? rv : rv2;
+  return rv;
 }
 
 bool MainThreadStopSyncLoopRunnable::WorkerRun(JSContext* aCx,
@@ -491,11 +539,15 @@ WorkerControlRunnable::WorkerControlRunnable(WorkerPrivate* aWorkerPrivate,
 #endif
 
 nsresult WorkerControlRunnable::Cancel() {
+  // We need to check first if cancel is called twice
+  nsresult rv = WorkerRunnable::Cancel();
+  NS_ENSURE_SUCCESS(rv, rv);
+
   if (NS_FAILED(Run())) {
     NS_WARNING("WorkerControlRunnable::Run() failed.");
   }
 
-  return WorkerRunnable::Cancel();
+  return NS_OK;
 }
 
 bool WorkerControlRunnable::DispatchInternal() {
@@ -521,6 +573,8 @@ WorkerMainThreadRunnable::WorkerMainThreadRunnable(
   mWorkerPrivate->AssertIsOnWorkerThread();
 }
 
+WorkerMainThreadRunnable::~WorkerMainThreadRunnable() = default;
+
 void WorkerMainThreadRunnable::Dispatch(WorkerStatus aFailStatus,
                                         mozilla::ErrorResult& aRv) {
   mWorkerPrivate->AssertIsOnWorkerThread();
@@ -529,7 +583,7 @@ void WorkerMainThreadRunnable::Dispatch(WorkerStatus aFailStatus,
 
   AutoSyncLoopHolder syncLoop(mWorkerPrivate, aFailStatus);
 
-  mSyncLoopTarget = syncLoop.GetEventTarget();
+  mSyncLoopTarget = syncLoop.GetSerialEventTarget();
   if (!mSyncLoopTarget) {
     // SyncLoop creation can fail if the worker is shutting down.
     aRv.Throw(NS_ERROR_DOM_INVALID_STATE_ERR);
@@ -541,7 +595,7 @@ void WorkerMainThreadRunnable::Dispatch(WorkerStatus aFailStatus,
       NS_SUCCEEDED(rv),
       "Should only fail after xpcom-shutdown-threads and we're gone by then");
 
-  bool success = syncLoop.Run();
+  bool success = NS_SUCCEEDED(syncLoop.Run());
 
   Telemetry::Accumulate(
       Telemetry::SYNC_WORKER_OPERATION, mTelemetryKey,
@@ -559,11 +613,18 @@ NS_IMETHODIMP
 WorkerMainThreadRunnable::Run() {
   AssertIsOnMainThread();
 
+  // This shouldn't be necessary once we're better about making sure no workers
+  // are created during shutdown in earlier phases.
+  if (AppShutdown::IsInOrBeyond(ShutdownPhase::XPCOMShutdownThreads)) {
+    return NS_ERROR_ILLEGAL_DURING_SHUTDOWN;
+  }
+
   bool runResult = MainThreadRun();
 
   RefPtr<MainThreadStopSyncLoopRunnable> response =
       new MainThreadStopSyncLoopRunnable(mWorkerPrivate,
-                                         mSyncLoopTarget.forget(), runResult);
+                                         std::move(mSyncLoopTarget),
+                                         runResult ? NS_OK : NS_ERROR_FAILURE);
 
   MOZ_ALWAYS_TRUE(response->Dispatch());
 
@@ -593,27 +654,31 @@ void WorkerSameThreadRunnable::PostDispatch(WorkerPrivate* aWorkerPrivate,
   }
 }
 
-WorkerProxyToMainThreadRunnable::WorkerProxyToMainThreadRunnable(
-    WorkerPrivate* aWorkerPrivate)
-    : mozilla::Runnable("dom::WorkerProxyToMainThreadRunnable"),
-      mWorkerPrivate(aWorkerPrivate) {
-  MOZ_ASSERT(mWorkerPrivate);
-  mWorkerPrivate->AssertIsOnWorkerThread();
-}
+WorkerProxyToMainThreadRunnable::WorkerProxyToMainThreadRunnable()
+    : mozilla::Runnable("dom::WorkerProxyToMainThreadRunnable") {}
 
-WorkerProxyToMainThreadRunnable::~WorkerProxyToMainThreadRunnable() {}
+WorkerProxyToMainThreadRunnable::~WorkerProxyToMainThreadRunnable() = default;
 
-bool WorkerProxyToMainThreadRunnable::Dispatch() {
-  mWorkerPrivate->AssertIsOnWorkerThread();
+bool WorkerProxyToMainThreadRunnable::Dispatch(WorkerPrivate* aWorkerPrivate) {
+  MOZ_ASSERT(aWorkerPrivate);
+  aWorkerPrivate->AssertIsOnWorkerThread();
 
-  if (NS_WARN_IF(!HoldWorker())) {
-    RunBackOnWorkerThreadForCleanup();
+  RefPtr<StrongWorkerRef> workerRef = StrongWorkerRef::Create(
+      aWorkerPrivate, "WorkerProxyToMainThreadRunnable");
+  if (NS_WARN_IF(!workerRef)) {
+    RunBackOnWorkerThreadForCleanup(aWorkerPrivate);
     return false;
   }
 
-  if (NS_WARN_IF(NS_FAILED(mWorkerPrivate->DispatchToMainThread(this)))) {
+  MOZ_ASSERT(!mWorkerRef);
+  mWorkerRef = new ThreadSafeWorkerRef(workerRef);
+
+  if (ForMessaging()
+          ? NS_WARN_IF(NS_FAILED(
+                aWorkerPrivate->DispatchToMainThreadForMessaging(this)))
+          : NS_WARN_IF(NS_FAILED(aWorkerPrivate->DispatchToMainThread(this)))) {
     ReleaseWorker();
-    RunBackOnWorkerThreadForCleanup();
+    RunBackOnWorkerThreadForCleanup(aWorkerPrivate);
     return false;
   }
 
@@ -623,7 +688,7 @@ bool WorkerProxyToMainThreadRunnable::Dispatch() {
 NS_IMETHODIMP
 WorkerProxyToMainThreadRunnable::Run() {
   AssertIsOnMainThread();
-  RunOnMainThread();
+  RunOnMainThread(mWorkerRef->Private());
   PostDispatchOnMainThread();
   return NS_OK;
 }
@@ -653,7 +718,7 @@ void WorkerProxyToMainThreadRunnable::PostDispatchOnMainThread() {
       aWorkerPrivate->AssertIsOnWorkerThread();
 
       if (mRunnable) {
-        mRunnable->RunBackOnWorkerThreadForCleanup();
+        mRunnable->RunBackOnWorkerThreadForCleanup(aWorkerPrivate);
 
         // Let's release the worker thread.
         mRunnable->ReleaseWorker();
@@ -664,44 +729,28 @@ void WorkerProxyToMainThreadRunnable::PostDispatchOnMainThread() {
     }
 
    private:
-    ~ReleaseRunnable() {}
+    ~ReleaseRunnable() = default;
   };
 
   RefPtr<WorkerControlRunnable> runnable =
-      new ReleaseRunnable(mWorkerPrivate, this);
+      new ReleaseRunnable(mWorkerRef->Private(), this);
   Unused << NS_WARN_IF(!runnable->Dispatch());
 }
 
-bool WorkerProxyToMainThreadRunnable::HoldWorker() {
-  mWorkerPrivate->AssertIsOnWorkerThread();
-  MOZ_ASSERT(!mWorkerHolder);
+void WorkerProxyToMainThreadRunnable::ReleaseWorker() { mWorkerRef = nullptr; }
 
-  class SimpleWorkerHolder final : public WorkerHolder {
-   public:
-    SimpleWorkerHolder()
-        : WorkerHolder("WorkerProxyToMainThreadRunnable::SimpleWorkerHolder") {}
-
-    bool Notify(WorkerStatus aStatus) override {
-      // We don't care about the notification. We just want to keep the
-      // mWorkerPrivate alive.
-      return true;
+bool WorkerDebuggeeRunnable::PreDispatch(WorkerPrivate* aWorkerPrivate) {
+  if (mBehavior == ParentThreadUnchangedBusyCount) {
+    RefPtr<StrongWorkerRef> strongRef = StrongWorkerRef::Create(
+        aWorkerPrivate, "WorkerDebuggeeRunnable::mSender");
+    if (!strongRef) {
+      return false;
     }
-  };
 
-  UniquePtr<WorkerHolder> workerHolder(new SimpleWorkerHolder());
-  if (NS_WARN_IF(!workerHolder->HoldWorker(mWorkerPrivate, Canceling))) {
-    return false;
+    mSender = new ThreadSafeWorkerRef(strongRef);
   }
 
-  mWorkerHolder = Move(workerHolder);
-  return true;
+  return WorkerRunnable::PreDispatch(aWorkerPrivate);
 }
 
-void WorkerProxyToMainThreadRunnable::ReleaseWorker() {
-  mWorkerPrivate->AssertIsOnWorkerThread();
-  MOZ_ASSERT(mWorkerHolder);
-  mWorkerHolder = nullptr;
-}
-
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
