@@ -116,11 +116,16 @@ static nsresult createAndAddToResult(nsAtom* aName, const nsAString& aValue,
   nsresult rv = text->SetText(aValue, false);
   NS_ENSURE_SUCCESS(rv, rv);
 
-  rv = elem->AppendChildTo(text, false);
-  NS_ENSURE_SUCCESS(rv, rv);
+  ErrorResult error;
+  elem->AppendChildTo(text, false, error);
+  if (error.Failed()) {
+    return error.StealNSResult();
+  }
 
-  rv = aResultHolder->AppendChildTo(elem, false);
-  NS_ENSURE_SUCCESS(rv, rv);
+  aResultHolder->AppendChildTo(elem, false, error);
+  if (error.Failed()) {
+    return error.StealNSResult();
+  }
 
   UniquePtr<txXPathNode> xpathNode(
       txXPathNativeNode::createXPathNode(elem, true));
@@ -330,15 +335,14 @@ nsresult txEXSLTFunctionCall::evaluate(txIEvalContext* aContext,
       rv = aContext->recycler()->getNodeSet(getter_AddRefs(resultSet));
       NS_ENSURE_SUCCESS(rv, rv);
 
-      nsTHashtable<nsStringHashKey> hash;
+      nsTHashSet<nsString> hash;
 
       int32_t i, len = nodes->size();
       for (i = 0; i < len; ++i) {
         nsAutoString str;
         const txXPathNode& node = nodes->get(i);
         txXPathNodeUtils::appendNodeValue(node, str);
-        if (!hash.GetEntry(str)) {
-          hash.PutEntry(str);
+        if (hash.EnsureInserted(str)) {
           rv = resultSet->append(node);
           NS_ENSURE_SUCCESS(rv, rv);
         }
@@ -527,7 +531,7 @@ nsresult txEXSLTFunctionCall::evaluate(txIEvalContext* aContext,
         nsAutoString str;
         txXPathNodeUtils::appendNodeValue(nodes->get(i), str);
         double val = txDouble::toDouble(str);
-        if (mozilla::IsNaN(val)) {
+        if (std::isnan(val)) {
           res = UnspecifiedNaN<double>();
           break;
         }
@@ -564,7 +568,7 @@ nsresult txEXSLTFunctionCall::evaluate(txIEvalContext* aContext,
         const txXPathNode& node = nodes->get(i);
         txXPathNodeUtils::appendNodeValue(node, str);
         double val = txDouble::toDouble(str);
-        if (mozilla::IsNaN(val)) {
+        if (std::isnan(val)) {
           resultSet->clear();
           break;
         }
@@ -616,8 +620,7 @@ nsresult txEXSLTFunctionCall::evaluate(txIEvalContext* aContext,
       return NS_OK;
     }
     default: {
-      aContext->receiveError(NS_LITERAL_STRING("Internal error"),
-                             NS_ERROR_UNEXPECTED);
+      aContext->receiveError(u"Internal error"_ns, NS_ERROR_UNEXPECTED);
       return NS_ERROR_UNEXPECTED;
     }
   }
@@ -667,7 +670,7 @@ nsresult txEXSLTRegExFunctionCall::evaluate(txIEvalContext* aContext,
   }
 
   nsCOMPtr<txIEXSLTFunctions> funcs =
-      do_ImportModule("resource://gre/modules/txEXSLTRegExFunctions.jsm");
+      do_ImportESModule("resource://gre/modules/txEXSLTRegExFunctions.sys.mjs");
   MOZ_ALWAYS_TRUE(funcs);
 
   switch (mType) {
@@ -721,8 +724,7 @@ nsresult txEXSLTRegExFunctionCall::evaluate(txIEvalContext* aContext,
       return NS_OK;
     }
     default: {
-      aContext->receiveError(NS_LITERAL_STRING("Internal error"),
-                             NS_ERROR_UNEXPECTED);
+      aContext->receiveError(u"Internal error"_ns, NS_ERROR_UNEXPECTED);
       return NS_ERROR_UNEXPECTED;
     }
   }
@@ -762,39 +764,39 @@ extern nsresult TX_ConstructEXSLTFunction(nsAtom* aName, int32_t aNamespaceID,
 }
 
 extern bool TX_InitEXSLTFunction() {
-#define EXSLT_FUNCS(NS, CLASS, ...)                                           \
-  {                                                                           \
-    int32_t nsid = txNamespaceManager::getNamespaceID(NS_LITERAL_STRING(NS)); \
-    if (nsid == kNameSpaceID_Unknown) {                                       \
-      return false;                                                           \
-    }                                                                         \
-    MOZ_FOR_EACH(EXSLT_FUNC, (nsid, CLASS, ), (__VA_ARGS__))                  \
+#define EXSLT_FUNCS(NS, CLASS, ...)                                         \
+  {                                                                         \
+    int32_t nsid = txNamespaceManager::getNamespaceID(nsLiteralString(NS)); \
+    if (nsid == kNameSpaceID_Unknown) {                                     \
+      return false;                                                         \
+    }                                                                       \
+    MOZ_FOR_EACH(EXSLT_FUNC, (nsid, CLASS, ), (__VA_ARGS__))                \
   }
 
 #define EXSLT_FUNC(NS, CLASS, ...)                      \
   descriptTable[txEXSLTType::MOZ_ARG_1 __VA_ARGS__] = { \
       MOZ_ARGS_AFTER_1 __VA_ARGS__, CLASS::Create, NS};
 
-  EXSLT_FUNCS("http://exslt.org/common", txEXSLTFunctionCall,
+  EXSLT_FUNCS(u"http://exslt.org/common", txEXSLTFunctionCall,
               (NODE_SET, 1, 1, Expr::NODESET_RESULT, nsGkAtoms::nodeSet),
               (OBJECT_TYPE, 1, 1, Expr::STRING_RESULT, nsGkAtoms::objectType))
 
-  EXSLT_FUNCS("http://exslt.org/dates-and-times", txEXSLTFunctionCall,
+  EXSLT_FUNCS(u"http://exslt.org/dates-and-times", txEXSLTFunctionCall,
               (DATE_TIME, 0, 0, Expr::STRING_RESULT, nsGkAtoms::dateTime))
 
-  EXSLT_FUNCS("http://exslt.org/math", txEXSLTFunctionCall,
+  EXSLT_FUNCS(u"http://exslt.org/math", txEXSLTFunctionCall,
               (MAX, 1, 1, Expr::NUMBER_RESULT, nsGkAtoms::max),
               (MIN, 1, 1, Expr::NUMBER_RESULT, nsGkAtoms::min),
               (HIGHEST, 1, 1, Expr::NODESET_RESULT, nsGkAtoms::highest),
               (LOWEST, 1, 1, Expr::NODESET_RESULT, nsGkAtoms::lowest))
 
-  EXSLT_FUNCS("http://exslt.org/regular-expressions", txEXSLTRegExFunctionCall,
+  EXSLT_FUNCS(u"http://exslt.org/regular-expressions", txEXSLTRegExFunctionCall,
               (MATCH, 2, 3, Expr::NODESET_RESULT, nsGkAtoms::match),
               (REPLACE, 4, 4, Expr::STRING_RESULT, nsGkAtoms::replace),
               (TEST, 2, 3, Expr::BOOLEAN_RESULT, nsGkAtoms::test))
 
   EXSLT_FUNCS(
-      "http://exslt.org/sets", txEXSLTFunctionCall,
+      u"http://exslt.org/sets", txEXSLTFunctionCall,
       (DIFFERENCE_, 2, 2, Expr::NODESET_RESULT, nsGkAtoms::difference),
       (DISTINCT, 1, 1, Expr::NODESET_RESULT, nsGkAtoms::distinct),
       (HAS_SAME_NODE, 2, 2, Expr::BOOLEAN_RESULT, nsGkAtoms::hasSameNode),
@@ -802,7 +804,7 @@ extern bool TX_InitEXSLTFunction() {
       (LEADING, 2, 2, Expr::NODESET_RESULT, nsGkAtoms::leading),
       (TRAILING, 2, 2, Expr::NODESET_RESULT, nsGkAtoms::trailing))
 
-  EXSLT_FUNCS("http://exslt.org/strings", txEXSLTFunctionCall,
+  EXSLT_FUNCS(u"http://exslt.org/strings", txEXSLTFunctionCall,
               (CONCAT, 1, 1, Expr::STRING_RESULT, nsGkAtoms::concat),
               (SPLIT, 1, 2, Expr::STRING_RESULT, nsGkAtoms::split),
               (TOKENIZE, 1, 2, Expr::STRING_RESULT, nsGkAtoms::tokenize))
