@@ -1,4 +1,4 @@
-/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
+/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
 /* This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
@@ -17,6 +17,7 @@
 #include "txStylesheet.h"
 #include "txXPathTreeWalker.h"
 #include "nsTArray.h"
+#include "mozilla/Result.h"
 
 class txAOutputHandlerFactory;
 class txAXMLEventHandler;
@@ -26,13 +27,15 @@ class txLoadedDocumentEntry : public nsStringHashKey {
  public:
   explicit txLoadedDocumentEntry(KeyTypePointer aStr)
       : nsStringHashKey(aStr), mLoadResult(NS_OK) {}
-  txLoadedDocumentEntry(const txLoadedDocumentEntry& aToCopy)
-      : nsStringHashKey(aToCopy) {
+  txLoadedDocumentEntry(txLoadedDocumentEntry&& aOther)
+      : nsStringHashKey(std::move(aOther)),
+        mDocument(std::move(aOther.mDocument)),
+        mLoadResult(std::move(aOther.mLoadResult)) {
     NS_ERROR("We're horked.");
   }
   ~txLoadedDocumentEntry() {
     if (mDocument) {
-      txXPathNodeUtils::release(mDocument);
+      txXPathNodeUtils::release(mDocument.get());
     }
   }
   bool LoadingFailed() {
@@ -42,7 +45,7 @@ class txLoadedDocumentEntry : public nsStringHashKey {
     return NS_FAILED(mLoadResult);
   }
 
-  nsAutoPtr<txXPathNode> mDocument;
+  mozilla::UniquePtr<txXPathNode> mDocument;
   nsresult mLoadResult;
 };
 
@@ -50,11 +53,11 @@ class txLoadedDocumentsHash : public nsTHashtable<txLoadedDocumentEntry> {
  public:
   txLoadedDocumentsHash() : nsTHashtable<txLoadedDocumentEntry>(4) {}
   ~txLoadedDocumentsHash();
-  MOZ_MUST_USE nsresult init(const txXPathNode& aSource);
+  [[nodiscard]] nsresult init(const txXPathNode& aSource);
 
  private:
   friend class txExecutionState;
-  nsAutoPtr<txXPathNode> mSourceDocument;
+  mozilla::UniquePtr<txXPathNode> mSourceDocument;
 };
 
 class txExecutionState : public txIMatchContext {
@@ -79,7 +82,7 @@ class txExecutionState : public txIMatchContext {
   };
 
   // Stack functions
-  nsresult pushEvalContext(txIEvalContext* aContext);
+  void pushEvalContext(txIEvalContext* aContext);
   txIEvalContext* popEvalContext();
 
   /**
@@ -89,9 +92,9 @@ class txExecutionState : public txIMatchContext {
    */
   void popAndDeleteEvalContextUntil(txIEvalContext* aContext);
 
-  nsresult pushBool(bool aBool);
+  void pushBool(bool aBool);
   bool popBool();
-  nsresult pushResultHandler(txAXMLEventHandler* aHandler);
+  void pushResultHandler(txAXMLEventHandler* aHandler);
   txAXMLEventHandler* popResultHandler();
   void pushTemplateRule(txStylesheet::ImportFrame* aFrame,
                         const txExpandedName& aMode, txParameterMap* aParams);
@@ -113,17 +116,18 @@ class txExecutionState : public txIMatchContext {
   }
 
   // state-modification functions
-  txInstruction* getNextInstruction();
+  mozilla::Result<txInstruction*, nsresult> getNextInstruction();
   nsresult runTemplate(txInstruction* aInstruction);
   nsresult runTemplate(txInstruction* aInstruction, txInstruction* aReturnTo);
   void gotoInstruction(txInstruction* aNext);
   void returnFromTemplate();
   nsresult bindVariable(const txExpandedName& aName, txAExprResult* aValue);
   void removeVariable(const txExpandedName& aName);
+  void stopProcessing() { mStopProcessing = true; }
 
   txAXMLEventHandler* mOutputHandler;
   txAXMLEventHandler* mResultHandler;
-  nsAutoPtr<txAXMLEventHandler> mObsoleteHandler;
+  mozilla::UniquePtr<txAXMLEventHandler> mObsoleteHandler;
   txAOutputHandlerFactory* mOutputHandlerFactory;
 
   RefPtr<txParameterMap> mTemplateParams;
@@ -154,6 +158,7 @@ class txExecutionState : public txIMatchContext {
   txKeyHash mKeyHash;
   RefPtr<txResultRecycler> mRecycler;
   bool mDisableLoads;
+  bool mStopProcessing = false;
 
   static const int32_t kMaxRecursionDepth;
 };
