@@ -1,4 +1,4 @@
-/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
+/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
 /* This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
@@ -14,8 +14,7 @@
 
 #include "prdtoa.h"
 
-#define INVALID_PARAM_VALUE \
-  NS_LITERAL_STRING("invalid parameter value for function")
+#define INVALID_PARAM_VALUE u"invalid parameter value for function"_ns
 
 const char16_t txFormatNumberFunctionCall::FORMAT_QUOTE = '\'';
 
@@ -31,6 +30,15 @@ txFormatNumberFunctionCall::txFormatNumberFunctionCall(
     txStylesheet* aStylesheet, txNamespaceMap* aMappings)
     : mStylesheet(aStylesheet), mMappings(aMappings) {}
 
+void txFormatNumberFunctionCall::ReportInvalidArg(txIEvalContext* aContext) {
+  nsAutoString err(INVALID_PARAM_VALUE);
+#ifdef TX_TO_STRING
+  err.AppendLiteral(": ");
+  toString(err);
+#endif
+  aContext->receiveError(err, NS_ERROR_XPATH_INVALID_ARG);
+}
+
 /*
  * Evaluates this Expr based on the given context node and processor state
  * @param context the context node for evaluation of this Expr
@@ -65,7 +73,7 @@ nsresult txFormatNumberFunctionCall::evaluate(txIEvalContext* aContext,
 
   txDecimalFormat* format = mStylesheet->getDecimalFormat(formatName);
   if (!format) {
-    nsAutoString err(NS_LITERAL_STRING("unknown decimal format"));
+    nsAutoString err(u"unknown decimal format"_ns);
 #ifdef TX_TO_STRING
     err.AppendLiteral(" for: ");
     toString(err);
@@ -75,7 +83,7 @@ nsresult txFormatNumberFunctionCall::evaluate(txIEvalContext* aContext,
   }
 
   // Special cases
-  if (mozilla::IsNaN(value)) {
+  if (std::isnan(value)) {
     return aContext->recycler()->getStringResult(format->mNaN, aResult);
   }
 
@@ -135,24 +143,14 @@ nsresult txFormatNumberFunctionCall::evaluate(txIEvalContext* aContext,
             if (multiplier == 1)
               multiplier = 100;
             else {
-              nsAutoString err(INVALID_PARAM_VALUE);
-#ifdef TX_TO_STRING
-              err.AppendLiteral(": ");
-              toString(err);
-#endif
-              aContext->receiveError(err, NS_ERROR_XPATH_INVALID_ARG);
+              ReportInvalidArg(aContext);
               return NS_ERROR_XPATH_INVALID_ARG;
             }
           } else if (c == format->mPerMille) {
             if (multiplier == 1)
               multiplier = 1000;
             else {
-              nsAutoString err(INVALID_PARAM_VALUE);
-#ifdef TX_TO_STRING
-              err.AppendLiteral(": ");
-              toString(err);
-#endif
-              aContext->receiveError(err, NS_ERROR_XPATH_INVALID_ARG);
+              ReportInvalidArg(aContext);
               return NS_ERROR_XPATH_INVALID_ARG;
             }
           } else if (c == format->mDecimalSeparator ||
@@ -225,12 +223,7 @@ nsresult txFormatNumberFunctionCall::evaluate(txIEvalContext* aContext,
   // Did we manage to parse the entire formatstring and was it valid
   if ((c != format->mPatternSeparator && pos < formatLen) || inQuote ||
       groupSize == 0) {
-    nsAutoString err(INVALID_PARAM_VALUE);
-#ifdef TX_TO_STRING
-    err.AppendLiteral(": ");
-    toString(err);
-#endif
-    aContext->receiveError(err, NS_ERROR_XPATH_INVALID_ARG);
+    ReportInvalidArg(aContext);
     return NS_ERROR_XPATH_INVALID_ARG;
   }
 
@@ -246,6 +239,11 @@ nsresult txFormatNumberFunctionCall::evaluate(txIEvalContext* aContext,
     return aContext->recycler()->getStringResult(format->mInfinity, aResult);
   }
 
+  // Make sure the multiplier didn't push value to infinity.
+  if (value == mozilla::PositiveInfinity<double>()) {
+    return aContext->recycler()->getStringResult(format->mInfinity, aResult);
+  }
+
   // Prefix
   nsAutoString res(prefix);
 
@@ -255,12 +253,12 @@ nsresult txFormatNumberFunctionCall::evaluate(txIEvalContext* aContext,
   else
     bufsize = 1 + 30;
 
-  char* buf = new char[bufsize];
+  auto buf = mozilla::MakeUnique<char[]>(bufsize);
   int bufIntDigits, sign;
   char* endp;
-  PR_dtoa(value, 0, 0, &bufIntDigits, &sign, &endp, buf, bufsize - 1);
+  PR_dtoa(value, 0, 0, &bufIntDigits, &sign, &endp, buf.get(), bufsize - 1);
 
-  int buflen = endp - buf;
+  int buflen = endp - buf.get();
   int intDigits;
   intDigits = bufIntDigits > minIntegerSize ? bufIntDigits : minIntegerSize;
 
@@ -276,7 +274,23 @@ nsresult txFormatNumberFunctionCall::evaluate(txIEvalContext* aContext,
   bool carry = (0 <= i + 1) && (i + 1 < buflen) && (buf[i + 1] >= '5');
   bool hasFraction = false;
 
-  uint32_t resPos = res.Length() - 1;
+  // The number of characters in res that we haven't filled in.
+  mozilla::CheckedUint32 resRemain = mozilla::CheckedUint32(res.Length());
+
+#define CHECKED_SET_CHAR(c)                                           \
+  --resRemain;                                                        \
+  if (!resRemain.isValid() || !res.SetCharAt(c, resRemain.value())) { \
+    ReportInvalidArg(aContext);                                       \
+    return NS_ERROR_XPATH_INVALID_ARG;                                \
+  }
+
+#define CHECKED_TRUNCATE()             \
+  --resRemain;                         \
+  if (!resRemain.isValid()) {          \
+    ReportInvalidArg(aContext);        \
+    return NS_ERROR_XPATH_INVALID_ARG; \
+  }                                    \
+  res.Truncate(resRemain.value());
 
   // Fractions
   for (; i >= bufIntDigits; --i) {
@@ -294,17 +308,17 @@ nsresult txFormatNumberFunctionCall::evaluate(txIEvalContext* aContext,
 
     if (hasFraction || digit != 0 || i < bufIntDigits + minFractionSize) {
       hasFraction = true;
-      res.SetCharAt((char16_t)(digit + format->mZeroDigit), resPos--);
+      CHECKED_SET_CHAR((char16_t)(digit + format->mZeroDigit));
     } else {
-      res.Truncate(resPos--);
+      CHECKED_TRUNCATE();
     }
   }
 
   // Decimal separator
   if (hasFraction) {
-    res.SetCharAt(format->mDecimalSeparator, resPos--);
+    CHECKED_SET_CHAR(format->mDecimalSeparator);
   } else {
-    res.Truncate(resPos--);
+    CHECKED_TRUNCATE();
   }
 
   // Integer digits
@@ -322,17 +336,20 @@ nsresult txFormatNumberFunctionCall::evaluate(txIEvalContext* aContext,
     }
 
     if (i != 0 && i % groupSize == 0) {
-      res.SetCharAt(format->mGroupingSeparator, resPos--);
+      CHECKED_SET_CHAR(format->mGroupingSeparator);
     }
 
-    res.SetCharAt((char16_t)(digit + format->mZeroDigit), resPos--);
+    CHECKED_SET_CHAR((char16_t)(digit + format->mZeroDigit));
   }
 
+#undef CHECKED_SET_CHAR
+#undef CHECKED_TRUNCATE
+
   if (carry) {
     if (i % groupSize == 0) {
-      res.Insert(format->mGroupingSeparator, resPos + 1);
+      res.Insert(format->mGroupingSeparator, resRemain.value());
     }
-    res.Insert((char16_t)(1 + format->mZeroDigit), resPos + 1);
+    res.Insert((char16_t)(1 + format->mZeroDigit), resRemain.value());
   }
 
   if (!hasFraction && !intDigits && !carry) {
@@ -341,8 +358,6 @@ nsresult txFormatNumberFunctionCall::evaluate(txIEvalContext* aContext,
     res.Append(format->mZeroDigit);
   }
 
-  delete[] buf;
-
   // Build suffix
   res.Append(suffix);
 
@@ -358,10 +373,8 @@ bool txFormatNumberFunctionCall::isSensitiveTo(ContextSensitivity aContext) {
 }
 
 #ifdef TX_TO_STRING
-nsresult txFormatNumberFunctionCall::getNameAtom(nsAtom** aAtom) {
-  *aAtom = nsGkAtoms::formatNumber;
-  NS_ADDREF(*aAtom);
-  return NS_OK;
+void txFormatNumberFunctionCall::appendName(nsAString& aDest) {
+  aDest.Append(nsGkAtoms::formatNumber->GetUTF16String());
 }
 #endif
 
@@ -371,7 +384,7 @@ nsresult txFormatNumberFunctionCall::getNameAtom(nsAtom** aAtom) {
  */
 
 txDecimalFormat::txDecimalFormat()
-    : mInfinity(NS_LITERAL_STRING("Infinity")), mNaN(NS_LITERAL_STRING("NaN")) {
+    : mInfinity(u"Infinity"_ns), mNaN(u"NaN"_ns) {
   mDecimalSeparator = '.';
   mGroupingSeparator = ',';
   mMinusSign = '-';
