@@ -1,49 +1,45 @@
-/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
+/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
 /* This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
-#include "nsCOMArray.h"
 #include "nsIAuthPrompt.h"
-#include "nsIDOMNode.h"
-#include "nsIDOMDocument.h"
-#include "nsIDocument.h"
+#include "mozilla/dom/Document.h"
 #include "nsIExpatSink.h"
-#include "nsIChannelEventSink.h"
 #include "nsIInterfaceRequestor.h"
 #include "nsILoadGroup.h"
-#include "nsIParser.h"
+#include "nsParser.h"
 #include "nsCharsetSource.h"
 #include "nsIRequestObserver.h"
-#include "nsIScriptSecurityManager.h"
 #include "nsContentPolicyUtils.h"
 #include "nsIStreamConverterService.h"
 #include "nsSyncLoadService.h"
+#include "nsIHttpChannel.h"
 #include "nsIURI.h"
 #include "nsIPrincipal.h"
 #include "nsIWindowWatcher.h"
 #include "nsIXMLContentSink.h"
 #include "nsMimeTypes.h"
 #include "nsNetUtil.h"
-#include "nsParserCIID.h"
 #include "nsGkAtoms.h"
 #include "txLog.h"
 #include "txMozillaXSLTProcessor.h"
 #include "txStylesheetCompiler.h"
 #include "txXMLUtils.h"
 #include "nsAttrName.h"
+#include "nsComponentManagerUtils.h"
 #include "nsIScriptError.h"
-#include "nsIURL.h"
 #include "nsError.h"
 #include "mozilla/Attributes.h"
 #include "mozilla/dom/Element.h"
+#include "mozilla/dom/Text.h"
 #include "mozilla/Encoding.h"
 #include "mozilla/UniquePtr.h"
+#include "ReferrerInfo.h"
 
 using namespace mozilla;
-using mozilla::net::ReferrerPolicy;
-
-static NS_DEFINE_CID(kCParserCID, NS_PARSER_CID);
+using mozilla::dom::Document;
+using mozilla::dom::ReferrerPolicy;
 
 static void getSpec(nsIChannel* aChannel, nsAString& aSpec) {
   if (!aChannel) {
@@ -66,7 +62,7 @@ class txStylesheetSink final : public nsIXMLContentSink,
                                public nsIStreamListener,
                                public nsIInterfaceRequestor {
  public:
-  txStylesheetSink(txStylesheetCompiler* aCompiler, nsIParser* aParser);
+  txStylesheetSink(txStylesheetCompiler* aCompiler, nsParser* aParser);
 
   NS_DECL_ISUPPORTS
   NS_DECL_NSIEXPATSINK
@@ -78,7 +74,7 @@ class txStylesheetSink final : public nsIXMLContentSink,
   NS_IMETHOD WillParse(void) override { return NS_OK; }
   NS_IMETHOD DidBuildModel(bool aTerminated) override;
   NS_IMETHOD WillInterrupt(void) override { return NS_OK; }
-  NS_IMETHOD WillResume(void) override { return NS_OK; }
+  void WillResume() override {}
   NS_IMETHOD SetParser(nsParserBase* aParser) override { return NS_OK; }
   virtual void FlushPendingNotifications(mozilla::FlushType aType) override {}
   virtual void SetDocumentCharset(NotNull<const Encoding*> aEncoding) override {
@@ -88,21 +84,22 @@ class txStylesheetSink final : public nsIXMLContentSink,
  private:
   RefPtr<txStylesheetCompiler> mCompiler;
   nsCOMPtr<nsIStreamListener> mListener;
-  nsCOMPtr<nsIParser> mParser;
+  RefPtr<nsParser> mParser;
   bool mCheckedForXML;
 
  protected:
-  ~txStylesheetSink() {}
+  ~txStylesheetSink() = default;
 
   // This exists solely to suppress a warning from nsDerivedSafe
   txStylesheetSink();
 };
 
 txStylesheetSink::txStylesheetSink(txStylesheetCompiler* aCompiler,
-                                   nsIParser* aParser)
-    : mCompiler(aCompiler), mParser(aParser), mCheckedForXML(false) {
-  mListener = do_QueryInterface(aParser);
-}
+                                   nsParser* aParser)
+    : mCompiler(aCompiler),
+      mListener(aParser),
+      mParser(aParser),
+      mCheckedForXML(false) {}
 
 NS_IMPL_ISUPPORTS(txStylesheetSink, nsIXMLContentSink, nsIContentSink,
                   nsIExpatSink, nsIStreamListener, nsIRequestObserver,
@@ -111,9 +108,9 @@ NS_IMPL_ISUPPORTS(txStylesheetSink, nsIXMLContentSink, nsIContentSink,
 NS_IMETHODIMP
 txStylesheetSink::HandleStartElement(const char16_t* aName,
                                      const char16_t** aAtts,
-                                     uint32_t aAttsCount,
-                                     uint32_t aLineNumber) {
-  NS_PRECONDITION(aAttsCount % 2 == 0, "incorrect aAttsCount");
+                                     uint32_t aAttsCount, uint32_t aLineNumber,
+                                     uint32_t aColumnNumber) {
+  MOZ_ASSERT(aAttsCount % 2 == 0, "incorrect aAttsCount");
 
   nsresult rv = mCompiler->startElement(aName, aAtts, aAttsCount / 2);
   if (NS_FAILED(rv)) {
@@ -182,7 +179,7 @@ NS_IMETHODIMP
 txStylesheetSink::ReportError(const char16_t* aErrorText,
                               const char16_t* aSourceText,
                               nsIScriptError* aError, bool* _retval) {
-  NS_PRECONDITION(aError && aSourceText && aErrorText, "Check arguments!!!");
+  MOZ_ASSERT(aError && aSourceText && aErrorText, "Check arguments!!!");
 
   // The expat driver should report the error.
   *_retval = true;
@@ -198,31 +195,27 @@ txStylesheetSink::DidBuildModel(bool aTerminated) {
 }
 
 NS_IMETHODIMP
-txStylesheetSink::OnDataAvailable(nsIRequest* aRequest, nsISupports* aContext,
+txStylesheetSink::OnDataAvailable(nsIRequest* aRequest,
                                   nsIInputStream* aInputStream,
                                   uint64_t aOffset, uint32_t aCount) {
   if (!mCheckedForXML) {
-    nsCOMPtr<nsIDTD> dtd;
-    mParser->GetDTD(getter_AddRefs(dtd));
-    if (dtd) {
-      mCheckedForXML = true;
-      if (!(dtd->GetType() & NS_IPARSER_FLAG_XML)) {
-        nsCOMPtr<nsIChannel> channel = do_QueryInterface(aRequest);
-        nsAutoString spec;
-        getSpec(channel, spec);
-        mCompiler->cancel(NS_ERROR_XSLT_WRONG_MIME_TYPE, nullptr, spec.get());
+    Maybe<bool> isForXML = mParser->IsForParsingXML();
+    mCheckedForXML = isForXML.isSome();
+    if (mCheckedForXML && !isForXML.value()) {
+      nsCOMPtr<nsIChannel> channel = do_QueryInterface(aRequest);
+      nsAutoString spec;
+      getSpec(channel, spec);
+      mCompiler->cancel(NS_ERROR_XSLT_WRONG_MIME_TYPE, nullptr, spec.get());
 
-        return NS_ERROR_XSLT_WRONG_MIME_TYPE;
-      }
+      return NS_ERROR_XSLT_WRONG_MIME_TYPE;
     }
   }
 
-  return mListener->OnDataAvailable(aRequest, mParser, aInputStream, aOffset,
-                                    aCount);
+  return mListener->OnDataAvailable(aRequest, aInputStream, aOffset, aCount);
 }
 
 NS_IMETHODIMP
-txStylesheetSink::OnStartRequest(nsIRequest* aRequest, nsISupports* aContext) {
+txStylesheetSink::OnStartRequest(nsIRequest* aRequest) {
   int32_t charsetSource = kCharsetFromDocTypeDefault;
 
   nsCOMPtr<nsIChannel> channel = do_QueryInterface(aRequest);
@@ -241,7 +234,7 @@ txStylesheetSink::OnStartRequest(nsIRequest* aRequest, nsISupports* aContext) {
     encoding = UTF_8_ENCODING;
   }
 
-  mParser->SetDocumentCharset(WrapNotNull(encoding), charsetSource);
+  mParser->SetDocumentCharset(WrapNotNull(encoding), charsetSource, false);
 
   nsAutoCString contentType;
   channel->GetContentType(contentType);
@@ -250,8 +243,7 @@ txStylesheetSink::OnStartRequest(nsIRequest* aRequest, nsISupports* aContext) {
   // sniffing themselves.
   nsCOMPtr<nsIURI> uri;
   channel->GetURI(getter_AddRefs(uri));
-  bool sniff;
-  if (NS_SUCCEEDED(uri->SchemeIs("file", &sniff)) && sniff &&
+  if (uri->SchemeIs("file") &&
       contentType.EqualsLiteral(UNKNOWN_CONTENT_TYPE)) {
     nsresult rv;
     nsCOMPtr<nsIStreamConverterService> serv =
@@ -259,19 +251,19 @@ txStylesheetSink::OnStartRequest(nsIRequest* aRequest, nsISupports* aContext) {
     if (NS_SUCCEEDED(rv)) {
       nsCOMPtr<nsIStreamListener> converter;
       rv = serv->AsyncConvertData(UNKNOWN_CONTENT_TYPE, "*/*", mListener,
-                                  mParser, getter_AddRefs(converter));
+                                  NS_ISUPPORTS_CAST(nsIParser*, mParser),
+                                  getter_AddRefs(converter));
       if (NS_SUCCEEDED(rv)) {
         mListener = converter;
       }
     }
   }
 
-  return mListener->OnStartRequest(aRequest, mParser);
+  return mListener->OnStartRequest(aRequest);
 }
 
 NS_IMETHODIMP
-txStylesheetSink::OnStopRequest(nsIRequest* aRequest, nsISupports* aContext,
-                                nsresult aStatusCode) {
+txStylesheetSink::OnStopRequest(nsIRequest* aRequest, nsresult aStatusCode) {
   bool success = true;
 
   nsCOMPtr<nsIHttpChannel> httpChannel = do_QueryInterface(aRequest);
@@ -286,9 +278,8 @@ txStylesheetSink::OnStopRequest(nsIRequest* aRequest, nsISupports* aContext,
     //     nsIHTMLContentSink.
     result = NS_ERROR_XSLT_NETWORK_ERROR;
   } else if (!mCheckedForXML) {
-    nsCOMPtr<nsIDTD> dtd;
-    mParser->GetDTD(getter_AddRefs(dtd));
-    if (dtd && !(dtd->GetType() & NS_IPARSER_FLAG_XML)) {
+    Maybe<bool> isForXML = mParser->IsForParsingXML();
+    if (isForXML.isSome() && !isForXML.value()) {
       result = NS_ERROR_XSLT_WRONG_MIME_TYPE;
     }
   }
@@ -300,7 +291,7 @@ txStylesheetSink::OnStopRequest(nsIRequest* aRequest, nsISupports* aContext,
     mCompiler->cancel(result, nullptr, spec.get());
   }
 
-  nsresult rv = mListener->OnStopRequest(aRequest, mParser, aStatusCode);
+  nsresult rv = mListener->OnStopRequest(aRequest, aStatusCode);
   mListener = nullptr;
   mParser = nullptr;
   return rv;
@@ -332,7 +323,7 @@ txStylesheetSink::GetInterface(const nsIID& aIID, void** aResult) {
 class txCompileObserver final : public txACompileObserver {
  public:
   txCompileObserver(txMozillaXSLTProcessor* aProcessor,
-                    nsIDocument* aLoaderDocument);
+                    Document* aLoaderDocument);
 
   TX_DECL_ACOMPILEOBSERVER
   NS_INLINE_DECL_REFCOUNTING(txCompileObserver, override)
@@ -343,17 +334,17 @@ class txCompileObserver final : public txACompileObserver {
 
  private:
   RefPtr<txMozillaXSLTProcessor> mProcessor;
-  nsCOMPtr<nsIDocument> mLoaderDocument;
+  nsCOMPtr<Document> mLoaderDocument;
 
   // This exists solely to suppress a warning from nsDerivedSafe
   txCompileObserver();
 
   // Private destructor, to discourage deletion outside of Release():
-  ~txCompileObserver() {}
+  ~txCompileObserver() = default;
 };
 
 txCompileObserver::txCompileObserver(txMozillaXSLTProcessor* aProcessor,
-                                     nsIDocument* aLoaderDocument)
+                                     Document* aLoaderDocument)
     : mProcessor(aProcessor), mLoaderDocument(aLoaderDocument) {}
 
 nsresult txCompileObserver::loadURI(const nsAString& aUri,
@@ -374,7 +365,7 @@ nsresult txCompileObserver::loadURI(const nsAString& aUri,
 
   OriginAttributes attrs;
   nsCOMPtr<nsIPrincipal> referrerPrincipal =
-      BasePrincipal::CreateCodebasePrincipal(referrerUri, attrs);
+      BasePrincipal::CreateContentPrincipal(referrerUri, attrs);
   NS_ENSURE_TRUE(referrerPrincipal, NS_ERROR_FAILURE);
 
   return startLoad(uri, aCompiler, referrerPrincipal, aReferrerPolicy);
@@ -404,34 +395,28 @@ nsresult txCompileObserver::startLoad(nsIURI* aUri,
   nsresult rv = NS_NewChannelWithTriggeringPrincipal(
       getter_AddRefs(channel), aUri, mLoaderDocument,
       aReferrerPrincipal,  // triggeringPrincipal
-      nsILoadInfo::SEC_REQUIRE_CORS_DATA_INHERITS, nsIContentPolicy::TYPE_XSLT,
+      nsILoadInfo::SEC_REQUIRE_CORS_INHERITS_SEC_CONTEXT,
+      nsIContentPolicy::TYPE_XSLT,
       nullptr,  // aPerformanceStorage
       loadGroup);
 
   NS_ENSURE_SUCCESS(rv, rv);
 
-  channel->SetContentType(NS_LITERAL_CSTRING("text/xml"));
+  channel->SetContentType("text/xml"_ns);
 
   nsCOMPtr<nsIHttpChannel> httpChannel(do_QueryInterface(channel));
   if (httpChannel) {
-    DebugOnly<nsresult> rv;
-    rv = httpChannel->SetRequestHeader(NS_LITERAL_CSTRING("Accept"),
-                                       NS_LITERAL_CSTRING("*/*"), false);
-    MOZ_ASSERT(NS_SUCCEEDED(rv));
-
-    nsCOMPtr<nsIURI> referrerURI;
-    aReferrerPrincipal->GetURI(getter_AddRefs(referrerURI));
-    if (referrerURI) {
-      rv = httpChannel->SetReferrerWithPolicy(referrerURI, aReferrerPolicy);
+    nsCOMPtr<nsIReferrerInfo> referrerInfo;
+    nsresult rv = aReferrerPrincipal->CreateReferrerInfo(
+        aReferrerPolicy, getter_AddRefs(referrerInfo));
+    if (NS_SUCCEEDED(rv)) {
+      rv = httpChannel->SetReferrerInfoWithoutClone(referrerInfo);
       MOZ_ASSERT(NS_SUCCEEDED(rv));
     }
   }
 
-  nsCOMPtr<nsIParser> parser = do_CreateInstance(kCParserCID, &rv);
-  NS_ENSURE_SUCCESS(rv, rv);
-
+  RefPtr<nsParser> parser = new nsParser();
   RefPtr<txStylesheetSink> sink = new txStylesheetSink(aCompiler, parser);
-  NS_ENSURE_TRUE(sink, NS_ERROR_OUT_OF_MEMORY);
 
   channel->SetNotificationCallbacks(sink);
 
@@ -439,11 +424,11 @@ nsresult txCompileObserver::startLoad(nsIURI* aUri,
   parser->SetContentSink(sink);
   parser->Parse(aUri);
 
-  return channel->AsyncOpen2(sink);
+  return channel->AsyncOpen(sink);
 }
 
 nsresult TX_LoadSheet(nsIURI* aUri, txMozillaXSLTProcessor* aProcessor,
-                      nsIDocument* aLoaderDocument,
+                      Document* aLoaderDocument,
                       ReferrerPolicy aReferrerPolicy) {
   nsIPrincipal* principal = aLoaderDocument->NodePrincipal();
 
@@ -453,11 +438,9 @@ nsresult TX_LoadSheet(nsIURI* aUri, txMozillaXSLTProcessor* aProcessor,
 
   RefPtr<txCompileObserver> observer =
       new txCompileObserver(aProcessor, aLoaderDocument);
-  NS_ENSURE_TRUE(observer, NS_ERROR_OUT_OF_MEMORY);
 
   RefPtr<txStylesheetCompiler> compiler = new txStylesheetCompiler(
       NS_ConvertUTF8toUTF16(spec), aReferrerPolicy, observer);
-  NS_ENSURE_TRUE(compiler, NS_ERROR_OUT_OF_MEMORY);
 
   return observer->startLoad(aUri, compiler, principal, aReferrerPolicy);
 }
@@ -504,12 +487,12 @@ static nsresult handleNode(nsINode* aNode, txStylesheetCompiler* aCompiler) {
 
     rv = aCompiler->endElement();
     NS_ENSURE_SUCCESS(rv, rv);
-  } else if (aNode->IsNodeOfType(nsINode::eTEXT)) {
+  } else if (dom::Text* text = aNode->GetAsText()) {
     nsAutoString chars;
-    static_cast<nsIContent*>(aNode)->AppendTextTo(chars);
+    text->AppendTextTo(chars);
     rv = aCompiler->characters(chars);
     NS_ENSURE_SUCCESS(rv, rv);
-  } else if (aNode->IsNodeOfType(nsINode::eDOCUMENT)) {
+  } else if (aNode->IsDocument()) {
     for (nsIContent* child = aNode->GetFirstChild(); child;
          child = child->GetNextSibling()) {
       rv = handleNode(child, aCompiler);
@@ -529,7 +512,7 @@ class txSyncCompileObserver final : public txACompileObserver {
 
  private:
   // Private destructor, to discourage deletion outside of Release():
-  ~txSyncCompileObserver() {}
+  ~txSyncCompileObserver() = default;
 
   RefPtr<txMozillaXSLTProcessor> mProcessor;
 };
@@ -554,7 +537,7 @@ nsresult txSyncCompileObserver::loadURI(const nsAString& aUri,
   NS_ENSURE_SUCCESS(rv, rv);
 
   nsCOMPtr<nsIPrincipal> referrerPrincipal =
-      BasePrincipal::CreateCodebasePrincipal(referrerUri, OriginAttributes());
+      BasePrincipal::CreateContentPrincipal(referrerUri, OriginAttributes());
   NS_ENSURE_TRUE(referrerPrincipal, NS_ERROR_FAILURE);
 
   // This is probably called by js, a loadGroup for the channel doesn't
@@ -563,17 +546,18 @@ nsresult txSyncCompileObserver::loadURI(const nsAString& aUri,
   if (mProcessor) {
     source = mProcessor->GetSourceContentModel();
   }
-  nsAutoSyncOperation sync(source ? source->OwnerDoc() : nullptr);
-  nsCOMPtr<nsIDOMDocument> document;
+  dom::nsAutoSyncOperation sync(source ? source->OwnerDoc() : nullptr,
+                                dom::SyncOperationBehavior::eSuspendInput);
+  nsCOMPtr<Document> document;
 
   rv = nsSyncLoadService::LoadDocument(
       uri, nsIContentPolicy::TYPE_XSLT, referrerPrincipal,
-      nsILoadInfo::SEC_REQUIRE_CORS_DATA_INHERITS, nullptr, false,
+      nsILoadInfo::SEC_REQUIRE_CORS_INHERITS_SEC_CONTEXT, nullptr,
+      source ? source->OwnerDoc()->CookieJarSettings() : nullptr, false,
       aReferrerPolicy, getter_AddRefs(document));
   NS_ENSURE_SUCCESS(rv, rv);
 
-  nsCOMPtr<nsIDocument> doc = do_QueryInterface(document);
-  rv = handleNode(doc, aCompiler);
+  rv = handleNode(document, aCompiler);
   if (NS_FAILED(rv)) {
     nsAutoCString spec;
     uri->GetSpec(spec);
@@ -594,19 +578,13 @@ nsresult TX_CompileStylesheet(nsINode* aNode,
                               txMozillaXSLTProcessor* aProcessor,
                               txStylesheet** aStylesheet) {
   // If we move GetBaseURI to nsINode this can be simplified.
-  nsCOMPtr<nsIDocument> doc = aNode->OwnerDoc();
+  nsCOMPtr<Document> doc = aNode->OwnerDoc();
 
-  nsCOMPtr<nsIURI> uri;
-  if (aNode->IsContent()) {
-    uri = aNode->AsContent()->GetBaseURI();
-  } else {
-    NS_ASSERTION(aNode->IsNodeOfType(nsINode::eDOCUMENT), "not a doc");
-    uri = static_cast<nsIDocument*>(aNode)->GetBaseURI();
-  }
-  NS_ENSURE_TRUE(uri, NS_ERROR_FAILURE);
+  nsIURI* nodeBaseURI = aNode->GetBaseURI();
+  NS_ENSURE_TRUE(nodeBaseURI, NS_ERROR_FAILURE);
 
   nsAutoCString spec;
-  uri->GetSpec(spec);
+  nodeBaseURI->GetSpec(spec);
   NS_ConvertUTF8toUTF16 baseURI(spec);
 
   nsIURI* docUri = doc->GetDocumentURI();
@@ -614,18 +592,17 @@ nsresult TX_CompileStylesheet(nsINode* aNode,
 
   // We need to remove the ref, a URI with a ref would mean that we have an
   // embedded stylesheet.
-  docUri->CloneIgnoringRef(getter_AddRefs(uri));
+  nsCOMPtr<nsIURI> uri;
+  NS_GetURIWithoutRef(docUri, getter_AddRefs(uri));
   NS_ENSURE_TRUE(uri, NS_ERROR_FAILURE);
 
   uri->GetSpec(spec);
   NS_ConvertUTF8toUTF16 stylesheetURI(spec);
 
   RefPtr<txSyncCompileObserver> obs = new txSyncCompileObserver(aProcessor);
-  NS_ENSURE_TRUE(obs, NS_ERROR_OUT_OF_MEMORY);
 
   RefPtr<txStylesheetCompiler> compiler =
       new txStylesheetCompiler(stylesheetURI, doc->GetReferrerPolicy(), obs);
-  NS_ENSURE_TRUE(compiler, NS_ERROR_OUT_OF_MEMORY);
 
   compiler->setBaseURI(baseURI);
 
