@@ -1,4 +1,4 @@
-/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
+/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
 /* This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
@@ -7,14 +7,13 @@
 #define TRANSFRMX_TXSTYLESHEETCOMPILER_H
 
 #include "mozilla/Attributes.h"
+#include "mozilla/UniquePtr.h"
 #include "txStack.h"
 #include "txXSLTPatterns.h"
 #include "txExpr.h"
 #include "txIXPathContext.h"
-#include "nsAutoPtr.h"
 #include "txStylesheet.h"
 #include "nsTArray.h"
-#include "mozilla/net/ReferrerPolicy.h"
 
 extern bool TX_XSLTFunctionAvailable(nsAtom* aName, int32_t aNameSpaceID);
 
@@ -26,7 +25,6 @@ class txNamespaceMap;
 class txToplevelItem;
 class txPushNewContext;
 class txStylesheetCompiler;
-class txInScopeVariable;
 
 class txElementContext : public txObject {
  public:
@@ -41,12 +39,14 @@ class txElementContext : public txObject {
   int32_t mDepth;
 };
 
+using mozilla::dom::ReferrerPolicy;
+
 class txACompileObserver {
  public:
   NS_INLINE_DECL_PURE_VIRTUAL_REFCOUNTING
 
   virtual nsresult loadURI(const nsAString& aUri, const nsAString& aReferrerUri,
-                           mozilla::net::ReferrerPolicy aReferrerPolicy,
+                           ReferrerPolicy aReferrerPolicy,
                            txStylesheetCompiler* aCompiler) = 0;
   virtual void onDoneCompiling(txStylesheetCompiler* aCompiler,
                                nsresult aResult,
@@ -56,19 +56,26 @@ class txACompileObserver {
 
 #define TX_DECL_ACOMPILEOBSERVER                                          \
   nsresult loadURI(const nsAString& aUri, const nsAString& aReferrerUri,  \
-                   mozilla::net::ReferrerPolicy aReferrerPolicy,          \
+                   ReferrerPolicy aReferrerPolicy,                        \
                    txStylesheetCompiler* aCompiler) override;             \
   void onDoneCompiling(txStylesheetCompiler* aCompiler, nsresult aResult, \
                        const char16_t* aErrorText = nullptr,              \
                        const char16_t* aParam = nullptr) override;
 
+class txInScopeVariable {
+ public:
+  explicit txInScopeVariable(const txExpandedName& aName)
+      : mName(aName), mLevel(1) {}
+  txExpandedName mName;
+  int32_t mLevel;
+};
+
 class txStylesheetCompilerState : public txIParseContext {
  public:
   explicit txStylesheetCompilerState(txACompileObserver* aObserver);
   ~txStylesheetCompilerState();
 
-  nsresult init(const nsAString& aStylesheetURI,
-                mozilla::net::ReferrerPolicy aReferrerPolicy,
+  nsresult init(const nsAString& aStylesheetURI, ReferrerPolicy aReferrerPolicy,
                 txStylesheet* aStylesheet, txListIterator* aInsertPosition);
 
   // Embedded stylesheets state
@@ -87,29 +94,35 @@ class txStylesheetCompilerState : public txIParseContext {
     eCheckParam,
     ePushNullTemplateRule
   };
-  nsresult pushHandlerTable(txHandlerTable* aTable);
+  void pushHandlerTable(txHandlerTable* aTable);
   void popHandlerTable();
-  nsresult pushSorter(txPushNewContext* aSorter);
+  void pushSorter(txPushNewContext* aSorter);
   void popSorter();
-  nsresult pushChooseGotoList();
+  void pushChooseGotoList();
   void popChooseGotoList();
-  nsresult pushObject(txObject* aObject);
+  void pushObject(txObject* aObject);
   txObject* popObject();
-  nsresult pushPtr(void* aPtr, enumStackType aType);
+  void pushPtr(void* aPtr, enumStackType aType);
   void* popPtr(enumStackType aType);
 
   // stylesheet functions
-  nsresult addToplevelItem(txToplevelItem* aItem);
+  void addToplevelItem(txToplevelItem* aItem);
   nsresult openInstructionContainer(txInstructionContainer* aContainer);
   void closeInstructionContainer();
-  nsresult addInstruction(nsAutoPtr<txInstruction>&& aInstruction);
+  txInstruction* addInstruction(
+      mozilla::UniquePtr<txInstruction>&& aInstruction);
+  template <class T>
+  T* addInstruction(mozilla::UniquePtr<T> aInstruction) {
+    return static_cast<T*>(addInstruction(
+        mozilla::UniquePtr<txInstruction>(std::move(aInstruction))));
+  }
   nsresult loadIncludedStylesheet(const nsAString& aURI);
   nsresult loadImportedStylesheet(const nsAString& aURI,
                                   txStylesheet::ImportFrame* aFrame);
 
   // misc
-  nsresult addGotoTarget(txInstruction** aTargetPointer);
-  nsresult addVariable(const txExpandedName& aName);
+  void addGotoTarget(txInstruction** aTargetPointer);
+  void addVariable(const txExpandedName& aName);
 
   // txIParseContext
   nsresult resolveNamespacePrefix(nsAtom* aPrefix, int32_t& aID) override;
@@ -132,20 +145,18 @@ class txStylesheetCompilerState : public txIParseContext {
     return aResult != NS_ERROR_XSLT_CALL_TO_KEY_NOT_ALLOWED && fcp();
   }
 
-  static void shutdown();
-
   RefPtr<txStylesheet> mStylesheet;
   txHandlerTable* mHandlerTable;
-  nsAutoPtr<txElementContext> mElementContext;
+  mozilla::UniquePtr<txElementContext> mElementContext;
   txPushNewContext* mSorter;
-  nsAutoPtr<txList> mChooseGotoList;
+  mozilla::UniquePtr<txList> mChooseGotoList;
   bool mDOE;
   bool mSearchingForFallback;
   uint16_t mDisAllowed;
 
  protected:
   RefPtr<txACompileObserver> mObserver;
-  nsTArray<txInScopeVariable*> mInScopeVariables;
+  nsTArray<txInScopeVariable> mInScopeVariables;
   nsTArray<txStylesheetCompiler*> mChildCompilerList;
   // embed info, target information is the ID
   nsString mTarget;
@@ -158,10 +169,10 @@ class txStylesheetCompilerState : public txIParseContext {
   nsTArray<enumStackType> mTypeStack;
 
  private:
-  txInstruction** mNextInstrPtr;
+  mozilla::UniquePtr<txInstruction>* mNextInstrPtr;
   txListIterator mToplevelIterator;
   nsTArray<txInstruction**> mGotoTargetPointers;
-  mozilla::net::ReferrerPolicy mReferrerPolicy;
+  ReferrerPolicy mReferrerPolicy;
 };
 
 struct txStylesheetAttr {
@@ -177,12 +188,12 @@ class txStylesheetCompiler final : private txStylesheetCompilerState,
   friend class txStylesheetCompilerState;
   friend bool TX_XSLTFunctionAvailable(nsAtom* aName, int32_t aNameSpaceID);
   txStylesheetCompiler(const nsAString& aStylesheetURI,
-                       mozilla::net::ReferrerPolicy aReferrerPolicy,
+                       ReferrerPolicy aReferrerPolicy,
                        txACompileObserver* aObserver);
   txStylesheetCompiler(const nsAString& aStylesheetURI,
                        txStylesheet* aStylesheet,
                        txListIterator* aInsertPosition,
-                       mozilla::net::ReferrerPolicy aReferrerPolicy,
+                       ReferrerPolicy aReferrerPolicy,
                        txACompileObserver* aObserver);
 
   void setBaseURI(const nsString& aBaseURI);
@@ -206,7 +217,7 @@ class txStylesheetCompiler final : private txStylesheetCompilerState,
 
  private:
   // Private destructor, to discourage deletion outside of Release():
-  ~txStylesheetCompiler() {}
+  ~txStylesheetCompiler() = default;
 
   nsresult startElementInternal(int32_t aNamespaceID, nsAtom* aLocalName,
                                 nsAtom* aPrefix, txStylesheetAttr* aAttributes,
@@ -220,12 +231,4 @@ class txStylesheetCompiler final : private txStylesheetCompilerState,
   nsresult mStatus;
 };
 
-class txInScopeVariable {
- public:
-  explicit txInScopeVariable(const txExpandedName& aName)
-      : mName(aName), mLevel(1) {}
-  txExpandedName mName;
-  int32_t mLevel;
-};
-
 #endif
