@@ -18,7 +18,7 @@ struct BroadcastListener {
 };
 
 struct BroadcasterMapEntry : public PLDHashEntryHdr {
-  Element* mBroadcaster;  // [WEAK]
+  mozilla::dom::Element* mBroadcaster;  // [WEAK]
   nsTArray<BroadcastListener*>
       mListeners;  // [OWNING] of BroadcastListener objects
 };
@@ -26,10 +26,9 @@ struct BroadcasterMapEntry : public PLDHashEntryHdr {
 struct nsAttrNameInfo {
   nsAttrNameInfo(int32_t aNamespaceID, nsAtom* aName, nsAtom* aPrefix)
       : mNamespaceID(aNamespaceID), mName(aName), mPrefix(aPrefix) {}
-  nsAttrNameInfo(const nsAttrNameInfo& aOther)
-      : mNamespaceID(aOther.mNamespaceID),
-        mName(aOther.mName),
-        mPrefix(aOther.mPrefix) {}
+  nsAttrNameInfo(const nsAttrNameInfo& aOther) = delete;
+  nsAttrNameInfo(nsAttrNameInfo&& aOther) = default;
+
   int32_t mNamespaceID;
   RefPtr<nsAtom> mName;
   RefPtr<nsAtom> mPrefix;
@@ -61,10 +60,51 @@ static bool CanBroadcast(int32_t aNameSpaceID, nsAtom* aAttribute) {
   return true;
 }
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 static LazyLogModule sXULBroadCastManager("XULBroadcastManager");
 
+class XULBroadcastManager::nsDelayedBroadcastUpdate {
+ public:
+  nsDelayedBroadcastUpdate(Element* aBroadcaster, Element* aListener,
+                           const nsAString& aAttr)
+      : mBroadcaster(aBroadcaster),
+        mListener(aListener),
+        mAttr(aAttr),
+        mSetAttr(false),
+        mNeedsAttrChange(false) {}
+
+  nsDelayedBroadcastUpdate(Element* aBroadcaster, Element* aListener,
+                           nsAtom* aAttrName, const nsAString& aAttr,
+                           bool aSetAttr, bool aNeedsAttrChange)
+      : mBroadcaster(aBroadcaster),
+        mListener(aListener),
+        mAttr(aAttr),
+        mAttrName(aAttrName),
+        mSetAttr(aSetAttr),
+        mNeedsAttrChange(aNeedsAttrChange) {}
+
+  nsDelayedBroadcastUpdate(const nsDelayedBroadcastUpdate& aOther) = delete;
+  nsDelayedBroadcastUpdate(nsDelayedBroadcastUpdate&& aOther) = default;
+
+  RefPtr<Element> mBroadcaster;
+  RefPtr<Element> mListener;
+  // Note if mAttrName isn't used, this is the name of the attr, otherwise
+  // this is the value of the attribute.
+  nsString mAttr;
+  RefPtr<nsAtom> mAttrName;
+  bool mSetAttr;
+  bool mNeedsAttrChange;
+
+  class Comparator {
+   public:
+    static bool Equals(const nsDelayedBroadcastUpdate& a,
+                       const nsDelayedBroadcastUpdate& b) {
+      return a.mBroadcaster == b.mBroadcaster && a.mListener == b.mListener &&
+             a.mAttrName == b.mAttrName;
+    }
+  };
+};
+
 /* static */
 bool XULBroadcastManager::MayNeedListener(const Element& aElement) {
   if (aElement.NodeInfo()->Equals(nsGkAtoms::observes, kNameSpaceID_XUL)) {
@@ -95,8 +135,7 @@ void XULBroadcastManager::SynchronizeBroadcastListener(Element* aBroadcaster,
                                                        Element* aListener,
                                                        const nsAString& aAttr) {
   if (!nsContentUtils::IsSafeToRunScript()) {
-    nsDelayedBroadcastUpdate delayedUpdate(aBroadcaster, aListener, aAttr);
-    mDelayedBroadcasters.AppendElement(delayedUpdate);
+    mDelayedBroadcasters.EmplaceBack(aBroadcaster, aListener, aAttr);
     MaybeBroadcast();
     return;
   }
@@ -260,7 +299,7 @@ nsresult XULBroadcastManager::ExecuteOnBroadcastHandlerFor(
   // observer. We need to find the observer in order to
   // execute the handler.
 
-  for (nsIContent* child = aListener->GetFirstChild(); child;
+  for (nsCOMPtr<nsIContent> child = aListener->GetFirstChild(); child;
        child = child->GetNextSibling()) {
     // Look for an <observes> element beneath the listener. This
     // ought to have an |element| attribute that refers to
@@ -293,8 +332,7 @@ nsresult XULBroadcastManager::ExecuteOnBroadcastHandlerFor(
     // |onbroadcast| event handler
     WidgetEvent event(true, eXULBroadcast);
 
-    RefPtr<nsPresContext> presContext = mDocument->GetPresContext();
-    if (presContext) {
+    if (RefPtr<nsPresContext> presContext = mDocument->GetPresContext()) {
       // Handle the DOM event
       nsEventStatus status = nsEventStatus_eIgnore;
       EventDispatcher::Dispatch(child, presContext, &event, nullptr, &status);
@@ -351,7 +389,8 @@ void XULBroadcastManager::AttributeChanged(Element* aElement,
               mDelayedAttrChangeBroadcasts.RemoveElementAt(index);
             }
 
-            mDelayedAttrChangeBroadcasts.AppendElement(delayedUpdate);
+            mDelayedAttrChangeBroadcasts.AppendElement(
+                std::move(delayedUpdate));
           }
         }
       }
@@ -375,10 +414,9 @@ void XULBroadcastManager::MaybeBroadcast() {
     if (!mHandlingDelayedAttrChange) {
       mHandlingDelayedAttrChange = true;
       for (uint32_t i = 0; i < mDelayedAttrChangeBroadcasts.Length(); ++i) {
-        nsAtom* attrName = mDelayedAttrChangeBroadcasts[i].mAttrName;
+        RefPtr<nsAtom> attrName = mDelayedAttrChangeBroadcasts[i].mAttrName;
+        RefPtr<Element> listener = mDelayedAttrChangeBroadcasts[i].mListener;
         if (mDelayedAttrChangeBroadcasts[i].mNeedsAttrChange) {
-          nsCOMPtr<Element> listener =
-              mDelayedAttrChangeBroadcasts[i].mListener;
           const nsString& value = mDelayedAttrChangeBroadcasts[i].mAttr;
           if (mDelayedAttrChangeBroadcasts[i].mSetAttr) {
             listener->SetAttr(kNameSpaceID_None, attrName, value, true);
@@ -386,9 +424,9 @@ void XULBroadcastManager::MaybeBroadcast() {
             listener->UnsetAttr(kNameSpaceID_None, attrName, true);
           }
         }
-        ExecuteOnBroadcastHandlerFor(
-            mDelayedAttrChangeBroadcasts[i].mBroadcaster,
-            mDelayedAttrChangeBroadcasts[i].mListener, attrName);
+        RefPtr<Element> broadcaster =
+            mDelayedAttrChangeBroadcasts[i].mBroadcaster;
+        ExecuteOnBroadcastHandlerFor(broadcaster, listener, attrName);
       }
       mDelayedAttrChangeBroadcasts.Clear();
       mHandlingDelayedAttrChange = false;
@@ -398,8 +436,8 @@ void XULBroadcastManager::MaybeBroadcast() {
     if (length) {
       bool oldValue = mHandlingDelayedBroadcasters;
       mHandlingDelayedBroadcasters = true;
-      nsTArray<nsDelayedBroadcastUpdate> delayedBroadcasters;
-      mDelayedBroadcasters.SwapElements(delayedBroadcasters);
+      nsTArray<nsDelayedBroadcastUpdate> delayedBroadcasters =
+          std::move(mDelayedBroadcasters);
       for (uint32_t i = 0; i < length; ++i) {
         SynchronizeBroadcastListener(delayedBroadcasters[i].mBroadcaster,
                                      delayedBroadcasters[i].mListener,
@@ -551,5 +589,4 @@ nsresult XULBroadcastManager::RemoveListener(Element* aElement) {
   return UpdateListenerHookup(aElement, eHookupRemove);
 }
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
