@@ -46,12 +46,6 @@ class nsXULPopupListener : public nsIDOMEventListener {
   virtual void ClosePopup();
 
  private:
-#ifndef NS_CONTEXT_MENU_IS_MOUSEUP
-  // When a context menu is opened, focus the target of the contextmenu event.
-  MOZ_CAN_RUN_SCRIPT_BOUNDARY nsresult
-  FireFocusOnTargetContent(nsIContent* aTargetContent, bool aIsTouch);
-#endif
-
   // |mElement| is the node to which this listener is attached.
   RefPtr<mozilla::dom::Element> mElement;
 
