@@ -474,7 +474,9 @@ static void SetPaintPattern(SkPaint& aPaint, const Pattern& aPattern,
       const LinearGradientPattern& pat =
           static_cast<const LinearGradientPattern&>(aPattern);
       GradientStopsSkia* stops =
-          static_cast<GradientStopsSkia*>(pat.mStops.get());
+          pat.mStops && pat.mStops->GetBackendType() == BackendType::SKIA
+              ? static_cast<GradientStopsSkia*>(pat.mStops.get())
+              : nullptr;
       if (!stops || stops->mCount < 2 || !pat.mBegin.IsFinite() ||
           !pat.mEnd.IsFinite() || pat.mBegin == pat.mEnd) {
         aPaint.setColor(SK_ColorTRANSPARENT);
@@ -506,7 +508,9 @@ static void SetPaintPattern(SkPaint& aPaint, const Pattern& aPattern,
       const RadialGradientPattern& pat =
           static_cast<const RadialGradientPattern&>(aPattern);
       GradientStopsSkia* stops =
-          static_cast<GradientStopsSkia*>(pat.mStops.get());
+          pat.mStops && pat.mStops->GetBackendType() == BackendType::SKIA
+              ? static_cast<GradientStopsSkia*>(pat.mStops.get())
+              : nullptr;
       if (!stops || stops->mCount < 2 || !pat.mCenter1.IsFinite() ||
           !std::isfinite(pat.mRadius1) || !pat.mCenter2.IsFinite() ||
           !std::isfinite(pat.mRadius2) ||
@@ -541,7 +545,9 @@ static void SetPaintPattern(SkPaint& aPaint, const Pattern& aPattern,
       const ConicGradientPattern& pat =
           static_cast<const ConicGradientPattern&>(aPattern);
       GradientStopsSkia* stops =
-          static_cast<GradientStopsSkia*>(pat.mStops.get());
+          pat.mStops && pat.mStops->GetBackendType() == BackendType::SKIA
+              ? static_cast<GradientStopsSkia*>(pat.mStops.get())
+              : nullptr;
       if (!stops || stops->mCount < 2 || !pat.mCenter.IsFinite() ||
           !std::isfinite(pat.mAngle)) {
         aPaint.setColor(SK_ColorTRANSPARENT);
@@ -747,6 +753,9 @@ DrawTargetType DrawTargetSkia::GetType() const {
 void DrawTargetSkia::DrawFilter(FilterNode* aNode, const Rect& aSourceRect,
                                 const Point& aDestPoint,
                                 const DrawOptions& aOptions) {
+  if (!aNode || aNode->GetBackendType() != FILTER_BACKEND_SOFTWARE) {
+    return;
+  }
   FilterNodeSoftware* filter = static_cast<FilterNodeSoftware*>(aNode);
   filter->Draw(this, aSourceRect, aDestPoint, aOptions);
 }
@@ -1389,7 +1398,7 @@ void DrawTargetSkia::Mask(const Pattern& aSource, const Pattern& aMask,
   SkPaint maskPaint;
   SetPaintPattern(maskPaint, aMask, lock);
 
-  sk_sp<SkShader> maskShader(maskPaint.getShader());
+  sk_sp<SkShader> maskShader(maskPaint.refShader());
   if (!maskShader && maskPaint.getAlpha() != 0xFF) {
     if (maskPaint.getAlpha() == 0) {
       return;
