@@ -5,6 +5,7 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "ScaledFontDWrite.h"
+#include "gfxDWriteCommon.h"
 #include "UnscaledFontDWrite.h"
 #include "PathD2D.h"
 #include "gfxFont.h"
@@ -12,6 +13,7 @@
 #include "mozilla/FontPropertyTypes.h"
 #include "mozilla/webrender/WebRenderTypes.h"
 #include "HelpersD2D.h"
+#include "StackArray.h"
 
 #include "dwrite_3.h"
 
@@ -27,20 +29,14 @@
 #  include "dw-extra.h"
 #endif
 
-using namespace std;
-
-#ifdef USE_SKIA
-#  include "PathSkia.h"
-#  include "skia/include/core/SkPaint.h"
-#  include "skia/include/core/SkPath.h"
-#  include "skia/include/ports/SkTypeface_win.h"
-#endif
+#include "PathSkia.h"
+#include "skia/include/core/SkPaint.h"
+#include "skia/include/core/SkPath.h"
+#include "skia/include/ports/SkTypeface_win.h"
 
 #include <vector>
 
-#ifdef USE_CAIRO_SCALED_FONT
-#  include "cairo-win32.h"
-#endif
+#include "cairo-win32.h"
 
 #include "HelpersWinFonts.h"
 
@@ -78,7 +74,7 @@ static bool DoGrayscale(IDWriteFontFace* aDWFace, Float ppem) {
     }
     gaspRange* ranges = (gaspRange*)(tableData + 4);
     for (int i = 0; i < numRanges; i++) {
-      if (readShort((char*)&ranges[i].maxPPEM) > ppem) {
+      if (readShort((char*)&ranges[i].maxPPEM) >= ppem) {
         if (!(readShort((char*)&ranges[i].behavior) & GASP_DOGRAY)) {
           aDWFace->ReleaseFontTable(tableContext);
           return false;
@@ -91,57 +87,21 @@ static bool DoGrayscale(IDWriteFontFace* aDWFace, Float ppem) {
   return true;
 }
 
-static inline DWRITE_FONT_STRETCH DWriteFontStretchFromStretch(
-    FontStretch aStretch) {
-  if (aStretch == FontStretch::UltraCondensed()) {
-    return DWRITE_FONT_STRETCH_ULTRA_CONDENSED;
-  }
-  if (aStretch == FontStretch::ExtraCondensed()) {
-    return DWRITE_FONT_STRETCH_EXTRA_CONDENSED;
-  }
-  if (aStretch == FontStretch::Condensed()) {
-    return DWRITE_FONT_STRETCH_CONDENSED;
-  }
-  if (aStretch == FontStretch::SemiCondensed()) {
-    return DWRITE_FONT_STRETCH_SEMI_CONDENSED;
-  }
-  if (aStretch == FontStretch::Normal()) {
-    return DWRITE_FONT_STRETCH_NORMAL;
-  }
-  if (aStretch == FontStretch::SemiExpanded()) {
-    return DWRITE_FONT_STRETCH_SEMI_EXPANDED;
-  }
-  if (aStretch == FontStretch::Expanded()) {
-    return DWRITE_FONT_STRETCH_EXPANDED;
-  }
-  if (aStretch == FontStretch::ExtraExpanded()) {
-    return DWRITE_FONT_STRETCH_EXTRA_EXPANDED;
-  }
-  if (aStretch == FontStretch::UltraExpanded()) {
-    return DWRITE_FONT_STRETCH_ULTRA_EXPANDED;
-  }
-  return DWRITE_FONT_STRETCH_UNDEFINED;
-}
-
 ScaledFontDWrite::ScaledFontDWrite(IDWriteFontFace* aFontFace,
                                    const RefPtr<UnscaledFont>& aUnscaledFont,
                                    Float aSize, bool aUseEmbeddedBitmap,
-                                   bool aForceGDIMode,
-                                   IDWriteRenderingParams* aParams,
-                                   Float aGamma, Float aContrast,
+                                   bool aUseMultistrikeBold, bool aGDIForced,
                                    const gfxFontStyle* aStyle)
     : ScaledFontBase(aUnscaledFont, aSize),
       mFontFace(aFontFace),
       mUseEmbeddedBitmap(aUseEmbeddedBitmap),
-      mForceGDIMode(aForceGDIMode),
-      mParams(aParams),
-      mGamma(aGamma),
-      mContrast(aContrast) {
+      mUseMultistrikeBold(aUseMultistrikeBold),
+      mGDIForced(aGDIForced) {
   if (aStyle) {
     mStyle = SkFontStyle(aStyle->weight.ToIntRounded(),
                          DWriteFontStretchFromStretch(aStyle->stretch),
                          // FIXME(jwatt): also use kOblique_Slant
-                         aStyle->style == FontSlantStyle::Normal()
+                         aStyle->style == FontSlantStyle::NORMAL
                              ? SkFontStyle::kUpright_Slant
                              : SkFontStyle::kItalic_Slant);
   }
@@ -149,13 +109,13 @@ ScaledFontDWrite::ScaledFontDWrite(IDWriteFontFace* aFontFace,
 
 already_AddRefed<Path> ScaledFontDWrite::GetPathForGlyphs(
     const GlyphBuffer& aBuffer, const DrawTarget* aTarget) {
-  if (aTarget->GetBackendType() != BackendType::DIRECT2D &&
-      aTarget->GetBackendType() != BackendType::DIRECT2D1_1) {
+  RefPtr<PathBuilder> pathBuilder = aTarget->CreatePathBuilder();
+
+  if (pathBuilder->GetBackendType() != BackendType::DIRECT2D &&
+      pathBuilder->GetBackendType() != BackendType::DIRECT2D1_1) {
     return ScaledFontBase::GetPathForGlyphs(aBuffer, aTarget);
   }
 
-  RefPtr<PathBuilder> pathBuilder = aTarget->CreatePathBuilder();
-
   PathBuilderD2D* pathBuilderD2D =
       static_cast<PathBuilderD2D*>(pathBuilder.get());
 
@@ -164,40 +124,53 @@ already_AddRefed<Path> ScaledFontDWrite::GetPathForGlyphs(
   return pathBuilder->Finish();
 }
 
-#ifdef USE_SKIA
 SkTypeface* ScaledFontDWrite::CreateSkTypeface() {
   RefPtr<IDWriteFactory> factory = Factory::GetDWriteFactory();
   if (!factory) {
     return nullptr;
   }
 
-  Float gamma = mGamma;
+  auto& settings = DWriteSettings();
+  Float gamma = settings.Gamma();
   // Skia doesn't support a gamma value outside of 0-4, so default to 2.2
   if (gamma < 0.0f || gamma > 4.0f) {
     gamma = 2.2f;
   }
 
-  Float contrast = mContrast;
+  Float contrast = settings.EnhancedContrast();
   // Skia doesn't support a contrast value outside of 0-1, so default to 1.0
   if (contrast < 0.0f || contrast > 1.0f) {
     contrast = 1.0f;
   }
 
+  Float clearTypeLevel = settings.ClearTypeLevel();
+  if (clearTypeLevel < 0.0f || clearTypeLevel > 1.0f) {
+    clearTypeLevel = 1.0f;
+  }
+
   return SkCreateTypefaceFromDWriteFont(factory, mFontFace, mStyle,
-                                        mForceGDIMode, gamma, contrast);
+                                        (int)settings.RenderingMode(), gamma,
+                                        contrast, clearTypeLevel);
+}
+
+void ScaledFontDWrite::SetupSkFontDrawOptions(SkFont& aFont) {
+  if (ForceGDIMode()) {
+    aFont.setEmbeddedBitmaps(true);
+    aFont.setSubpixel(false);
+  } else {
+    aFont.setEmbeddedBitmaps(UseEmbeddedBitmaps());
+    aFont.setSubpixel(true);
+  }
+}
+
+bool ScaledFontDWrite::MayUseBitmaps() {
+  return ForceGDIMode() || UseEmbeddedBitmaps();
 }
-#endif
 
 void ScaledFontDWrite::CopyGlyphsToBuilder(const GlyphBuffer& aBuffer,
                                            PathBuilder* aBuilder,
                                            const Matrix* aTransformHint) {
   BackendType backendType = aBuilder->GetBackendType();
-  if (backendType == BackendType::CAPTURE) {
-    StreamingGeometrySink sink(aBuilder);
-    CopyGlyphsToSink(aBuffer, &sink);
-    return;
-  }
-
   if (backendType != BackendType::DIRECT2D &&
       backendType != BackendType::DIRECT2D1_1) {
     ScaledFontBase::CopyGlyphsToBuilder(aBuffer, aBuilder, aTransformHint);
@@ -214,34 +187,6 @@ void ScaledFontDWrite::CopyGlyphsToBuilder(const GlyphBuffer& aBuffer,
   CopyGlyphsToSink(aBuffer, pathBuilderD2D->GetSink());
 }
 
-void ScaledFontDWrite::GetGlyphDesignMetrics(const uint16_t* aGlyphs,
-                                             uint32_t aNumGlyphs,
-                                             GlyphMetrics* aGlyphMetrics) {
-  DWRITE_FONT_METRICS fontMetrics;
-  mFontFace->GetMetrics(&fontMetrics);
-
-  vector<DWRITE_GLYPH_METRICS> metrics(aNumGlyphs);
-  mFontFace->GetDesignGlyphMetrics(aGlyphs, aNumGlyphs, &metrics.front());
-
-  Float scaleFactor = mSize / fontMetrics.designUnitsPerEm;
-
-  for (uint32_t i = 0; i < aNumGlyphs; i++) {
-    aGlyphMetrics[i].mXBearing = metrics[i].leftSideBearing * scaleFactor;
-    aGlyphMetrics[i].mXAdvance = metrics[i].advanceWidth * scaleFactor;
-    aGlyphMetrics[i].mYBearing =
-        (metrics[i].topSideBearing - metrics[i].verticalOriginY) * scaleFactor;
-    aGlyphMetrics[i].mYAdvance = metrics[i].advanceHeight * scaleFactor;
-    aGlyphMetrics[i].mWidth =
-        (metrics[i].advanceWidth - metrics[i].leftSideBearing -
-         metrics[i].rightSideBearing) *
-        scaleFactor;
-    aGlyphMetrics[i].mHeight =
-        (metrics[i].advanceHeight - metrics[i].topSideBearing -
-         metrics[i].bottomSideBearing) *
-        scaleFactor;
-  }
-}
-
 void ScaledFontDWrite::CopyGlyphsToSink(const GlyphBuffer& aBuffer,
                                         ID2D1SimplifiedGeometrySink* aSink) {
   std::vector<UINT16> indices;
@@ -305,7 +250,8 @@ bool UnscaledFontDWrite::GetFontFileData(FontFileDataOutput aDataCallback,
   }
 
   RefPtr<IDWriteFontFileStream> stream;
-  hr = loader->CreateStreamFromKey(referenceKey, refKeySize, getter_AddRefs(stream));
+  hr = loader->CreateStreamFromKey(referenceKey, refKeySize,
+                                   getter_AddRefs(stream));
   if (FAILED(hr)) {
     return false;
   }
@@ -317,19 +263,26 @@ bool UnscaledFontDWrite::GetFontFileData(FontFileDataOutput aDataCallback,
     return false;
   }
 
-  uint32_t fileSize = static_cast<uint32_t>(fileSize64);
-  const void* fragmentStart;
-  void* context;
-  hr = stream->ReadFileFragment(&fragmentStart, 0, fileSize, &context);
+  // Try to catch any device memory exceptions that may occur while attempting
+  // to read the file fragment.
+  void* context = nullptr;
+  hr = E_FAIL;
+  MOZ_SEH_TRY {
+    uint32_t fileSize = static_cast<uint32_t>(fileSize64);
+    const void* fragmentStart = nullptr;
+    hr = stream->ReadFileFragment(&fragmentStart, 0, fileSize, &context);
+    if (SUCCEEDED(hr)) {
+      aDataCallback((uint8_t*)fragmentStart, fileSize, mFontFace->GetIndex(),
+                    aBaton);
+    }
+  }
+  MOZ_SEH_EXCEPT(EXCEPTION_EXECUTE_HANDLER) {
+    gfxCriticalNote << "Exception occurred reading DWrite font file data";
+  }
   if (FAILED(hr)) {
     return false;
   }
-
-  aDataCallback((uint8_t*)fragmentStart, fileSize, mFontFace->GetIndex(),
-                aBaton);
-
   stream->ReleaseFileFragment(context);
-
   return true;
 }
 
@@ -390,12 +343,12 @@ static bool GetFontFileName(RefPtr<IDWriteFontFace> aFontFace,
     gfxDebug() << "Invalid file \"" << aFileName.data() << "\" for WR font";
     return false;
   }
-  aFileName.pop_back();
+  // We leave the null terminator at the end of the returned file name.
   return true;
 }
 
-bool UnscaledFontDWrite::GetWRFontDescriptor(WRFontDescriptorOutput aCb,
-                                             void* aBaton) {
+bool UnscaledFontDWrite::GetFontDescriptor(FontDescriptorOutput aCb,
+                                           void* aBaton) {
   if (!mFont) {
     return false;
   }
@@ -413,23 +366,61 @@ bool UnscaledFontDWrite::GetWRFontDescriptor(WRFontDescriptorOutput aCb,
 
 ScaledFontDWrite::InstanceData::InstanceData(
     const wr::FontInstanceOptions* aOptions,
-    const wr::FontInstancePlatformOptions* aPlatformOptions)
-    : mUseEmbeddedBitmap(false),
-      mForceGDIMode(false),
-      mGamma(2.2f),
-      mContrast(1.0f) {
+    const wr::FontInstancePlatformOptions* aPlatformOptions) {
   if (aOptions) {
-    if (aOptions->flags & wr::FontInstanceFlags_EMBEDDED_BITMAPS) {
+    if (aOptions->flags & wr::FontInstanceFlags::EMBEDDED_BITMAPS) {
       mUseEmbeddedBitmap = true;
     }
-    if (aOptions->flags & wr::FontInstanceFlags_FORCE_GDI) {
-      mForceGDIMode = true;
+    if (aOptions->flags & wr::FontInstanceFlags::SYNTHETIC_BOLD) {
+      mUseBoldSimulation = true;
+    }
+    if (aOptions->flags & wr::FontInstanceFlags::MULTISTRIKE_BOLD) {
+      mUseMultistrikeBold = true;
+    }
+    if (aOptions->flags & wr::FontInstanceFlags::FORCE_GDI) {
+      mGDIForced = true;
     }
   }
-  if (aPlatformOptions) {
-    mGamma = aPlatformOptions->gamma / 100.0f;
-    mContrast = aPlatformOptions->contrast / 100.0f;
+}
+
+bool ScaledFontDWrite::HasVariationSettings() {
+  RefPtr<IDWriteFontFace5> ff5;
+  mFontFace->QueryInterface(__uuidof(IDWriteFontFace5),
+                            (void**)getter_AddRefs(ff5));
+  if (!ff5 || !ff5->HasVariations()) {
+    return false;
   }
+
+  uint32_t count = ff5->GetFontAxisValueCount();
+  if (!count) {
+    return false;
+  }
+
+  RefPtr<IDWriteFontResource> res;
+  if (FAILED(ff5->GetFontResource(getter_AddRefs(res)))) {
+    return false;
+  }
+
+  std::vector<DWRITE_FONT_AXIS_VALUE> defaults(count);
+  if (FAILED(res->GetDefaultFontAxisValues(defaults.data(), count))) {
+    return false;
+  }
+
+  std::vector<DWRITE_FONT_AXIS_VALUE> values(count);
+  if (FAILED(ff5->GetFontAxisValues(values.data(), count))) {
+    return false;
+  }
+
+  for (uint32_t i = 0; i < count; i++) {
+    DWRITE_FONT_AXIS_ATTRIBUTES attr = res->GetFontAxisAttributes(i);
+    if (attr & DWRITE_FONT_AXIS_ATTRIBUTES_VARIABLE) {
+      if (values[i].value != defaults[i].value) {
+        return true;
+      }
+    }
+  }
+
+  return false;
 }
 
 // Helper for ScaledFontDWrite::GetFontInstanceData: if the font has variation
@@ -490,31 +481,46 @@ bool ScaledFontDWrite::GetWRFontInstanceOptions(
     Maybe<wr::FontInstancePlatformOptions>* aOutPlatformOptions,
     std::vector<FontVariation>* aOutVariations) {
   wr::FontInstanceOptions options;
-  bool useSubpixel = !mParams || mParams->GetClearTypeLevel() != 0.0f;
-  options.render_mode = wr::ToFontRenderMode(GetDefaultAAMode(), useSubpixel);
+  options.render_mode = wr::ToFontRenderMode(GetDefaultAAMode());
   options.flags = wr::FontInstanceFlags{0};
-  if (mFontFace->GetSimulations() & DWRITE_FONT_SIMULATIONS_BOLD) {
-    options.flags |= wr::FontInstanceFlags_SYNTHETIC_BOLD;
+  if (HasBoldSimulation()) {
+    options.flags |= wr::FontInstanceFlags::SYNTHETIC_BOLD;
+  }
+  if (UseMultistrikeBold()) {
+    options.flags |= wr::FontInstanceFlags::MULTISTRIKE_BOLD;
   }
   if (UseEmbeddedBitmaps()) {
-    options.flags |= wr::FontInstanceFlags_EMBEDDED_BITMAPS;
+    options.flags |= wr::FontInstanceFlags::EMBEDDED_BITMAPS;
   }
   if (ForceGDIMode()) {
-    options.flags |= wr::FontInstanceFlags_FORCE_GDI;
+    options.flags |= wr::FontInstanceFlags::FORCE_GDI;
   } else {
-    options.flags |= wr::FontInstanceFlags_SUBPIXEL_POSITION;
+    options.flags |= wr::FontInstanceFlags::SUBPIXEL_POSITION;
+  }
+  auto& settings = DWriteSettings();
+  switch (settings.RenderingMode()) {
+    case DWRITE_RENDERING_MODE_CLEARTYPE_NATURAL_SYMMETRIC:
+      options.flags |= wr::FontInstanceFlags::FORCE_SYMMETRIC;
+      break;
+    case DWRITE_RENDERING_MODE_CLEARTYPE_NATURAL:
+      options.flags |= wr::FontInstanceFlags::NO_SYMMETRIC;
+      break;
+    default:
+      break;
   }
   if (Factory::GetBGRSubpixelOrder()) {
-    options.flags |= wr::FontInstanceFlags_SUBPIXEL_BGR;
+    options.flags |= wr::FontInstanceFlags::SUBPIXEL_BGR;
   }
-  options.bg_color = wr::ToColorU(Color());
+  options.bg_color = wr::ToColorU(DeviceColor());
   options.synthetic_italics =
       wr::DegreesToSyntheticItalics(GetSyntheticObliqueAngle());
 
   wr::FontInstancePlatformOptions platformOptions;
-  platformOptions.gamma = uint16_t(std::round(mGamma * 100.0f));
+  platformOptions.gamma = uint16_t(std::round(settings.Gamma() * 100.0f));
   platformOptions.contrast =
-      uint16_t(std::round(std::min(mContrast, 1.0f) * 100.0f));
+      uint8_t(std::round(std::min(settings.EnhancedContrast(), 1.0f) * 100.0f));
+  platformOptions.cleartype_level =
+      uint8_t(std::round(std::min(settings.ClearTypeLevel(), 1.0f) * 100.0f));
 
   *aOutOptions = Some(options);
   *aOutPlatformOptions = Some(platformOptions);
@@ -524,46 +530,107 @@ bool ScaledFontDWrite::GetWRFontInstanceOptions(
   return true;
 }
 
+DWriteSettings& ScaledFontDWrite::DWriteSettings() const {
+  return DWriteSettings::Get(mGDIForced);
+}
+
 // Helper for UnscaledFontDWrite::CreateScaledFont: create a clone of the
 // given IDWriteFontFace, with specified variation-axis values applied.
 // Returns nullptr in case of failure.
 static already_AddRefed<IDWriteFontFace5> CreateFaceWithVariations(
-    IDWriteFontFace* aFace, const FontVariation* aVariations,
-    uint32_t aNumVariations) {
+    IDWriteFontFace* aFace, DWRITE_FONT_SIMULATIONS aSimulations,
+    const FontVariation* aVariations = nullptr, uint32_t aNumVariations = 0) {
   auto makeDWriteAxisTag = [](uint32_t aTag) {
     return DWRITE_MAKE_FONT_AXIS_TAG((aTag >> 24) & 0xff, (aTag >> 16) & 0xff,
                                      (aTag >> 8) & 0xff, aTag & 0xff);
   };
 
-  RefPtr<IDWriteFontFace5> ff5;
-  aFace->QueryInterface(__uuidof(IDWriteFontFace5),
-                        (void**)getter_AddRefs(ff5));
-  if (!ff5) {
+  MOZ_SEH_TRY {
+    RefPtr<IDWriteFontFace5> ff5;
+    aFace->QueryInterface(__uuidof(IDWriteFontFace5),
+                          (void**)getter_AddRefs(ff5));
+    if (!ff5) {
+      return nullptr;
+    }
+
+    RefPtr<IDWriteFontResource> res;
+    if (FAILED(ff5->GetFontResource(getter_AddRefs(res)))) {
+      return nullptr;
+    }
+
+    std::vector<DWRITE_FONT_AXIS_VALUE> fontAxisValues;
+    if (aNumVariations) {
+      fontAxisValues.reserve(aNumVariations);
+      for (uint32_t i = 0; i < aNumVariations; i++) {
+        DWRITE_FONT_AXIS_VALUE axisValue = {
+            makeDWriteAxisTag(aVariations[i].mTag), aVariations[i].mValue};
+        fontAxisValues.push_back(axisValue);
+      }
+    } else {
+      uint32_t count = ff5->GetFontAxisValueCount();
+      if (count) {
+        fontAxisValues.resize(count);
+        if (FAILED(ff5->GetFontAxisValues(fontAxisValues.data(), count))) {
+          fontAxisValues.clear();
+        }
+      }
+    }
+
+    RefPtr<IDWriteFontFace5> newFace;
+    if (FAILED(res->CreateFontFace(aSimulations, fontAxisValues.data(),
+                                   fontAxisValues.size(),
+                                   getter_AddRefs(newFace)))) {
+      return nullptr;
+    }
+    return newFace.forget();
+  }
+  MOZ_SEH_EXCEPT(EXCEPTION_EXECUTE_HANDLER) {
+    gfxCriticalNote << "Exception occurred initializing variation face";
     return nullptr;
   }
+}
 
-  DWRITE_FONT_SIMULATIONS sims = aFace->GetSimulations();
-  RefPtr<IDWriteFontResource> res;
-  if (FAILED(ff5->GetFontResource(getter_AddRefs(res)))) {
-    return nullptr;
+bool UnscaledFontDWrite::InitBold() {
+  if (mFontFaceBold) {
+    return true;
   }
 
-  std::vector<DWRITE_FONT_AXIS_VALUE> fontAxisValues;
-  fontAxisValues.reserve(aNumVariations);
-  for (uint32_t i = 0; i < aNumVariations; i++) {
-    DWRITE_FONT_AXIS_VALUE axisValue = {makeDWriteAxisTag(aVariations[i].mTag),
-                                        aVariations[i].mValue};
-    fontAxisValues.push_back(axisValue);
+  DWRITE_FONT_SIMULATIONS sims = mFontFace->GetSimulations();
+  if (sims & DWRITE_FONT_SIMULATIONS_BOLD) {
+    mFontFaceBold = mFontFace;
+    return true;
   }
+  sims |= DWRITE_FONT_SIMULATIONS_BOLD;
 
-  RefPtr<IDWriteFontFace5> newFace;
-  if (FAILED(res->CreateFontFace(sims, fontAxisValues.data(),
-                                 fontAxisValues.size(),
-                                 getter_AddRefs(newFace)))) {
-    return nullptr;
+  RefPtr<IDWriteFontFace5> ff5 = CreateFaceWithVariations(mFontFace, sims);
+  if (ff5) {
+    mFontFaceBold = ff5;
+  } else {
+    MOZ_SEH_TRY {
+      UINT32 numFiles = 0;
+      if (FAILED(mFontFace->GetFiles(&numFiles, nullptr))) {
+        return false;
+      }
+      StackArray<IDWriteFontFile*, 1> files(numFiles);
+      if (FAILED(mFontFace->GetFiles(&numFiles, files.data()))) {
+        return false;
+      }
+      HRESULT hr = Factory::GetDWriteFactory()->CreateFontFace(
+          mFontFace->GetType(), numFiles, files.data(), mFontFace->GetIndex(),
+          sims, getter_AddRefs(mFontFaceBold));
+      for (UINT32 i = 0; i < numFiles; ++i) {
+        files[i]->Release();
+      }
+      if (FAILED(hr) || !mFontFaceBold) {
+        return false;
+      }
+    }
+    MOZ_SEH_EXCEPT(EXCEPTION_EXECUTE_HANDLER) {
+      gfxCriticalNote << "Exception occurred initializing bold face";
+      return false;
+    }
   }
-
-  return newFace.forget();
+  return true;
 }
 
 already_AddRefed<ScaledFont> UnscaledFontDWrite::CreateScaledFont(
@@ -578,12 +645,21 @@ already_AddRefed<ScaledFont> UnscaledFontDWrite::CreateScaledFont(
       *reinterpret_cast<const ScaledFontDWrite::InstanceData*>(aInstanceData);
 
   IDWriteFontFace* face = mFontFace;
+  if (instanceData.mUseBoldSimulation) {
+    if (!InitBold()) {
+      gfxWarning() << "Failed creating bold IDWriteFontFace.";
+      return nullptr;
+    }
+    face = mFontFaceBold;
+  }
+  DWRITE_FONT_SIMULATIONS sims = face->GetSimulations();
 
   // If variations are required, we create a separate IDWriteFontFace5 with
   // the requested settings applied.
   RefPtr<IDWriteFontFace5> ff5;
   if (aNumVariations) {
-    ff5 = CreateFaceWithVariations(mFontFace, aVariations, aNumVariations);
+    ff5 =
+        CreateFaceWithVariations(mFontFace, sims, aVariations, aNumVariations);
     if (ff5) {
       face = ff5;
     } else {
@@ -591,17 +667,9 @@ already_AddRefed<ScaledFont> UnscaledFontDWrite::CreateScaledFont(
     }
   }
 
-  RefPtr<ScaledFontBase> scaledFont = new ScaledFontDWrite(
+  return MakeAndAddRef<ScaledFontDWrite>(
       face, this, aGlyphSize, instanceData.mUseEmbeddedBitmap,
-      instanceData.mForceGDIMode, nullptr, instanceData.mGamma,
-      instanceData.mContrast);
-
-  if (mNeedsCairo && !scaledFont->PopulateCairoScaledFont()) {
-    gfxWarning() << "Unable to create cairo scaled font DWrite font.";
-    return nullptr;
-  }
-
-  return scaledFont.forget();
+      instanceData.mUseMultistrikeBold, instanceData.mGDIForced, nullptr);
 }
 
 already_AddRefed<ScaledFont> UnscaledFontDWrite::CreateScaledFontFromWRFont(
@@ -616,23 +684,86 @@ already_AddRefed<ScaledFont> UnscaledFontDWrite::CreateScaledFontFromWRFont(
 AntialiasMode ScaledFontDWrite::GetDefaultAAMode() {
   AntialiasMode defaultMode = GetSystemDefaultAAMode();
 
-  if (defaultMode == AntialiasMode::GRAY) {
-    if (!DoGrayscale(mFontFace, mSize)) {
-      defaultMode = AntialiasMode::NONE;
-    }
+  switch (defaultMode) {
+    case AntialiasMode::SUBPIXEL:
+    case AntialiasMode::DEFAULT:
+      if (DWriteSettings().ClearTypeLevel() == 0.0f) {
+        defaultMode = AntialiasMode::GRAY;
+      }
+      break;
+    case AntialiasMode::GRAY:
+      if (!DoGrayscale(mFontFace, mSize)) {
+        defaultMode = AntialiasMode::NONE;
+      }
+      break;
+    case AntialiasMode::NONE:
+      break;
   }
+
   return defaultMode;
 }
 
-#ifdef USE_CAIRO_SCALED_FONT
-cairo_font_face_t* ScaledFontDWrite::GetCairoFontFace() {
+cairo_font_face_t* ScaledFontDWrite::CreateCairoFontFace(
+    cairo_font_options_t* aFontOptions) {
   if (!mFontFace) {
     return nullptr;
   }
 
   return cairo_dwrite_font_face_create_for_dwrite_fontface(nullptr, mFontFace);
 }
-#endif
+
+void ScaledFontDWrite::PrepareCairoScaledFont(cairo_scaled_font_t* aFont) {
+  if (mGDIForced) {
+    cairo_dwrite_scaled_font_set_force_GDI_classic(aFont, true);
+  }
+}
+
+already_AddRefed<UnscaledFont> UnscaledFontDWrite::CreateFromFontDescriptor(
+    const uint8_t* aData, uint32_t aDataLength, uint32_t aIndex) {
+  // Note that despite the type of aData here, it actually points to a 16-bit
+  // Windows font file path (hence the cast to WCHAR* below).
+  if (aDataLength == 0) {
+    gfxWarning() << "DWrite font descriptor is truncated.";
+    return nullptr;
+  }
+
+  RefPtr<IDWriteFactory> factory = Factory::GetDWriteFactory();
+  if (!factory) {
+    return nullptr;
+  }
+
+  MOZ_SEH_TRY {
+    RefPtr<IDWriteFontFile> fontFile;
+    HRESULT hr = factory->CreateFontFileReference((const WCHAR*)aData, nullptr,
+                                                  getter_AddRefs(fontFile));
+    if (FAILED(hr)) {
+      return nullptr;
+    }
+    BOOL isSupported;
+    DWRITE_FONT_FILE_TYPE fileType;
+    DWRITE_FONT_FACE_TYPE faceType;
+    UINT32 numFaces;
+    hr = fontFile->Analyze(&isSupported, &fileType, &faceType, &numFaces);
+    if (FAILED(hr) || !isSupported || aIndex >= numFaces) {
+      return nullptr;
+    }
+    IDWriteFontFile* fontFiles[1] = {fontFile.get()};
+    RefPtr<IDWriteFontFace> fontFace;
+    hr = factory->CreateFontFace(faceType, 1, fontFiles, aIndex,
+                                 DWRITE_FONT_SIMULATIONS_NONE,
+                                 getter_AddRefs(fontFace));
+    if (FAILED(hr)) {
+      return nullptr;
+    }
+    RefPtr unscaledFont = new UnscaledFontDWrite(fontFace, nullptr);
+    return unscaledFont.forget();
+  }
+  MOZ_SEH_EXCEPT(EXCEPTION_EXECUTE_HANDLER) {
+    gfxCriticalNote << "Exception occurred creating unscaledFont for "
+                    << NS_ConvertUTF16toUTF8((const char16_t*)aData).get();
+    return nullptr;
+  }
+}
 
 }  // namespace gfx
 }  // namespace mozilla
