@@ -5,16 +5,30 @@
  * found in the LICENSE file.
  */
 
-#include "SkPolyUtils.h"
+#include "src/utils/SkPolyUtils.h"
 
+#include "include/core/SkRect.h"
+#include "include/core/SkTypes.h"
+#include "include/private/base/SkDebug.h"
+#include "include/private/base/SkFloatingPoint.h"
+#include "include/private/base/SkMalloc.h"
+#include "include/private/base/SkTArray.h"
+#include "include/private/base/SkTDArray.h"
+#include "include/private/base/SkTemplates.h"
+#include "src/base/SkTDPQueue.h"
+#include "src/base/SkTInternalLList.h"
+#include "src/base/SkVx.h"
+#include "src/core/SkPointPriv.h"
+#include "src/core/SkRectPriv.h"
+
+#include <algorithm>
+#include <cstdint>
 #include <limits>
+#include <new>
 
-#include "SkNx.h"
-#include "SkPointPriv.h"
-#include "SkTArray.h"
-#include "SkTemplates.h"
-#include "SkTDPQueue.h"
-#include "SkTInternalLList.h"
+using namespace skia_private;
+
+#if !defined(SK_ENABLE_OPTIMIZE_SIZE)
 
 //////////////////////////////////////////////////////////////////////////////////
 // Helper data structures and functions
@@ -79,9 +93,16 @@ static inline bool outside_interval(SkScalar numer, SkScalar denom, bool denomPo
            (!denomPositive && (numer > 0 || numer < denom));
 }
 
+// special zero-length test when we're using vdotv as a denominator
+static inline bool zero_length(const SkPoint& v, SkScalar vdotv) {
+    return !(SkScalarsAreFinite(v.fX, v.fY) && vdotv);
+}
+
 // Compute the intersection 'p' between segments s0 and s1, if any.
 // 's' is the parametric value for the intersection along 's0' & 't' is the same for 's1'.
 // Returns false if there is no intersection.
+// If the length squared of a segment is 0, then we treat the segment as degenerate
+// and use only the first endpoint for tests.
 static bool compute_intersection(const OffsetSegment& s0, const OffsetSegment& s1,
                                  SkPoint* p, SkScalar* s, SkScalar* t) {
     const SkVector& v0 = s0.fV;
@@ -98,9 +119,11 @@ static bool compute_intersection(const OffsetSegment& s0, const OffsetSegment& s
         }
 
         // Check for zero-length segments
-        if (!SkPointPriv::CanNormalize(v0.fX, v0.fY)) {
+        SkScalar v0dotv0 = v0.dot(v0);
+        if (zero_length(v0, v0dotv0)) {
             // Both are zero-length
-            if (!SkPointPriv::CanNormalize(v1.fX, v1.fY)) {
+            SkScalar v1dotv1 = v1.dot(v1);
+            if (zero_length(v1, v1dotv1)) {
                 // Check if they're the same point
                 if (!SkPointPriv::CanNormalize(w.fX, w.fY)) {
                     *p = s0.fP0;
@@ -108,12 +131,13 @@ static bool compute_intersection(const OffsetSegment& s0, const OffsetSegment& s
                     *t = 0;
                     return true;
                 } else {
+                    // Intersection is indeterminate
                     return false;
                 }
             }
             // Otherwise project segment0's origin onto segment1
             tNumer = v1.dot(-w);
-            denom = v1.dot(v1);
+            denom = v1dotv1;
             if (outside_interval(tNumer, denom, true)) {
                 return false;
             }
@@ -121,12 +145,13 @@ static bool compute_intersection(const OffsetSegment& s0, const OffsetSegment& s
         } else {
             // Project segment1's endpoints onto segment0
             sNumer = v0.dot(w);
-            denom = v0.dot(v0);
+            denom = v0dotv0;
             tNumer = 0;
             if (outside_interval(sNumer, denom, true)) {
                 // The first endpoint doesn't lie on segment0
                 // If segment1 is degenerate, then there's no collision
-                if (!SkPointPriv::CanNormalize(v1.fX, v1.fY)) {
+                SkScalar v1dotv1 = v1.dot(v1);
+                if (zero_length(v1, v1dotv1)) {
                     return false;
                 }
 
@@ -143,7 +168,7 @@ static bool compute_intersection(const OffsetSegment& s0, const OffsetSegment& s
                     // otherwise project segment0's endpoint onto segment1 instead
                     sNumer = 0;
                     tNumer = v1.dot(-w);
-                    denom = v1.dot(v1);
+                    denom = v1dotv1;
                 }
             }
         }
@@ -173,17 +198,17 @@ bool SkIsConvexPolygon(const SkPoint* polygonVerts, int polygonSize) {
         return false;
     }
 
-    SkScalar lastArea = 0;
     SkScalar lastPerpDot = 0;
+    int xSignChangeCount = 0;
+    int ySignChangeCount = 0;
 
     int prevIndex = polygonSize - 1;
     int currIndex = 0;
     int nextIndex = 1;
-    SkPoint origin = polygonVerts[0];
     SkVector v0 = polygonVerts[currIndex] - polygonVerts[prevIndex];
+    SkScalar lastVx = v0.fX;
+    SkScalar lastVy = v0.fY;
     SkVector v1 = polygonVerts[nextIndex] - polygonVerts[currIndex];
-    SkVector w0 = polygonVerts[currIndex] - origin;
-    SkVector w1 = polygonVerts[nextIndex] - origin;
     for (int i = 0; i < polygonSize; ++i) {
         if (!polygonVerts[i].isFinite()) {
             return false;
@@ -198,23 +223,27 @@ bool SkIsConvexPolygon(const SkPoint* polygonVerts, int polygonSize) {
             lastPerpDot = perpDot;
         }
 
-        // If the signed area ever flips it's concave
-        // TODO: see if we can verify convexity only with signed area
-        SkScalar quadArea = w0.cross(w1);
-        if (quadArea*lastArea < 0) {
+        // Check that the signs of the edge vectors don't change more than twice per coordinate
+        if (lastVx*v1.fX < 0) {
+            xSignChangeCount++;
+        }
+        if (lastVy*v1.fY < 0) {
+            ySignChangeCount++;
+        }
+        if (xSignChangeCount > 2 || ySignChangeCount > 2) {
             return false;
         }
-        if (0 != quadArea) {
-            lastArea = quadArea;
-        }
-
         prevIndex = currIndex;
         currIndex = nextIndex;
         nextIndex = (currIndex + 1) % polygonSize;
+        if (v1.fX != 0) {
+            lastVx = v1.fX;
+        }
+        if (v1.fY != 0) {
+            lastVy = v1.fY;
+        }
         v0 = v1;
         v1 = polygonVerts[nextIndex] - polygonVerts[currIndex];
-        w0 = w1;
-        w1 = polygonVerts[nextIndex] - origin;
     }
 
     return true;
@@ -326,7 +355,7 @@ bool SkInsetConvexPolygon(const SkPoint* inputPolygonVerts, int inputPolygonSize
     // insetting close to zero just returns the original poly
     if (inset <= SK_ScalarNearlyZero) {
         for (int i = 0; i < inputPolygonSize; ++i) {
-            *insetPolygon->push() = inputPolygonVerts[i];
+            *insetPolygon->append() = inputPolygonVerts[i];
         }
         return true;
     }
@@ -338,7 +367,7 @@ bool SkInsetConvexPolygon(const SkPoint* inputPolygonVerts, int inputPolygonSize
     }
 
     // set up
-    SkAutoSTMalloc<64, OffsetEdge> edgeData(inputPolygonSize);
+    AutoSTMalloc<64, OffsetEdge> edgeData(inputPolygonSize);
     int prev = inputPolygonSize - 1;
     for (int curr = 0; curr < inputPolygonSize; prev = curr, ++curr) {
         int next = (curr + 1) % inputPolygonSize;
@@ -431,16 +460,16 @@ bool SkInsetConvexPolygon(const SkPoint* inputPolygonVerts, int inputPolygonSize
 
     static constexpr SkScalar kCleanupTolerance = 0.01f;
     if (insetVertexCount >= 0) {
-        insetPolygon->setReserve(insetVertexCount);
+        insetPolygon->reserve(insetVertexCount);
     }
     int currIndex = 0;
-    *insetPolygon->push() = head->fIntersection;
+    *insetPolygon->append() = head->fIntersection;
     currEdge = head->fNext;
     while (currEdge != head) {
         if (!SkPointPriv::EqualsWithinTolerance(currEdge->fIntersection,
                                                 (*insetPolygon)[currIndex],
                                                 kCleanupTolerance)) {
-            *insetPolygon->push() = currEdge->fIntersection;
+            *insetPolygon->append() = currEdge->fIntersection;
             currIndex++;
         }
         currEdge = currEdge->fNext;
@@ -449,10 +478,10 @@ bool SkInsetConvexPolygon(const SkPoint* inputPolygonVerts, int inputPolygonSize
     if (currIndex >= 1 &&
         SkPointPriv::EqualsWithinTolerance((*insetPolygon)[0], (*insetPolygon)[currIndex],
                                             kCleanupTolerance)) {
-        insetPolygon->pop();
+        insetPolygon->pop_back();
     }
 
-    return SkIsConvexPolygon(insetPolygon->begin(), insetPolygon->count());
+    return SkIsConvexPolygon(insetPolygon->begin(), insetPolygon->size());
 }
 
 ///////////////////////////////////////////////////////////////////////////////////////////
@@ -481,7 +510,13 @@ bool SkComputeRadialSteps(const SkVector& v1, const SkVector& v2, SkScalar offse
     int steps = SkScalarRoundToInt(floatSteps);
 
     SkScalar dTheta = steps > 0 ? theta / steps : 0;
-    *rotSin = SkScalarSinCos(dTheta, rotCos);
+    *rotSin = SkScalarSin(dTheta);
+    *rotCos = SkScalarCos(dTheta);
+    // Our offset may be so large that we end up with a tiny dTheta, in which case we
+    // lose precision when computing rotSin and rotCos.
+    if (steps > 0 && (*rotSin == 0 || *rotCos == 1)) {
+        return false;
+    }
     *n = steps;
     return true;
 }
@@ -715,9 +750,17 @@ public:
                 curr->fAbove = pred;
                 curr->fBelow = succ;
                 if (pred) {
+                    if (pred->fSegment.fP0 == curr->fSegment.fP0 &&
+                        pred->fSegment.fV == curr->fSegment.fV) {
+                        return false;
+                    }
                     pred->fBelow = curr;
                 }
                 if (succ) {
+                    if (succ->fSegment.fP0 == curr->fSegment.fP0 &&
+                        succ->fSegment.fV == curr->fSegment.fV) {
+                        return false;
+                    }
                     succ->fAbove = curr;
                 }
                 if (IsRed(parent)) {
@@ -884,7 +927,7 @@ public:
                 } else {
                     ActiveEdge *s = parent->fChild[!last];
 
-                    if (s != NULL) {
+                    if (s != nullptr) {
                         if (!IsRed(s->fChild[!last]) && !IsRed(s->fChild[last])) {
                             // color flip
                             parent->fRed = false;
@@ -1051,16 +1094,16 @@ bool SkIsSimplePolygon(const SkPoint* polygon, int polygonSize) {
         return false;
     }
 
-    // need to be able to represent all the vertices in the 16-bit indices
-    if (polygonSize > std::numeric_limits<uint16_t>::max()) {
-        return false;
-    }
-
     // If it's convex, it's simple
     if (SkIsConvexPolygon(polygon, polygonSize)) {
         return true;
     }
 
+    // practically speaking, it takes too long to process large polygons
+    if (polygonSize > 2048) {
+        return false;
+    }
+
     SkTDPQueue <Vertex, Vertex::Left> vertexQueue(polygonSize);
     for (int i = 0; i < polygonSize; ++i) {
         Vertex newVertex;
@@ -1072,6 +1115,10 @@ bool SkIsSimplePolygon(const SkPoint* polygon, int polygonSize) {
         newVertex.fPrevIndex = (i - 1 + polygonSize) % polygonSize;
         newVertex.fNextIndex = (i + 1) % polygonSize;
         newVertex.fFlags = 0;
+        // The two edges adjacent to this vertex are the same, so polygon is not simple
+        if (polygon[newVertex.fPrevIndex] == polygon[newVertex.fNextIndex]) {
+            return false;
+        }
         if (left(polygon[newVertex.fPrevIndex], polygon[i])) {
             newVertex.fFlags |= kPrevLeft_VertexFlag;
         }
@@ -1145,7 +1192,8 @@ static bool is_reflex_vertex(const SkPoint* inputPolygonVerts, int winding, SkSc
     return (side*winding*offset < 0);
 }
 
-bool SkOffsetSimplePolygon(const SkPoint* inputPolygonVerts, int inputPolygonSize, SkScalar offset,
+bool SkOffsetSimplePolygon(const SkPoint* inputPolygonVerts, int inputPolygonSize,
+                           const SkRect& bounds, SkScalar offset,
                            SkTDArray<SkPoint>* offsetPolygon, SkTDArray<int>* polygonIndices) {
     if (inputPolygonSize < 3) {
         return false;
@@ -1160,11 +1208,19 @@ bool SkOffsetSimplePolygon(const SkPoint* inputPolygonVerts, int inputPolygonSiz
         return false;
     }
 
+    // can't inset more than the half bounds of the polygon
+    if (offset > std::min(SkTAbs(SkRectPriv::HalfWidth(bounds)),
+                          SkTAbs(SkRectPriv::HalfHeight(bounds)))) {
+        return false;
+    }
+
     // offsetting close to zero just returns the original poly
     if (SkScalarNearlyZero(offset)) {
         for (int i = 0; i < inputPolygonSize; ++i) {
-            *offsetPolygon->push() = inputPolygonVerts[i];
-            *polygonIndices->push() = i;
+            *offsetPolygon->append() = inputPolygonVerts[i];
+            if (polygonIndices) {
+                *polygonIndices->append() = i;
+            }
         }
         return true;
     }
@@ -1176,7 +1232,7 @@ bool SkOffsetSimplePolygon(const SkPoint* inputPolygonVerts, int inputPolygonSiz
     }
 
     // build normals
-    SkAutoSTMalloc<64, SkVector> normals(inputPolygonSize);
+    AutoSTMalloc<64, SkVector> normals(inputPolygonSize);
     unsigned int numEdges = 0;
     for (int currIndex = 0, prevIndex = inputPolygonSize - 1;
          currIndex < inputPolygonSize;
@@ -1199,7 +1255,7 @@ bool SkOffsetSimplePolygon(const SkPoint* inputPolygonVerts, int inputPolygonSiz
                                           &rotSin, &rotCos, &numSteps)) {
                     return false;
                 }
-                numEdges += SkTMax(numSteps, 1);
+                numEdges += std::max(numSteps, 1);
             }
         }
         numEdges++;
@@ -1212,7 +1268,7 @@ bool SkOffsetSimplePolygon(const SkPoint* inputPolygonVerts, int inputPolygonSiz
                                   &rotSin, &rotCos, &numSteps)) {
             return false;
         }
-        numEdges += SkTMax(numSteps, 1);
+        numEdges += std::max(numSteps, 1);
     }
 
     // Make sure we don't overflow the max array count.
@@ -1239,7 +1295,7 @@ bool SkOffsetSimplePolygon(const SkPoint* inputPolygonVerts, int inputPolygonSiz
                                       &rotSin, &rotCos, &numSteps)) {
                 return false;
             }
-            auto currEdge = edgeData.push_back_n(SkTMax(numSteps, 1));
+            auto currEdge = edgeData.push_back_n(std::max(numSteps, 1));
             for (int i = 0; i < numSteps - 1; ++i) {
                 SkVector currNormal = SkVector::Make(prevNormal.fX*rotCos - prevNormal.fY*rotSin,
                                                      prevNormal.fY*rotCos + prevNormal.fX*rotSin);
@@ -1284,7 +1340,7 @@ bool SkOffsetSimplePolygon(const SkPoint* inputPolygonVerts, int inputPolygonSiz
     edgeData[0].fPrev = prevEdge;
 
     // now clip edges
-    SkASSERT(edgeData.count() == (int)numEdges);
+    SkASSERT(edgeData.size() == (int)numEdges);
     auto head = &edgeData[0];
     auto currEdge = head;
     unsigned int offsetVertexCount = numEdges;
@@ -1375,20 +1431,20 @@ bool SkOffsetSimplePolygon(const SkPoint* inputPolygonVerts, int inputPolygonSiz
     }
 
     static constexpr SkScalar kCleanupTolerance = 0.01f;
-    offsetPolygon->setReserve(offsetVertexCount);
+    offsetPolygon->reserve(offsetVertexCount);
     int currIndex = 0;
-    *offsetPolygon->push() = head->fIntersection;
+    *offsetPolygon->append() = head->fIntersection;
     if (polygonIndices) {
-        *polygonIndices->push() = head->fIndex;
+        *polygonIndices->append() = head->fIndex;
     }
     currEdge = head->fNext;
     while (currEdge != head) {
         if (!SkPointPriv::EqualsWithinTolerance(currEdge->fIntersection,
                                                 (*offsetPolygon)[currIndex],
                                                 kCleanupTolerance)) {
-            *offsetPolygon->push() = currEdge->fIntersection;
+            *offsetPolygon->append() = currEdge->fIntersection;
             if (polygonIndices) {
-                *polygonIndices->push() = currEdge->fIndex;
+                *polygonIndices->append() = currEdge->fIndex;
             }
             currIndex++;
         }
@@ -1398,17 +1454,17 @@ bool SkOffsetSimplePolygon(const SkPoint* inputPolygonVerts, int inputPolygonSiz
     if (currIndex >= 1 &&
         SkPointPriv::EqualsWithinTolerance((*offsetPolygon)[0], (*offsetPolygon)[currIndex],
                                             kCleanupTolerance)) {
-        offsetPolygon->pop();
+        offsetPolygon->pop_back();
         if (polygonIndices) {
-            polygonIndices->pop();
+            polygonIndices->pop_back();
         }
     }
 
     // check winding of offset polygon (it should be same as the original polygon)
-    SkScalar offsetWinding = SkGetPolygonWinding(offsetPolygon->begin(), offsetPolygon->count());
+    SkScalar offsetWinding = SkGetPolygonWinding(offsetPolygon->begin(), offsetPolygon->size());
 
     return (winding*offsetWinding > 0 &&
-            SkIsSimplePolygon(offsetPolygon->begin(), offsetPolygon->count()));
+            SkIsSimplePolygon(offsetPolygon->begin(), offsetPolygon->size()));
 }
 
 //////////////////////////////////////////////////////////////////////////////////////////
@@ -1427,13 +1483,13 @@ struct TriangulationVertex {
 
 static void compute_triangle_bounds(const SkPoint& p0, const SkPoint& p1, const SkPoint& p2,
                                     SkRect* bounds) {
-    Sk4s min, max;
-    min = max = Sk4s(p0.fX, p0.fY, p0.fX, p0.fY);
-    Sk4s xy(p1.fX, p1.fY, p2.fX, p2.fY);
-    min = Sk4s::Min(min, xy);
-    max = Sk4s::Max(max, xy);
-    bounds->set(SkTMin(min[0], min[2]), SkTMin(min[1], min[3]),
-                SkTMax(max[0], max[2]), SkTMax(max[1], max[3]));
+    skvx::float4 min, max;
+    min = max = skvx::float4(p0.fX, p0.fY, p0.fX, p0.fY);
+    skvx::float4 xy(p1.fX, p1.fY, p2.fX, p2.fY);
+    min = skvx::min(min, xy);
+    max = skvx::max(max, xy);
+    bounds->setLTRB(std::min(min[0], min[2]), std::min(min[1], min[3]),
+                    std::max(max[0], max[2]), std::max(max[1], max[3]));
 }
 
 // test to see if point p is in triangle p0p1p2.
@@ -1480,7 +1536,7 @@ public:
         if (!SkScalarIsFinite(hCount)) {
             return false;
         }
-        fHCount = SkTMax(SkTMin(SkScalarRoundToInt(hCount), vertexCount), 1);
+        fHCount = std::max(std::min(SkScalarRoundToInt(hCount), vertexCount), 1);
         fVCount = vertexCount/fHCount;
         fGridConversion.set(sk_ieee_float_divide(fHCount - 0.001f, width),
                             sk_ieee_float_divide(fVCount - 0.001f, height));
@@ -1488,8 +1544,8 @@ public:
             return false;
         }
 
-        fGrid.setCount(fHCount*fVCount);
-        for (int i = 0; i < fGrid.count(); ++i) {
+        fGrid.resize(fHCount*fVCount);
+        for (int i = 0; i < fGrid.size(); ++i) {
             fGrid[i].reset();
         }
 
@@ -1597,13 +1653,13 @@ bool SkTriangulateSimplePolygon(const SkPoint* polygonVerts, uint16_t* indexMap,
     }
 
     // Set up vertices
-    SkAutoSTMalloc<64, TriangulationVertex> triangulationVertices(polygonSize);
+    AutoSTArray<64, TriangulationVertex> triangulationVertices(polygonSize);
     int prevIndex = polygonSize - 1;
     SkVector v0 = polygonVerts[0] - polygonVerts[prevIndex];
     for (int currIndex = 0; currIndex < polygonSize; ++currIndex) {
         int nextIndex = (currIndex + 1) % polygonSize;
 
-        SkDEBUGCODE(memset(&triangulationVertices[currIndex], 0, sizeof(TriangulationVertex)));
+        triangulationVertices[currIndex] = TriangulationVertex{};
         triangulationVertices[currIndex].fPosition = polygonVerts[currIndex];
         triangulationVertices[currIndex].fIndex = currIndex;
         triangulationVertices[currIndex].fPrevIndex = prevIndex;
@@ -1654,7 +1710,7 @@ bool SkTriangulateSimplePolygon(const SkPoint* polygonVerts, uint16_t* indexMap,
     // In the worst case this is an n^2 algorithm. We can cut down the search space somewhat by
     // noting that only convex vertices can be potential ears, and we only need to check whether
     // any reflex vertices lie inside the ear.
-    triangleIndices->setReserve(triangleIndices->count() + 3 * (polygonSize - 2));
+    triangleIndices->reserve(triangleIndices->size() + 3 * (polygonSize - 2));
     int vertexCount = polygonSize;
     while (vertexCount > 3) {
         bool success = false;
@@ -1708,121 +1764,11 @@ bool SkTriangulateSimplePolygon(const SkPoint* polygonVerts, uint16_t* indexMap,
     for (SkTInternalLList<TriangulationVertex>::Iter vertexIter = convexList.begin();
          vertexIter != convexList.end(); ++vertexIter) {
         TriangulationVertex* vertex = *vertexIter;
-        *triangleIndices->push() = indexMap[vertex->fIndex];
+        *triangleIndices->append() = indexMap[vertex->fIndex];
     }
 
     return true;
 }
 
-///////////
-
-static double crs(SkVector a, SkVector b) {
-    return a.fX * b.fY - a.fY * b.fX;
-}
-
-static int sign(SkScalar v) {
-    return v < 0 ? -1 : (v > 0);
-}
-
-struct SignTracker {
-    int fSign;
-    int fSignChanges;
-
-    void reset() {
-        fSign = 0;
-        fSignChanges = 0;
-    }
-
-    void init(int s) {
-        SkASSERT(fSignChanges == 0);
-        SkASSERT(s == 1 || s == -1 || s == 0);
-        fSign = s;
-        fSignChanges = 1;
-    }
-
-    void update(int s) {
-        if (s) {
-            if (fSign != s) {
-                fSignChanges += 1;
-                fSign = s;
-            }
-        }
-    }
-};
-
-struct ConvexTracker {
-    SkVector    fFirst, fPrev;
-    SignTracker fDSign, fCSign;
-    int         fVecCounter;
-    bool        fIsConcave;
-
-    ConvexTracker() { this->reset(); }
-
-    void reset() {
-        fPrev = {0, 0};
-        fDSign.reset();
-        fCSign.reset();
-        fVecCounter = 0;
-        fIsConcave = false;
-    }
-
-    void addVec(SkPoint p1, SkPoint p0) {
-        this->addVec(p1 - p0);
-    }
-    void addVec(SkVector v) {
-        if (v.fX == 0 && v.fY == 0) {
-            return;
-        }
-
-        fVecCounter += 1;
-        if (fVecCounter == 1) {
-            fFirst = fPrev = v;
-            fDSign.update(sign(v.fX));
-            return;
-        }
-
-        SkScalar d = v.fX;
-        SkScalar c = crs(fPrev, v);
-        int sign_c;
-        if (c) {
-            sign_c = sign(c);
-        } else {
-            if (d >= 0) {
-                sign_c = fCSign.fSign;
-            } else {
-                sign_c = -fCSign.fSign;
-            }
-        }
-
-        fDSign.update(sign(d));
-        fCSign.update(sign_c);
-        fPrev = v;
-
-        if (fDSign.fSignChanges > 3 || fCSign.fSignChanges > 1) {
-            fIsConcave = true;
-        }
-    }
-
-    void finalCross() {
-        this->addVec(fFirst);
-    }
-};
-
-bool SkIsPolyConvex_experimental(const SkPoint pts[], int count) {
-    if (count <= 3) {
-        return true;
-    }
-
-    ConvexTracker tracker;
-
-    for (int i = 0; i < count - 1; ++i) {
-        tracker.addVec(pts[i + 1], pts[i]);
-        if (tracker.fIsConcave) {
-            return false;
-        }
-    }
-    tracker.addVec(pts[0], pts[count - 1]);
-    tracker.finalCross();
-    return !tracker.fIsConcave;
-}
+#endif // !defined(SK_ENABLE_OPTIMIZE_SIZE)
 
