@@ -12,7 +12,6 @@
 #include "mozilla/gfx/Blur.h"
 #include "mozilla/gfx/PathHelpers.h"
 #include "mozilla/Maybe.h"
-#include "mozilla/SystemGroup.h"
 #include "nsExpirationTracker.h"
 #include "nsClassHashtable.h"
 #include "gfxUtils.h"
@@ -22,29 +21,28 @@
 using namespace mozilla;
 using namespace mozilla::gfx;
 
-gfxAlphaBoxBlur::~gfxAlphaBoxBlur() {}
+gfxAlphaBoxBlur::~gfxAlphaBoxBlur() = default;
 
-already_AddRefed<gfxContext> gfxAlphaBoxBlur::Init(gfxContext* aDestinationCtx,
-                                                   const gfxRect& aRect,
-                                                   const IntSize& aSpreadRadius,
-                                                   const IntSize& aBlurRadius,
-                                                   const gfxRect* aDirtyRect,
-                                                   const gfxRect* aSkipRect,
-                                                   bool aUseHardwareAccel) {
+UniquePtr<gfxContext> gfxAlphaBoxBlur::Init(gfxContext* aDestinationCtx,
+                                            const gfxRect& aRect,
+                                            const IntSize& aSpreadRadius,
+                                            const IntSize& aBlurRadius,
+                                            const gfxRect* aDirtyRect,
+                                            const gfxRect* aSkipRect,
+                                            bool aUseHardwareAccel) {
   DrawTarget* refDT = aDestinationCtx->GetDrawTarget();
   Maybe<Rect> dirtyRect = aDirtyRect ? Some(ToRect(*aDirtyRect)) : Nothing();
   Maybe<Rect> skipRect = aSkipRect ? Some(ToRect(*aSkipRect)) : Nothing();
   RefPtr<DrawTarget> dt = InitDrawTarget(
       refDT, ToRect(aRect), aSpreadRadius, aBlurRadius,
       dirtyRect.ptrOr(nullptr), skipRect.ptrOr(nullptr), aUseHardwareAccel);
-  if (!dt) {
+  if (!dt || !dt->IsValid()) {
     return nullptr;
   }
 
-  RefPtr<gfxContext> context = gfxContext::CreateOrNull(dt);
-  MOZ_ASSERT(context);  // already checked for target above
+  auto context = MakeUnique<gfxContext>(dt);
   context->SetMatrix(Matrix::Translation(-mBlur.GetRect().TopLeft()));
-  return context.forget();
+  return context;
 }
 
 already_AddRefed<DrawTarget> gfxAlphaBoxBlur::InitDrawTarget(
@@ -71,18 +69,8 @@ already_AddRefed<DrawTarget> gfxAlphaBoxBlur::InitDrawTarget(
     mAccelerated = true;
   }
 
-  if (aReferenceDT->IsCaptureDT()) {
-    if (mAccelerated) {
-      mDrawTarget = Factory::CreateCaptureDrawTarget(backend, mBlur.GetSize(),
-                                                     SurfaceFormat::A8);
-    } else {
-      mDrawTarget = Factory::CreateCaptureDrawTargetForData(
-          backend, mBlur.GetSize(), SurfaceFormat::A8, mBlur.GetStride(),
-          blurDataSize);
-    }
-  } else if (mAccelerated) {
-    // Note: CreateShadowDrawTarget is only implemented for Cairo, so we don't
-    // care about mimicking this in the DrawTargetCapture case.
+  if (mAccelerated) {
+    // Note: CreateShadowDrawTarget is only implemented for Cairo.
     mDrawTarget = aReferenceDT->CreateShadowDrawTarget(
         mBlur.GetSize(), SurfaceFormat::A8,
         AlphaBoxBlur::CalculateBlurSigma(aBlurRadius.width));
@@ -125,7 +113,7 @@ already_AddRefed<DrawTarget> gfxAlphaBoxBlur::InitDrawTarget(
 }
 
 already_AddRefed<SourceSurface> gfxAlphaBoxBlur::DoBlur(
-    const Color* aShadowColor, IntPoint* aOutTopLeft) {
+    const sRGBColor* aShadowColor, IntPoint* aOutTopLeft) {
   if (aOutTopLeft) {
     *aOutTopLeft = mBlur.GetRect().TopLeft();
   }
@@ -142,13 +130,12 @@ already_AddRefed<SourceSurface> gfxAlphaBoxBlur::DoBlur(
       return nullptr;
     }
     blurDT->DrawSurfaceWithShadow(
-        blurMask, Point(0, 0), Color(1, 1, 1), Point(0, 0),
-        AlphaBoxBlur::CalculateBlurSigma(mBlur.GetBlurRadius().width),
+        blurMask, Point(0, 0),
+        ShadowOptions(
+            DeviceColor::MaskOpaqueWhite(), Point(0, 0),
+            AlphaBoxBlur::CalculateBlurSigma(mBlur.GetBlurRadius().width)),
         CompositionOp::OP_OVER);
     blurMask = blurDT->Snapshot();
-  } else if (mDrawTarget->IsCaptureDT()) {
-    mDrawTarget->Blur(mBlur);
-    blurMask = mDrawTarget->Snapshot();
   }
 
   if (!aShadowColor) {
@@ -167,7 +154,7 @@ already_AddRefed<SourceSurface> gfxAlphaBoxBlur::DoBlur(
 }
 
 void gfxAlphaBoxBlur::Paint(gfxContext* aDestinationCtx) {
-  if ((mDrawTarget && !mDrawTarget->IsCaptureDT()) && !mAccelerated && !mData) {
+  if (mDrawTarget && !mAccelerated && !mData) {
     return;
   }
 
@@ -225,7 +212,7 @@ struct BlurCacheKey : public PLDHashEntryHdr {
 
   IntSize mMinSize;
   IntSize mBlurRadius;
-  Color mShadowColor;
+  sRGBColor mShadowColor;
   BackendType mBackend;
   RectCornerRadii mCornerRadii;
   bool mIsInset;
@@ -234,8 +221,8 @@ struct BlurCacheKey : public PLDHashEntryHdr {
   IntSize mInnerMinSize;
 
   BlurCacheKey(const IntSize& aMinSize, const IntSize& aBlurRadius,
-               const RectCornerRadii* aCornerRadii, const Color& aShadowColor,
-               BackendType aBackendType)
+               const RectCornerRadii* aCornerRadii,
+               const sRGBColor& aShadowColor, BackendType aBackendType)
       : BlurCacheKey(aMinSize, IntSize(0, 0), aBlurRadius, aCornerRadii,
                      aShadowColor, false, aBackendType) {}
 
@@ -252,7 +239,7 @@ struct BlurCacheKey : public PLDHashEntryHdr {
                         const IntSize& aInnerMinSize,
                         const IntSize& aBlurRadius,
                         const RectCornerRadii* aCornerRadii,
-                        const Color& aShadowColor, bool aIsInset,
+                        const sRGBColor& aShadowColor, bool aIsInset,
                         BackendType aBackendType)
       : mMinSize(aOuterMinSize),
         mBlurRadius(aBlurRadius),
@@ -337,9 +324,7 @@ struct BlurCacheData {
 class BlurCache final : public nsExpirationTracker<BlurCacheData, 4> {
  public:
   BlurCache()
-      : nsExpirationTracker<BlurCacheData, 4>(
-            GENERATION_MS, "BlurCache",
-            SystemGroup::EventTargetFor(TaskCategory::Other)) {}
+      : nsExpirationTracker<BlurCacheData, 4>(GENERATION_MS, "BlurCache") {}
 
   virtual void NotifyExpired(BlurCacheData* aObject) override {
     RemoveObject(aObject);
@@ -348,7 +333,8 @@ class BlurCache final : public nsExpirationTracker<BlurCacheData, 4> {
 
   BlurCacheData* Lookup(const IntSize& aMinSize, const IntSize& aBlurRadius,
                         const RectCornerRadii* aCornerRadii,
-                        const Color& aShadowColor, BackendType aBackendType) {
+                        const sRGBColor& aShadowColor,
+                        BackendType aBackendType) {
     BlurCacheData* blur = mHashEntries.Get(BlurCacheKey(
         aMinSize, aBlurRadius, aCornerRadii, aShadowColor, aBackendType));
     if (blur) {
@@ -362,7 +348,7 @@ class BlurCache final : public nsExpirationTracker<BlurCacheData, 4> {
                                       const IntSize& aInnerMinSize,
                                       const IntSize& aBlurRadius,
                                       const RectCornerRadii* aCornerRadii,
-                                      const Color& aShadowColor,
+                                      const sRGBColor& aShadowColor,
                                       BackendType aBackendType) {
     bool insetBoxShadow = true;
     BlurCacheKey key(aOuterMinSize, aInnerMinSize, aBlurRadius, aCornerRadii,
@@ -375,20 +361,17 @@ class BlurCache final : public nsExpirationTracker<BlurCacheData, 4> {
     return blur;
   }
 
-  // Returns true if we successfully register the blur in the cache, false
-  // otherwise.
-  bool RegisterEntry(BlurCacheData* aValue) {
-    nsresult rv = AddObject(aValue);
+  void RegisterEntry(UniquePtr<BlurCacheData> aValue) {
+    nsresult rv = AddObject(aValue.get());
     if (NS_FAILED(rv)) {
       // We are OOM, and we cannot track this object. We don't want stall
       // entries in the hash table (since the expiration tracker is responsible
       // for removing the cache entries), so we avoid putting that entry in the
-      // table, which is a good things considering we are short on memory
+      // table, which is a good thing considering we are short on memory
       // anyway, we probably don't want to retain things.
-      return false;
+      return;
     }
-    mHashEntries.Put(aValue->mKey, aValue);
-    return true;
+    mHashEntries.InsertOrUpdate(aValue->mKey, std::move(aValue));
   }
 
  protected:
@@ -409,7 +392,7 @@ static IntSize ComputeMinSizeForShadowShape(const RectCornerRadii* aCornerRadii,
   Size cornerSize(0, 0);
   if (aCornerRadii) {
     const RectCornerRadii& corners = *aCornerRadii;
-    NS_FOR_CSS_FULL_CORNERS(i) {
+    for (const auto i : mozilla::AllPhysicalCorners()) {
       cornerSize.width = std::max(cornerSize.width, corners[i].width);
       cornerSize.height = std::max(cornerSize.height, corners[i].height);
     }
@@ -446,22 +429,20 @@ static IntSize ComputeMinSizeForShadowShape(const RectCornerRadii* aCornerRadii,
 static void CacheBlur(DrawTarget* aDT, const IntSize& aMinSize,
                       const IntSize& aBlurRadius,
                       const RectCornerRadii* aCornerRadii,
-                      const Color& aShadowColor, const IntMargin& aBlurMargin,
-                      SourceSurface* aBoxShadow) {
-  BlurCacheKey key(aMinSize, aBlurRadius, aCornerRadii, aShadowColor,
-                   aDT->GetBackendType());
-  BlurCacheData* data =
-      new BlurCacheData(aBoxShadow, aBlurMargin, std::move(key));
-  if (!gBlurCache->RegisterEntry(data)) {
-    delete data;
-  }
+                      const sRGBColor& aShadowColor,
+                      const IntMargin& aBlurMargin, SourceSurface* aBoxShadow) {
+  gBlurCache->RegisterEntry(MakeUnique<BlurCacheData>(
+      aBoxShadow, aBlurMargin,
+      BlurCacheKey(aMinSize, aBlurRadius, aCornerRadii, aShadowColor,
+                   aDT->GetBackendType())));
 }
 
 // Blurs a small surface and creates the colored box shadow.
 static already_AddRefed<SourceSurface> CreateBoxShadow(
     DrawTarget* aDestDrawTarget, const IntSize& aMinSize,
     const RectCornerRadii* aCornerRadii, const IntSize& aBlurRadius,
-    const Color& aShadowColor, bool aMirrorCorners, IntMargin& aOutBlurMargin) {
+    const sRGBColor& aShadowColor, bool aMirrorCorners,
+    IntMargin& aOutBlurMargin) {
   gfxAlphaBoxBlur blur;
   Rect minRect(Point(0, 0), Size(aMinSize));
   Rect blurRect(minRect);
@@ -479,7 +460,7 @@ static already_AddRefed<SourceSurface> CreateBoxShadow(
     return nullptr;
   }
 
-  ColorPattern black(Color(0.f, 0.f, 0.f, 1.f));
+  ColorPattern black(DeviceColor::MaskOpaqueBlack());
 
   if (aCornerRadii) {
     RefPtr<Path> roundedRect =
@@ -506,8 +487,8 @@ static already_AddRefed<SourceSurface> CreateBoxShadow(
 static already_AddRefed<SourceSurface> GetBlur(
     gfxContext* aDestinationCtx, const IntSize& aRectSize,
     const IntSize& aBlurRadius, const RectCornerRadii* aCornerRadii,
-    const Color& aShadowColor, bool aMirrorCorners, IntMargin& aOutBlurMargin,
-    IntMargin& aOutSlice, IntSize& aOutMinSize) {
+    const sRGBColor& aShadowColor, bool aMirrorCorners,
+    IntMargin& aOutBlurMargin, IntMargin& aOutSlice, IntSize& aOutMinSize) {
   if (!gBlurCache) {
     gBlurCache = new BlurCache();
   }
@@ -877,7 +858,7 @@ void gfxAlphaBoxBlur::BlurRectangle(gfxContext* aDestinationCtx,
                                     const gfxRect& aRect,
                                     const RectCornerRadii* aCornerRadii,
                                     const gfxPoint& aBlurStdDev,
-                                    const Color& aShadowColor,
+                                    const sRGBColor& aShadowColor,
                                     const gfxRect& aDirtyRect,
                                     const gfxRect& aSkipRect) {
   if (!RectIsInt32Safe(ToRect(aRect))) {
@@ -978,7 +959,7 @@ static already_AddRefed<Path> GetBoxShadowInsetPath(
 
 static void FillDestinationPath(
     gfxContext* aDestinationCtx, const Rect& aDestinationRect,
-    const Rect& aShadowClipRect, const Color& aShadowColor,
+    const Rect& aShadowClipRect, const sRGBColor& aShadowColor,
     const RectCornerRadii* aInnerClipRadii = nullptr) {
   // When there is no blur radius, fill the path onto the destination
   // surface.
@@ -995,22 +976,21 @@ static void CacheInsetBlur(const IntSize& aMinOuterSize,
                            const IntSize& aMinInnerSize,
                            const IntSize& aBlurRadius,
                            const RectCornerRadii* aCornerRadii,
-                           const Color& aShadowColor, BackendType aBackendType,
+                           const sRGBColor& aShadowColor,
+                           BackendType aBackendType,
                            SourceSurface* aBoxShadow) {
   bool isInsetBlur = true;
   BlurCacheKey key(aMinOuterSize, aMinInnerSize, aBlurRadius, aCornerRadii,
                    aShadowColor, isInsetBlur, aBackendType);
   IntMargin blurMargin(0, 0, 0, 0);
-  BlurCacheData* data =
-      new BlurCacheData(aBoxShadow, blurMargin, std::move(key));
-  if (!gBlurCache->RegisterEntry(data)) {
-    delete data;
-  }
+
+  gBlurCache->RegisterEntry(
+      MakeUnique<BlurCacheData>(aBoxShadow, blurMargin, std::move(key)));
 }
 
 already_AddRefed<SourceSurface> gfxAlphaBoxBlur::GetInsetBlur(
     const Rect& aOuterRect, const Rect& aWhitespaceRect, bool aIsDestRect,
-    const Color& aShadowColor, const IntSize& aBlurRadius,
+    const sRGBColor& aShadowColor, const IntSize& aBlurRadius,
     const RectCornerRadii* aInnerClipRadii, DrawTarget* aDestDrawTarget,
     bool aMirrorCorners) {
   if (!gBlurCache) {
@@ -1063,7 +1043,7 @@ already_AddRefed<SourceSurface> gfxAlphaBoxBlur::GetInsetBlur(
   RefPtr<Path> maskPath = GetBoxShadowInsetPath(
       minDrawTarget, aOuterRect, aWhitespaceRect, aInnerClipRadii);
 
-  ColorPattern black(Color(0.f, 0.f, 0.f, 1.f));
+  ColorPattern black(DeviceColor::MaskOpaqueBlack());
   minDrawTarget->Fill(maskPath, black);
 
   // Blur and fill in with the color we actually wanted
@@ -1115,7 +1095,7 @@ static void GetBlurMargins(const RectCornerRadii* aInnerClipRadii,
   Size cornerSize(0, 0);
   if (aInnerClipRadii) {
     const RectCornerRadii& corners = *aInnerClipRadii;
-    NS_FOR_CSS_FULL_CORNERS(i) {
+    for (const auto i : mozilla::AllPhysicalCorners()) {
       cornerSize.width = std::max(cornerSize.width, corners[i].width);
       cornerSize.height = std::max(cornerSize.height, corners[i].height);
     }
@@ -1168,7 +1148,7 @@ static bool GetInsetBoxShadowRects(const Margin& aBlurMargin,
 void gfxAlphaBoxBlur::BlurInsetBox(
     gfxContext* aDestinationCtx, const Rect& aDestinationRect,
     const Rect& aShadowClipRect, const IntSize& aBlurRadius,
-    const Color& aShadowColor, const RectCornerRadii* aInnerClipRadii,
+    const sRGBColor& aShadowColor, const RectCornerRadii* aInnerClipRadii,
     const Rect& aSkipRect, const Point& aShadowOffset) {
   if ((aBlurRadius.width == 0 && aBlurRadius.height == 0) ||
       aShadowClipRect.IsEmpty()) {
