@@ -7,7 +7,10 @@
 #include "VRProcessChild.h"
 
 #include "mozilla/BackgroundHangMonitor.h"
+#include "mozilla/GeckoArgs.h"
 #include "mozilla/ipc/IOThreadChild.h"
+#include "mozilla/ipc/ProcessUtils.h"
+#include "mozilla/StaticPrefs_dom.h"
 
 using namespace mozilla;
 using namespace mozilla::gfx;
@@ -15,9 +18,6 @@ using mozilla::ipc::IOThreadChild;
 
 StaticRefPtr<VRParent> sVRParent;
 
-VRProcessChild::VRProcessChild(ProcessId aParentPid)
-    : ProcessChild(aParentPid) {}
-
 VRProcessChild::~VRProcessChild() { sVRParent = nullptr; }
 
 /*static*/
@@ -27,19 +27,18 @@ VRParent* VRProcessChild::GetVRParent() {
 }
 
 bool VRProcessChild::Init(int aArgc, char* aArgv[]) {
-  char* parentBuildID = nullptr;
-  for (int i = 1; i < aArgc; i++) {
-    if (!aArgv[i]) {
-      continue;
-    }
-    if (strcmp(aArgv[i], "-parentBuildID") == 0) {
-      parentBuildID = aArgv[i + 1];
-    }
+  Maybe<const char*> parentBuildID =
+      geckoargs::sParentBuildID.Get(aArgc, aArgv);
+  if (parentBuildID.isNothing()) {
+    return false;
+  }
+
+  if (!ProcessChild::InitPrefs(aArgc, aArgv)) {
+    return false;
   }
 
   sVRParent = new VRParent();
-  sVRParent->Init(ParentPid(), parentBuildID, IOThreadChild::message_loop(),
-                  IOThreadChild::channel());
+  sVRParent->Init(TakeInitialEndpoint(), *parentBuildID);
 
   return true;
 }
@@ -47,4 +46,4 @@ bool VRProcessChild::Init(int aArgc, char* aArgv[]) {
 void VRProcessChild::CleanUp() {
   sVRParent = nullptr;
   NS_ShutdownXPCOM(nullptr);
-}
\ No newline at end of file
+}
