@@ -5,12 +5,35 @@
 
 #include "gtest/gtest.h"
 #include "mozilla/Preferences.h"
+#include "mozilla/intl/Locale.h"
 #include "mozilla/intl/LocaleService.h"
-#include "mozilla/intl/MozLocale.h"
-#include "nsIToolkitChromeRegistry.h"
+#include "mozilla/intl/Collator.h"
 
 using namespace mozilla::intl;
 
+TEST(Intl_Locale_LocaleService, CanonicalizeLanguageId)
+{
+  nsCString locale("en-US.POSIX");
+  ASSERT_TRUE(LocaleService::CanonicalizeLanguageId(locale));
+  ASSERT_TRUE(locale.EqualsLiteral("en-US"));
+
+  locale.AssignLiteral("en-US_POSIX");
+  ASSERT_TRUE(LocaleService::CanonicalizeLanguageId(locale));
+  ASSERT_TRUE(locale.EqualsLiteral("en-US-posix"));
+
+  locale.AssignLiteral("en-US-POSIX");
+  ASSERT_TRUE(LocaleService::CanonicalizeLanguageId(locale));
+  ASSERT_TRUE(locale.EqualsLiteral("en-US-posix"));
+
+  locale.AssignLiteral("C");
+  ASSERT_FALSE(LocaleService::CanonicalizeLanguageId(locale));
+  ASSERT_TRUE(locale.EqualsLiteral("und"));
+
+  locale.AssignLiteral("");
+  ASSERT_FALSE(LocaleService::CanonicalizeLanguageId(locale));
+  ASSERT_TRUE(locale.EqualsLiteral("und"));
+}
+
 TEST(Intl_Locale_LocaleService, GetAppLocalesAsBCP47)
 {
   nsTArray<nsCString> appLocales;
@@ -27,7 +50,7 @@ TEST(Intl_Locale_LocaleService, GetAppLocalesAsLangTags)
   ASSERT_FALSE(appLocales.IsEmpty());
 }
 
-TEST(Intl_Locale_LocaleService, GetAppLocalesAsLangTags_lastIsEnUS)
+TEST(Intl_Locale_LocaleService, GetAppLocalesAsLangTags_lastIsPresent)
 {
   nsAutoCString lastFallbackLocale;
   LocaleService::GetInstance()->GetLastFallbackLocale(lastFallbackLocale);
@@ -35,8 +58,7 @@ TEST(Intl_Locale_LocaleService, GetAppLocalesAsLangTags_lastIsEnUS)
   nsTArray<nsCString> appLocales;
   LocaleService::GetInstance()->GetAppLocalesAsLangTags(appLocales);
 
-  int32_t len = appLocales.Length();
-  ASSERT_TRUE(appLocales[len - 1].Equals(lastFallbackLocale));
+  ASSERT_TRUE(appLocales.Contains(lastFallbackLocale));
 }
 
 TEST(Intl_Locale_LocaleService, GetAppLocaleAsLangTag)
@@ -61,7 +83,7 @@ TEST(Intl_Locale_LocaleService, GetRegionalPrefsLocales)
 
 TEST(Intl_Locale_LocaleService, GetWebExposedLocales)
 {
-  const nsTArray<nsCString> spoofLocale{NS_LITERAL_CSTRING("de")};
+  const nsTArray<nsCString> spoofLocale{"de"_ns};
   LocaleService::GetInstance()->SetAvailableLocales(spoofLocale);
   LocaleService::GetInstance()->SetRequestedLocales(spoofLocale);
 
@@ -70,17 +92,17 @@ TEST(Intl_Locale_LocaleService, GetWebExposedLocales)
   mozilla::Preferences::SetInt("privacy.spoof_english", 0);
   LocaleService::GetInstance()->GetWebExposedLocales(pvLocales);
   ASSERT_TRUE(pvLocales.Length() > 0);
-  ASSERT_TRUE(pvLocales[0].Equals(NS_LITERAL_CSTRING("de")));
+  ASSERT_TRUE(pvLocales[0].Equals("de"_ns));
 
   mozilla::Preferences::SetCString("intl.locale.privacy.web_exposed", "zh-TW");
   LocaleService::GetInstance()->GetWebExposedLocales(pvLocales);
   ASSERT_TRUE(pvLocales.Length() > 0);
-  ASSERT_TRUE(pvLocales[0].Equals(NS_LITERAL_CSTRING("zh-TW")));
+  ASSERT_TRUE(pvLocales[0].Equals("zh-TW"_ns));
 
   mozilla::Preferences::SetInt("privacy.spoof_english", 2);
   LocaleService::GetInstance()->GetWebExposedLocales(pvLocales);
   ASSERT_EQ(1u, pvLocales.Length());
-  ASSERT_TRUE(pvLocales[0].Equals(NS_LITERAL_CSTRING("en-US")));
+  ASSERT_TRUE(pvLocales[0].Equals("en-US"_ns));
 }
 
 TEST(Intl_Locale_LocaleService, GetRequestedLocales)
@@ -116,12 +138,36 @@ TEST(Intl_Locale_LocaleService, GetDefaultLocale)
   LocaleService::GetInstance()->GetDefaultLocale(locStr);
 
   ASSERT_FALSE(locStr.IsEmpty());
-  ASSERT_TRUE(Locale(locStr).IsWellFormed());
+  Locale loc;
+  ASSERT_TRUE(LocaleParser::TryParse(locStr, loc).isOk());
 }
 
 TEST(Intl_Locale_LocaleService, IsAppLocaleRTL)
 {
-  // For now we can only test if the method doesn't crash.
-  LocaleService::GetInstance()->IsAppLocaleRTL();
-  ASSERT_TRUE(true);
+  mozilla::Preferences::SetCString("intl.l10n.pseudo", "bidi");
+  ASSERT_TRUE(LocaleService::GetInstance()->IsAppLocaleRTL());
+  mozilla::Preferences::ClearUser("intl.l10n.pseudo");
+}
+
+TEST(Intl_Locale_LocaleService, TryCreateComponent)
+{
+  {
+    // Create a Collator with the app locale.
+    auto result = LocaleService::GetInstance()->TryCreateComponent<Collator>();
+    ASSERT_TRUE(result.isOk());
+  }
+  {
+    // Create a Collator with the "en" locale.
+    auto result =
+        LocaleService::GetInstance()->TryCreateComponentWithLocale<Collator>(
+            "en");
+    ASSERT_TRUE(result.isOk());
+  }
+  {
+    // Fallback to the app locale when an invalid one is used.
+    auto result =
+        LocaleService::GetInstance()->TryCreateComponentWithLocale<Collator>(
+            "$invalidName");
+    ASSERT_TRUE(result.isOk());
+  }
 }
