@@ -9,6 +9,7 @@
 #include "nsIFile.h"
 #include "nsIObserverService.h"
 #include "mozilla/DebugOnly.h"
+#include "mozilla/Logging.h"
 #include "mozilla/Omnijar.h"
 #include "mozilla/Unused.h"
 
@@ -20,19 +21,27 @@
 
 using namespace mozilla;
 
+static LazyLogModule gJarLog("nsJAR");
+
+#ifdef LOG
+#  undef LOG
+#endif
+#ifdef LOG_ENABLED
+#  undef LOG_ENABLED
+#endif
+
+#define LOG(args) MOZ_LOG(gJarLog, mozilla::LogLevel::Debug, args)
+#define LOG_ENABLED() MOZ_LOG_TEST(gJarLog, mozilla::LogLevel::Debug)
+
 //----------------------------------------------
 // nsJAR constructor/destructor
 //----------------------------------------------
 
 // The following initialization makes a guess of 10 entries per jarfile.
 nsJAR::nsJAR()
-    : mZip(new nsZipArchive()),
-      mReleaseTime(PR_INTERVAL_NO_TIMEOUT),
-      mCache(nullptr),
+    : mReleaseTime(PR_INTERVAL_NO_TIMEOUT),
       mLock("nsJAR::mLock"),
-      mMtime(0),
-      mOpened(false),
-      mSkipArchiveClosing(false) {}
+      mCache(nullptr) {}
 
 nsJAR::~nsJAR() { Close(); }
 
@@ -50,7 +59,7 @@ MozExternalRefCountType nsJAR::Release(void) {
   if (mRefCnt == 2) {  // don't use a lock too frequently
     // Use a mutex here to guarantee mCache is not racing and the target
     // instance is still valid to increase ref-count.
-    MutexAutoLock lock(mLock);
+    RecursiveMutexAutoLock lock(mLock);
     cache = mCache;
     mCache = nullptr;
   }
@@ -79,76 +88,71 @@ MozExternalRefCountType nsJAR::Release(void) {
 NS_IMETHODIMP
 nsJAR::Open(nsIFile* zipFile) {
   NS_ENSURE_ARG_POINTER(zipFile);
-  if (mOpened) return NS_ERROR_FAILURE;  // Already open!
+  RecursiveMutexAutoLock lock(mLock);
+  LOG(("Open[%p] %s", this, zipFile->HumanReadablePath().get()));
+  if (mZip) return NS_ERROR_FAILURE;  // Already open!
 
   mZipFile = zipFile;
   mOuterZipEntry.Truncate();
-  mOpened = true;
 
   // The omnijar is special, it is opened early on and closed late
-  // this avoids reopening it
   RefPtr<nsZipArchive> zip = mozilla::Omnijar::GetReader(zipFile);
-  if (zip) {
-    mZip = zip;
-    mSkipArchiveClosing = true;
-    return NS_OK;
+  if (!zip) {
+    zip = nsZipArchive::OpenArchive(zipFile);
   }
-  return mZip->OpenArchive(zipFile);
+  mZip = zip;
+  return mZip ? NS_OK : NS_ERROR_FAILURE;
 }
 
 NS_IMETHODIMP
 nsJAR::OpenInner(nsIZipReader* aZipReader, const nsACString& aZipEntry) {
+  nsresult rv;
+
+  LOG(("OpenInner[%p] %s", this, PromiseFlatCString(aZipEntry).get()));
   NS_ENSURE_ARG_POINTER(aZipReader);
-  if (mOpened) return NS_ERROR_FAILURE;  // Already open!
+
+  nsCOMPtr<nsIFile> zipFile;
+  rv = aZipReader->GetFile(getter_AddRefs(zipFile));
+  NS_ENSURE_SUCCESS(rv, rv);
 
-  nsJAR* outerJAR = static_cast<nsJAR*>(aZipReader);
   RefPtr<nsZipArchive> innerZip =
-      mozilla::Omnijar::GetInnerReader(outerJAR->mZipFile, aZipEntry);
+      mozilla::Omnijar::GetInnerReader(zipFile, aZipEntry);
   if (innerZip) {
-    mOpened = true;
+    RecursiveMutexAutoLock lock(mLock);
+    if (mZip) {
+      return NS_ERROR_FAILURE;
+    }
     mZip = innerZip;
-    mSkipArchiveClosing = true;
     return NS_OK;
   }
 
   bool exist;
-  nsresult rv = aZipReader->HasEntry(aZipEntry, &exist);
+  rv = aZipReader->HasEntry(aZipEntry, &exist);
   NS_ENSURE_SUCCESS(rv, rv);
   NS_ENSURE_TRUE(exist, NS_ERROR_FILE_NOT_FOUND);
 
-  rv = aZipReader->GetFile(getter_AddRefs(mZipFile));
-  NS_ENSURE_SUCCESS(rv, rv);
-
-  mOpened = true;
+  RefPtr<nsZipHandle> handle;
+  {
+    nsJAR* outerJAR = static_cast<nsJAR*>(aZipReader);
+    RecursiveMutexAutoLock outerLock(outerJAR->mLock);
+    rv = nsZipHandle::Init(outerJAR->mZip.get(),
+                           PromiseFlatCString(aZipEntry).get(),
+                           getter_AddRefs(handle));
+    NS_ENSURE_SUCCESS(rv, rv);
+  }
 
+  RecursiveMutexAutoLock lock(mLock);
+  MOZ_ASSERT(!mZip, "Another thread tried to open this nsJAR racily!");
+  mZipFile = zipFile.forget();
   mOuterZipEntry.Assign(aZipEntry);
-
-  RefPtr<nsZipHandle> handle;
-  rv = nsZipHandle::Init(static_cast<nsJAR*>(aZipReader)->mZip.get(),
-                         PromiseFlatCString(aZipEntry).get(),
-                         getter_AddRefs(handle));
-  if (NS_FAILED(rv)) return rv;
-
-  return mZip->OpenArchive(handle);
-}
-
-NS_IMETHODIMP
-nsJAR::OpenMemory(void* aData, uint32_t aLength) {
-  NS_ENSURE_ARG_POINTER(aData);
-  if (mOpened) return NS_ERROR_FAILURE;  // Already open!
-
-  mOpened = true;
-
-  RefPtr<nsZipHandle> handle;
-  nsresult rv = nsZipHandle::Init(static_cast<uint8_t*>(aData), aLength,
-                                  getter_AddRefs(handle));
-  if (NS_FAILED(rv)) return rv;
-
-  return mZip->OpenArchive(handle);
+  mZip = nsZipArchive::OpenArchive(handle);
+  return mZip ? NS_OK : NS_ERROR_FAILURE;
 }
 
 NS_IMETHODIMP
 nsJAR::GetFile(nsIFile** result) {
+  RecursiveMutexAutoLock lock(mLock);
+  LOG(("GetFile[%p]", this));
   *result = mZipFile;
   NS_IF_ADDREF(*result);
   return NS_OK;
@@ -156,24 +160,22 @@ nsJAR::GetFile(nsIFile** result) {
 
 NS_IMETHODIMP
 nsJAR::Close() {
-  if (!mOpened) {
+  RecursiveMutexAutoLock lock(mLock);
+  LOG(("Close[%p]", this));
+  if (!mZip) {
     return NS_ERROR_FAILURE;  // Never opened or already closed.
   }
 
-  mOpened = false;
-
-  if (mSkipArchiveClosing) {
-    // Reset state, but don't close the omnijar because we did not open it.
-    mSkipArchiveClosing = false;
-    mZip = new nsZipArchive();
-    return NS_OK;
-  }
-
-  return mZip->CloseArchive();
+  mZip = nullptr;
+  return NS_OK;
 }
 
 NS_IMETHODIMP
 nsJAR::Test(const nsACString& aEntryName) {
+  RecursiveMutexAutoLock lock(mLock);
+  if (!mZip) {
+    return NS_ERROR_FAILURE;
+  }
   return mZip->Test(
       aEntryName.IsEmpty() ? nullptr : PromiseFlatCString(aEntryName).get());
 }
@@ -182,10 +184,14 @@ NS_IMETHODIMP
 nsJAR::Extract(const nsACString& aEntryName, nsIFile* outFile) {
   // nsZipArchive and zlib are not thread safe
   // we need to use a lock to prevent bug #51267
-  MutexAutoLock lock(mLock);
+  RecursiveMutexAutoLock lock(mLock);
+  if (!mZip) {
+    return NS_ERROR_FAILURE;
+  }
 
+  LOG(("Extract[%p] %s", this, PromiseFlatCString(aEntryName).get()));
   nsZipItem* item = mZip->GetItem(PromiseFlatCString(aEntryName).get());
-  NS_ENSURE_TRUE(item, NS_ERROR_FILE_TARGET_DOES_NOT_EXIST);
+  NS_ENSURE_TRUE(item, NS_ERROR_FILE_NOT_FOUND);
 
   // Remove existing file or directory so we set permissions correctly.
   // If it's a directory that already exists and contains files, throw
@@ -219,8 +225,13 @@ nsJAR::Extract(const nsACString& aEntryName, nsIFile* outFile) {
 
 NS_IMETHODIMP
 nsJAR::GetEntry(const nsACString& aEntryName, nsIZipEntry** result) {
+  RecursiveMutexAutoLock lock(mLock);
+  LOG(("GetEntry[%p] %s", this, PromiseFlatCString(aEntryName).get()));
+  if (!mZip) {
+    return NS_ERROR_FAILURE;
+  }
   nsZipItem* zipItem = mZip->GetItem(PromiseFlatCString(aEntryName).get());
-  NS_ENSURE_TRUE(zipItem, NS_ERROR_FILE_TARGET_DOES_NOT_EXIST);
+  NS_ENSURE_TRUE(zipItem, NS_ERROR_FILE_NOT_FOUND);
 
   nsJARItem* jarItem = new nsJARItem(zipItem);
 
@@ -230,6 +241,11 @@ nsJAR::GetEntry(const nsACString& aEntryName, nsIZipEntry** result) {
 
 NS_IMETHODIMP
 nsJAR::HasEntry(const nsACString& aEntryName, bool* result) {
+  RecursiveMutexAutoLock lock(mLock);
+  LOG(("HasEntry[%p] %s", this, PromiseFlatCString(aEntryName).get()));
+  if (!mZip) {
+    return NS_ERROR_FAILURE;
+  }
   *result = mZip->GetItem(PromiseFlatCString(aEntryName).get()) != nullptr;
   return NS_OK;
 }
@@ -238,6 +254,11 @@ NS_IMETHODIMP
 nsJAR::FindEntries(const nsACString& aPattern,
                    nsIUTF8StringEnumerator** result) {
   NS_ENSURE_ARG_POINTER(result);
+  RecursiveMutexAutoLock lock(mLock);
+  LOG(("FindEntries[%p] %s", this, PromiseFlatCString(aPattern).get()));
+  if (!mZip) {
+    return NS_ERROR_FAILURE;
+  }
 
   nsZipFind* find;
   nsresult rv = mZip->FindInit(
@@ -260,14 +281,21 @@ nsJAR::GetInputStreamWithSpec(const nsACString& aJarDirSpec,
                               const nsACString& aEntryName,
                               nsIInputStream** result) {
   NS_ENSURE_ARG_POINTER(result);
+  RecursiveMutexAutoLock lock(mLock);
+  if (!mZip) {
+    return NS_ERROR_FAILURE;
+  }
 
+  LOG(("GetInputStreamWithSpec[%p] %s %s", this,
+       PromiseFlatCString(aJarDirSpec).get(),
+       PromiseFlatCString(aEntryName).get()));
   // Watch out for the jar:foo.zip!/ (aDir is empty) top-level special case!
   nsZipItem* item = nullptr;
   const nsCString& entry = PromiseFlatCString(aEntryName);
   if (*entry.get()) {
     // First check if item exists in jar
     item = mZip->GetItem(entry.get());
-    if (!item) return NS_ERROR_FILE_TARGET_DOES_NOT_EXIST;
+    if (!item) return NS_ERROR_FILE_NOT_FOUND;
   }
   nsJARInputStream* jis = new nsJARInputStream();
   // addref now so we can call InitFile/InitDirectory()
@@ -277,7 +305,8 @@ nsJAR::GetInputStreamWithSpec(const nsACString& aJarDirSpec,
   if (!item || item->IsDirectory()) {
     rv = jis->InitDirectory(this, aJarDirSpec, entry.get());
   } else {
-    rv = jis->InitFile(this, item);
+    RefPtr<nsZipHandle> fd = mZip->GetFD();
+    rv = jis->InitFile(fd, mZip->GetData(item), item);
   }
   if (NS_FAILED(rv)) {
     NS_RELEASE(*result);
@@ -285,13 +314,27 @@ nsJAR::GetInputStreamWithSpec(const nsACString& aJarDirSpec,
   return rv;
 }
 
-nsresult nsJAR::GetJarPath(nsACString& aResult) {
+nsresult nsJAR::GetFullJarPath(nsACString& aResult) {
+  RecursiveMutexAutoLock lock(mLock);
   NS_ENSURE_ARG_POINTER(mZipFile);
 
-  return mZipFile->GetPersistentDescriptor(aResult);
+  nsresult rv = mZipFile->GetPersistentDescriptor(aResult);
+  if (NS_FAILED(rv)) {
+    return rv;
+  }
+
+  if (mOuterZipEntry.IsEmpty()) {
+    aResult.InsertLiteral("file:", 0);
+  } else {
+    aResult.InsertLiteral("jar:", 0);
+    aResult.AppendLiteral("!/");
+    aResult.Append(mOuterZipEntry);
+  }
+  return NS_OK;
 }
 
 nsresult nsJAR::GetNSPRFileDesc(PRFileDesc** aNSPRFileDesc) {
+  RecursiveMutexAutoLock lock(mLock);
   if (!aNSPRFileDesc) {
     return NS_ERROR_ILLEGAL_VALUE;
   }
@@ -317,7 +360,7 @@ nsresult nsJAR::LoadEntry(const nsACString& aFilename, nsCString& aBuf) {
   nsresult rv;
   nsCOMPtr<nsIInputStream> manifestStream;
   rv = GetInputStream(aFilename, getter_AddRefs(manifestStream));
-  if (NS_FAILED(rv)) return NS_ERROR_FILE_TARGET_DOES_NOT_EXIST;
+  if (NS_FAILED(rv)) return NS_ERROR_FILE_NOT_FOUND;
 
   //-- Read the manifest file into memory
   char* buf;
@@ -350,7 +393,7 @@ int32_t nsJAR::ReadLine(const char** src) {
   //--Moves pointer to beginning of next line and returns line length
   //  not including CR/LF.
   int32_t length;
-  char* eol = PL_strpbrk(*src, "\r\n");
+  const char* eol = strpbrk(*src, "\r\n");
 
   if (eol == nullptr)  // Probably reached end of file before newline
   {
@@ -380,7 +423,7 @@ nsJAREnumerator::HasMore(bool* aResult) {
   if (!mName) {
     NS_ASSERTION(mFind, "nsJAREnumerator: Missing zipFind.");
     nsresult rv = mFind->FindNext(&mName, &mNameLen);
-    if (rv == NS_ERROR_FILE_TARGET_DOES_NOT_EXIST) {
+    if (rv == NS_ERROR_FILE_NOT_FOUND) {
       *aResult = false;  // No more matches available
       return NS_OK;
     }
@@ -530,6 +573,7 @@ nsZipReaderCache::nsZipReaderCache()
 
 NS_IMETHODIMP
 nsZipReaderCache::Init(uint32_t cacheSize) {
+  MutexAutoLock lock(mLock);
   mCacheSize = cacheSize;
 
   // Register as a memory pressure observer
@@ -769,20 +813,14 @@ nsresult nsZipReaderCache::ReleaseZip(nsJAR* zip) {
 
   // remove from hashtable
   nsAutoCString uri;
-  rv = oldest->GetJarPath(uri);
-  if (NS_FAILED(rv)) return rv;
-
-  if (oldest->mOuterZipEntry.IsEmpty()) {
-    uri.InsertLiteral("file:", 0);
-  } else {
-    uri.InsertLiteral("jar:", 0);
-    uri.AppendLiteral("!/");
-    uri.Append(oldest->mOuterZipEntry);
+  rv = oldest->GetFullJarPath(uri);
+  if (NS_FAILED(rv)) {
+    return rv;
   }
 
-  // Retrieving and removing the JAR must be done without an extra AddRef
+  // Retrieving and removing the JAR should be done without an extra AddRef
   // and Release, or we'll trigger nsJAR::Release's magic refcount 1 case
-  // an extra time and trigger a deadlock.
+  // an extra time.
   RefPtr<nsJAR> removed;
   mZips.Remove(uri, getter_AddRefs(removed));
   NS_ASSERTION(removed, "botched");
