@@ -5,9 +5,6 @@
 
 /*
  * This module implements a simple archive extractor for the PKZIP format.
- *
- * The underlying nsZipArchive is NOT thread-safe. Do not pass references
- * or pointers to it across thread boundaries.
  */
 
 #define READTYPE int32_t
@@ -15,7 +12,6 @@
 #include "nsISupportsUtils.h"
 #include "mozilla/MmapFaultHandler.h"
 #include "prio.h"
-#include "plstr.h"
 #include "mozilla/Attributes.h"
 #include "mozilla/Logging.h"
 #include "mozilla/MemUtils.h"
@@ -68,6 +64,18 @@
 
 using namespace mozilla;
 
+static LazyLogModule gZipLog("nsZipArchive");
+
+#ifdef LOG
+#  undef LOG
+#endif
+#ifdef LOG_ENABLED
+#  undef LOG_ENABLED
+#endif
+
+#define LOG(args) MOZ_LOG(gZipLog, mozilla::LogLevel::Debug, args)
+#define LOG_ENABLED() MOZ_LOG_TEST(gZipLog, mozilla::LogLevel::Debug)
+
 static const uint32_t kMaxNameLength = PATH_MAX; /* Maximum name length */
 // For synthetic zip entries. Date/time corresponds to 1980-01-01 00:00.
 static const uint16_t kSyntheticTime = 0;
@@ -111,8 +119,8 @@ class ZipArchiveLogger {
       file = PR_ImportFile((PROsfd)handle);
       if (!file) return;
 #else
-      rv = logFile->OpenNSPRFileDesc(PR_WRONLY | PR_CREATE_FILE | PR_APPEND,
-                                     0644, &file);
+      rv = logFile->OpenNSPRFileDesc(
+          PR_WRONLY | PR_CREATE_FILE | PR_APPEND | PR_SYNC, 0644, &file);
       if (NS_FAILED(rv)) return;
 #endif
       mFd = file;
@@ -143,8 +151,8 @@ class ZipArchiveLogger {
 
  private:
   static StaticMutex sLock;
-  int mRefCnt;
-  PRFileDesc* mFd;
+  int mRefCnt MOZ_GUARDED_BY(sLock);
+  PRFileDesc* mFd MOZ_GUARDED_BY(sLock);
 };
 
 StaticMutex ZipArchiveLogger::sLock;
@@ -181,6 +189,7 @@ nsresult nsZipHandle::Init(nsIFile* file, nsZipHandle** ret, PRFileDesc** aFd) {
 #if defined(XP_WIN)
   flags |= nsIFile::OS_READAHEAD;
 #endif
+  LOG(("ZipHandle::Init %s", file->HumanReadablePath().get()));
   nsresult rv = file->OpenNSPRFileDesc(flags, 0000, &fd.rwget());
   if (NS_FAILED(rv)) return rv;
 
@@ -232,6 +241,7 @@ nsresult nsZipHandle::Init(nsZipArchive* zip, const char* entry,
   RefPtr<nsZipHandle> handle = new nsZipHandle();
   if (!handle) return NS_ERROR_OUT_OF_MEMORY;
 
+  LOG(("ZipHandle::Init entry %s", entry));
   handle->mBuf = MakeUnique<nsZipItemPtr<uint8_t>>(zip, entry);
   if (!handle->mBuf) return NS_ERROR_OUT_OF_MEMORY;
 
@@ -249,20 +259,6 @@ nsresult nsZipHandle::Init(nsZipArchive* zip, const char* entry,
   return NS_OK;
 }
 
-nsresult nsZipHandle::Init(const uint8_t* aData, uint32_t aLen,
-                           nsZipHandle** aRet) {
-  RefPtr<nsZipHandle> handle = new nsZipHandle();
-
-  handle->mFileStart = aData;
-  handle->mTotalLen = aLen;
-  nsresult rv = handle->findDataStart();
-  if (NS_FAILED(rv)) {
-    return rv;
-  }
-  handle.forget(aRet);
-  return NS_OK;
-}
-
 // This function finds the start of the ZIP data. If the file is a regular ZIP,
 // this is just the start of the file. If the file is a CRX file, the start of
 // the data is after the CRX header.
@@ -298,6 +294,7 @@ nsresult nsZipHandle::findDataStart() {
       mFileData = mFileStart + headerSize;
       return NS_OK;
     }
+    return NS_ERROR_FILE_CORRUPTED;
   }
   mLen = mTotalLen;
   mFileData = mFileStart;
@@ -338,65 +335,20 @@ nsZipHandle::~nsZipHandle() {
 //---------------------------------------------
 //  nsZipArchive::OpenArchive
 //---------------------------------------------
-nsresult nsZipArchive::OpenArchive(nsZipHandle* aZipHandle, PRFileDesc* aFd) {
-  mFd = aZipHandle;
-
-  //-- get table of contents for archive
-  nsresult rv = BuildFileList(aFd);
-  if (NS_SUCCEEDED(rv)) {
-    if (aZipHandle->mFile && XRE_IsParentProcess()) {
-      static char* env = PR_GetEnv("MOZ_JAR_LOG_FILE");
-      if (env) {
-        mUseZipLog = true;
-
-        zipLog.Init(env);
-        // We only log accesses in jar/zip archives within the NS_GRE_DIR
-        // and/or the APK on Android. For the former, we log the archive path
-        // relative to NS_GRE_DIR, and for the latter, the nested-archive
-        // path within the APK. This makes the path match the path of the
-        // archives relative to the packaged dist/$APP_NAME directory in a
-        // build.
-        if (aZipHandle->mFile.IsZip()) {
-          // Nested archive, likely omni.ja in APK.
-          aZipHandle->mFile.GetPath(mURI);
-        } else if (nsDirectoryService::gService) {
-          // We can reach here through the initialization of Omnijar from
-          // XRE_InitCommandLine, which happens before the directory service
-          // is initialized. When that happens, it means the opened archive is
-          // the APK, and we don't care to log that one, so we just skip
-          // when the directory service is not initialized.
-          nsCOMPtr<nsIFile> dir = aZipHandle->mFile.GetBaseFile();
-          nsCOMPtr<nsIFile> gre_dir;
-          nsAutoCString path;
-          if (NS_SUCCEEDED(nsDirectoryService::gService->Get(
-                  NS_GRE_DIR, NS_GET_IID(nsIFile), getter_AddRefs(gre_dir)))) {
-            nsAutoCString leaf;
-            nsCOMPtr<nsIFile> parent;
-            while (NS_SUCCEEDED(dir->GetNativeLeafName(leaf)) &&
-                   NS_SUCCEEDED(dir->GetParent(getter_AddRefs(parent)))) {
-              if (!parent) {
-                break;
-              }
-              dir = parent;
-              if (path.Length()) {
-                path.Insert('/', 0);
-              }
-              path.Insert(leaf, 0);
-              bool equals;
-              if (NS_SUCCEEDED(dir->Equals(gre_dir, &equals)) && equals) {
-                mURI.Assign(path);
-                break;
-              }
-            }
-          }
-        }
-      }
-    }
+/* static */
+already_AddRefed<nsZipArchive> nsZipArchive::OpenArchive(
+    nsZipHandle* aZipHandle, PRFileDesc* aFd) {
+  nsresult rv;
+  RefPtr<nsZipArchive> self(new nsZipArchive(aZipHandle, aFd, rv));
+  LOG(("ZipHandle::OpenArchive[%p]", self.get()));
+  if (NS_FAILED(rv)) {
+    self = nullptr;
   }
-  return rv;
+  return self.forget();
 }
 
-nsresult nsZipArchive::OpenArchive(nsIFile* aFile) {
+/* static */
+already_AddRefed<nsZipArchive> nsZipArchive::OpenArchive(nsIFile* aFile) {
   RefPtr<nsZipHandle> handle;
 #if defined(XP_WIN)
   mozilla::AutoFDClose fd;
@@ -404,7 +356,7 @@ nsresult nsZipArchive::OpenArchive(nsIFile* aFile) {
 #else
   nsresult rv = nsZipHandle::Init(aFile, getter_AddRefs(handle));
 #endif
-  if (NS_FAILED(rv)) return rv;
+  if (NS_FAILED(rv)) return nullptr;
 
 #if defined(XP_WIN)
   return OpenArchive(handle, fd.get());
@@ -422,7 +374,7 @@ nsresult nsZipArchive::Test(const char* aEntryName) {
   if (aEntryName)  // only test specified item
   {
     currItem = GetItem(aEntryName);
-    if (!currItem) return NS_ERROR_FILE_TARGET_DOES_NOT_EXIST;
+    if (!currItem) return NS_ERROR_FILE_NOT_FOUND;
     //-- don't test (synthetic) directory items
     if (currItem->IsDirectory()) return NS_OK;
     return ExtractFile(currItem, 0, 0);
@@ -441,34 +393,13 @@ nsresult nsZipArchive::Test(const char* aEntryName) {
   return NS_OK;
 }
 
-//---------------------------------------------
-//  nsZipArchive::CloseArchive
-//---------------------------------------------
-nsresult nsZipArchive::CloseArchive() {
-  MutexAutoLock lock(mLock);
-  if (mFd) {
-    mArena.Clear();
-    mFd = nullptr;
-  }
-
-  // CAUTION:
-  // We don't need to delete each of the nsZipItem as the memory for
-  // the zip item and the filename it holds are both allocated from the Arena.
-  // Hence, destroying the Arena is like destroying all the memory
-  // for all the nsZipItem in one shot. But if the ~nsZipItem is doing
-  // anything more than cleaning up memory, we should start calling it.
-  // Let us also cleanup the mFiles table for re-use on the next 'open' call
-  memset(mFiles, 0, sizeof(mFiles));
-  mBuiltSynthetics = false;
-  return NS_OK;
-}
-
 //---------------------------------------------
 // nsZipArchive::GetItem
 //---------------------------------------------
 nsZipItem* nsZipArchive::GetItem(const char* aEntryName) {
   MutexAutoLock lock(mLock);
 
+  LOG(("ZipHandle::GetItem[%p] %s", this, aEntryName));
   if (aEntryName) {
     uint32_t len = strlen(aEntryName);
     //-- If the request is for a directory, make sure that synthetic entries
@@ -505,6 +436,8 @@ nsZipItem* nsZipArchive::GetItem(const char* aEntryName) {
 //---------------------------------------------
 nsresult nsZipArchive::ExtractFile(nsZipItem* item, nsIFile* outFile,
                                    PRFileDesc* aFd) {
+  MutexAutoLock lock(mLock);
+  LOG(("ZipHandle::ExtractFile[%p]", this));
   if (!item) return NS_ERROR_ILLEGAL_VALUE;
   if (!mFd) return NS_ERROR_FAILURE;
 
@@ -554,6 +487,7 @@ nsresult nsZipArchive::FindInit(const char* aPattern, nsZipFind** aFind) {
 
   MutexAutoLock lock(mLock);
 
+  LOG(("ZipHandle::FindInit[%p]", this));
   // null out param in case an error happens
   *aFind = nullptr;
 
@@ -584,13 +518,13 @@ nsresult nsZipArchive::FindInit(const char* aPattern, nsZipFind** aFind) {
         return NS_ERROR_ILLEGAL_VALUE;
     }
 
-    pattern = PL_strdup(aPattern);
+    pattern = strdup(aPattern);
     if (!pattern) return NS_ERROR_OUT_OF_MEMORY;
   }
 
   *aFind = new nsZipFind(this, pattern, regExp);
   if (!*aFind) {
-    PL_strfree(pattern);
+    free(pattern);
     return NS_ERROR_OUT_OF_MEMORY;
   }
 
@@ -629,11 +563,13 @@ nsresult nsZipFind::FindNext(const char** aResult, uint16_t* aNameLen) {
       // Need also to return the name length, as it is NOT zero-terminatdd...
       *aResult = mItem->Name();
       *aNameLen = mItem->nameLength;
+      LOG(("ZipHandle::FindNext[%p] %s", this, *aResult));
       return NS_OK;
     }
   }
   MMAP_FAULT_HANDLER_CATCH(NS_ERROR_FAILURE)
-  return NS_ERROR_FILE_TARGET_DOES_NOT_EXIST;
+  LOG(("ZipHandle::FindNext[%p] not found %s", this, mPattern));
+  return NS_ERROR_FILE_NOT_FOUND;
 }
 
 //***********************************************************
@@ -651,8 +587,12 @@ nsZipItem* nsZipArchive::CreateZipItem() {
 //---------------------------------------------
 //  nsZipArchive::BuildFileList
 //---------------------------------------------
-nsresult nsZipArchive::BuildFileList(PRFileDesc* aFd) {
-  MutexAutoLock lock(mLock);
+nsresult nsZipArchive::BuildFileList(PRFileDesc* aFd)
+    MOZ_NO_THREAD_SAFETY_ANALYSIS {
+  // We're only called from the constructor, but need to call
+  // CreateZipItem(), which touches locked data, and modify mFiles.  Turn
+  // off thread-safety, which normally doesn't apply for constructors
+  // anyways
 
   // Get archive size using end pos
   const uint8_t* buf;
@@ -660,6 +600,7 @@ nsresult nsZipArchive::BuildFileList(PRFileDesc* aFd) {
   const uint8_t* endp = startp + mFd->mLen;
   MMAP_FAULT_HANDLER_BEGIN_HANDLE(mFd)
   uint32_t centralOffset = 4;
+  LOG(("ZipHandle::BuildFileList[%p]", this));
   // Only perform readahead in the parent process. Children processes
   // don't need readahead when the file has already been readahead by
   // the parent process, and readahead only really happens for omni.ja,
@@ -728,6 +669,10 @@ nsresult nsZipArchive::BuildFileList(PRFileDesc* aFd) {
     item->isSynthetic = false;
 
     // Add item to file table
+#ifdef DEBUG
+    nsDependentCSubstring name(item->Name(), namelen);
+    LOG(("   %s", PromiseFlatCString(name).get()));
+#endif
     uint32_t hash = HashName(item->Name(), namelen);
     item->next = mFiles[hash];
     mFiles[hash] = item;
@@ -739,18 +684,6 @@ nsresult nsZipArchive::BuildFileList(PRFileDesc* aFd) {
     return NS_ERROR_FILE_CORRUPTED;
   }
 
-  // Make the comment available for consumers.
-  if ((endp >= buf) && (endp - buf >= ZIPEND_SIZE)) {
-    ZipEnd* zipend = (ZipEnd*)buf;
-
-    buf += ZIPEND_SIZE;
-    uint16_t commentlen = xtoint(zipend->commentfield_len);
-    if (endp - buf >= commentlen) {
-      mCommentPtr = (const char*)buf;
-      mCommentLen = commentlen;
-    }
-  }
-
   MMAP_FAULT_HANDLER_CATCH(NS_ERROR_FAILURE)
   return NS_OK;
 }
@@ -822,25 +755,35 @@ nsresult nsZipArchive::BuildSynthetics() {
   return NS_OK;
 }
 
-nsZipHandle* nsZipArchive::GetFD() {
-  if (!mFd) return nullptr;
-  return mFd.get();
-}
+//---------------------------------------------
+// nsZipArchive::GetFD
+//---------------------------------------------
+nsZipHandle* nsZipArchive::GetFD() const { return mFd.get(); }
 
 //---------------------------------------------
 // nsZipArchive::GetDataOffset
+// Returns 0 on an error; 0 is not a valid result for any success case
 //---------------------------------------------
 uint32_t nsZipArchive::GetDataOffset(nsZipItem* aItem) {
   MOZ_ASSERT(aItem);
+  MOZ_DIAGNOSTIC_ASSERT(mFd);
 
   uint32_t offset;
   MMAP_FAULT_HANDLER_BEGIN_HANDLE(mFd)
   //-- read local header to get variable length values and calculate
   //-- the real data offset
   uint32_t len = mFd->mLen;
+  MOZ_DIAGNOSTIC_ASSERT(len <= UINT32_MAX, "mLen > 2GB");
   const uint8_t* data = mFd->mFileData;
   offset = aItem->LocalOffset();
-  if (len < ZIPLOCAL_SIZE || offset > len - ZIPLOCAL_SIZE) return 0;
+  if (len < ZIPLOCAL_SIZE || offset > len - ZIPLOCAL_SIZE) {
+    return 0;
+  }
+  // Check there's enough space for the signature
+  if (offset > mFd->mLen) {
+    NS_WARNING("Corrupt local offset in JAR file");
+    return 0;
+  }
 
   // -- check signature before using the structure, in case the zip file is
   // corrupt
@@ -852,8 +795,14 @@ uint32_t nsZipArchive::GetDataOffset(nsZipItem* aItem) {
   //--       the offset accurately we need the _local_ extralen.
   offset += ZIPLOCAL_SIZE + xtoint(Local->filename_len) +
             xtoint(Local->extrafield_len);
+  // Check data points inside the file.
+  if (offset > mFd->mLen) {
+    NS_WARNING("Corrupt data offset in JAR file");
+    return 0;
+  }
 
   MMAP_FAULT_HANDLER_CATCH(0)
+  // can't be 0
   return offset;
 }
 
@@ -861,7 +810,10 @@ uint32_t nsZipArchive::GetDataOffset(nsZipItem* aItem) {
 // nsZipArchive::GetData
 //---------------------------------------------
 const uint8_t* nsZipArchive::GetData(nsZipItem* aItem) {
-  MOZ_ASSERT(aItem);
+  MOZ_DIAGNOSTIC_ASSERT(aItem);
+  if (!aItem) {
+    return nullptr;
+  }
   uint32_t offset = GetDataOffset(aItem);
 
   MMAP_FAULT_HANDLER_BEGIN_HANDLE(mFd)
@@ -876,14 +828,6 @@ const uint8_t* nsZipArchive::GetData(nsZipItem* aItem) {
   return mFd->mFileData + offset;
 }
 
-// nsZipArchive::GetComment
-bool nsZipArchive::GetComment(nsACString& aComment) {
-  MMAP_FAULT_HANDLER_BEGIN_BUFFER(mCommentPtr, mCommentLen)
-  aComment.Assign(mCommentPtr, mCommentLen);
-  MMAP_FAULT_HANDLER_CATCH(false)
-  return true;
-}
-
 //---------------------------------------------
 // nsZipArchive::SizeOfMapping
 //---------------------------------------------
@@ -893,22 +837,73 @@ int64_t nsZipArchive::SizeOfMapping() { return mFd ? mFd->SizeOfMapping() : 0; }
 // nsZipArchive constructor and destructor
 //------------------------------------------
 
-nsZipArchive::nsZipArchive()
-    : mRefCnt(0),
-      mCommentPtr(nullptr),
-      mCommentLen(0),
-      mBuiltSynthetics(false),
-      mUseZipLog(false) {
+nsZipArchive::nsZipArchive(nsZipHandle* aZipHandle, PRFileDesc* aFd,
+                           nsresult& aRv)
+    : mRefCnt(0), mFd(aZipHandle), mUseZipLog(false), mBuiltSynthetics(false) {
   // initialize the table to nullptr
   memset(mFiles, 0, sizeof(mFiles));
+  MOZ_DIAGNOSTIC_ASSERT(aZipHandle);
+
+  //-- get table of contents for archive
+  aRv = BuildFileList(aFd);
+  if (NS_FAILED(aRv)) {
+    return;  // whomever created us must destroy us in this case
+  }
+  if (aZipHandle->mFile && XRE_IsParentProcess()) {
+    static char* env = PR_GetEnv("MOZ_JAR_LOG_FILE");
+    if (env) {
+      mUseZipLog = true;
+
+      zipLog.Init(env);
+      // We only log accesses in jar/zip archives within the NS_GRE_DIR
+      // and/or the APK on Android. For the former, we log the archive path
+      // relative to NS_GRE_DIR, and for the latter, the nested-archive
+      // path within the APK. This makes the path match the path of the
+      // archives relative to the packaged dist/$APP_NAME directory in a
+      // build.
+      if (aZipHandle->mFile.IsZip()) {
+        // Nested archive, likely omni.ja in APK.
+        aZipHandle->mFile.GetPath(mURI);
+      } else if (nsDirectoryService::gService) {
+        // We can reach here through the initialization of Omnijar from
+        // XRE_InitCommandLine, which happens before the directory service
+        // is initialized. When that happens, it means the opened archive is
+        // the APK, and we don't care to log that one, so we just skip
+        // when the directory service is not initialized.
+        nsCOMPtr<nsIFile> dir = aZipHandle->mFile.GetBaseFile();
+        nsCOMPtr<nsIFile> gre_dir;
+        nsAutoCString path;
+        if (NS_SUCCEEDED(nsDirectoryService::gService->Get(
+                NS_GRE_DIR, NS_GET_IID(nsIFile), getter_AddRefs(gre_dir)))) {
+          nsAutoCString leaf;
+          nsCOMPtr<nsIFile> parent;
+          while (NS_SUCCEEDED(dir->GetNativeLeafName(leaf)) &&
+                 NS_SUCCEEDED(dir->GetParent(getter_AddRefs(parent)))) {
+            if (!parent) {
+              break;
+            }
+            dir = parent;
+            if (path.Length()) {
+              path.Insert('/', 0);
+            }
+            path.Insert(leaf, 0);
+            bool equals;
+            if (NS_SUCCEEDED(dir->Equals(gre_dir, &equals)) && equals) {
+              mURI.Assign(path);
+              break;
+            }
+          }
+        }
+      }
+    }
+  }
 }
 
 NS_IMPL_ADDREF(nsZipArchive)
 NS_IMPL_RELEASE(nsZipArchive)
 
 nsZipArchive::~nsZipArchive() {
-  CloseArchive();
-
+  LOG(("Closing nsZipArchive[%p]", this));
   if (mUseZipLog) {
     zipLog.Release();
   }
@@ -928,7 +923,7 @@ nsZipFind::nsZipFind(nsZipArchive* aZip, char* aPattern, bool aRegExp)
 }
 
 nsZipFind::~nsZipFind() {
-  PL_strfree(mPattern);
+  free(mPattern);
 
   MOZ_COUNT_DTOR(nsZipFind);
 }
